/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

// Create group operation settings
type CreateGroupOperation struct {
	GroupId               string   `json:"GroupId,omitempty"`
	LogicalProcessorCount uint32   `json:"LogicalProcessorCount,omitempty"`
	LogicalProcessors     []uint32 `json:"LogicalProcessors,omitempty"`
}
