/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type Container struct {
	GuestOs *GuestOs `json:"GuestOs,omitempty"`

	Storage *Storage `json:"Storage,omitempty"`

	MappedDirectories []MappedDirectory `json:"MappedDirectories,omitempty"`

	MappedPipes []MappedPipe `json:"MappedPipes,omitempty"`

	Memory *Memory `json:"Memory,omitempty"`

	Processor *Processor `json:"Processor,omitempty"`

	Networking *Networking `json:"Networking,omitempty"`

	HvSocket *HvSocket `json:"HvSocket,omitempty"`

	ContainerCredentialGuard *ContainerCredentialGuardState `json:"ContainerCredentialGuard,omitempty"`

	RegistryChanges *RegistryChanges `json:"RegistryChanges,omitempty"`

	AssignedDevices []Device `json:"AssignedDevices,omitempty"`
}
