// Code generated by mkstdlib.go. DO NOT EDIT.

package imports

var stdlib = map[string]map[string]bool{
	"archive/tar": map[string]bool{
		"ErrFieldTooLong":    true,
		"ErrHeader":          true,
		"ErrWriteAfterClose": true,
		"ErrWriteTooLong":    true,
		"FileInfoHeader":     true,
		"Format":             true,
		"FormatGNU":          true,
		"FormatPAX":          true,
		"FormatUSTAR":        true,
		"FormatUnknown":      true,
		"Header":             true,
		"NewReader":          true,
		"NewWriter":          true,
		"Reader":             true,
		"TypeBlock":          true,
		"TypeChar":           true,
		"TypeCont":           true,
		"TypeDir":            true,
		"TypeFifo":           true,
		"TypeGNULongLink":    true,
		"TypeGNULongName":    true,
		"TypeGNUSparse":      true,
		"TypeLink":           true,
		"TypeReg":            true,
		"TypeRegA":           true,
		"TypeSymlink":        true,
		"TypeXGlobalHeader":  true,
		"TypeXHeader":        true,
		"Writer":             true,
	},
	"archive/zip": map[string]bool{
		"Compressor":           true,
		"Decompressor":         true,
		"Deflate":              true,
		"ErrAlgorithm":         true,
		"ErrChecksum":          true,
		"ErrFormat":            true,
		"File":                 true,
		"FileHeader":           true,
		"FileInfoHeader":       true,
		"NewReader":            true,
		"NewWriter":            true,
		"OpenReader":           true,
		"ReadCloser":           true,
		"Reader":               true,
		"RegisterCompressor":   true,
		"RegisterDecompressor": true,
		"Store":                true,
		"Writer":               true,
	},
	"bufio": map[string]bool{
		"ErrAdvanceTooFar":     true,
		"ErrBufferFull":        true,
		"ErrFinalToken":        true,
		"ErrInvalidUnreadByte": true,
		"ErrInvalidUnreadRune": true,
		"ErrNegativeAdvance":   true,
		"ErrNegativeCount":     true,
		"ErrTooLong":           true,
		"MaxScanTokenSize":     true,
		"NewReadWriter":        true,
		"NewReader":            true,
		"NewReaderSize":        true,
		"NewScanner":           true,
		"NewWriter":            true,
		"NewWriterSize":        true,
		"ReadWriter":           true,
		"Reader":               true,
		"ScanBytes":            true,
		"ScanLines":            true,
		"ScanRunes":            true,
		"ScanWords":            true,
		"Scanner":              true,
		"SplitFunc":            true,
		"Writer":               true,
	},
	"bytes": map[string]bool{
		"Buffer":          true,
		"Compare":         true,
		"Contains":        true,
		"ContainsAny":     true,
		"ContainsRune":    true,
		"Count":           true,
		"Equal":           true,
		"EqualFold":       true,
		"ErrTooLarge":     true,
		"Fields":          true,
		"FieldsFunc":      true,
		"HasPrefix":       true,
		"HasSuffix":       true,
		"Index":           true,
		"IndexAny":        true,
		"IndexByte":       true,
		"IndexFunc":       true,
		"IndexRune":       true,
		"Join":            true,
		"LastIndex":       true,
		"LastIndexAny":    true,
		"LastIndexByte":   true,
		"LastIndexFunc":   true,
		"Map":             true,
		"MinRead":         true,
		"NewBuffer":       true,
		"NewBufferString": true,
		"NewReader":       true,
		"Reader":          true,
		"Repeat":          true,
		"Replace":         true,
		"ReplaceAll":      true,
		"Runes":           true,
		"Split":           true,
		"SplitAfter":      true,
		"SplitAfterN":     true,
		"SplitN":          true,
		"Title":           true,
		"ToLower":         true,
		"ToLowerSpecial":  true,
		"ToTitle":         true,
		"ToTitleSpecial":  true,
		"ToUpper":         true,
		"ToUpperSpecial":  true,
		"Trim":            true,
		"TrimFunc":        true,
		"TrimLeft":        true,
		"TrimLeftFunc":    true,
		"TrimPrefix":      true,
		"TrimRight":       true,
		"TrimRightFunc":   true,
		"TrimSpace":       true,
		"TrimSuffix":      true,
	},
	"compress/bzip2": map[string]bool{
		"NewReader":       true,
		"StructuralError": true,
	},
	"compress/flate": map[string]bool{
		"BestCompression":    true,
		"BestSpeed":          true,
		"CorruptInputError":  true,
		"DefaultCompression": true,
		"HuffmanOnly":        true,
		"InternalError":      true,
		"NewReader":          true,
		"NewReaderDict":      true,
		"NewWriter":          true,
		"NewWriterDict":      true,
		"NoCompression":      true,
		"ReadError":          true,
		"Reader":             true,
		"Resetter":           true,
		"WriteError":         true,
		"Writer":             true,
	},
	"compress/gzip": map[string]bool{
		"BestCompression":    true,
		"BestSpeed":          true,
		"DefaultCompression": true,
		"ErrChecksum":        true,
		"ErrHeader":          true,
		"Header":             true,
		"HuffmanOnly":        true,
		"NewReader":          true,
		"NewWriter":          true,
		"NewWriterLevel":     true,
		"NoCompression":      true,
		"Reader":             true,
		"Writer":             true,
	},
	"compress/lzw": map[string]bool{
		"LSB":       true,
		"MSB":       true,
		"NewReader": true,
		"NewWriter": true,
		"Order":     true,
	},
	"compress/zlib": map[string]bool{
		"BestCompression":    true,
		"BestSpeed":          true,
		"DefaultCompression": true,
		"ErrChecksum":        true,
		"ErrDictionary":      true,
		"ErrHeader":          true,
		"HuffmanOnly":        true,
		"NewReader":          true,
		"NewReaderDict":      true,
		"NewWriter":          true,
		"NewWriterLevel":     true,
		"NewWriterLevelDict": true,
		"NoCompression":      true,
		"Resetter":           true,
		"Writer":             true,
	},
	"container/heap": map[string]bool{
		"Fix":       true,
		"Init":      true,
		"Interface": true,
		"Pop":       true,
		"Push":      true,
		"Remove":    true,
	},
	"container/list": map[string]bool{
		"Element": true,
		"List":    true,
		"New":     true,
	},
	"container/ring": map[string]bool{
		"New":  true,
		"Ring": true,
	},
	"context": map[string]bool{
		"Background":       true,
		"CancelFunc":       true,
		"Canceled":         true,
		"Context":          true,
		"DeadlineExceeded": true,
		"TODO":             true,
		"WithCancel":       true,
		"WithDeadline":     true,
		"WithTimeout":      true,
		"WithValue":        true,
	},
	"crypto": map[string]bool{
		"BLAKE2b_256":   true,
		"BLAKE2b_384":   true,
		"BLAKE2b_512":   true,
		"BLAKE2s_256":   true,
		"Decrypter":     true,
		"DecrypterOpts": true,
		"Hash":          true,
		"MD4":           true,
		"MD5":           true,
		"MD5SHA1":       true,
		"PrivateKey":    true,
		"PublicKey":     true,
		"RIPEMD160":     true,
		"RegisterHash":  true,
		"SHA1":          true,
		"SHA224":        true,
		"SHA256":        true,
		"SHA384":        true,
		"SHA3_224":      true,
		"SHA3_256":      true,
		"SHA3_384":      true,
		"SHA3_512":      true,
		"SHA512":        true,
		"SHA512_224":    true,
		"SHA512_256":    true,
		"Signer":        true,
		"SignerOpts":    true,
	},
	"crypto/aes": map[string]bool{
		"BlockSize":    true,
		"KeySizeError": true,
		"NewCipher":    true,
	},
	"crypto/cipher": map[string]bool{
		"AEAD":                true,
		"Block":               true,
		"BlockMode":           true,
		"NewCBCDecrypter":     true,
		"NewCBCEncrypter":     true,
		"NewCFBDecrypter":     true,
		"NewCFBEncrypter":     true,
		"NewCTR":              true,
		"NewGCM":              true,
		"NewGCMWithNonceSize": true,
		"NewGCMWithTagSize":   true,
		"NewOFB":              true,
		"Stream":              true,
		"StreamReader":        true,
		"StreamWriter":        true,
	},
	"crypto/des": map[string]bool{
		"BlockSize":          true,
		"KeySizeError":       true,
		"NewCipher":          true,
		"NewTripleDESCipher": true,
	},
	"crypto/dsa": map[string]bool{
		"ErrInvalidPublicKey": true,
		"GenerateKey":         true,
		"GenerateParameters":  true,
		"L1024N160":           true,
		"L2048N224":           true,
		"L2048N256":           true,
		"L3072N256":           true,
		"ParameterSizes":      true,
		"Parameters":          true,
		"PrivateKey":          true,
		"PublicKey":           true,
		"Sign":                true,
		"Verify":              true,
	},
	"crypto/ecdsa": map[string]bool{
		"GenerateKey": true,
		"PrivateKey":  true,
		"PublicKey":   true,
		"Sign":        true,
		"Verify":      true,
	},
	"crypto/elliptic": map[string]bool{
		"Curve":       true,
		"CurveParams": true,
		"GenerateKey": true,
		"Marshal":     true,
		"P224":        true,
		"P256":        true,
		"P384":        true,
		"P521":        true,
		"Unmarshal":   true,
	},
	"crypto/hmac": map[string]bool{
		"Equal": true,
		"New":   true,
	},
	"crypto/md5": map[string]bool{
		"BlockSize": true,
		"New":       true,
		"Size":      true,
		"Sum":       true,
	},
	"crypto/rand": map[string]bool{
		"Int":    true,
		"Prime":  true,
		"Read":   true,
		"Reader": true,
	},
	"crypto/rc4": map[string]bool{
		"Cipher":       true,
		"KeySizeError": true,
		"NewCipher":    true,
	},
	"crypto/rsa": map[string]bool{
		"CRTValue":                  true,
		"DecryptOAEP":               true,
		"DecryptPKCS1v15":           true,
		"DecryptPKCS1v15SessionKey": true,
		"EncryptOAEP":               true,
		"EncryptPKCS1v15":           true,
		"ErrDecryption":             true,
		"ErrMessageTooLong":         true,
		"ErrVerification":           true,
		"GenerateKey":               true,
		"GenerateMultiPrimeKey":     true,
		"OAEPOptions":               true,
		"PKCS1v15DecryptOptions":    true,
		"PSSOptions":                true,
		"PSSSaltLengthAuto":         true,
		"PSSSaltLengthEqualsHash":   true,
		"PrecomputedValues":         true,
		"PrivateKey":                true,
		"PublicKey":                 true,
		"SignPKCS1v15":              true,
		"SignPSS":                   true,
		"VerifyPKCS1v15":            true,
		"VerifyPSS":                 true,
	},
	"crypto/sha1": map[string]bool{
		"BlockSize": true,
		"New":       true,
		"Size":      true,
		"Sum":       true,
	},
	"crypto/sha256": map[string]bool{
		"BlockSize": true,
		"New":       true,
		"New224":    true,
		"Size":      true,
		"Size224":   true,
		"Sum224":    true,
		"Sum256":    true,
	},
	"crypto/sha512": map[string]bool{
		"BlockSize":  true,
		"New":        true,
		"New384":     true,
		"New512_224": true,
		"New512_256": true,
		"Size":       true,
		"Size224":    true,
		"Size256":    true,
		"Size384":    true,
		"Sum384":     true,
		"Sum512":     true,
		"Sum512_224": true,
		"Sum512_256": true,
	},
	"crypto/subtle": map[string]bool{
		"ConstantTimeByteEq":   true,
		"ConstantTimeCompare":  true,
		"ConstantTimeCopy":     true,
		"ConstantTimeEq":       true,
		"ConstantTimeLessOrEq": true,
		"ConstantTimeSelect":   true,
	},
	"crypto/tls": map[string]bool{
		"Certificate":                          true,
		"CertificateRequestInfo":               true,
		"Client":                               true,
		"ClientAuthType":                       true,
		"ClientHelloInfo":                      true,
		"ClientSessionCache":                   true,
		"ClientSessionState":                   true,
		"Config":                               true,
		"Conn":                                 true,
		"ConnectionState":                      true,
		"CurveID":                              true,
		"CurveP256":                            true,
		"CurveP384":                            true,
		"CurveP521":                            true,
		"Dial":                                 true,
		"DialWithDialer":                       true,
		"ECDSAWithP256AndSHA256":               true,
		"ECDSAWithP384AndSHA384":               true,
		"ECDSAWithP521AndSHA512":               true,
		"ECDSAWithSHA1":                        true,
		"Listen":                               true,
		"LoadX509KeyPair":                      true,
		"NewLRUClientSessionCache":             true,
		"NewListener":                          true,
		"NoClientCert":                         true,
		"PKCS1WithSHA1":                        true,
		"PKCS1WithSHA256":                      true,
		"PKCS1WithSHA384":                      true,
		"PKCS1WithSHA512":                      true,
		"PSSWithSHA256":                        true,
		"PSSWithSHA384":                        true,
		"PSSWithSHA512":                        true,
		"RecordHeaderError":                    true,
		"RenegotiateFreelyAsClient":            true,
		"RenegotiateNever":                     true,
		"RenegotiateOnceAsClient":              true,
		"RenegotiationSupport":                 true,
		"RequestClientCert":                    true,
		"RequireAndVerifyClientCert":           true,
		"RequireAnyClientCert":                 true,
		"Server":                               true,
		"SignatureScheme":                      true,
		"TLS_AES_128_GCM_SHA256":               true,
		"TLS_AES_256_GCM_SHA384":               true,
		"TLS_CHACHA20_POLY1305_SHA256":         true,
		"TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA": true,
		"TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256": true,
		"TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256": true,
		"TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA":    true,
		"TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384": true,
		"TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305":  true,
		"TLS_ECDHE_ECDSA_WITH_RC4_128_SHA":        true,
		"TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA":     true,
		"TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA":      true,
		"TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256":   true,
		"TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256":   true,
		"TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA":      true,
		"TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384":   true,
		"TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305":    true,
		"TLS_ECDHE_RSA_WITH_RC4_128_SHA":          true,
		"TLS_FALLBACK_SCSV":                       true,
		"TLS_RSA_WITH_3DES_EDE_CBC_SHA":           true,
		"TLS_RSA_WITH_AES_128_CBC_SHA":            true,
		"TLS_RSA_WITH_AES_128_CBC_SHA256":         true,
		"TLS_RSA_WITH_AES_128_GCM_SHA256":         true,
		"TLS_RSA_WITH_AES_256_CBC_SHA":            true,
		"TLS_RSA_WITH_AES_256_GCM_SHA384":         true,
		"TLS_RSA_WITH_RC4_128_SHA":                true,
		"VerifyClientCertIfGiven":                 true,
		"VersionSSL30":                            true,
		"VersionTLS10":                            true,
		"VersionTLS11":                            true,
		"VersionTLS12":                            true,
		"VersionTLS13":                            true,
		"X25519":                                  true,
		"X509KeyPair":                             true,
	},
	"crypto/x509": map[string]bool{
		"CANotAuthorizedForExtKeyUsage":             true,
		"CANotAuthorizedForThisName":                true,
		"CertPool":                                  true,
		"Certificate":                               true,
		"CertificateInvalidError":                   true,
		"CertificateRequest":                        true,
		"ConstraintViolationError":                  true,
		"CreateCertificate":                         true,
		"CreateCertificateRequest":                  true,
		"DSA":                                       true,
		"DSAWithSHA1":                               true,
		"DSAWithSHA256":                             true,
		"DecryptPEMBlock":                           true,
		"ECDSA":                                     true,
		"ECDSAWithSHA1":                             true,
		"ECDSAWithSHA256":                           true,
		"ECDSAWithSHA384":                           true,
		"ECDSAWithSHA512":                           true,
		"EncryptPEMBlock":                           true,
		"ErrUnsupportedAlgorithm":                   true,
		"Expired":                                   true,
		"ExtKeyUsage":                               true,
		"ExtKeyUsageAny":                            true,
		"ExtKeyUsageClientAuth":                     true,
		"ExtKeyUsageCodeSigning":                    true,
		"ExtKeyUsageEmailProtection":                true,
		"ExtKeyUsageIPSECEndSystem":                 true,
		"ExtKeyUsageIPSECTunnel":                    true,
		"ExtKeyUsageIPSECUser":                      true,
		"ExtKeyUsageMicrosoftCommercialCodeSigning": true,
		"ExtKeyUsageMicrosoftKernelCodeSigning":     true,
		"ExtKeyUsageMicrosoftServerGatedCrypto":     true,
		"ExtKeyUsageNetscapeServerGatedCrypto":      true,
		"ExtKeyUsageOCSPSigning":                    true,
		"ExtKeyUsageServerAuth":                     true,
		"ExtKeyUsageTimeStamping":                   true,
		"HostnameError":                             true,
		"IncompatibleUsage":                         true,
		"IncorrectPasswordError":                    true,
		"InsecureAlgorithmError":                    true,
		"InvalidReason":                             true,
		"IsEncryptedPEMBlock":                       true,
		"KeyUsage":                                  true,
		"KeyUsageCRLSign":                           true,
		"KeyUsageCertSign":                          true,
		"KeyUsageContentCommitment":                 true,
		"KeyUsageDataEncipherment":                  true,
		"KeyUsageDecipherOnly":                      true,
		"KeyUsageDigitalSignature":                  true,
		"KeyUsageEncipherOnly":                      true,
		"KeyUsageKeyAgreement":                      true,
		"KeyUsageKeyEncipherment":                   true,
		"MD2WithRSA":                                true,
		"MD5WithRSA":                                true,
		"MarshalECPrivateKey":                       true,
		"MarshalPKCS1PrivateKey":                    true,
		"MarshalPKCS1PublicKey":                     true,
		"MarshalPKCS8PrivateKey":                    true,
		"MarshalPKIXPublicKey":                      true,
		"NameConstraintsWithoutSANs":                true,
		"NameMismatch":                              true,
		"NewCertPool":                               true,
		"NotAuthorizedToSign":                       true,
		"PEMCipher":                                 true,
		"PEMCipher3DES":                             true,
		"PEMCipherAES128":                           true,
		"PEMCipherAES192":                           true,
		"PEMCipherAES256":                           true,
		"PEMCipherDES":                              true,
		"ParseCRL":                                  true,
		"ParseCertificate":                          true,
		"ParseCertificateRequest":                   true,
		"ParseCertificates":                         true,
		"ParseDERCRL":                               true,
		"ParseECPrivateKey":                         true,
		"ParsePKCS1PrivateKey":                      true,
		"ParsePKCS1PublicKey":                       true,
		"ParsePKCS8PrivateKey":                      true,
		"ParsePKIXPublicKey":                        true,
		"PublicKeyAlgorithm":                        true,
		"RSA":                                       true,
		"SHA1WithRSA":                               true,
		"SHA256WithRSA":                             true,
		"SHA256WithRSAPSS":                          true,
		"SHA384WithRSA":                             true,
		"SHA384WithRSAPSS":                          true,
		"SHA512WithRSA":                             true,
		"SHA512WithRSAPSS":                          true,
		"SignatureAlgorithm":                        true,
		"SystemCertPool":                            true,
		"SystemRootsError":                          true,
		"TooManyConstraints":                        true,
		"TooManyIntermediates":                      true,
		"UnconstrainedName":                         true,
		"UnhandledCriticalExtension":                true,
		"UnknownAuthorityError":                     true,
		"UnknownPublicKeyAlgorithm":                 true,
		"UnknownSignatureAlgorithm":                 true,
		"VerifyOptions":                             true,
	},
	"crypto/x509/pkix": map[string]bool{
		"AlgorithmIdentifier":          true,
		"AttributeTypeAndValue":        true,
		"AttributeTypeAndValueSET":     true,
		"CertificateList":              true,
		"Extension":                    true,
		"Name":                         true,
		"RDNSequence":                  true,
		"RelativeDistinguishedNameSET": true,
		"RevokedCertificate":           true,
		"TBSCertificateList":           true,
	},
	"database/sql": map[string]bool{
		"ColumnType":           true,
		"Conn":                 true,
		"DB":                   true,
		"DBStats":              true,
		"Drivers":              true,
		"ErrConnDone":          true,
		"ErrNoRows":            true,
		"ErrTxDone":            true,
		"IsolationLevel":       true,
		"LevelDefault":         true,
		"LevelLinearizable":    true,
		"LevelReadCommitted":   true,
		"LevelReadUncommitted": true,
		"LevelRepeatableRead":  true,
		"LevelSerializable":    true,
		"LevelSnapshot":        true,
		"LevelWriteCommitted":  true,
		"Named":                true,
		"NamedArg":             true,
		"NullBool":             true,
		"NullFloat64":          true,
		"NullInt64":            true,
		"NullString":           true,
		"Open":                 true,
		"OpenDB":               true,
		"Out":                  true,
		"RawBytes":             true,
		"Register":             true,
		"Result":               true,
		"Row":                  true,
		"Rows":                 true,
		"Scanner":              true,
		"Stmt":                 true,
		"Tx":                   true,
		"TxOptions":            true,
	},
	"database/sql/driver": map[string]bool{
		"Bool":                           true,
		"ColumnConverter":                true,
		"Conn":                           true,
		"ConnBeginTx":                    true,
		"ConnPrepareContext":             true,
		"Connector":                      true,
		"DefaultParameterConverter":      true,
		"Driver":                         true,
		"DriverContext":                  true,
		"ErrBadConn":                     true,
		"ErrRemoveArgument":              true,
		"ErrSkip":                        true,
		"Execer":                         true,
		"ExecerContext":                  true,
		"Int32":                          true,
		"IsScanValue":                    true,
		"IsValue":                        true,
		"IsolationLevel":                 true,
		"NamedValue":                     true,
		"NamedValueChecker":              true,
		"NotNull":                        true,
		"Null":                           true,
		"Pinger":                         true,
		"Queryer":                        true,
		"QueryerContext":                 true,
		"Result":                         true,
		"ResultNoRows":                   true,
		"Rows":                           true,
		"RowsAffected":                   true,
		"RowsColumnTypeDatabaseTypeName": true,
		"RowsColumnTypeLength":           true,
		"RowsColumnTypeNullable":         true,
		"RowsColumnTypePrecisionScale":   true,
		"RowsColumnTypeScanType":         true,
		"RowsNextResultSet":              true,
		"SessionResetter":                true,
		"Stmt":                           true,
		"StmtExecContext":                true,
		"StmtQueryContext":               true,
		"String":                         true,
		"Tx":                             true,
		"TxOptions":                      true,
		"Value":                          true,
		"ValueConverter":                 true,
		"Valuer":                         true,
	},
	"debug/dwarf": map[string]bool{
		"AddrType":                  true,
		"ArrayType":                 true,
		"Attr":                      true,
		"AttrAbstractOrigin":        true,
		"AttrAccessibility":         true,
		"AttrAddrClass":             true,
		"AttrAllocated":             true,
		"AttrArtificial":            true,
		"AttrAssociated":            true,
		"AttrBaseTypes":             true,
		"AttrBitOffset":             true,
		"AttrBitSize":               true,
		"AttrByteSize":              true,
		"AttrCallColumn":            true,
		"AttrCallFile":              true,
		"AttrCallLine":              true,
		"AttrCalling":               true,
		"AttrCommonRef":             true,
		"AttrCompDir":               true,
		"AttrConstValue":            true,
		"AttrContainingType":        true,
		"AttrCount":                 true,
		"AttrDataLocation":          true,
		"AttrDataMemberLoc":         true,
		"AttrDeclColumn":            true,
		"AttrDeclFile":              true,
		"AttrDeclLine":              true,
		"AttrDeclaration":           true,
		"AttrDefaultValue":          true,
		"AttrDescription":           true,
		"AttrDiscr":                 true,
		"AttrDiscrList":             true,
		"AttrDiscrValue":            true,
		"AttrEncoding":              true,
		"AttrEntrypc":               true,
		"AttrExtension":             true,
		"AttrExternal":              true,
		"AttrFrameBase":             true,
		"AttrFriend":                true,
		"AttrHighpc":                true,
		"AttrIdentifierCase":        true,
		"AttrImport":                true,
		"AttrInline":                true,
		"AttrIsOptional":            true,
		"AttrLanguage":              true,
		"AttrLocation":              true,
		"AttrLowerBound":            true,
		"AttrLowpc":                 true,
		"AttrMacroInfo":             true,
		"AttrName":                  true,
		"AttrNamelistItem":          true,
		"AttrOrdering":              true,
		"AttrPriority":              true,
		"AttrProducer":              true,
		"AttrPrototyped":            true,
		"AttrRanges":                true,
		"AttrReturnAddr":            true,
		"AttrSegment":               true,
		"AttrSibling":               true,
		"AttrSpecification":         true,
		"AttrStartScope":            true,
		"AttrStaticLink":            true,
		"AttrStmtList":              true,
		"AttrStride":                true,
		"AttrStrideSize":            true,
		"AttrStringLength":          true,
		"AttrTrampoline":            true,
		"AttrType":                  true,
		"AttrUpperBound":            true,
		"AttrUseLocation":           true,
		"AttrUseUTF8":               true,
		"AttrVarParam":              true,
		"AttrVirtuality":            true,
		"AttrVisibility":            true,
		"AttrVtableElemLoc":         true,
		"BasicType":                 true,
		"BoolType":                  true,
		"CharType":                  true,
		"Class":                     true,
		"ClassAddress":              true,
		"ClassBlock":                true,
		"ClassConstant":             true,
		"ClassExprLoc":              true,
		"ClassFlag":                 true,
		"ClassLinePtr":              true,
		"ClassLocListPtr":           true,
		"ClassMacPtr":               true,
		"ClassRangeListPtr":         true,
		"ClassReference":            true,
		"ClassReferenceAlt":         true,
		"ClassReferenceSig":         true,
		"ClassString":               true,
		"ClassStringAlt":            true,
		"ClassUnknown":              true,
		"CommonType":                true,
		"ComplexType":               true,
		"Data":                      true,
		"DecodeError":               true,
		"DotDotDotType":             true,
		"Entry":                     true,
		"EnumType":                  true,
		"EnumValue":                 true,
		"ErrUnknownPC":              true,
		"Field":                     true,
		"FloatType":                 true,
		"FuncType":                  true,
		"IntType":                   true,
		"LineEntry":                 true,
		"LineFile":                  true,
		"LineReader":                true,
		"LineReaderPos":             true,
		"New":                       true,
		"Offset":                    true,
		"PtrType":                   true,
		"QualType":                  true,
		"Reader":                    true,
		"StructField":               true,
		"StructType":                true,
		"Tag":                       true,
		"TagAccessDeclaration":      true,
		"TagArrayType":              true,
		"TagBaseType":               true,
		"TagCatchDwarfBlock":        true,
		"TagClassType":              true,
		"TagCommonDwarfBlock":       true,
		"TagCommonInclusion":        true,
		"TagCompileUnit":            true,
		"TagCondition":              true,
		"TagConstType":              true,
		"TagConstant":               true,
		"TagDwarfProcedure":         true,
		"TagEntryPoint":             true,
		"TagEnumerationType":        true,
		"TagEnumerator":             true,
		"TagFileType":               true,
		"TagFormalParameter":        true,
		"TagFriend":                 true,
		"TagImportedDeclaration":    true,
		"TagImportedModule":         true,
		"TagImportedUnit":           true,
		"TagInheritance":            true,
		"TagInlinedSubroutine":      true,
		"TagInterfaceType":          true,
		"TagLabel":                  true,
		"TagLexDwarfBlock":          true,
		"TagMember":                 true,
		"TagModule":                 true,
		"TagMutableType":            true,
		"TagNamelist":               true,
		"TagNamelistItem":           true,
		"TagNamespace":              true,
		"TagPackedType":             true,
		"TagPartialUnit":            true,
		"TagPointerType":            true,
		"TagPtrToMemberType":        true,
		"TagReferenceType":          true,
		"TagRestrictType":           true,
		"TagRvalueReferenceType":    true,
		"TagSetType":                true,
		"TagSharedType":             true,
		"TagStringType":             true,
		"TagStructType":             true,
		"TagSubprogram":             true,
		"TagSubrangeType":           true,
		"TagSubroutineType":         true,
		"TagTemplateAlias":          true,
		"TagTemplateTypeParameter":  true,
		"TagTemplateValueParameter": true,
		"TagThrownType":             true,
		"TagTryDwarfBlock":          true,
		"TagTypeUnit":               true,
		"TagTypedef":                true,
		"TagUnionType":              true,
		"TagUnspecifiedParameters":  true,
		"TagUnspecifiedType":        true,
		"TagVariable":               true,
		"TagVariant":                true,
		"TagVariantPart":            true,
		"TagVolatileType":           true,
		"TagWithStmt":               true,
		"Type":                      true,
		"TypedefType":               true,
		"UcharType":                 true,
		"UintType":                  true,
		"UnspecifiedType":           true,
		"VoidType":                  true,
	},
	"debug/elf": map[string]bool{
		"ARM_MAGIC_TRAMP_NUMBER":             true,
		"COMPRESS_HIOS":                      true,
		"COMPRESS_HIPROC":                    true,
		"COMPRESS_LOOS":                      true,
		"COMPRESS_LOPROC":                    true,
		"COMPRESS_ZLIB":                      true,
		"Chdr32":                             true,
		"Chdr64":                             true,
		"Class":                              true,
		"CompressionType":                    true,
		"DF_BIND_NOW":                        true,
		"DF_ORIGIN":                          true,
		"DF_STATIC_TLS":                      true,
		"DF_SYMBOLIC":                        true,
		"DF_TEXTREL":                         true,
		"DT_BIND_NOW":                        true,
		"DT_DEBUG":                           true,
		"DT_ENCODING":                        true,
		"DT_FINI":                            true,
		"DT_FINI_ARRAY":                      true,
		"DT_FINI_ARRAYSZ":                    true,
		"DT_FLAGS":                           true,
		"DT_HASH":                            true,
		"DT_HIOS":                            true,
		"DT_HIPROC":                          true,
		"DT_INIT":                            true,
		"DT_INIT_ARRAY":                      true,
		"DT_INIT_ARRAYSZ":                    true,
		"DT_JMPREL":                          true,
		"DT_LOOS":                            true,
		"DT_LOPROC":                          true,
		"DT_NEEDED":                          true,
		"DT_NULL":                            true,
		"DT_PLTGOT":                          true,
		"DT_PLTREL":                          true,
		"DT_PLTRELSZ":                        true,
		"DT_PREINIT_ARRAY":                   true,
		"DT_PREINIT_ARRAYSZ":                 true,
		"DT_REL":                             true,
		"DT_RELA":                            true,
		"DT_RELAENT":                         true,
		"DT_RELASZ":                          true,
		"DT_RELENT":                          true,
		"DT_RELSZ":                           true,
		"DT_RPATH":                           true,
		"DT_RUNPATH":                         true,
		"DT_SONAME":                          true,
		"DT_STRSZ":                           true,
		"DT_STRTAB":                          true,
		"DT_SYMBOLIC":                        true,
		"DT_SYMENT":                          true,
		"DT_SYMTAB":                          true,
		"DT_TEXTREL":                         true,
		"DT_VERNEED":                         true,
		"DT_VERNEEDNUM":                      true,
		"DT_VERSYM":                          true,
		"Data":                               true,
		"Dyn32":                              true,
		"Dyn64":                              true,
		"DynFlag":                            true,
		"DynTag":                             true,
		"EI_ABIVERSION":                      true,
		"EI_CLASS":                           true,
		"EI_DATA":                            true,
		"EI_NIDENT":                          true,
		"EI_OSABI":                           true,
		"EI_PAD":                             true,
		"EI_VERSION":                         true,
		"ELFCLASS32":                         true,
		"ELFCLASS64":                         true,
		"ELFCLASSNONE":                       true,
		"ELFDATA2LSB":                        true,
		"ELFDATA2MSB":                        true,
		"ELFDATANONE":                        true,
		"ELFMAG":                             true,
		"ELFOSABI_86OPEN":                    true,
		"ELFOSABI_AIX":                       true,
		"ELFOSABI_ARM":                       true,
		"ELFOSABI_AROS":                      true,
		"ELFOSABI_CLOUDABI":                  true,
		"ELFOSABI_FENIXOS":                   true,
		"ELFOSABI_FREEBSD":                   true,
		"ELFOSABI_HPUX":                      true,
		"ELFOSABI_HURD":                      true,
		"ELFOSABI_IRIX":                      true,
		"ELFOSABI_LINUX":                     true,
		"ELFOSABI_MODESTO":                   true,
		"ELFOSABI_NETBSD":                    true,
		"ELFOSABI_NONE":                      true,
		"ELFOSABI_NSK":                       true,
		"ELFOSABI_OPENBSD":                   true,
		"ELFOSABI_OPENVMS":                   true,
		"ELFOSABI_SOLARIS":                   true,
		"ELFOSABI_STANDALONE":                true,
		"ELFOSABI_TRU64":                     true,
		"EM_386":                             true,
		"EM_486":                             true,
		"EM_56800EX":                         true,
		"EM_68HC05":                          true,
		"EM_68HC08":                          true,
		"EM_68HC11":                          true,
		"EM_68HC12":                          true,
		"EM_68HC16":                          true,
		"EM_68K":                             true,
		"EM_78KOR":                           true,
		"EM_8051":                            true,
		"EM_860":                             true,
		"EM_88K":                             true,
		"EM_960":                             true,
		"EM_AARCH64":                         true,
		"EM_ALPHA":                           true,
		"EM_ALPHA_STD":                       true,
		"EM_ALTERA_NIOS2":                    true,
		"EM_AMDGPU":                          true,
		"EM_ARC":                             true,
		"EM_ARCA":                            true,
		"EM_ARC_COMPACT":                     true,
		"EM_ARC_COMPACT2":                    true,
		"EM_ARM":                             true,
		"EM_AVR":                             true,
		"EM_AVR32":                           true,
		"EM_BA1":                             true,
		"EM_BA2":                             true,
		"EM_BLACKFIN":                        true,
		"EM_BPF":                             true,
		"EM_C166":                            true,
		"EM_CDP":                             true,
		"EM_CE":                              true,
		"EM_CLOUDSHIELD":                     true,
		"EM_COGE":                            true,
		"EM_COLDFIRE":                        true,
		"EM_COOL":                            true,
		"EM_COREA_1ST":                       true,
		"EM_COREA_2ND":                       true,
		"EM_CR":                              true,
		"EM_CR16":                            true,
		"EM_CRAYNV2":                         true,
		"EM_CRIS":                            true,
		"EM_CRX":                             true,
		"EM_CSR_KALIMBA":                     true,
		"EM_CUDA":                            true,
		"EM_CYPRESS_M8C":                     true,
		"EM_D10V":                            true,
		"EM_D30V":                            true,
		"EM_DSP24":                           true,
		"EM_DSPIC30F":                        true,
		"EM_DXP":                             true,
		"EM_ECOG1":                           true,
		"EM_ECOG16":                          true,
		"EM_ECOG1X":                          true,
		"EM_ECOG2":                           true,
		"EM_ETPU":                            true,
		"EM_EXCESS":                          true,
		"EM_F2MC16":                          true,
		"EM_FIREPATH":                        true,
		"EM_FR20":                            true,
		"EM_FR30":                            true,
		"EM_FT32":                            true,
		"EM_FX66":                            true,
		"EM_H8S":                             true,
		"EM_H8_300":                          true,
		"EM_H8_300H":                         true,
		"EM_H8_500":                          true,
		"EM_HUANY":                           true,
		"EM_IA_64":                           true,
		"EM_INTEL205":                        true,
		"EM_INTEL206":                        true,
		"EM_INTEL207":                        true,
		"EM_INTEL208":                        true,
		"EM_INTEL209":                        true,
		"EM_IP2K":                            true,
		"EM_JAVELIN":                         true,
		"EM_K10M":                            true,
		"EM_KM32":                            true,
		"EM_KMX16":                           true,
		"EM_KMX32":                           true,
		"EM_KMX8":                            true,
		"EM_KVARC":                           true,
		"EM_L10M":                            true,
		"EM_LANAI":                           true,
		"EM_LATTICEMICO32":                   true,
		"EM_M16C":                            true,
		"EM_M32":                             true,
		"EM_M32C":                            true,
		"EM_M32R":                            true,
		"EM_MANIK":                           true,
		"EM_MAX":                             true,
		"EM_MAXQ30":                          true,
		"EM_MCHP_PIC":                        true,
		"EM_MCST_ELBRUS":                     true,
		"EM_ME16":                            true,
		"EM_METAG":                           true,
		"EM_MICROBLAZE":                      true,
		"EM_MIPS":                            true,
		"EM_MIPS_RS3_LE":                     true,
		"EM_MIPS_RS4_BE":                     true,
		"EM_MIPS_X":                          true,
		"EM_MMA":                             true,
		"EM_MMDSP_PLUS":                      true,
		"EM_MMIX":                            true,
		"EM_MN10200":                         true,
		"EM_MN10300":                         true,
		"EM_MOXIE":                           true,
		"EM_MSP430":                          true,
		"EM_NCPU":                            true,
		"EM_NDR1":                            true,
		"EM_NDS32":                           true,
		"EM_NONE":                            true,
		"EM_NORC":                            true,
		"EM_NS32K":                           true,
		"EM_OPEN8":                           true,
		"EM_OPENRISC":                        true,
		"EM_PARISC":                          true,
		"EM_PCP":                             true,
		"EM_PDP10":                           true,
		"EM_PDP11":                           true,
		"EM_PDSP":                            true,
		"EM_PJ":                              true,
		"EM_PPC":                             true,
		"EM_PPC64":                           true,
		"EM_PRISM":                           true,
		"EM_QDSP6":                           true,
		"EM_R32C":                            true,
		"EM_RCE":                             true,
		"EM_RH32":                            true,
		"EM_RISCV":                           true,
		"EM_RL78":                            true,
		"EM_RS08":                            true,
		"EM_RX":                              true,
		"EM_S370":                            true,
		"EM_S390":                            true,
		"EM_SCORE7":                          true,
		"EM_SEP":                             true,
		"EM_SE_C17":                          true,
		"EM_SE_C33":                          true,
		"EM_SH":                              true,
		"EM_SHARC":                           true,
		"EM_SLE9X":                           true,
		"EM_SNP1K":                           true,
		"EM_SPARC":                           true,
		"EM_SPARC32PLUS":                     true,
		"EM_SPARCV9":                         true,
		"EM_ST100":                           true,
		"EM_ST19":                            true,
		"EM_ST200":                           true,
		"EM_ST7":                             true,
		"EM_ST9PLUS":                         true,
		"EM_STARCORE":                        true,
		"EM_STM8":                            true,
		"EM_STXP7X":                          true,
		"EM_SVX":                             true,
		"EM_TILE64":                          true,
		"EM_TILEGX":                          true,
		"EM_TILEPRO":                         true,
		"EM_TINYJ":                           true,
		"EM_TI_ARP32":                        true,
		"EM_TI_C2000":                        true,
		"EM_TI_C5500":                        true,
		"EM_TI_C6000":                        true,
		"EM_TI_PRU":                          true,
		"EM_TMM_GPP":                         true,
		"EM_TPC":                             true,
		"EM_TRICORE":                         true,
		"EM_TRIMEDIA":                        true,
		"EM_TSK3000":                         true,
		"EM_UNICORE":                         true,
		"EM_V800":                            true,
		"EM_V850":                            true,
		"EM_VAX":                             true,
		"EM_VIDEOCORE":                       true,
		"EM_VIDEOCORE3":                      true,
		"EM_VIDEOCORE5":                      true,
		"EM_VISIUM":                          true,
		"EM_VPP500":                          true,
		"EM_X86_64":                          true,
		"EM_XCORE":                           true,
		"EM_XGATE":                           true,
		"EM_XIMO16":                          true,
		"EM_XTENSA":                          true,
		"EM_Z80":                             true,
		"EM_ZSP":                             true,
		"ET_CORE":                            true,
		"ET_DYN":                             true,
		"ET_EXEC":                            true,
		"ET_HIOS":                            true,
		"ET_HIPROC":                          true,
		"ET_LOOS":                            true,
		"ET_LOPROC":                          true,
		"ET_NONE":                            true,
		"ET_REL":                             true,
		"EV_CURRENT":                         true,
		"EV_NONE":                            true,
		"ErrNoSymbols":                       true,
		"File":                               true,
		"FileHeader":                         true,
		"FormatError":                        true,
		"Header32":                           true,
		"Header64":                           true,
		"ImportedSymbol":                     true,
		"Machine":                            true,
		"NT_FPREGSET":                        true,
		"NT_PRPSINFO":                        true,
		"NT_PRSTATUS":                        true,
		"NType":                              true,
		"NewFile":                            true,
		"OSABI":                              true,
		"Open":                               true,
		"PF_MASKOS":                          true,
		"PF_MASKPROC":                        true,
		"PF_R":                               true,
		"PF_W":                               true,
		"PF_X":                               true,
		"PT_DYNAMIC":                         true,
		"PT_HIOS":                            true,
		"PT_HIPROC":                          true,
		"PT_INTERP":                          true,
		"PT_LOAD":                            true,
		"PT_LOOS":                            true,
		"PT_LOPROC":                          true,
		"PT_NOTE":                            true,
		"PT_NULL":                            true,
		"PT_PHDR":                            true,
		"PT_SHLIB":                           true,
		"PT_TLS":                             true,
		"Prog":                               true,
		"Prog32":                             true,
		"Prog64":                             true,
		"ProgFlag":                           true,
		"ProgHeader":                         true,
		"ProgType":                           true,
		"R_386":                              true,
		"R_386_16":                           true,
		"R_386_32":                           true,
		"R_386_32PLT":                        true,
		"R_386_8":                            true,
		"R_386_COPY":                         true,
		"R_386_GLOB_DAT":                     true,
		"R_386_GOT32":                        true,
		"R_386_GOT32X":                       true,
		"R_386_GOTOFF":                       true,
		"R_386_GOTPC":                        true,
		"R_386_IRELATIVE":                    true,
		"R_386_JMP_SLOT":                     true,
		"R_386_NONE":                         true,
		"R_386_PC16":                         true,
		"R_386_PC32":                         true,
		"R_386_PC8":                          true,
		"R_386_PLT32":                        true,
		"R_386_RELATIVE":                     true,
		"R_386_SIZE32":                       true,
		"R_386_TLS_DESC":                     true,
		"R_386_TLS_DESC_CALL":                true,
		"R_386_TLS_DTPMOD32":                 true,
		"R_386_TLS_DTPOFF32":                 true,
		"R_386_TLS_GD":                       true,
		"R_386_TLS_GD_32":                    true,
		"R_386_TLS_GD_CALL":                  true,
		"R_386_TLS_GD_POP":                   true,
		"R_386_TLS_GD_PUSH":                  true,
		"R_386_TLS_GOTDESC":                  true,
		"R_386_TLS_GOTIE":                    true,
		"R_386_TLS_IE":                       true,
		"R_386_TLS_IE_32":                    true,
		"R_386_TLS_LDM":                      true,
		"R_386_TLS_LDM_32":                   true,
		"R_386_TLS_LDM_CALL":                 true,
		"R_386_TLS_LDM_POP":                  true,
		"R_386_TLS_LDM_PUSH":                 true,
		"R_386_TLS_LDO_32":                   true,
		"R_386_TLS_LE":                       true,
		"R_386_TLS_LE_32":                    true,
		"R_386_TLS_TPOFF":                    true,
		"R_386_TLS_TPOFF32":                  true,
		"R_390":                              true,
		"R_390_12":                           true,
		"R_390_16":                           true,
		"R_390_20":                           true,
		"R_390_32":                           true,
		"R_390_64":                           true,
		"R_390_8":                            true,
		"R_390_COPY":                         true,
		"R_390_GLOB_DAT":                     true,
		"R_390_GOT12":                        true,
		"R_390_GOT16":                        true,
		"R_390_GOT20":                        true,
		"R_390_GOT32":                        true,
		"R_390_GOT64":                        true,
		"R_390_GOTENT":                       true,
		"R_390_GOTOFF":                       true,
		"R_390_GOTOFF16":                     true,
		"R_390_GOTOFF64":                     true,
		"R_390_GOTPC":                        true,
		"R_390_GOTPCDBL":                     true,
		"R_390_GOTPLT12":                     true,
		"R_390_GOTPLT16":                     true,
		"R_390_GOTPLT20":                     true,
		"R_390_GOTPLT32":                     true,
		"R_390_GOTPLT64":                     true,
		"R_390_GOTPLTENT":                    true,
		"R_390_GOTPLTOFF16":                  true,
		"R_390_GOTPLTOFF32":                  true,
		"R_390_GOTPLTOFF64":                  true,
		"R_390_JMP_SLOT":                     true,
		"R_390_NONE":                         true,
		"R_390_PC16":                         true,
		"R_390_PC16DBL":                      true,
		"R_390_PC32":                         true,
		"R_390_PC32DBL":                      true,
		"R_390_PC64":                         true,
		"R_390_PLT16DBL":                     true,
		"R_390_PLT32":                        true,
		"R_390_PLT32DBL":                     true,
		"R_390_PLT64":                        true,
		"R_390_RELATIVE":                     true,
		"R_390_TLS_DTPMOD":                   true,
		"R_390_TLS_DTPOFF":                   true,
		"R_390_TLS_GD32":                     true,
		"R_390_TLS_GD64":                     true,
		"R_390_TLS_GDCALL":                   true,
		"R_390_TLS_GOTIE12":                  true,
		"R_390_TLS_GOTIE20":                  true,
		"R_390_TLS_GOTIE32":                  true,
		"R_390_TLS_GOTIE64":                  true,
		"R_390_TLS_IE32":                     true,
		"R_390_TLS_IE64":                     true,
		"R_390_TLS_IEENT":                    true,
		"R_390_TLS_LDCALL":                   true,
		"R_390_TLS_LDM32":                    true,
		"R_390_TLS_LDM64":                    true,
		"R_390_TLS_LDO32":                    true,
		"R_390_TLS_LDO64":                    true,
		"R_390_TLS_LE32":                     true,
		"R_390_TLS_LE64":                     true,
		"R_390_TLS_LOAD":                     true,
		"R_390_TLS_TPOFF":                    true,
		"R_AARCH64":                          true,
		"R_AARCH64_ABS16":                    true,
		"R_AARCH64_ABS32":                    true,
		"R_AARCH64_ABS64":                    true,
		"R_AARCH64_ADD_ABS_LO12_NC":          true,
		"R_AARCH64_ADR_GOT_PAGE":             true,
		"R_AARCH64_ADR_PREL_LO21":            true,
		"R_AARCH64_ADR_PREL_PG_HI21":         true,
		"R_AARCH64_ADR_PREL_PG_HI21_NC":      true,
		"R_AARCH64_CALL26":                   true,
		"R_AARCH64_CONDBR19":                 true,
		"R_AARCH64_COPY":                     true,
		"R_AARCH64_GLOB_DAT":                 true,
		"R_AARCH64_GOT_LD_PREL19":            true,
		"R_AARCH64_IRELATIVE":                true,
		"R_AARCH64_JUMP26":                   true,
		"R_AARCH64_JUMP_SLOT":                true,
		"R_AARCH64_LD64_GOTOFF_LO15":         true,
		"R_AARCH64_LD64_GOTPAGE_LO15":        true,
		"R_AARCH64_LD64_GOT_LO12_NC":         true,
		"R_AARCH64_LDST128_ABS_LO12_NC":      true,
		"R_AARCH64_LDST16_ABS_LO12_NC":       true,
		"R_AARCH64_LDST32_ABS_LO12_NC":       true,
		"R_AARCH64_LDST64_ABS_LO12_NC":       true,
		"R_AARCH64_LDST8_ABS_LO12_NC":        true,
		"R_AARCH64_LD_PREL_LO19":             true,
		"R_AARCH64_MOVW_SABS_G0":             true,
		"R_AARCH64_MOVW_SABS_G1":             true,
		"R_AARCH64_MOVW_SABS_G2":             true,
		"R_AARCH64_MOVW_UABS_G0":             true,
		"R_AARCH64_MOVW_UABS_G0_NC":          true,
		"R_AARCH64_MOVW_UABS_G1":             true,
		"R_AARCH64_MOVW_UABS_G1_NC":          true,
		"R_AARCH64_MOVW_UABS_G2":             true,
		"R_AARCH64_MOVW_UABS_G2_NC":          true,
		"R_AARCH64_MOVW_UABS_G3":             true,
		"R_AARCH64_NONE":                     true,
		"R_AARCH64_NULL":                     true,
		"R_AARCH64_P32_ABS16":                true,
		"R_AARCH64_P32_ABS32":                true,
		"R_AARCH64_P32_ADD_ABS_LO12_NC":      true,
		"R_AARCH64_P32_ADR_GOT_PAGE":         true,
		"R_AARCH64_P32_ADR_PREL_LO21":        true,
		"R_AARCH64_P32_ADR_PREL_PG_HI21":     true,
		"R_AARCH64_P32_CALL26":               true,
		"R_AARCH64_P32_CONDBR19":             true,
		"R_AARCH64_P32_COPY":                 true,
		"R_AARCH64_P32_GLOB_DAT":             true,
		"R_AARCH64_P32_GOT_LD_PREL19":        true,
		"R_AARCH64_P32_IRELATIVE":            true,
		"R_AARCH64_P32_JUMP26":               true,
		"R_AARCH64_P32_JUMP_SLOT":            true,
		"R_AARCH64_P32_LD32_GOT_LO12_NC":     true,
		"R_AARCH64_P32_LDST128_ABS_LO12_NC":  true,
		"R_AARCH64_P32_LDST16_ABS_LO12_NC":   true,
		"R_AARCH64_P32_LDST32_ABS_LO12_NC":   true,
		"R_AARCH64_P32_LDST64_ABS_LO12_NC":   true,
		"R_AARCH64_P32_LDST8_ABS_LO12_NC":    true,
		"R_AARCH64_P32_LD_PREL_LO19":         true,
		"R_AARCH64_P32_MOVW_SABS_G0":         true,
		"R_AARCH64_P32_MOVW_UABS_G0":         true,
		"R_AARCH64_P32_MOVW_UABS_G0_NC":      true,
		"R_AARCH64_P32_MOVW_UABS_G1":         true,
		"R_AARCH64_P32_PREL16":               true,
		"R_AARCH64_P32_PREL32":               true,
		"R_AARCH64_P32_RELATIVE":             true,
		"R_AARCH64_P32_TLSDESC":              true,
		"R_AARCH64_P32_TLSDESC_ADD_LO12_NC":  true,
		"R_AARCH64_P32_TLSDESC_ADR_PAGE21":   true,
		"R_AARCH64_P32_TLSDESC_ADR_PREL21":   true,
		"R_AARCH64_P32_TLSDESC_CALL":         true,
		"R_AARCH64_P32_TLSDESC_LD32_LO12_NC": true,
		"R_AARCH64_P32_TLSDESC_LD_PREL19":    true,
		"R_AARCH64_P32_TLSGD_ADD_LO12_NC":    true,
		"R_AARCH64_P32_TLSGD_ADR_PAGE21":     true,
		"R_AARCH64_P32_TLSIE_ADR_GOTTPREL_PAGE21":   true,
		"R_AARCH64_P32_TLSIE_LD32_GOTTPREL_LO12_NC": true,
		"R_AARCH64_P32_TLSIE_LD_GOTTPREL_PREL19":    true,
		"R_AARCH64_P32_TLSLE_ADD_TPREL_HI12":        true,
		"R_AARCH64_P32_TLSLE_ADD_TPREL_LO12":        true,
		"R_AARCH64_P32_TLSLE_ADD_TPREL_LO12_NC":     true,
		"R_AARCH64_P32_TLSLE_MOVW_TPREL_G0":         true,
		"R_AARCH64_P32_TLSLE_MOVW_TPREL_G0_NC":      true,
		"R_AARCH64_P32_TLSLE_MOVW_TPREL_G1":         true,
		"R_AARCH64_P32_TLS_DTPMOD":                  true,
		"R_AARCH64_P32_TLS_DTPREL":                  true,
		"R_AARCH64_P32_TLS_TPREL":                   true,
		"R_AARCH64_P32_TSTBR14":                     true,
		"R_AARCH64_PREL16":                          true,
		"R_AARCH64_PREL32":                          true,
		"R_AARCH64_PREL64":                          true,
		"R_AARCH64_RELATIVE":                        true,
		"R_AARCH64_TLSDESC":                         true,
		"R_AARCH64_TLSDESC_ADD":                     true,
		"R_AARCH64_TLSDESC_ADD_LO12_NC":             true,
		"R_AARCH64_TLSDESC_ADR_PAGE21":              true,
		"R_AARCH64_TLSDESC_ADR_PREL21":              true,
		"R_AARCH64_TLSDESC_CALL":                    true,
		"R_AARCH64_TLSDESC_LD64_LO12_NC":            true,
		"R_AARCH64_TLSDESC_LDR":                     true,
		"R_AARCH64_TLSDESC_LD_PREL19":               true,
		"R_AARCH64_TLSDESC_OFF_G0_NC":               true,
		"R_AARCH64_TLSDESC_OFF_G1":                  true,
		"R_AARCH64_TLSGD_ADD_LO12_NC":               true,
		"R_AARCH64_TLSGD_ADR_PAGE21":                true,
		"R_AARCH64_TLSGD_ADR_PREL21":                true,
		"R_AARCH64_TLSGD_MOVW_G0_NC":                true,
		"R_AARCH64_TLSGD_MOVW_G1":                   true,
		"R_AARCH64_TLSIE_ADR_GOTTPREL_PAGE21":       true,
		"R_AARCH64_TLSIE_LD64_GOTTPREL_LO12_NC":     true,
		"R_AARCH64_TLSIE_LD_GOTTPREL_PREL19":        true,
		"R_AARCH64_TLSIE_MOVW_GOTTPREL_G0_NC":       true,
		"R_AARCH64_TLSIE_MOVW_GOTTPREL_G1":          true,
		"R_AARCH64_TLSLD_ADR_PAGE21":                true,
		"R_AARCH64_TLSLD_ADR_PREL21":                true,
		"R_AARCH64_TLSLD_LDST128_DTPREL_LO12":       true,
		"R_AARCH64_TLSLD_LDST128_DTPREL_LO12_NC":    true,
		"R_AARCH64_TLSLE_ADD_TPREL_HI12":            true,
		"R_AARCH64_TLSLE_ADD_TPREL_LO12":            true,
		"R_AARCH64_TLSLE_ADD_TPREL_LO12_NC":         true,
		"R_AARCH64_TLSLE_LDST128_TPREL_LO12":        true,
		"R_AARCH64_TLSLE_LDST128_TPREL_LO12_NC":     true,
		"R_AARCH64_TLSLE_MOVW_TPREL_G0":             true,
		"R_AARCH64_TLSLE_MOVW_TPREL_G0_NC":          true,
		"R_AARCH64_TLSLE_MOVW_TPREL_G1":             true,
		"R_AARCH64_TLSLE_MOVW_TPREL_G1_NC":          true,
		"R_AARCH64_TLSLE_MOVW_TPREL_G2":             true,
		"R_AARCH64_TLS_DTPMOD64":                    true,
		"R_AARCH64_TLS_DTPREL64":                    true,
		"R_AARCH64_TLS_TPREL64":                     true,
		"R_AARCH64_TSTBR14":                         true,
		"R_ALPHA":                                   true,
		"R_ALPHA_BRADDR":                            true,
		"R_ALPHA_COPY":                              true,
		"R_ALPHA_GLOB_DAT":                          true,
		"R_ALPHA_GPDISP":                            true,
		"R_ALPHA_GPREL32":                           true,
		"R_ALPHA_GPRELHIGH":                         true,
		"R_ALPHA_GPRELLOW":                          true,
		"R_ALPHA_GPVALUE":                           true,
		"R_ALPHA_HINT":                              true,
		"R_ALPHA_IMMED_BR_HI32":                     true,
		"R_ALPHA_IMMED_GP_16":                       true,
		"R_ALPHA_IMMED_GP_HI32":                     true,
		"R_ALPHA_IMMED_LO32":                        true,
		"R_ALPHA_IMMED_SCN_HI32":                    true,
		"R_ALPHA_JMP_SLOT":                          true,
		"R_ALPHA_LITERAL":                           true,
		"R_ALPHA_LITUSE":                            true,
		"R_ALPHA_NONE":                              true,
		"R_ALPHA_OP_PRSHIFT":                        true,
		"R_ALPHA_OP_PSUB":                           true,
		"R_ALPHA_OP_PUSH":                           true,
		"R_ALPHA_OP_STORE":                          true,
		"R_ALPHA_REFLONG":                           true,
		"R_ALPHA_REFQUAD":                           true,
		"R_ALPHA_RELATIVE":                          true,
		"R_ALPHA_SREL16":                            true,
		"R_ALPHA_SREL32":                            true,
		"R_ALPHA_SREL64":                            true,
		"R_ARM":                                     true,
		"R_ARM_ABS12":                               true,
		"R_ARM_ABS16":                               true,
		"R_ARM_ABS32":                               true,
		"R_ARM_ABS32_NOI":                           true,
		"R_ARM_ABS8":                                true,
		"R_ARM_ALU_PCREL_15_8":                      true,
		"R_ARM_ALU_PCREL_23_15":                     true,
		"R_ARM_ALU_PCREL_7_0":                       true,
		"R_ARM_ALU_PC_G0":                           true,
		"R_ARM_ALU_PC_G0_NC":                        true,
		"R_ARM_ALU_PC_G1":                           true,
		"R_ARM_ALU_PC_G1_NC":                        true,
		"R_ARM_ALU_PC_G2":                           true,
		"R_ARM_ALU_SBREL_19_12_NC":                  true,
		"R_ARM_ALU_SBREL_27_20_CK":                  true,
		"R_ARM_ALU_SB_G0":                           true,
		"R_ARM_ALU_SB_G0_NC":                        true,
		"R_ARM_ALU_SB_G1":                           true,
		"R_ARM_ALU_SB_G1_NC":                        true,
		"R_ARM_ALU_SB_G2":                           true,
		"R_ARM_AMP_VCALL9":                          true,
		"R_ARM_BASE_ABS":                            true,
		"R_ARM_CALL":                                true,
		"R_ARM_COPY":                                true,
		"R_ARM_GLOB_DAT":                            true,
		"R_ARM_GNU_VTENTRY":                         true,
		"R_ARM_GNU_VTINHERIT":                       true,
		"R_ARM_GOT32":                               true,
		"R_ARM_GOTOFF":                              true,
		"R_ARM_GOTOFF12":                            true,
		"R_ARM_GOTPC":                               true,
		"R_ARM_GOTRELAX":                            true,
		"R_ARM_GOT_ABS":                             true,
		"R_ARM_GOT_BREL12":                          true,
		"R_ARM_GOT_PREL":                            true,
		"R_ARM_IRELATIVE":                           true,
		"R_ARM_JUMP24":                              true,
		"R_ARM_JUMP_SLOT":                           true,
		"R_ARM_LDC_PC_G0":                           true,
		"R_ARM_LDC_PC_G1":                           true,
		"R_ARM_LDC_PC_G2":                           true,
		"R_ARM_LDC_SB_G0":                           true,
		"R_ARM_LDC_SB_G1":                           true,
		"R_ARM_LDC_SB_G2":                           true,
		"R_ARM_LDRS_PC_G0":                          true,
		"R_ARM_LDRS_PC_G1":                          true,
		"R_ARM_LDRS_PC_G2":                          true,
		"R_ARM_LDRS_SB_G0":                          true,
		"R_ARM_LDRS_SB_G1":                          true,
		"R_ARM_LDRS_SB_G2":                          true,
		"R_ARM_LDR_PC_G1":                           true,
		"R_ARM_LDR_PC_G2":                           true,
		"R_ARM_LDR_SBREL_11_10_NC":                  true,
		"R_ARM_LDR_SB_G0":                           true,
		"R_ARM_LDR_SB_G1":                           true,
		"R_ARM_LDR_SB_G2":                           true,
		"R_ARM_ME_TOO":                              true,
		"R_ARM_MOVT_ABS":                            true,
		"R_ARM_MOVT_BREL":                           true,
		"R_ARM_MOVT_PREL":                           true,
		"R_ARM_MOVW_ABS_NC":                         true,
		"R_ARM_MOVW_BREL":                           true,
		"R_ARM_MOVW_BREL_NC":                        true,
		"R_ARM_MOVW_PREL_NC":                        true,
		"R_ARM_NONE":                                true,
		"R_ARM_PC13":                                true,
		"R_ARM_PC24":                                true,
		"R_ARM_PLT32":                               true,
		"R_ARM_PLT32_ABS":                           true,
		"R_ARM_PREL31":                              true,
		"R_ARM_PRIVATE_0":                           true,
		"R_ARM_PRIVATE_1":                           true,
		"R_ARM_PRIVATE_10":                          true,
		"R_ARM_PRIVATE_11":                          true,
		"R_ARM_PRIVATE_12":                          true,
		"R_ARM_PRIVATE_13":                          true,
		"R_ARM_PRIVATE_14":                          true,
		"R_ARM_PRIVATE_15":                          true,
		"R_ARM_PRIVATE_2":                           true,
		"R_ARM_PRIVATE_3":                           true,
		"R_ARM_PRIVATE_4":                           true,
		"R_ARM_PRIVATE_5":                           true,
		"R_ARM_PRIVATE_6":                           true,
		"R_ARM_PRIVATE_7":                           true,
		"R_ARM_PRIVATE_8":                           true,
		"R_ARM_PRIVATE_9":                           true,
		"R_ARM_RABS32":                              true,
		"R_ARM_RBASE":                               true,
		"R_ARM_REL32":                               true,
		"R_ARM_REL32_NOI":                           true,
		"R_ARM_RELATIVE":                            true,
		"R_ARM_RPC24":                               true,
		"R_ARM_RREL32":                              true,
		"R_ARM_RSBREL32":                            true,
		"R_ARM_RXPC25":                              true,
		"R_ARM_SBREL31":                             true,
		"R_ARM_SBREL32":                             true,
		"R_ARM_SWI24":                               true,
		"R_ARM_TARGET1":                             true,
		"R_ARM_TARGET2":                             true,
		"R_ARM_THM_ABS5":                            true,
		"R_ARM_THM_ALU_ABS_G0_NC":                   true,
		"R_ARM_THM_ALU_ABS_G1_NC":                   true,
		"R_ARM_THM_ALU_ABS_G2_NC":                   true,
		"R_ARM_THM_ALU_ABS_G3":                      true,
		"R_ARM_THM_ALU_PREL_11_0":                   true,
		"R_ARM_THM_GOT_BREL12":                      true,
		"R_ARM_THM_JUMP11":                          true,
		"R_ARM_THM_JUMP19":                          true,
		"R_ARM_THM_JUMP24":                          true,
		"R_ARM_THM_JUMP6":                           true,
		"R_ARM_THM_JUMP8":                           true,
		"R_ARM_THM_MOVT_ABS":                        true,
		"R_ARM_THM_MOVT_BREL":                       true,
		"R_ARM_THM_MOVT_PREL":                       true,
		"R_ARM_THM_MOVW_ABS_NC":                     true,
		"R_ARM_THM_MOVW_BREL":                       true,
		"R_ARM_THM_MOVW_BREL_NC":                    true,
		"R_ARM_THM_MOVW_PREL_NC":                    true,
		"R_ARM_THM_PC12":                            true,
		"R_ARM_THM_PC22":                            true,
		"R_ARM_THM_PC8":                             true,
		"R_ARM_THM_RPC22":                           true,
		"R_ARM_THM_SWI8":                            true,
		"R_ARM_THM_TLS_CALL":                        true,
		"R_ARM_THM_TLS_DESCSEQ16":                   true,
		"R_ARM_THM_TLS_DESCSEQ32":                   true,
		"R_ARM_THM_XPC22":                           true,
		"R_ARM_TLS_CALL":                            true,
		"R_ARM_TLS_DESCSEQ":                         true,
		"R_ARM_TLS_DTPMOD32":                        true,
		"R_ARM_TLS_DTPOFF32":                        true,
		"R_ARM_TLS_GD32":                            true,
		"R_ARM_TLS_GOTDESC":                         true,
		"R_ARM_TLS_IE12GP":                          true,
		"R_ARM_TLS_IE32":                            true,
		"R_ARM_TLS_LDM32":                           true,
		"R_ARM_TLS_LDO12":                           true,
		"R_ARM_TLS_LDO32":                           true,
		"R_ARM_TLS_LE12":                            true,
		"R_ARM_TLS_LE32":                            true,
		"R_ARM_TLS_TPOFF32":                         true,
		"R_ARM_V4BX":                                true,
		"R_ARM_XPC25":                               true,
		"R_INFO":                                    true,
		"R_INFO32":                                  true,
		"R_MIPS":                                    true,
		"R_MIPS_16":                                 true,
		"R_MIPS_26":                                 true,
		"R_MIPS_32":                                 true,
		"R_MIPS_64":                                 true,
		"R_MIPS_ADD_IMMEDIATE":                      true,
		"R_MIPS_CALL16":                             true,
		"R_MIPS_CALL_HI16":                          true,
		"R_MIPS_CALL_LO16":                          true,
		"R_MIPS_DELETE":                             true,
		"R_MIPS_GOT16":                              true,
		"R_MIPS_GOT_DISP":                           true,
		"R_MIPS_GOT_HI16":                           true,
		"R_MIPS_GOT_LO16":                           true,
		"R_MIPS_GOT_OFST":                           true,
		"R_MIPS_GOT_PAGE":                           true,
		"R_MIPS_GPREL16":                            true,
		"R_MIPS_GPREL32":                            true,
		"R_MIPS_HI16":                               true,
		"R_MIPS_HIGHER":                             true,
		"R_MIPS_HIGHEST":                            true,
		"R_MIPS_INSERT_A":                           true,
		"R_MIPS_INSERT_B":                           true,
		"R_MIPS_JALR":                               true,
		"R_MIPS_LITERAL":                            true,
		"R_MIPS_LO16":                               true,
		"R_MIPS_NONE":                               true,
		"R_MIPS_PC16":                               true,
		"R_MIPS_PJUMP":                              true,
		"R_MIPS_REL16":                              true,
		"R_MIPS_REL32":                              true,
		"R_MIPS_RELGOT":                             true,
		"R_MIPS_SCN_DISP":                           true,
		"R_MIPS_SHIFT5":                             true,
		"R_MIPS_SHIFT6":                             true,
		"R_MIPS_SUB":                                true,
		"R_MIPS_TLS_DTPMOD32":                       true,
		"R_MIPS_TLS_DTPMOD64":                       true,
		"R_MIPS_TLS_DTPREL32":                       true,
		"R_MIPS_TLS_DTPREL64":                       true,
		"R_MIPS_TLS_DTPREL_HI16":                    true,
		"R_MIPS_TLS_DTPREL_LO16":                    true,
		"R_MIPS_TLS_GD":                             true,
		"R_MIPS_TLS_GOTTPREL":                       true,
		"R_MIPS_TLS_LDM":                            true,
		"R_MIPS_TLS_TPREL32":                        true,
		"R_MIPS_TLS_TPREL64":                        true,
		"R_MIPS_TLS_TPREL_HI16":                     true,
		"R_MIPS_TLS_TPREL_LO16":                     true,
		"R_PPC":                                     true,
		"R_PPC64":                                   true,
		"R_PPC64_ADDR14":                            true,
		"R_PPC64_ADDR14_BRNTAKEN":                   true,
		"R_PPC64_ADDR14_BRTAKEN":                    true,
		"R_PPC64_ADDR16":                            true,
		"R_PPC64_ADDR16_DS":                         true,
		"R_PPC64_ADDR16_HA":                         true,
		"R_PPC64_ADDR16_HI":                         true,
		"R_PPC64_ADDR16_HIGH":                       true,
		"R_PPC64_ADDR16_HIGHA":                      true,
		"R_PPC64_ADDR16_HIGHER":                     true,
		"R_PPC64_ADDR16_HIGHERA":                    true,
		"R_PPC64_ADDR16_HIGHEST":                    true,
		"R_PPC64_ADDR16_HIGHESTA":                   true,
		"R_PPC64_ADDR16_LO":                         true,
		"R_PPC64_ADDR16_LO_DS":                      true,
		"R_PPC64_ADDR24":                            true,
		"R_PPC64_ADDR32":                            true,
		"R_PPC64_ADDR64":                            true,
		"R_PPC64_ADDR64_LOCAL":                      true,
		"R_PPC64_DTPMOD64":                          true,
		"R_PPC64_DTPREL16":                          true,
		"R_PPC64_DTPREL16_DS":                       true,
		"R_PPC64_DTPREL16_HA":                       true,
		"R_PPC64_DTPREL16_HI":                       true,
		"R_PPC64_DTPREL16_HIGH":                     true,
		"R_PPC64_DTPREL16_HIGHA":                    true,
		"R_PPC64_DTPREL16_HIGHER":                   true,
		"R_PPC64_DTPREL16_HIGHERA":                  true,
		"R_PPC64_DTPREL16_HIGHEST":                  true,
		"R_PPC64_DTPREL16_HIGHESTA":                 true,
		"R_PPC64_DTPREL16_LO":                       true,
		"R_PPC64_DTPREL16_LO_DS":                    true,
		"R_PPC64_DTPREL64":                          true,
		"R_PPC64_ENTRY":                             true,
		"R_PPC64_GOT16":                             true,
		"R_PPC64_GOT16_DS":                          true,
		"R_PPC64_GOT16_HA":                          true,
		"R_PPC64_GOT16_HI":                          true,
		"R_PPC64_GOT16_LO":                          true,
		"R_PPC64_GOT16_LO_DS":                       true,
		"R_PPC64_GOT_DTPREL16_DS":                   true,
		"R_PPC64_GOT_DTPREL16_HA":                   true,
		"R_PPC64_GOT_DTPREL16_HI":                   true,
		"R_PPC64_GOT_DTPREL16_LO_DS":                true,
		"R_PPC64_GOT_TLSGD16":                       true,
		"R_PPC64_GOT_TLSGD16_HA":                    true,
		"R_PPC64_GOT_TLSGD16_HI":                    true,
		"R_PPC64_GOT_TLSGD16_LO":                    true,
		"R_PPC64_GOT_TLSLD16":                       true,
		"R_PPC64_GOT_TLSLD16_HA":                    true,
		"R_PPC64_GOT_TLSLD16_HI":                    true,
		"R_PPC64_GOT_TLSLD16_LO":                    true,
		"R_PPC64_GOT_TPREL16_DS":                    true,
		"R_PPC64_GOT_TPREL16_HA":                    true,
		"R_PPC64_GOT_TPREL16_HI":                    true,
		"R_PPC64_GOT_TPREL16_LO_DS":                 true,
		"R_PPC64_IRELATIVE":                         true,
		"R_PPC64_JMP_IREL":                          true,
		"R_PPC64_JMP_SLOT":                          true,
		"R_PPC64_NONE":                              true,
		"R_PPC64_PLT16_LO_DS":                       true,
		"R_PPC64_PLTGOT16":                          true,
		"R_PPC64_PLTGOT16_DS":                       true,
		"R_PPC64_PLTGOT16_HA":                       true,
		"R_PPC64_PLTGOT16_HI":                       true,
		"R_PPC64_PLTGOT16_LO":                       true,
		"R_PPC64_PLTGOT_LO_DS":                      true,
		"R_PPC64_REL14":                             true,
		"R_PPC64_REL14_BRNTAKEN":                    true,
		"R_PPC64_REL14_BRTAKEN":                     true,
		"R_PPC64_REL16":                             true,
		"R_PPC64_REL16DX_HA":                        true,
		"R_PPC64_REL16_HA":                          true,
		"R_PPC64_REL16_HI":                          true,
		"R_PPC64_REL16_LO":                          true,
		"R_PPC64_REL24":                             true,
		"R_PPC64_REL24_NOTOC":                       true,
		"R_PPC64_REL32":                             true,
		"R_PPC64_REL64":                             true,
		"R_PPC64_SECTOFF_DS":                        true,
		"R_PPC64_SECTOFF_LO_DS":                     true,
		"R_PPC64_TLS":                               true,
		"R_PPC64_TLSGD":                             true,
		"R_PPC64_TLSLD":                             true,
		"R_PPC64_TOC":                               true,
		"R_PPC64_TOC16":                             true,
		"R_PPC64_TOC16_DS":                          true,
		"R_PPC64_TOC16_HA":                          true,
		"R_PPC64_TOC16_HI":                          true,
		"R_PPC64_TOC16_LO":                          true,
		"R_PPC64_TOC16_LO_DS":                       true,
		"R_PPC64_TOCSAVE":                           true,
		"R_PPC64_TPREL16":                           true,
		"R_PPC64_TPREL16_DS":                        true,
		"R_PPC64_TPREL16_HA":                        true,
		"R_PPC64_TPREL16_HI":                        true,
		"R_PPC64_TPREL16_HIGH":                      true,
		"R_PPC64_TPREL16_HIGHA":                     true,
		"R_PPC64_TPREL16_HIGHER":                    true,
		"R_PPC64_TPREL16_HIGHERA":                   true,
		"R_PPC64_TPREL16_HIGHEST":                   true,
		"R_PPC64_TPREL16_HIGHESTA":                  true,
		"R_PPC64_TPREL16_LO":                        true,
		"R_PPC64_TPREL16_LO_DS":                     true,
		"R_PPC64_TPREL64":                           true,
		"R_PPC_ADDR14":                              true,
		"R_PPC_ADDR14_BRNTAKEN":                     true,
		"R_PPC_ADDR14_BRTAKEN":                      true,
		"R_PPC_ADDR16":                              true,
		"R_PPC_ADDR16_HA":                           true,
		"R_PPC_ADDR16_HI":                           true,
		"R_PPC_ADDR16_LO":                           true,
		"R_PPC_ADDR24":                              true,
		"R_PPC_ADDR32":                              true,
		"R_PPC_COPY":                                true,
		"R_PPC_DTPMOD32":                            true,
		"R_PPC_DTPREL16":                            true,
		"R_PPC_DTPREL16_HA":                         true,
		"R_PPC_DTPREL16_HI":                         true,
		"R_PPC_DTPREL16_LO":                         true,
		"R_PPC_DTPREL32":                            true,
		"R_PPC_EMB_BIT_FLD":                         true,
		"R_PPC_EMB_MRKREF":                          true,
		"R_PPC_EMB_NADDR16":                         true,
		"R_PPC_EMB_NADDR16_HA":                      true,
		"R_PPC_EMB_NADDR16_HI":                      true,
		"R_PPC_EMB_NADDR16_LO":                      true,
		"R_PPC_EMB_NADDR32":                         true,
		"R_PPC_EMB_RELSDA":                          true,
		"R_PPC_EMB_RELSEC16":                        true,
		"R_PPC_EMB_RELST_HA":                        true,
		"R_PPC_EMB_RELST_HI":                        true,
		"R_PPC_EMB_RELST_LO":                        true,
		"R_PPC_EMB_SDA21":                           true,
		"R_PPC_EMB_SDA2I16":                         true,
		"R_PPC_EMB_SDA2REL":                         true,
		"R_PPC_EMB_SDAI16":                          true,
		"R_PPC_GLOB_DAT":                            true,
		"R_PPC_GOT16":                               true,
		"R_PPC_GOT16_HA":                            true,
		"R_PPC_GOT16_HI":                            true,
		"R_PPC_GOT16_LO":                            true,
		"R_PPC_GOT_TLSGD16":                         true,
		"R_PPC_GOT_TLSGD16_HA":                      true,
		"R_PPC_GOT_TLSGD16_HI":                      true,
		"R_PPC_GOT_TLSGD16_LO":                      true,
		"R_PPC_GOT_TLSLD16":                         true,
		"R_PPC_GOT_TLSLD16_HA":                      true,
		"R_PPC_GOT_TLSLD16_HI":                      true,
		"R_PPC_GOT_TLSLD16_LO":                      true,
		"R_PPC_GOT_TPREL16":                         true,
		"R_PPC_GOT_TPREL16_HA":                      true,
		"R_PPC_GOT_TPREL16_HI":                      true,
		"R_PPC_GOT_TPREL16_LO":                      true,
		"R_PPC_JMP_SLOT":                            true,
		"R_PPC_LOCAL24PC":                           true,
		"R_PPC_NONE":                                true,
		"R_PPC_PLT16_HA":                            true,
		"R_PPC_PLT16_HI":                            true,
		"R_PPC_PLT16_LO":                            true,
		"R_PPC_PLT32":                               true,
		"R_PPC_PLTREL24":                            true,
		"R_PPC_PLTREL32":                            true,
		"R_PPC_REL14":                               true,
		"R_PPC_REL14_BRNTAKEN":                      true,
		"R_PPC_REL14_BRTAKEN":                       true,
		"R_PPC_REL24":                               true,
		"R_PPC_REL32":                               true,
		"R_PPC_RELATIVE":                            true,
		"R_PPC_SDAREL16":                            true,
		"R_PPC_SECTOFF":                             true,
		"R_PPC_SECTOFF_HA":                          true,
		"R_PPC_SECTOFF_HI":                          true,
		"R_PPC_SECTOFF_LO":                          true,
		"R_PPC_TLS":                                 true,
		"R_PPC_TPREL16":                             true,
		"R_PPC_TPREL16_HA":                          true,
		"R_PPC_TPREL16_HI":                          true,
		"R_PPC_TPREL16_LO":                          true,
		"R_PPC_TPREL32":                             true,
		"R_PPC_UADDR16":                             true,
		"R_PPC_UADDR32":                             true,
		"R_RISCV":                                   true,
		"R_RISCV_32":                                true,
		"R_RISCV_32_PCREL":                          true,
		"R_RISCV_64":                                true,
		"R_RISCV_ADD16":                             true,
		"R_RISCV_ADD32":                             true,
		"R_RISCV_ADD64":                             true,
		"R_RISCV_ADD8":                              true,
		"R_RISCV_ALIGN":                             true,
		"R_RISCV_BRANCH":                            true,
		"R_RISCV_CALL":                              true,
		"R_RISCV_CALL_PLT":                          true,
		"R_RISCV_COPY":                              true,
		"R_RISCV_GNU_VTENTRY":                       true,
		"R_RISCV_GNU_VTINHERIT":                     true,
		"R_RISCV_GOT_HI20":                          true,
		"R_RISCV_GPREL_I":                           true,
		"R_RISCV_GPREL_S":                           true,
		"R_RISCV_HI20":                              true,
		"R_RISCV_JAL":                               true,
		"R_RISCV_JUMP_SLOT":                         true,
		"R_RISCV_LO12_I":                            true,
		"R_RISCV_LO12_S":                            true,
		"R_RISCV_NONE":                              true,
		"R_RISCV_PCREL_HI20":                        true,
		"R_RISCV_PCREL_LO12_I":                      true,
		"R_RISCV_PCREL_LO12_S":                      true,
		"R_RISCV_RELATIVE":                          true,
		"R_RISCV_RELAX":                             true,
		"R_RISCV_RVC_BRANCH":                        true,
		"R_RISCV_RVC_JUMP":                          true,
		"R_RISCV_RVC_LUI":                           true,
		"R_RISCV_SET16":                             true,
		"R_RISCV_SET32":                             true,
		"R_RISCV_SET6":                              true,
		"R_RISCV_SET8":                              true,
		"R_RISCV_SUB16":                             true,
		"R_RISCV_SUB32":                             true,
		"R_RISCV_SUB6":                              true,
		"R_RISCV_SUB64":                             true,
		"R_RISCV_SUB8":                              true,
		"R_RISCV_TLS_DTPMOD32":                      true,
		"R_RISCV_TLS_DTPMOD64":                      true,
		"R_RISCV_TLS_DTPREL32":                      true,
		"R_RISCV_TLS_DTPREL64":                      true,
		"R_RISCV_TLS_GD_HI20":                       true,
		"R_RISCV_TLS_GOT_HI20":                      true,
		"R_RISCV_TLS_TPREL32":                       true,
		"R_RISCV_TLS_TPREL64":                       true,
		"R_RISCV_TPREL_ADD":                         true,
		"R_RISCV_TPREL_HI20":                        true,
		"R_RISCV_TPREL_I":                           true,
		"R_RISCV_TPREL_LO12_I":                      true,
		"R_RISCV_TPREL_LO12_S":                      true,
		"R_RISCV_TPREL_S":                           true,
		"R_SPARC":                                   true,
		"R_SPARC_10":                                true,
		"R_SPARC_11":                                true,
		"R_SPARC_13":                                true,
		"R_SPARC_16":                                true,
		"R_SPARC_22":                                true,
		"R_SPARC_32":                                true,
		"R_SPARC_5":                                 true,
		"R_SPARC_6":                                 true,
		"R_SPARC_64":                                true,
		"R_SPARC_7":                                 true,
		"R_SPARC_8":                                 true,
		"R_SPARC_COPY":                              true,
		"R_SPARC_DISP16":                            true,
		"R_SPARC_DISP32":                            true,
		"R_SPARC_DISP64":                            true,
		"R_SPARC_DISP8":                             true,
		"R_SPARC_GLOB_DAT":                          true,
		"R_SPARC_GLOB_JMP":                          true,
		"R_SPARC_GOT10":                             true,
		"R_SPARC_GOT13":                             true,
		"R_SPARC_GOT22":                             true,
		"R_SPARC_H44":                               true,
		"R_SPARC_HH22":                              true,
		"R_SPARC_HI22":                              true,
		"R_SPARC_HIPLT22":                           true,
		"R_SPARC_HIX22":                             true,
		"R_SPARC_HM10":                              true,
		"R_SPARC_JMP_SLOT":                          true,
		"R_SPARC_L44":                               true,
		"R_SPARC_LM22":                              true,
		"R_SPARC_LO10":                              true,
		"R_SPARC_LOPLT10":                           true,
		"R_SPARC_LOX10":                             true,
		"R_SPARC_M44":                               true,
		"R_SPARC_NONE":                              true,
		"R_SPARC_OLO10":                             true,
		"R_SPARC_PC10":                              true,
		"R_SPARC_PC22":                              true,
		"R_SPARC_PCPLT10":                           true,
		"R_SPARC_PCPLT22":                           true,
		"R_SPARC_PCPLT32":                           true,
		"R_SPARC_PC_HH22":                           true,
		"R_SPARC_PC_HM10":                           true,
		"R_SPARC_PC_LM22":                           true,
		"R_SPARC_PLT32":                             true,
		"R_SPARC_PLT64":                             true,
		"R_SPARC_REGISTER":                          true,
		"R_SPARC_RELATIVE":                          true,
		"R_SPARC_UA16":                              true,
		"R_SPARC_UA32":                              true,
		"R_SPARC_UA64":                              true,
		"R_SPARC_WDISP16":                           true,
		"R_SPARC_WDISP19":                           true,
		"R_SPARC_WDISP22":                           true,
		"R_SPARC_WDISP30":                           true,
		"R_SPARC_WPLT30":                            true,
		"R_SYM32":                                   true,
		"R_SYM64":                                   true,
		"R_TYPE32":                                  true,
		"R_TYPE64":                                  true,
		"R_X86_64":                                  true,
		"R_X86_64_16":                               true,
		"R_X86_64_32":                               true,
		"R_X86_64_32S":                              true,
		"R_X86_64_64":                               true,
		"R_X86_64_8":                                true,
		"R_X86_64_COPY":                             true,
		"R_X86_64_DTPMOD64":                         true,
		"R_X86_64_DTPOFF32":                         true,
		"R_X86_64_DTPOFF64":                         true,
		"R_X86_64_GLOB_DAT":                         true,
		"R_X86_64_GOT32":                            true,
		"R_X86_64_GOT64":                            true,
		"R_X86_64_GOTOFF64":                         true,
		"R_X86_64_GOTPC32":                          true,
		"R_X86_64_GOTPC32_TLSDESC":                  true,
		"R_X86_64_GOTPC64":                          true,
		"R_X86_64_GOTPCREL":                         true,
		"R_X86_64_GOTPCREL64":                       true,
		"R_X86_64_GOTPCRELX":                        true,
		"R_X86_64_GOTPLT64":                         true,
		"R_X86_64_GOTTPOFF":                         true,
		"R_X86_64_IRELATIVE":                        true,
		"R_X86_64_JMP_SLOT":                         true,
		"R_X86_64_NONE":                             true,
		"R_X86_64_PC16":                             true,
		"R_X86_64_PC32":                             true,
		"R_X86_64_PC32_BND":                         true,
		"R_X86_64_PC64":                             true,
		"R_X86_64_PC8":                              true,
		"R_X86_64_PLT32":                            true,
		"R_X86_64_PLT32_BND":                        true,
		"R_X86_64_PLTOFF64":                         true,
		"R_X86_64_RELATIVE":                         true,
		"R_X86_64_RELATIVE64":                       true,
		"R_X86_64_REX_GOTPCRELX":                    true,
		"R_X86_64_SIZE32":                           true,
		"R_X86_64_SIZE64":                           true,
		"R_X86_64_TLSDESC":                          true,
		"R_X86_64_TLSDESC_CALL":                     true,
		"R_X86_64_TLSGD":                            true,
		"R_X86_64_TLSLD":                            true,
		"R_X86_64_TPOFF32":                          true,
		"R_X86_64_TPOFF64":                          true,
		"Rel32":                                     true,
		"Rel64":                                     true,
		"Rela32":                                    true,
		"Rela64":                                    true,
		"SHF_ALLOC":                                 true,
		"SHF_COMPRESSED":                            true,
		"SHF_EXECINSTR":                             true,
		"SHF_GROUP":                                 true,
		"SHF_INFO_LINK":                             true,
		"SHF_LINK_ORDER":                            true,
		"SHF_MASKOS":                                true,
		"SHF_MASKPROC":                              true,
		"SHF_MERGE":                                 true,
		"SHF_OS_NONCONFORMING":                      true,
		"SHF_STRINGS":                               true,
		"SHF_TLS":                                   true,
		"SHF_WRITE":                                 true,
		"SHN_ABS":                                   true,
		"SHN_COMMON":                                true,
		"SHN_HIOS":                                  true,
		"SHN_HIPROC":                                true,
		"SHN_HIRESERVE":                             true,
		"SHN_LOOS":                                  true,
		"SHN_LOPROC":                                true,
		"SHN_LORESERVE":                             true,
		"SHN_UNDEF":                                 true,
		"SHN_XINDEX":                                true,
		"SHT_DYNAMIC":                               true,
		"SHT_DYNSYM":                                true,
		"SHT_FINI_ARRAY":                            true,
		"SHT_GNU_ATTRIBUTES":                        true,
		"SHT_GNU_HASH":                              true,
		"SHT_GNU_LIBLIST":                           true,
		"SHT_GNU_VERDEF":                            true,
		"SHT_GNU_VERNEED":                           true,
		"SHT_GNU_VERSYM":                            true,
		"SHT_GROUP":                                 true,
		"SHT_HASH":                                  true,
		"SHT_HIOS":                                  true,
		"SHT_HIPROC":                                true,
		"SHT_HIUSER":                                true,
		"SHT_INIT_ARRAY":                            true,
		"SHT_LOOS":                                  true,
		"SHT_LOPROC":                                true,
		"SHT_LOUSER":                                true,
		"SHT_NOBITS":                                true,
		"SHT_NOTE":                                  true,
		"SHT_NULL":                                  true,
		"SHT_PREINIT_ARRAY":                         true,
		"SHT_PROGBITS":                              true,
		"SHT_REL":                                   true,
		"SHT_RELA":                                  true,
		"SHT_SHLIB":                                 true,
		"SHT_STRTAB":                                true,
		"SHT_SYMTAB":                                true,
		"SHT_SYMTAB_SHNDX":                          true,
		"STB_GLOBAL":                                true,
		"STB_HIOS":                                  true,
		"STB_HIPROC":                                true,
		"STB_LOCAL":                                 true,
		"STB_LOOS":                                  true,
		"STB_LOPROC":                                true,
		"STB_WEAK":                                  true,
		"STT_COMMON":                                true,
		"STT_FILE":                                  true,
		"STT_FUNC":                                  true,
		"STT_HIOS":                                  true,
		"STT_HIPROC":                                true,
		"STT_LOOS":                                  true,
		"STT_LOPROC":                                true,
		"STT_NOTYPE":                                true,
		"STT_OBJECT":                                true,
		"STT_SECTION":                               true,
		"STT_TLS":                                   true,
		"STV_DEFAULT":                               true,
		"STV_HIDDEN":                                true,
		"STV_INTERNAL":                              true,
		"STV_PROTECTED":                             true,
		"ST_BIND":                                   true,
		"ST_INFO":                                   true,
		"ST_TYPE":                                   true,
		"ST_VISIBILITY":                             true,
		"Section":                                   true,
		"Section32":                                 true,
		"Section64":                                 true,
		"SectionFlag":                               true,
		"SectionHeader":                             true,
		"SectionIndex":                              true,
		"SectionType":                               true,
		"Sym32":                                     true,
		"Sym32Size":                                 true,
		"Sym64":                                     true,
		"Sym64Size":                                 true,
		"SymBind":                                   true,
		"SymType":                                   true,
		"SymVis":                                    true,
		"Symbol":                                    true,
		"Type":                                      true,
		"Version":                                   true,
	},
	"debug/gosym": map[string]bool{
		"DecodingError":    true,
		"Func":             true,
		"LineTable":        true,
		"NewLineTable":     true,
		"NewTable":         true,
		"Obj":              true,
		"Sym":              true,
		"Table":            true,
		"UnknownFileError": true,
		"UnknownLineError": true,
	},
	"debug/macho": map[string]bool{
		"ARM64_RELOC_ADDEND":              true,
		"ARM64_RELOC_BRANCH26":            true,
		"ARM64_RELOC_GOT_LOAD_PAGE21":     true,
		"ARM64_RELOC_GOT_LOAD_PAGEOFF12":  true,
		"ARM64_RELOC_PAGE21":              true,
		"ARM64_RELOC_PAGEOFF12":           true,
		"ARM64_RELOC_POINTER_TO_GOT":      true,
		"ARM64_RELOC_SUBTRACTOR":          true,
		"ARM64_RELOC_TLVP_LOAD_PAGE21":    true,
		"ARM64_RELOC_TLVP_LOAD_PAGEOFF12": true,
		"ARM64_RELOC_UNSIGNED":            true,
		"ARM_RELOC_BR24":                  true,
		"ARM_RELOC_HALF":                  true,
		"ARM_RELOC_HALF_SECTDIFF":         true,
		"ARM_RELOC_LOCAL_SECTDIFF":        true,
		"ARM_RELOC_PAIR":                  true,
		"ARM_RELOC_PB_LA_PTR":             true,
		"ARM_RELOC_SECTDIFF":              true,
		"ARM_RELOC_VANILLA":               true,
		"ARM_THUMB_32BIT_BRANCH":          true,
		"ARM_THUMB_RELOC_BR22":            true,
		"Cpu":                             true,
		"Cpu386":                          true,
		"CpuAmd64":                        true,
		"CpuArm":                          true,
		"CpuArm64":                        true,
		"CpuPpc":                          true,
		"CpuPpc64":                        true,
		"Dylib":                           true,
		"DylibCmd":                        true,
		"Dysymtab":                        true,
		"DysymtabCmd":                     true,
		"ErrNotFat":                       true,
		"FatArch":                         true,
		"FatArchHeader":                   true,
		"FatFile":                         true,
		"File":                            true,
		"FileHeader":                      true,
		"FlagAllModsBound":                true,
		"FlagAllowStackExecution":         true,
		"FlagAppExtensionSafe":            true,
		"FlagBindAtLoad":                  true,
		"FlagBindsToWeak":                 true,
		"FlagCanonical":                   true,
		"FlagDeadStrippableDylib":         true,
		"FlagDyldLink":                    true,
		"FlagForceFlat":                   true,
		"FlagHasTLVDescriptors":           true,
		"FlagIncrLink":                    true,
		"FlagLazyInit":                    true,
		"FlagNoFixPrebinding":             true,
		"FlagNoHeapExecution":             true,
		"FlagNoMultiDefs":                 true,
		"FlagNoReexportedDylibs":          true,
		"FlagNoUndefs":                    true,
		"FlagPIE":                         true,
		"FlagPrebindable":                 true,
		"FlagPrebound":                    true,
		"FlagRootSafe":                    true,
		"FlagSetuidSafe":                  true,
		"FlagSplitSegs":                   true,
		"FlagSubsectionsViaSymbols":       true,
		"FlagTwoLevel":                    true,
		"FlagWeakDefines":                 true,
		"FormatError":                     true,
		"GENERIC_RELOC_LOCAL_SECTDIFF":    true,
		"GENERIC_RELOC_PAIR":              true,
		"GENERIC_RELOC_PB_LA_PTR":         true,
		"GENERIC_RELOC_SECTDIFF":          true,
		"GENERIC_RELOC_TLV":               true,
		"GENERIC_RELOC_VANILLA":           true,
		"Load":                            true,
		"LoadBytes":                       true,
		"LoadCmd":                         true,
		"LoadCmdDylib":                    true,
		"LoadCmdDylinker":                 true,
		"LoadCmdDysymtab":                 true,
		"LoadCmdRpath":                    true,
		"LoadCmdSegment":                  true,
		"LoadCmdSegment64":                true,
		"LoadCmdSymtab":                   true,
		"LoadCmdThread":                   true,
		"LoadCmdUnixThread":               true,
		"Magic32":                         true,
		"Magic64":                         true,
		"MagicFat":                        true,
		"NewFatFile":                      true,
		"NewFile":                         true,
		"Nlist32":                         true,
		"Nlist64":                         true,
		"Open":                            true,
		"OpenFat":                         true,
		"Regs386":                         true,
		"RegsAMD64":                       true,
		"Reloc":                           true,
		"RelocTypeARM":                    true,
		"RelocTypeARM64":                  true,
		"RelocTypeGeneric":                true,
		"RelocTypeX86_64":                 true,
		"Rpath":                           true,
		"RpathCmd":                        true,
		"Section":                         true,
		"Section32":                       true,
		"Section64":                       true,
		"SectionHeader":                   true,
		"Segment":                         true,
		"Segment32":                       true,
		"Segment64":                       true,
		"SegmentHeader":                   true,
		"Symbol":                          true,
		"Symtab":                          true,
		"SymtabCmd":                       true,
		"Thread":                          true,
		"Type":                            true,
		"TypeBundle":                      true,
		"TypeDylib":                       true,
		"TypeExec":                        true,
		"TypeObj":                         true,
		"X86_64_RELOC_BRANCH":             true,
		"X86_64_RELOC_GOT":                true,
		"X86_64_RELOC_GOT_LOAD":           true,
		"X86_64_RELOC_SIGNED":             true,
		"X86_64_RELOC_SIGNED_1":           true,
		"X86_64_RELOC_SIGNED_2":           true,
		"X86_64_RELOC_SIGNED_4":           true,
		"X86_64_RELOC_SUBTRACTOR":         true,
		"X86_64_RELOC_TLV":                true,
		"X86_64_RELOC_UNSIGNED":           true,
	},
	"debug/pe": map[string]bool{
		"COFFSymbol":                           true,
		"COFFSymbolSize":                       true,
		"DataDirectory":                        true,
		"File":                                 true,
		"FileHeader":                           true,
		"FormatError":                          true,
		"IMAGE_DIRECTORY_ENTRY_ARCHITECTURE":   true,
		"IMAGE_DIRECTORY_ENTRY_BASERELOC":      true,
		"IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT":   true,
		"IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR": true,
		"IMAGE_DIRECTORY_ENTRY_DEBUG":          true,
		"IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT":   true,
		"IMAGE_DIRECTORY_ENTRY_EXCEPTION":      true,
		"IMAGE_DIRECTORY_ENTRY_EXPORT":         true,
		"IMAGE_DIRECTORY_ENTRY_GLOBALPTR":      true,
		"IMAGE_DIRECTORY_ENTRY_IAT":            true,
		"IMAGE_DIRECTORY_ENTRY_IMPORT":         true,
		"IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG":    true,
		"IMAGE_DIRECTORY_ENTRY_RESOURCE":       true,
		"IMAGE_DIRECTORY_ENTRY_SECURITY":       true,
		"IMAGE_DIRECTORY_ENTRY_TLS":            true,
		"IMAGE_FILE_MACHINE_AM33":              true,
		"IMAGE_FILE_MACHINE_AMD64":             true,
		"IMAGE_FILE_MACHINE_ARM":               true,
		"IMAGE_FILE_MACHINE_ARM64":             true,
		"IMAGE_FILE_MACHINE_ARMNT":             true,
		"IMAGE_FILE_MACHINE_EBC":               true,
		"IMAGE_FILE_MACHINE_I386":              true,
		"IMAGE_FILE_MACHINE_IA64":              true,
		"IMAGE_FILE_MACHINE_M32R":              true,
		"IMAGE_FILE_MACHINE_MIPS16":            true,
		"IMAGE_FILE_MACHINE_MIPSFPU":           true,
		"IMAGE_FILE_MACHINE_MIPSFPU16":         true,
		"IMAGE_FILE_MACHINE_POWERPC":           true,
		"IMAGE_FILE_MACHINE_POWERPCFP":         true,
		"IMAGE_FILE_MACHINE_R4000":             true,
		"IMAGE_FILE_MACHINE_SH3":               true,
		"IMAGE_FILE_MACHINE_SH3DSP":            true,
		"IMAGE_FILE_MACHINE_SH4":               true,
		"IMAGE_FILE_MACHINE_SH5":               true,
		"IMAGE_FILE_MACHINE_THUMB":             true,
		"IMAGE_FILE_MACHINE_UNKNOWN":           true,
		"IMAGE_FILE_MACHINE_WCEMIPSV2":         true,
		"ImportDirectory":                      true,
		"NewFile":                              true,
		"Open":                                 true,
		"OptionalHeader32":                     true,
		"OptionalHeader64":                     true,
		"Reloc":                                true,
		"Section":                              true,
		"SectionHeader":                        true,
		"SectionHeader32":                      true,
		"StringTable":                          true,
		"Symbol":                               true,
	},
	"debug/plan9obj": map[string]bool{
		"File":          true,
		"FileHeader":    true,
		"Magic386":      true,
		"Magic64":       true,
		"MagicAMD64":    true,
		"MagicARM":      true,
		"NewFile":       true,
		"Open":          true,
		"Section":       true,
		"SectionHeader": true,
		"Sym":           true,
	},
	"encoding": map[string]bool{
		"BinaryMarshaler":   true,
		"BinaryUnmarshaler": true,
		"TextMarshaler":     true,
		"TextUnmarshaler":   true,
	},
	"encoding/ascii85": map[string]bool{
		"CorruptInputError": true,
		"Decode":            true,
		"Encode":            true,
		"MaxEncodedLen":     true,
		"NewDecoder":        true,
		"NewEncoder":        true,
	},
	"encoding/asn1": map[string]bool{
		"BitString":            true,
		"ClassApplication":     true,
		"ClassContextSpecific": true,
		"ClassPrivate":         true,
		"ClassUniversal":       true,
		"Enumerated":           true,
		"Flag":                 true,
		"Marshal":              true,
		"MarshalWithParams":    true,
		"NullBytes":            true,
		"NullRawValue":         true,
		"ObjectIdentifier":     true,
		"RawContent":           true,
		"RawValue":             true,
		"StructuralError":      true,
		"SyntaxError":          true,
		"TagBitString":         true,
		"TagBoolean":           true,
		"TagEnum":              true,
		"TagGeneralString":     true,
		"TagGeneralizedTime":   true,
		"TagIA5String":         true,
		"TagInteger":           true,
		"TagNull":              true,
		"TagNumericString":     true,
		"TagOID":               true,
		"TagOctetString":       true,
		"TagPrintableString":   true,
		"TagSequence":          true,
		"TagSet":               true,
		"TagT61String":         true,
		"TagUTCTime":           true,
		"TagUTF8String":        true,
		"Unmarshal":            true,
		"UnmarshalWithParams":  true,
	},
	"encoding/base32": map[string]bool{
		"CorruptInputError": true,
		"Encoding":          true,
		"HexEncoding":       true,
		"NewDecoder":        true,
		"NewEncoder":        true,
		"NewEncoding":       true,
		"NoPadding":         true,
		"StdEncoding":       true,
		"StdPadding":        true,
	},
	"encoding/base64": map[string]bool{
		"CorruptInputError": true,
		"Encoding":          true,
		"NewDecoder":        true,
		"NewEncoder":        true,
		"NewEncoding":       true,
		"NoPadding":         true,
		"RawStdEncoding":    true,
		"RawURLEncoding":    true,
		"StdEncoding":       true,
		"StdPadding":        true,
		"URLEncoding":       true,
	},
	"encoding/binary": map[string]bool{
		"BigEndian":      true,
		"ByteOrder":      true,
		"LittleEndian":   true,
		"MaxVarintLen16": true,
		"MaxVarintLen32": true,
		"MaxVarintLen64": true,
		"PutUvarint":     true,
		"PutVarint":      true,
		"Read":           true,
		"ReadUvarint":    true,
		"ReadVarint":     true,
		"Size":           true,
		"Uvarint":        true,
		"Varint":         true,
		"Write":          true,
	},
	"encoding/csv": map[string]bool{
		"ErrBareQuote":     true,
		"ErrFieldCount":    true,
		"ErrQuote":         true,
		"ErrTrailingComma": true,
		"NewReader":        true,
		"NewWriter":        true,
		"ParseError":       true,
		"Reader":           true,
		"Writer":           true,
	},
	"encoding/gob": map[string]bool{
		"CommonType":   true,
		"Decoder":      true,
		"Encoder":      true,
		"GobDecoder":   true,
		"GobEncoder":   true,
		"NewDecoder":   true,
		"NewEncoder":   true,
		"Register":     true,
		"RegisterName": true,
	},
	"encoding/hex": map[string]bool{
		"Decode":           true,
		"DecodeString":     true,
		"DecodedLen":       true,
		"Dump":             true,
		"Dumper":           true,
		"Encode":           true,
		"EncodeToString":   true,
		"EncodedLen":       true,
		"ErrLength":        true,
		"InvalidByteError": true,
		"NewDecoder":       true,
		"NewEncoder":       true,
	},
	"encoding/json": map[string]bool{
		"Compact":               true,
		"Decoder":               true,
		"Delim":                 true,
		"Encoder":               true,
		"HTMLEscape":            true,
		"Indent":                true,
		"InvalidUTF8Error":      true,
		"InvalidUnmarshalError": true,
		"Marshal":               true,
		"MarshalIndent":         true,
		"Marshaler":             true,
		"MarshalerError":        true,
		"NewDecoder":            true,
		"NewEncoder":            true,
		"Number":                true,
		"RawMessage":            true,
		"SyntaxError":           true,
		"Token":                 true,
		"Unmarshal":             true,
		"UnmarshalFieldError":   true,
		"UnmarshalTypeError":    true,
		"Unmarshaler":           true,
		"UnsupportedTypeError":  true,
		"UnsupportedValueError": true,
		"Valid":                 true,
	},
	"encoding/pem": map[string]bool{
		"Block":          true,
		"Decode":         true,
		"Encode":         true,
		"EncodeToMemory": true,
	},
	"encoding/xml": map[string]bool{
		"Attr":                 true,
		"CharData":             true,
		"Comment":              true,
		"CopyToken":            true,
		"Decoder":              true,
		"Directive":            true,
		"Encoder":              true,
		"EndElement":           true,
		"Escape":               true,
		"EscapeText":           true,
		"HTMLAutoClose":        true,
		"HTMLEntity":           true,
		"Header":               true,
		"Marshal":              true,
		"MarshalIndent":        true,
		"Marshaler":            true,
		"MarshalerAttr":        true,
		"Name":                 true,
		"NewDecoder":           true,
		"NewEncoder":           true,
		"NewTokenDecoder":      true,
		"ProcInst":             true,
		"StartElement":         true,
		"SyntaxError":          true,
		"TagPathError":         true,
		"Token":                true,
		"TokenReader":          true,
		"Unmarshal":            true,
		"UnmarshalError":       true,
		"Unmarshaler":          true,
		"UnmarshalerAttr":      true,
		"UnsupportedTypeError": true,
	},
	"errors": map[string]bool{
		"New": true,
	},
	"expvar": map[string]bool{
		"Do":        true,
		"Float":     true,
		"Func":      true,
		"Get":       true,
		"Handler":   true,
		"Int":       true,
		"KeyValue":  true,
		"Map":       true,
		"NewFloat":  true,
		"NewInt":    true,
		"NewMap":    true,
		"NewString": true,
		"Publish":   true,
		"String":    true,
		"Var":       true,
	},
	"flag": map[string]bool{
		"Arg":             true,
		"Args":            true,
		"Bool":            true,
		"BoolVar":         true,
		"CommandLine":     true,
		"ContinueOnError": true,
		"Duration":        true,
		"DurationVar":     true,
		"ErrHelp":         true,
		"ErrorHandling":   true,
		"ExitOnError":     true,
		"Flag":            true,
		"FlagSet":         true,
		"Float64":         true,
		"Float64Var":      true,
		"Getter":          true,
		"Int":             true,
		"Int64":           true,
		"Int64Var":        true,
		"IntVar":          true,
		"Lookup":          true,
		"NArg":            true,
		"NFlag":           true,
		"NewFlagSet":      true,
		"PanicOnError":    true,
		"Parse":           true,
		"Parsed":          true,
		"PrintDefaults":   true,
		"Set":             true,
		"String":          true,
		"StringVar":       true,
		"Uint":            true,
		"Uint64":          true,
		"Uint64Var":       true,
		"UintVar":         true,
		"UnquoteUsage":    true,
		"Usage":           true,
		"Value":           true,
		"Var":             true,
		"Visit":           true,
		"VisitAll":        true,
	},
	"fmt": map[string]bool{
		"Errorf":     true,
		"Formatter":  true,
		"Fprint":     true,
		"Fprintf":    true,
		"Fprintln":   true,
		"Fscan":      true,
		"Fscanf":     true,
		"Fscanln":    true,
		"GoStringer": true,
		"Print":      true,
		"Printf":     true,
		"Println":    true,
		"Scan":       true,
		"ScanState":  true,
		"Scanf":      true,
		"Scanln":     true,
		"Scanner":    true,
		"Sprint":     true,
		"Sprintf":    true,
		"Sprintln":   true,
		"Sscan":      true,
		"Sscanf":     true,
		"Sscanln":    true,
		"State":      true,
		"Stringer":   true,
	},
	"go/ast": map[string]bool{
		"ArrayType":                  true,
		"AssignStmt":                 true,
		"Bad":                        true,
		"BadDecl":                    true,
		"BadExpr":                    true,
		"BadStmt":                    true,
		"BasicLit":                   true,
		"BinaryExpr":                 true,
		"BlockStmt":                  true,
		"BranchStmt":                 true,
		"CallExpr":                   true,
		"CaseClause":                 true,
		"ChanDir":                    true,
		"ChanType":                   true,
		"CommClause":                 true,
		"Comment":                    true,
		"CommentGroup":               true,
		"CommentMap":                 true,
		"CompositeLit":               true,
		"Con":                        true,
		"DeclStmt":                   true,
		"DeferStmt":                  true,
		"Ellipsis":                   true,
		"EmptyStmt":                  true,
		"ExprStmt":                   true,
		"Field":                      true,
		"FieldFilter":                true,
		"FieldList":                  true,
		"File":                       true,
		"FileExports":                true,
		"Filter":                     true,
		"FilterDecl":                 true,
		"FilterFile":                 true,
		"FilterFuncDuplicates":       true,
		"FilterImportDuplicates":     true,
		"FilterPackage":              true,
		"FilterUnassociatedComments": true,
		"ForStmt":                    true,
		"Fprint":                     true,
		"Fun":                        true,
		"FuncDecl":                   true,
		"FuncLit":                    true,
		"FuncType":                   true,
		"GenDecl":                    true,
		"GoStmt":                     true,
		"Ident":                      true,
		"IfStmt":                     true,
		"ImportSpec":                 true,
		"Importer":                   true,
		"IncDecStmt":                 true,
		"IndexExpr":                  true,
		"Inspect":                    true,
		"InterfaceType":              true,
		"IsExported":                 true,
		"KeyValueExpr":               true,
		"LabeledStmt":                true,
		"Lbl":                        true,
		"MapType":                    true,
		"MergeMode":                  true,
		"MergePackageFiles":          true,
		"NewCommentMap":              true,
		"NewIdent":                   true,
		"NewObj":                     true,
		"NewPackage":                 true,
		"NewScope":                   true,
		"Node":                       true,
		"NotNilFilter":               true,
		"ObjKind":                    true,
		"Object":                     true,
		"Package":                    true,
		"PackageExports":             true,
		"ParenExpr":                  true,
		"Pkg":                        true,
		"Print":                      true,
		"RECV":                       true,
		"RangeStmt":                  true,
		"ReturnStmt":                 true,
		"SEND":                       true,
		"Scope":                      true,
		"SelectStmt":                 true,
		"SelectorExpr":               true,
		"SendStmt":                   true,
		"SliceExpr":                  true,
		"SortImports":                true,
		"StarExpr":                   true,
		"StructType":                 true,
		"SwitchStmt":                 true,
		"Typ":                        true,
		"TypeAssertExpr":             true,
		"TypeSpec":                   true,
		"TypeSwitchStmt":             true,
		"UnaryExpr":                  true,
		"ValueSpec":                  true,
		"Var":                        true,
		"Visitor":                    true,
		"Walk":                       true,
	},
	"go/build": map[string]bool{
		"AllowBinary":          true,
		"ArchChar":             true,
		"Context":              true,
		"Default":              true,
		"FindOnly":             true,
		"IgnoreVendor":         true,
		"Import":               true,
		"ImportComment":        true,
		"ImportDir":            true,
		"ImportMode":           true,
		"IsLocalImport":        true,
		"MultiplePackageError": true,
		"NoGoError":            true,
		"Package":              true,
		"ToolDir":              true,
	},
	"go/constant": map[string]bool{
		"BinaryOp":        true,
		"BitLen":          true,
		"Bool":            true,
		"BoolVal":         true,
		"Bytes":           true,
		"Compare":         true,
		"Complex":         true,
		"Denom":           true,
		"Float":           true,
		"Float32Val":      true,
		"Float64Val":      true,
		"Imag":            true,
		"Int":             true,
		"Int64Val":        true,
		"Kind":            true,
		"MakeBool":        true,
		"MakeFloat64":     true,
		"MakeFromBytes":   true,
		"MakeFromLiteral": true,
		"MakeImag":        true,
		"MakeInt64":       true,
		"MakeString":      true,
		"MakeUint64":      true,
		"MakeUnknown":     true,
		"Num":             true,
		"Real":            true,
		"Shift":           true,
		"Sign":            true,
		"String":          true,
		"StringVal":       true,
		"ToComplex":       true,
		"ToFloat":         true,
		"ToInt":           true,
		"Uint64Val":       true,
		"UnaryOp":         true,
		"Unknown":         true,
	},
	"go/doc": map[string]bool{
		"AllDecls":        true,
		"AllMethods":      true,
		"Example":         true,
		"Examples":        true,
		"Filter":          true,
		"Func":            true,
		"IllegalPrefixes": true,
		"IsPredeclared":   true,
		"Mode":            true,
		"New":             true,
		"Note":            true,
		"Package":         true,
		"PreserveAST":     true,
		"Synopsis":        true,
		"ToHTML":          true,
		"ToText":          true,
		"Type":            true,
		"Value":           true,
	},
	"go/format": map[string]bool{
		"Node":   true,
		"Source": true,
	},
	"go/importer": map[string]bool{
		"Default":     true,
		"For":         true,
		"ForCompiler": true,
		"Lookup":      true,
	},
	"go/parser": map[string]bool{
		"AllErrors":         true,
		"DeclarationErrors": true,
		"ImportsOnly":       true,
		"Mode":              true,
		"PackageClauseOnly": true,
		"ParseComments":     true,
		"ParseDir":          true,
		"ParseExpr":         true,
		"ParseExprFrom":     true,
		"ParseFile":         true,
		"SpuriousErrors":    true,
		"Trace":             true,
	},
	"go/printer": map[string]bool{
		"CommentedNode": true,
		"Config":        true,
		"Fprint":        true,
		"Mode":          true,
		"RawFormat":     true,
		"SourcePos":     true,
		"TabIndent":     true,
		"UseSpaces":     true,
	},
	"go/scanner": map[string]bool{
		"Error":        true,
		"ErrorHandler": true,
		"ErrorList":    true,
		"Mode":         true,
		"PrintError":   true,
		"ScanComments": true,
		"Scanner":      true,
	},
	"go/token": map[string]bool{
		"ADD":            true,
		"ADD_ASSIGN":     true,
		"AND":            true,
		"AND_ASSIGN":     true,
		"AND_NOT":        true,
		"AND_NOT_ASSIGN": true,
		"ARROW":          true,
		"ASSIGN":         true,
		"BREAK":          true,
		"CASE":           true,
		"CHAN":           true,
		"CHAR":           true,
		"COLON":          true,
		"COMMA":          true,
		"COMMENT":        true,
		"CONST":          true,
		"CONTINUE":       true,
		"DEC":            true,
		"DEFAULT":        true,
		"DEFER":          true,
		"DEFINE":         true,
		"ELLIPSIS":       true,
		"ELSE":           true,
		"EOF":            true,
		"EQL":            true,
		"FALLTHROUGH":    true,
		"FLOAT":          true,
		"FOR":            true,
		"FUNC":           true,
		"File":           true,
		"FileSet":        true,
		"GEQ":            true,
		"GO":             true,
		"GOTO":           true,
		"GTR":            true,
		"HighestPrec":    true,
		"IDENT":          true,
		"IF":             true,
		"ILLEGAL":        true,
		"IMAG":           true,
		"IMPORT":         true,
		"INC":            true,
		"INT":            true,
		"INTERFACE":      true,
		"LAND":           true,
		"LBRACE":         true,
		"LBRACK":         true,
		"LEQ":            true,
		"LOR":            true,
		"LPAREN":         true,
		"LSS":            true,
		"Lookup":         true,
		"LowestPrec":     true,
		"MAP":            true,
		"MUL":            true,
		"MUL_ASSIGN":     true,
		"NEQ":            true,
		"NOT":            true,
		"NewFileSet":     true,
		"NoPos":          true,
		"OR":             true,
		"OR_ASSIGN":      true,
		"PACKAGE":        true,
		"PERIOD":         true,
		"Pos":            true,
		"Position":       true,
		"QUO":            true,
		"QUO_ASSIGN":     true,
		"RANGE":          true,
		"RBRACE":         true,
		"RBRACK":         true,
		"REM":            true,
		"REM_ASSIGN":     true,
		"RETURN":         true,
		"RPAREN":         true,
		"SELECT":         true,
		"SEMICOLON":      true,
		"SHL":            true,
		"SHL_ASSIGN":     true,
		"SHR":            true,
		"SHR_ASSIGN":     true,
		"STRING":         true,
		"STRUCT":         true,
		"SUB":            true,
		"SUB_ASSIGN":     true,
		"SWITCH":         true,
		"TYPE":           true,
		"Token":          true,
		"UnaryPrec":      true,
		"VAR":            true,
		"XOR":            true,
		"XOR_ASSIGN":     true,
	},
	"go/types": map[string]bool{
		"Array":                   true,
		"AssertableTo":            true,
		"AssignableTo":            true,
		"Basic":                   true,
		"BasicInfo":               true,
		"BasicKind":               true,
		"Bool":                    true,
		"Builtin":                 true,
		"Byte":                    true,
		"Chan":                    true,
		"ChanDir":                 true,
		"Checker":                 true,
		"Comparable":              true,
		"Complex128":              true,
		"Complex64":               true,
		"Config":                  true,
		"Const":                   true,
		"ConvertibleTo":           true,
		"DefPredeclaredTestFuncs": true,
		"Default":                 true,
		"Error":                   true,
		"Eval":                    true,
		"ExprString":              true,
		"FieldVal":                true,
		"Float32":                 true,
		"Float64":                 true,
		"Func":                    true,
		"Id":                      true,
		"Identical":               true,
		"IdenticalIgnoreTags":     true,
		"Implements":              true,
		"ImportMode":              true,
		"Importer":                true,
		"ImporterFrom":            true,
		"Info":                    true,
		"Initializer":             true,
		"Int":                     true,
		"Int16":                   true,
		"Int32":                   true,
		"Int64":                   true,
		"Int8":                    true,
		"Interface":               true,
		"Invalid":                 true,
		"IsBoolean":               true,
		"IsComplex":               true,
		"IsConstType":             true,
		"IsFloat":                 true,
		"IsInteger":               true,
		"IsInterface":             true,
		"IsNumeric":               true,
		"IsOrdered":               true,
		"IsString":                true,
		"IsUnsigned":              true,
		"IsUntyped":               true,
		"Label":                   true,
		"LookupFieldOrMethod":     true,
		"Map":                     true,
		"MethodExpr":              true,
		"MethodSet":               true,
		"MethodVal":               true,
		"MissingMethod":           true,
		"Named":                   true,
		"NewArray":                true,
		"NewChan":                 true,
		"NewChecker":              true,
		"NewConst":                true,
		"NewField":                true,
		"NewFunc":                 true,
		"NewInterface":            true,
		"NewInterfaceType":        true,
		"NewLabel":                true,
		"NewMap":                  true,
		"NewMethodSet":            true,
		"NewNamed":                true,
		"NewPackage":              true,
		"NewParam":                true,
		"NewPkgName":              true,
		"NewPointer":              true,
		"NewScope":                true,
		"NewSignature":            true,
		"NewSlice":                true,
		"NewStruct":               true,
		"NewTuple":                true,
		"NewTypeName":             true,
		"NewVar":                  true,
		"Nil":                     true,
		"ObjectString":            true,
		"Package":                 true,
		"PkgName":                 true,
		"Pointer":                 true,
		"Qualifier":               true,
		"RecvOnly":                true,
		"RelativeTo":              true,
		"Rune":                    true,
		"Scope":                   true,
		"Selection":               true,
		"SelectionKind":           true,
		"SelectionString":         true,
		"SendOnly":                true,
		"SendRecv":                true,
		"Signature":               true,
		"Sizes":                   true,
		"SizesFor":                true,
		"Slice":                   true,
		"StdSizes":                true,
		"String":                  true,
		"Struct":                  true,
		"Tuple":                   true,
		"Typ":                     true,
		"Type":                    true,
		"TypeAndValue":            true,
		"TypeName":                true,
		"TypeString":              true,
		"Uint":                    true,
		"Uint16":                  true,
		"Uint32":                  true,
		"Uint64":                  true,
		"Uint8":                   true,
		"Uintptr":                 true,
		"Universe":                true,
		"Unsafe":                  true,
		"UnsafePointer":           true,
		"UntypedBool":             true,
		"UntypedComplex":          true,
		"UntypedFloat":            true,
		"UntypedInt":              true,
		"UntypedNil":              true,
		"UntypedRune":             true,
		"UntypedString":           true,
		"Var":                     true,
		"WriteExpr":               true,
		"WriteSignature":          true,
		"WriteType":               true,
	},
	"hash": map[string]bool{
		"Hash":   true,
		"Hash32": true,
		"Hash64": true,
	},
	"hash/adler32": map[string]bool{
		"Checksum": true,
		"New":      true,
		"Size":     true,
	},
	"hash/crc32": map[string]bool{
		"Castagnoli":   true,
		"Checksum":     true,
		"ChecksumIEEE": true,
		"IEEE":         true,
		"IEEETable":    true,
		"Koopman":      true,
		"MakeTable":    true,
		"New":          true,
		"NewIEEE":      true,
		"Size":         true,
		"Table":        true,
		"Update":       true,
	},
	"hash/crc64": map[string]bool{
		"Checksum":  true,
		"ECMA":      true,
		"ISO":       true,
		"MakeTable": true,
		"New":       true,
		"Size":      true,
		"Table":     true,
		"Update":    true,
	},
	"hash/fnv": map[string]bool{
		"New128":  true,
		"New128a": true,
		"New32":   true,
		"New32a":  true,
		"New64":   true,
		"New64a":  true,
	},
	"html": map[string]bool{
		"EscapeString":   true,
		"UnescapeString": true,
	},
	"html/template": map[string]bool{
		"CSS":                  true,
		"ErrAmbigContext":      true,
		"ErrBadHTML":           true,
		"ErrBranchEnd":         true,
		"ErrEndContext":        true,
		"ErrNoSuchTemplate":    true,
		"ErrOutputContext":     true,
		"ErrPartialCharset":    true,
		"ErrPartialEscape":     true,
		"ErrPredefinedEscaper": true,
		"ErrRangeLoopReentry":  true,
		"ErrSlashAmbig":        true,
		"Error":                true,
		"ErrorCode":            true,
		"FuncMap":              true,
		"HTML":                 true,
		"HTMLAttr":             true,
		"HTMLEscape":           true,
		"HTMLEscapeString":     true,
		"HTMLEscaper":          true,
		"IsTrue":               true,
		"JS":                   true,
		"JSEscape":             true,
		"JSEscapeString":       true,
		"JSEscaper":            true,
		"JSStr":                true,
		"Must":                 true,
		"New":                  true,
		"OK":                   true,
		"ParseFiles":           true,
		"ParseGlob":            true,
		"Srcset":               true,
		"Template":             true,
		"URL":                  true,
		"URLQueryEscaper":      true,
	},
	"image": map[string]bool{
		"Alpha":                  true,
		"Alpha16":                true,
		"Black":                  true,
		"CMYK":                   true,
		"Config":                 true,
		"Decode":                 true,
		"DecodeConfig":           true,
		"ErrFormat":              true,
		"Gray":                   true,
		"Gray16":                 true,
		"Image":                  true,
		"NRGBA":                  true,
		"NRGBA64":                true,
		"NYCbCrA":                true,
		"NewAlpha":               true,
		"NewAlpha16":             true,
		"NewCMYK":                true,
		"NewGray":                true,
		"NewGray16":              true,
		"NewNRGBA":               true,
		"NewNRGBA64":             true,
		"NewNYCbCrA":             true,
		"NewPaletted":            true,
		"NewRGBA":                true,
		"NewRGBA64":              true,
		"NewUniform":             true,
		"NewYCbCr":               true,
		"Opaque":                 true,
		"Paletted":               true,
		"PalettedImage":          true,
		"Point":                  true,
		"Pt":                     true,
		"RGBA":                   true,
		"RGBA64":                 true,
		"Rect":                   true,
		"Rectangle":              true,
		"RegisterFormat":         true,
		"Transparent":            true,
		"Uniform":                true,
		"White":                  true,
		"YCbCr":                  true,
		"YCbCrSubsampleRatio":    true,
		"YCbCrSubsampleRatio410": true,
		"YCbCrSubsampleRatio411": true,
		"YCbCrSubsampleRatio420": true,
		"YCbCrSubsampleRatio422": true,
		"YCbCrSubsampleRatio440": true,
		"YCbCrSubsampleRatio444": true,
		"ZP":                     true,
		"ZR":                     true,
	},
	"image/color": map[string]bool{
		"Alpha":        true,
		"Alpha16":      true,
		"Alpha16Model": true,
		"AlphaModel":   true,
		"Black":        true,
		"CMYK":         true,
		"CMYKModel":    true,
		"CMYKToRGB":    true,
		"Color":        true,
		"Gray":         true,
		"Gray16":       true,
		"Gray16Model":  true,
		"GrayModel":    true,
		"Model":        true,
		"ModelFunc":    true,
		"NRGBA":        true,
		"NRGBA64":      true,
		"NRGBA64Model": true,
		"NRGBAModel":   true,
		"NYCbCrA":      true,
		"NYCbCrAModel": true,
		"Opaque":       true,
		"Palette":      true,
		"RGBA":         true,
		"RGBA64":       true,
		"RGBA64Model":  true,
		"RGBAModel":    true,
		"RGBToCMYK":    true,
		"RGBToYCbCr":   true,
		"Transparent":  true,
		"White":        true,
		"YCbCr":        true,
		"YCbCrModel":   true,
		"YCbCrToRGB":   true,
	},
	"image/color/palette": map[string]bool{
		"Plan9":   true,
		"WebSafe": true,
	},
	"image/draw": map[string]bool{
		"Draw":           true,
		"DrawMask":       true,
		"Drawer":         true,
		"FloydSteinberg": true,
		"Image":          true,
		"Op":             true,
		"Over":           true,
		"Quantizer":      true,
		"Src":            true,
	},
	"image/gif": map[string]bool{
		"Decode":             true,
		"DecodeAll":          true,
		"DecodeConfig":       true,
		"DisposalBackground": true,
		"DisposalNone":       true,
		"DisposalPrevious":   true,
		"Encode":             true,
		"EncodeAll":          true,
		"GIF":                true,
		"Options":            true,
	},
	"image/jpeg": map[string]bool{
		"Decode":           true,
		"DecodeConfig":     true,
		"DefaultQuality":   true,
		"Encode":           true,
		"FormatError":      true,
		"Options":          true,
		"Reader":           true,
		"UnsupportedError": true,
	},
	"image/png": map[string]bool{
		"BestCompression":    true,
		"BestSpeed":          true,
		"CompressionLevel":   true,
		"Decode":             true,
		"DecodeConfig":       true,
		"DefaultCompression": true,
		"Encode":             true,
		"Encoder":            true,
		"EncoderBuffer":      true,
		"EncoderBufferPool":  true,
		"FormatError":        true,
		"NoCompression":      true,
		"UnsupportedError":   true,
	},
	"index/suffixarray": map[string]bool{
		"Index": true,
		"New":   true,
	},
	"io": map[string]bool{
		"ByteReader":       true,
		"ByteScanner":      true,
		"ByteWriter":       true,
		"Closer":           true,
		"Copy":             true,
		"CopyBuffer":       true,
		"CopyN":            true,
		"EOF":              true,
		"ErrClosedPipe":    true,
		"ErrNoProgress":    true,
		"ErrShortBuffer":   true,
		"ErrShortWrite":    true,
		"ErrUnexpectedEOF": true,
		"LimitReader":      true,
		"LimitedReader":    true,
		"MultiReader":      true,
		"MultiWriter":      true,
		"NewSectionReader": true,
		"Pipe":             true,
		"PipeReader":       true,
		"PipeWriter":       true,
		"ReadAtLeast":      true,
		"ReadCloser":       true,
		"ReadFull":         true,
		"ReadSeeker":       true,
		"ReadWriteCloser":  true,
		"ReadWriteSeeker":  true,
		"ReadWriter":       true,
		"Reader":           true,
		"ReaderAt":         true,
		"ReaderFrom":       true,
		"RuneReader":       true,
		"RuneScanner":      true,
		"SectionReader":    true,
		"SeekCurrent":      true,
		"SeekEnd":          true,
		"SeekStart":        true,
		"Seeker":           true,
		"StringWriter":     true,
		"TeeReader":        true,
		"WriteCloser":      true,
		"WriteSeeker":      true,
		"WriteString":      true,
		"Writer":           true,
		"WriterAt":         true,
		"WriterTo":         true,
	},
	"io/ioutil": map[string]bool{
		"Discard":   true,
		"NopCloser": true,
		"ReadAll":   true,
		"ReadDir":   true,
		"ReadFile":  true,
		"TempDir":   true,
		"TempFile":  true,
		"WriteFile": true,
	},
	"log": map[string]bool{
		"Fatal":         true,
		"Fatalf":        true,
		"Fatalln":       true,
		"Flags":         true,
		"LUTC":          true,
		"Ldate":         true,
		"Llongfile":     true,
		"Lmicroseconds": true,
		"Logger":        true,
		"Lshortfile":    true,
		"LstdFlags":     true,
		"Ltime":         true,
		"New":           true,
		"Output":        true,
		"Panic":         true,
		"Panicf":        true,
		"Panicln":       true,
		"Prefix":        true,
		"Print":         true,
		"Printf":        true,
		"Println":       true,
		"SetFlags":      true,
		"SetOutput":     true,
		"SetPrefix":     true,
	},
	"log/syslog": map[string]bool{
		"Dial":         true,
		"LOG_ALERT":    true,
		"LOG_AUTH":     true,
		"LOG_AUTHPRIV": true,
		"LOG_CRIT":     true,
		"LOG_CRON":     true,
		"LOG_DAEMON":   true,
		"LOG_DEBUG":    true,
		"LOG_EMERG":    true,
		"LOG_ERR":      true,
		"LOG_FTP":      true,
		"LOG_INFO":     true,
		"LOG_KERN":     true,
		"LOG_LOCAL0":   true,
		"LOG_LOCAL1":   true,
		"LOG_LOCAL2":   true,
		"LOG_LOCAL3":   true,
		"LOG_LOCAL4":   true,
		"LOG_LOCAL5":   true,
		"LOG_LOCAL6":   true,
		"LOG_LOCAL7":   true,
		"LOG_LPR":      true,
		"LOG_MAIL":     true,
		"LOG_NEWS":     true,
		"LOG_NOTICE":   true,
		"LOG_SYSLOG":   true,
		"LOG_USER":     true,
		"LOG_UUCP":     true,
		"LOG_WARNING":  true,
		"New":          true,
		"NewLogger":    true,
		"Priority":     true,
		"Writer":       true,
	},
	"math": map[string]bool{
		"Abs":                    true,
		"Acos":                   true,
		"Acosh":                  true,
		"Asin":                   true,
		"Asinh":                  true,
		"Atan":                   true,
		"Atan2":                  true,
		"Atanh":                  true,
		"Cbrt":                   true,
		"Ceil":                   true,
		"Copysign":               true,
		"Cos":                    true,
		"Cosh":                   true,
		"Dim":                    true,
		"E":                      true,
		"Erf":                    true,
		"Erfc":                   true,
		"Erfcinv":                true,
		"Erfinv":                 true,
		"Exp":                    true,
		"Exp2":                   true,
		"Expm1":                  true,
		"Float32bits":            true,
		"Float32frombits":        true,
		"Float64bits":            true,
		"Float64frombits":        true,
		"Floor":                  true,
		"Frexp":                  true,
		"Gamma":                  true,
		"Hypot":                  true,
		"Ilogb":                  true,
		"Inf":                    true,
		"IsInf":                  true,
		"IsNaN":                  true,
		"J0":                     true,
		"J1":                     true,
		"Jn":                     true,
		"Ldexp":                  true,
		"Lgamma":                 true,
		"Ln10":                   true,
		"Ln2":                    true,
		"Log":                    true,
		"Log10":                  true,
		"Log10E":                 true,
		"Log1p":                  true,
		"Log2":                   true,
		"Log2E":                  true,
		"Logb":                   true,
		"Max":                    true,
		"MaxFloat32":             true,
		"MaxFloat64":             true,
		"MaxInt16":               true,
		"MaxInt32":               true,
		"MaxInt64":               true,
		"MaxInt8":                true,
		"MaxUint16":              true,
		"MaxUint32":              true,
		"MaxUint64":              true,
		"MaxUint8":               true,
		"Min":                    true,
		"MinInt16":               true,
		"MinInt32":               true,
		"MinInt64":               true,
		"MinInt8":                true,
		"Mod":                    true,
		"Modf":                   true,
		"NaN":                    true,
		"Nextafter":              true,
		"Nextafter32":            true,
		"Phi":                    true,
		"Pi":                     true,
		"Pow":                    true,
		"Pow10":                  true,
		"Remainder":              true,
		"Round":                  true,
		"RoundToEven":            true,
		"Signbit":                true,
		"Sin":                    true,
		"Sincos":                 true,
		"Sinh":                   true,
		"SmallestNonzeroFloat32": true,
		"SmallestNonzeroFloat64": true,
		"Sqrt":                   true,
		"Sqrt2":                  true,
		"SqrtE":                  true,
		"SqrtPhi":                true,
		"SqrtPi":                 true,
		"Tan":                    true,
		"Tanh":                   true,
		"Trunc":                  true,
		"Y0":                     true,
		"Y1":                     true,
		"Yn":                     true,
	},
	"math/big": map[string]bool{
		"Above":         true,
		"Accuracy":      true,
		"AwayFromZero":  true,
		"Below":         true,
		"ErrNaN":        true,
		"Exact":         true,
		"Float":         true,
		"Int":           true,
		"Jacobi":        true,
		"MaxBase":       true,
		"MaxExp":        true,
		"MaxPrec":       true,
		"MinExp":        true,
		"NewFloat":      true,
		"NewInt":        true,
		"NewRat":        true,
		"ParseFloat":    true,
		"Rat":           true,
		"RoundingMode":  true,
		"ToNearestAway": true,
		"ToNearestEven": true,
		"ToNegativeInf": true,
		"ToPositiveInf": true,
		"ToZero":        true,
		"Word":          true,
	},
	"math/bits": map[string]bool{
		"Add":             true,
		"Add32":           true,
		"Add64":           true,
		"Div":             true,
		"Div32":           true,
		"Div64":           true,
		"LeadingZeros":    true,
		"LeadingZeros16":  true,
		"LeadingZeros32":  true,
		"LeadingZeros64":  true,
		"LeadingZeros8":   true,
		"Len":             true,
		"Len16":           true,
		"Len32":           true,
		"Len64":           true,
		"Len8":            true,
		"Mul":             true,
		"Mul32":           true,
		"Mul64":           true,
		"OnesCount":       true,
		"OnesCount16":     true,
		"OnesCount32":     true,
		"OnesCount64":     true,
		"OnesCount8":      true,
		"Reverse":         true,
		"Reverse16":       true,
		"Reverse32":       true,
		"Reverse64":       true,
		"Reverse8":        true,
		"ReverseBytes":    true,
		"ReverseBytes16":  true,
		"ReverseBytes32":  true,
		"ReverseBytes64":  true,
		"RotateLeft":      true,
		"RotateLeft16":    true,
		"RotateLeft32":    true,
		"RotateLeft64":    true,
		"RotateLeft8":     true,
		"Sub":             true,
		"Sub32":           true,
		"Sub64":           true,
		"TrailingZeros":   true,
		"TrailingZeros16": true,
		"TrailingZeros32": true,
		"TrailingZeros64": true,
		"TrailingZeros8":  true,
		"UintSize":        true,
	},
	"math/cmplx": map[string]bool{
		"Abs":   true,
		"Acos":  true,
		"Acosh": true,
		"Asin":  true,
		"Asinh": true,
		"Atan":  true,
		"Atanh": true,
		"Conj":  true,
		"Cos":   true,
		"Cosh":  true,
		"Cot":   true,
		"Exp":   true,
		"Inf":   true,
		"IsInf": true,
		"IsNaN": true,
		"Log":   true,
		"Log10": true,
		"NaN":   true,
		"Phase": true,
		"Polar": true,
		"Pow":   true,
		"Rect":  true,
		"Sin":   true,
		"Sinh":  true,
		"Sqrt":  true,
		"Tan":   true,
		"Tanh":  true,
	},
	"math/rand": map[string]bool{
		"ExpFloat64":  true,
		"Float32":     true,
		"Float64":     true,
		"Int":         true,
		"Int31":       true,
		"Int31n":      true,
		"Int63":       true,
		"Int63n":      true,
		"Intn":        true,
		"New":         true,
		"NewSource":   true,
		"NewZipf":     true,
		"NormFloat64": true,
		"Perm":        true,
		"Rand":        true,
		"Read":        true,
		"Seed":        true,
		"Shuffle":     true,
		"Source":      true,
		"Source64":    true,
		"Uint32":      true,
		"Uint64":      true,
		"Zipf":        true,
	},
	"mime": map[string]bool{
		"AddExtensionType":         true,
		"BEncoding":                true,
		"ErrInvalidMediaParameter": true,
		"ExtensionsByType":         true,
		"FormatMediaType":          true,
		"ParseMediaType":           true,
		"QEncoding":                true,
		"TypeByExtension":          true,
		"WordDecoder":              true,
		"WordEncoder":              true,
	},
	"mime/multipart": map[string]bool{
		"ErrMessageTooLarge": true,
		"File":               true,
		"FileHeader":         true,
		"Form":               true,
		"NewReader":          true,
		"NewWriter":          true,
		"Part":               true,
		"Reader":             true,
		"Writer":             true,
	},
	"mime/quotedprintable": map[string]bool{
		"NewReader": true,
		"NewWriter": true,
		"Reader":    true,
		"Writer":    true,
	},
	"net": map[string]bool{
		"Addr":                       true,
		"AddrError":                  true,
		"Buffers":                    true,
		"CIDRMask":                   true,
		"Conn":                       true,
		"DNSConfigError":             true,
		"DNSError":                   true,
		"DefaultResolver":            true,
		"Dial":                       true,
		"DialIP":                     true,
		"DialTCP":                    true,
		"DialTimeout":                true,
		"DialUDP":                    true,
		"DialUnix":                   true,
		"Dialer":                     true,
		"ErrWriteToConnected":        true,
		"Error":                      true,
		"FileConn":                   true,
		"FileListener":               true,
		"FilePacketConn":             true,
		"FlagBroadcast":              true,
		"FlagLoopback":               true,
		"FlagMulticast":              true,
		"FlagPointToPoint":           true,
		"FlagUp":                     true,
		"Flags":                      true,
		"HardwareAddr":               true,
		"IP":                         true,
		"IPAddr":                     true,
		"IPConn":                     true,
		"IPMask":                     true,
		"IPNet":                      true,
		"IPv4":                       true,
		"IPv4Mask":                   true,
		"IPv4allrouter":              true,
		"IPv4allsys":                 true,
		"IPv4bcast":                  true,
		"IPv4len":                    true,
		"IPv4zero":                   true,
		"IPv6interfacelocalallnodes": true,
		"IPv6len":                    true,
		"IPv6linklocalallnodes":      true,
		"IPv6linklocalallrouters":    true,
		"IPv6loopback":               true,
		"IPv6unspecified":            true,
		"IPv6zero":                   true,
		"Interface":                  true,
		"InterfaceAddrs":             true,
		"InterfaceByIndex":           true,
		"InterfaceByName":            true,
		"Interfaces":                 true,
		"InvalidAddrError":           true,
		"JoinHostPort":               true,
		"Listen":                     true,
		"ListenConfig":               true,
		"ListenIP":                   true,
		"ListenMulticastUDP":         true,
		"ListenPacket":               true,
		"ListenTCP":                  true,
		"ListenUDP":                  true,
		"ListenUnix":                 true,
		"ListenUnixgram":             true,
		"Listener":                   true,
		"LookupAddr":                 true,
		"LookupCNAME":                true,
		"LookupHost":                 true,
		"LookupIP":                   true,
		"LookupMX":                   true,
		"LookupNS":                   true,
		"LookupPort":                 true,
		"LookupSRV":                  true,
		"LookupTXT":                  true,
		"MX":                         true,
		"NS":                         true,
		"OpError":                    true,
		"PacketConn":                 true,
		"ParseCIDR":                  true,
		"ParseError":                 true,
		"ParseIP":                    true,
		"ParseMAC":                   true,
		"Pipe":                       true,
		"ResolveIPAddr":              true,
		"ResolveTCPAddr":             true,
		"ResolveUDPAddr":             true,
		"ResolveUnixAddr":            true,
		"Resolver":                   true,
		"SRV":                        true,
		"SplitHostPort":              true,
		"TCPAddr":                    true,
		"TCPConn":                    true,
		"TCPListener":                true,
		"UDPAddr":                    true,
		"UDPConn":                    true,
		"UnixAddr":                   true,
		"UnixConn":                   true,
		"UnixListener":               true,
		"UnknownNetworkError":        true,
	},
	"net/http": map[string]bool{
		"CanonicalHeaderKey":                  true,
		"Client":                              true,
		"CloseNotifier":                       true,
		"ConnState":                           true,
		"Cookie":                              true,
		"CookieJar":                           true,
		"DefaultClient":                       true,
		"DefaultMaxHeaderBytes":               true,
		"DefaultMaxIdleConnsPerHost":          true,
		"DefaultServeMux":                     true,
		"DefaultTransport":                    true,
		"DetectContentType":                   true,
		"Dir":                                 true,
		"ErrAbortHandler":                     true,
		"ErrBodyNotAllowed":                   true,
		"ErrBodyReadAfterClose":               true,
		"ErrContentLength":                    true,
		"ErrHandlerTimeout":                   true,
		"ErrHeaderTooLong":                    true,
		"ErrHijacked":                         true,
		"ErrLineTooLong":                      true,
		"ErrMissingBoundary":                  true,
		"ErrMissingContentLength":             true,
		"ErrMissingFile":                      true,
		"ErrNoCookie":                         true,
		"ErrNoLocation":                       true,
		"ErrNotMultipart":                     true,
		"ErrNotSupported":                     true,
		"ErrServerClosed":                     true,
		"ErrShortBody":                        true,
		"ErrSkipAltProtocol":                  true,
		"ErrUnexpectedTrailer":                true,
		"ErrUseLastResponse":                  true,
		"ErrWriteAfterFlush":                  true,
		"Error":                               true,
		"File":                                true,
		"FileServer":                          true,
		"FileSystem":                          true,
		"Flusher":                             true,
		"Get":                                 true,
		"Handle":                              true,
		"HandleFunc":                          true,
		"Handler":                             true,
		"HandlerFunc":                         true,
		"Head":                                true,
		"Header":                              true,
		"Hijacker":                            true,
		"ListenAndServe":                      true,
		"ListenAndServeTLS":                   true,
		"LocalAddrContextKey":                 true,
		"MaxBytesReader":                      true,
		"MethodConnect":                       true,
		"MethodDelete":                        true,
		"MethodGet":                           true,
		"MethodHead":                          true,
		"MethodOptions":                       true,
		"MethodPatch":                         true,
		"MethodPost":                          true,
		"MethodPut":                           true,
		"MethodTrace":                         true,
		"NewFileTransport":                    true,
		"NewRequest":                          true,
		"NewServeMux":                         true,
		"NoBody":                              true,
		"NotFound":                            true,
		"NotFoundHandler":                     true,
		"ParseHTTPVersion":                    true,
		"ParseTime":                           true,
		"Post":                                true,
		"PostForm":                            true,
		"ProtocolError":                       true,
		"ProxyFromEnvironment":                true,
		"ProxyURL":                            true,
		"PushOptions":                         true,
		"Pusher":                              true,
		"ReadRequest":                         true,
		"ReadResponse":                        true,
		"Redirect":                            true,
		"RedirectHandler":                     true,
		"Request":                             true,
		"Response":                            true,
		"ResponseWriter":                      true,
		"RoundTripper":                        true,
		"SameSite":                            true,
		"SameSiteDefaultMode":                 true,
		"SameSiteLaxMode":                     true,
		"SameSiteStrictMode":                  true,
		"Serve":                               true,
		"ServeContent":                        true,
		"ServeFile":                           true,
		"ServeMux":                            true,
		"ServeTLS":                            true,
		"Server":                              true,
		"ServerContextKey":                    true,
		"SetCookie":                           true,
		"StateActive":                         true,
		"StateClosed":                         true,
		"StateHijacked":                       true,
		"StateIdle":                           true,
		"StateNew":                            true,
		"StatusAccepted":                      true,
		"StatusAlreadyReported":               true,
		"StatusBadGateway":                    true,
		"StatusBadRequest":                    true,
		"StatusConflict":                      true,
		"StatusContinue":                      true,
		"StatusCreated":                       true,
		"StatusExpectationFailed":             true,
		"StatusFailedDependency":              true,
		"StatusForbidden":                     true,
		"StatusFound":                         true,
		"StatusGatewayTimeout":                true,
		"StatusGone":                          true,
		"StatusHTTPVersionNotSupported":       true,
		"StatusIMUsed":                        true,
		"StatusInsufficientStorage":           true,
		"StatusInternalServerError":           true,
		"StatusLengthRequired":                true,
		"StatusLocked":                        true,
		"StatusLoopDetected":                  true,
		"StatusMethodNotAllowed":              true,
		"StatusMisdirectedRequest":            true,
		"StatusMovedPermanently":              true,
		"StatusMultiStatus":                   true,
		"StatusMultipleChoices":               true,
		"StatusNetworkAuthenticationRequired": true,
		"StatusNoContent":                     true,
		"StatusNonAuthoritativeInfo":          true,
		"StatusNotAcceptable":                 true,
		"StatusNotExtended":                   true,
		"StatusNotFound":                      true,
		"StatusNotImplemented":                true,
		"StatusNotModified":                   true,
		"StatusOK":                            true,
		"StatusPartialContent":                true,
		"StatusPaymentRequired":               true,
		"StatusPermanentRedirect":             true,
		"StatusPreconditionFailed":            true,
		"StatusPreconditionRequired":          true,
		"StatusProcessing":                    true,
		"StatusProxyAuthRequired":             true,
		"StatusRequestEntityTooLarge":         true,
		"StatusRequestHeaderFieldsTooLarge":   true,
		"StatusRequestTimeout":                true,
		"StatusRequestURITooLong":             true,
		"StatusRequestedRangeNotSatisfiable":  true,
		"StatusResetContent":                  true,
		"StatusSeeOther":                      true,
		"StatusServiceUnavailable":            true,
		"StatusSwitchingProtocols":            true,
		"StatusTeapot":                        true,
		"StatusTemporaryRedirect":             true,
		"StatusText":                          true,
		"StatusTooEarly":                      true,
		"StatusTooManyRequests":               true,
		"StatusUnauthorized":                  true,
		"StatusUnavailableForLegalReasons":    true,
		"StatusUnprocessableEntity":           true,
		"StatusUnsupportedMediaType":          true,
		"StatusUpgradeRequired":               true,
		"StatusUseProxy":                      true,
		"StatusVariantAlsoNegotiates":         true,
		"StripPrefix":                         true,
		"TimeFormat":                          true,
		"TimeoutHandler":                      true,
		"TrailerPrefix":                       true,
		"Transport":                           true,
	},
	"net/http/cgi": map[string]bool{
		"Handler":        true,
		"Request":        true,
		"RequestFromMap": true,
		"Serve":          true,
	},
	"net/http/cookiejar": map[string]bool{
		"Jar":              true,
		"New":              true,
		"Options":          true,
		"PublicSuffixList": true,
	},
	"net/http/fcgi": map[string]bool{
		"ErrConnClosed":     true,
		"ErrRequestAborted": true,
		"ProcessEnv":        true,
		"Serve":             true,
	},
	"net/http/httptest": map[string]bool{
		"DefaultRemoteAddr":  true,
		"NewRecorder":        true,
		"NewRequest":         true,
		"NewServer":          true,
		"NewTLSServer":       true,
		"NewUnstartedServer": true,
		"ResponseRecorder":   true,
		"Server":             true,
	},
	"net/http/httptrace": map[string]bool{
		"ClientTrace":        true,
		"ContextClientTrace": true,
		"DNSDoneInfo":        true,
		"DNSStartInfo":       true,
		"GotConnInfo":        true,
		"WithClientTrace":    true,
		"WroteRequestInfo":   true,
	},
	"net/http/httputil": map[string]bool{
		"BufferPool":                true,
		"ClientConn":                true,
		"DumpRequest":               true,
		"DumpRequestOut":            true,
		"DumpResponse":              true,
		"ErrClosed":                 true,
		"ErrLineTooLong":            true,
		"ErrPersistEOF":             true,
		"ErrPipeline":               true,
		"NewChunkedReader":          true,
		"NewChunkedWriter":          true,
		"NewClientConn":             true,
		"NewProxyClientConn":        true,
		"NewServerConn":             true,
		"NewSingleHostReverseProxy": true,
		"ReverseProxy":              true,
		"ServerConn":                true,
	},
	"net/http/pprof": map[string]bool{
		"Cmdline": true,
		"Handler": true,
		"Index":   true,
		"Profile": true,
		"Symbol":  true,
		"Trace":   true,
	},
	"net/mail": map[string]bool{
		"Address":             true,
		"AddressParser":       true,
		"ErrHeaderNotPresent": true,
		"Header":              true,
		"Message":             true,
		"ParseAddress":        true,
		"ParseAddressList":    true,
		"ParseDate":           true,
		"ReadMessage":         true,
	},
	"net/rpc": map[string]bool{
		"Accept":             true,
		"Call":               true,
		"Client":             true,
		"ClientCodec":        true,
		"DefaultDebugPath":   true,
		"DefaultRPCPath":     true,
		"DefaultServer":      true,
		"Dial":               true,
		"DialHTTP":           true,
		"DialHTTPPath":       true,
		"ErrShutdown":        true,
		"HandleHTTP":         true,
		"NewClient":          true,
		"NewClientWithCodec": true,
		"NewServer":          true,
		"Register":           true,
		"RegisterName":       true,
		"Request":            true,
		"Response":           true,
		"ServeCodec":         true,
		"ServeConn":          true,
		"ServeRequest":       true,
		"Server":             true,
		"ServerCodec":        true,
		"ServerError":        true,
	},
	"net/rpc/jsonrpc": map[string]bool{
		"Dial":           true,
		"NewClient":      true,
		"NewClientCodec": true,
		"NewServerCodec": true,
		"ServeConn":      true,
	},
	"net/smtp": map[string]bool{
		"Auth":        true,
		"CRAMMD5Auth": true,
		"Client":      true,
		"Dial":        true,
		"NewClient":   true,
		"PlainAuth":   true,
		"SendMail":    true,
		"ServerInfo":  true,
	},
	"net/textproto": map[string]bool{
		"CanonicalMIMEHeaderKey": true,
		"Conn":                   true,
		"Dial":                   true,
		"Error":                  true,
		"MIMEHeader":             true,
		"NewConn":                true,
		"NewReader":              true,
		"NewWriter":              true,
		"Pipeline":               true,
		"ProtocolError":          true,
		"Reader":                 true,
		"TrimBytes":              true,
		"TrimString":             true,
		"Writer":                 true,
	},
	"net/url": map[string]bool{
		"Error":            true,
		"EscapeError":      true,
		"InvalidHostError": true,
		"Parse":            true,
		"ParseQuery":       true,
		"ParseRequestURI":  true,
		"PathEscape":       true,
		"PathUnescape":     true,
		"QueryEscape":      true,
		"QueryUnescape":    true,
		"URL":              true,
		"User":             true,
		"UserPassword":     true,
		"Userinfo":         true,
		"Values":           true,
	},
	"os": map[string]bool{
		"Args":              true,
		"Chdir":             true,
		"Chmod":             true,
		"Chown":             true,
		"Chtimes":           true,
		"Clearenv":          true,
		"Create":            true,
		"DevNull":           true,
		"Environ":           true,
		"ErrClosed":         true,
		"ErrExist":          true,
		"ErrInvalid":        true,
		"ErrNoDeadline":     true,
		"ErrNotExist":       true,
		"ErrPermission":     true,
		"Executable":        true,
		"Exit":              true,
		"Expand":            true,
		"ExpandEnv":         true,
		"File":              true,
		"FileInfo":          true,
		"FileMode":          true,
		"FindProcess":       true,
		"Getegid":           true,
		"Getenv":            true,
		"Geteuid":           true,
		"Getgid":            true,
		"Getgroups":         true,
		"Getpagesize":       true,
		"Getpid":            true,
		"Getppid":           true,
		"Getuid":            true,
		"Getwd":             true,
		"Hostname":          true,
		"Interrupt":         true,
		"IsExist":           true,
		"IsNotExist":        true,
		"IsPathSeparator":   true,
		"IsPermission":      true,
		"IsTimeout":         true,
		"Kill":              true,
		"Lchown":            true,
		"Link":              true,
		"LinkError":         true,
		"LookupEnv":         true,
		"Lstat":             true,
		"Mkdir":             true,
		"MkdirAll":          true,
		"ModeAppend":        true,
		"ModeCharDevice":    true,
		"ModeDevice":        true,
		"ModeDir":           true,
		"ModeExclusive":     true,
		"ModeIrregular":     true,
		"ModeNamedPipe":     true,
		"ModePerm":          true,
		"ModeSetgid":        true,
		"ModeSetuid":        true,
		"ModeSocket":        true,
		"ModeSticky":        true,
		"ModeSymlink":       true,
		"ModeTemporary":     true,
		"ModeType":          true,
		"NewFile":           true,
		"NewSyscallError":   true,
		"O_APPEND":          true,
		"O_CREATE":          true,
		"O_EXCL":            true,
		"O_RDONLY":          true,
		"O_RDWR":            true,
		"O_SYNC":            true,
		"O_TRUNC":           true,
		"O_WRONLY":          true,
		"Open":              true,
		"OpenFile":          true,
		"PathError":         true,
		"PathListSeparator": true,
		"PathSeparator":     true,
		"Pipe":              true,
		"ProcAttr":          true,
		"Process":           true,
		"ProcessState":      true,
		"Readlink":          true,
		"Remove":            true,
		"RemoveAll":         true,
		"Rename":            true,
		"SEEK_CUR":          true,
		"SEEK_END":          true,
		"SEEK_SET":          true,
		"SameFile":          true,
		"Setenv":            true,
		"Signal":            true,
		"StartProcess":      true,
		"Stat":              true,
		"Stderr":            true,
		"Stdin":             true,
		"Stdout":            true,
		"Symlink":           true,
		"SyscallError":      true,
		"TempDir":           true,
		"Truncate":          true,
		"Unsetenv":          true,
		"UserCacheDir":      true,
		"UserHomeDir":       true,
	},
	"os/exec": map[string]bool{
		"Cmd":            true,
		"Command":        true,
		"CommandContext": true,
		"ErrNotFound":    true,
		"Error":          true,
		"ExitError":      true,
		"LookPath":       true,
	},
	"os/signal": map[string]bool{
		"Ignore":  true,
		"Ignored": true,
		"Notify":  true,
		"Reset":   true,
		"Stop":    true,
	},
	"os/user": map[string]bool{
		"Current":             true,
		"Group":               true,
		"Lookup":              true,
		"LookupGroup":         true,
		"LookupGroupId":       true,
		"LookupId":            true,
		"UnknownGroupError":   true,
		"UnknownGroupIdError": true,
		"UnknownUserError":    true,
		"UnknownUserIdError":  true,
		"User":                true,
	},
	"path": map[string]bool{
		"Base":          true,
		"Clean":         true,
		"Dir":           true,
		"ErrBadPattern": true,
		"Ext":           true,
		"IsAbs":         true,
		"Join":          true,
		"Match":         true,
		"Split":         true,
	},
	"path/filepath": map[string]bool{
		"Abs":           true,
		"Base":          true,
		"Clean":         true,
		"Dir":           true,
		"ErrBadPattern": true,
		"EvalSymlinks":  true,
		"Ext":           true,
		"FromSlash":     true,
		"Glob":          true,
		"HasPrefix":     true,
		"IsAbs":         true,
		"Join":          true,
		"ListSeparator": true,
		"Match":         true,
		"Rel":           true,
		"Separator":     true,
		"SkipDir":       true,
		"Split":         true,
		"SplitList":     true,
		"ToSlash":       true,
		"VolumeName":    true,
		"Walk":          true,
		"WalkFunc":      true,
	},
	"plugin": map[string]bool{
		"Open":   true,
		"Plugin": true,
		"Symbol": true,
	},
	"reflect": map[string]bool{
		"Append":          true,
		"AppendSlice":     true,
		"Array":           true,
		"ArrayOf":         true,
		"Bool":            true,
		"BothDir":         true,
		"Chan":            true,
		"ChanDir":         true,
		"ChanOf":          true,
		"Complex128":      true,
		"Complex64":       true,
		"Copy":            true,
		"DeepEqual":       true,
		"Float32":         true,
		"Float64":         true,
		"Func":            true,
		"FuncOf":          true,
		"Indirect":        true,
		"Int":             true,
		"Int16":           true,
		"Int32":           true,
		"Int64":           true,
		"Int8":            true,
		"Interface":       true,
		"Invalid":         true,
		"Kind":            true,
		"MakeChan":        true,
		"MakeFunc":        true,
		"MakeMap":         true,
		"MakeMapWithSize": true,
		"MakeSlice":       true,
		"Map":             true,
		"MapIter":         true,
		"MapOf":           true,
		"Method":          true,
		"New":             true,
		"NewAt":           true,
		"Ptr":             true,
		"PtrTo":           true,
		"RecvDir":         true,
		"Select":          true,
		"SelectCase":      true,
		"SelectDefault":   true,
		"SelectDir":       true,
		"SelectRecv":      true,
		"SelectSend":      true,
		"SendDir":         true,
		"Slice":           true,
		"SliceHeader":     true,
		"SliceOf":         true,
		"String":          true,
		"StringHeader":    true,
		"Struct":          true,
		"StructField":     true,
		"StructOf":        true,
		"StructTag":       true,
		"Swapper":         true,
		"TypeOf":          true,
		"Uint":            true,
		"Uint16":          true,
		"Uint32":          true,
		"Uint64":          true,
		"Uint8":           true,
		"Uintptr":         true,
		"UnsafePointer":   true,
		"Value":           true,
		"ValueError":      true,
		"ValueOf":         true,
		"Zero":            true,
	},
	"regexp": map[string]bool{
		"Compile":          true,
		"CompilePOSIX":     true,
		"Match":            true,
		"MatchReader":      true,
		"MatchString":      true,
		"MustCompile":      true,
		"MustCompilePOSIX": true,
		"QuoteMeta":        true,
		"Regexp":           true,
	},
	"regexp/syntax": map[string]bool{
		"ClassNL":                  true,
		"Compile":                  true,
		"DotNL":                    true,
		"EmptyBeginLine":           true,
		"EmptyBeginText":           true,
		"EmptyEndLine":             true,
		"EmptyEndText":             true,
		"EmptyNoWordBoundary":      true,
		"EmptyOp":                  true,
		"EmptyOpContext":           true,
		"EmptyWordBoundary":        true,
		"ErrInternalError":         true,
		"ErrInvalidCharClass":      true,
		"ErrInvalidCharRange":      true,
		"ErrInvalidEscape":         true,
		"ErrInvalidNamedCapture":   true,
		"ErrInvalidPerlOp":         true,
		"ErrInvalidRepeatOp":       true,
		"ErrInvalidRepeatSize":     true,
		"ErrInvalidUTF8":           true,
		"ErrMissingBracket":        true,
		"ErrMissingParen":          true,
		"ErrMissingRepeatArgument": true,
		"ErrTrailingBackslash":     true,
		"ErrUnexpectedParen":       true,
		"Error":                    true,
		"ErrorCode":                true,
		"Flags":                    true,
		"FoldCase":                 true,
		"Inst":                     true,
		"InstAlt":                  true,
		"InstAltMatch":             true,
		"InstCapture":              true,
		"InstEmptyWidth":           true,
		"InstFail":                 true,
		"InstMatch":                true,
		"InstNop":                  true,
		"InstOp":                   true,
		"InstRune":                 true,
		"InstRune1":                true,
		"InstRuneAny":              true,
		"InstRuneAnyNotNL":         true,
		"IsWordChar":               true,
		"Literal":                  true,
		"MatchNL":                  true,
		"NonGreedy":                true,
		"OneLine":                  true,
		"Op":                       true,
		"OpAlternate":              true,
		"OpAnyChar":                true,
		"OpAnyCharNotNL":           true,
		"OpBeginLine":              true,
		"OpBeginText":              true,
		"OpCapture":                true,
		"OpCharClass":              true,
		"OpConcat":                 true,
		"OpEmptyMatch":             true,
		"OpEndLine":                true,
		"OpEndText":                true,
		"OpLiteral":                true,
		"OpNoMatch":                true,
		"OpNoWordBoundary":         true,
		"OpPlus":                   true,
		"OpQuest":                  true,
		"OpRepeat":                 true,
		"OpStar":                   true,
		"OpWordBoundary":           true,
		"POSIX":                    true,
		"Parse":                    true,
		"Perl":                     true,
		"PerlX":                    true,
		"Prog":                     true,
		"Regexp":                   true,
		"Simple":                   true,
		"UnicodeGroups":            true,
		"WasDollar":                true,
	},
	"runtime": map[string]bool{
		"BlockProfile":            true,
		"BlockProfileRecord":      true,
		"Breakpoint":              true,
		"CPUProfile":              true,
		"Caller":                  true,
		"Callers":                 true,
		"CallersFrames":           true,
		"Compiler":                true,
		"Error":                   true,
		"Frame":                   true,
		"Frames":                  true,
		"Func":                    true,
		"FuncForPC":               true,
		"GC":                      true,
		"GOARCH":                  true,
		"GOMAXPROCS":              true,
		"GOOS":                    true,
		"GOROOT":                  true,
		"Goexit":                  true,
		"GoroutineProfile":        true,
		"Gosched":                 true,
		"KeepAlive":               true,
		"LockOSThread":            true,
		"MemProfile":              true,
		"MemProfileRate":          true,
		"MemProfileRecord":        true,
		"MemStats":                true,
		"MutexProfile":            true,
		"NumCPU":                  true,
		"NumCgoCall":              true,
		"NumGoroutine":            true,
		"ReadMemStats":            true,
		"ReadTrace":               true,
		"SetBlockProfileRate":     true,
		"SetCPUProfileRate":       true,
		"SetCgoTraceback":         true,
		"SetFinalizer":            true,
		"SetMutexProfileFraction": true,
		"Stack":                   true,
		"StackRecord":             true,
		"StartTrace":              true,
		"StopTrace":               true,
		"ThreadCreateProfile":     true,
		"TypeAssertionError":      true,
		"UnlockOSThread":          true,
		"Version":                 true,
	},
	"runtime/debug": map[string]bool{
		"BuildInfo":       true,
		"FreeOSMemory":    true,
		"GCStats":         true,
		"Module":          true,
		"PrintStack":      true,
		"ReadBuildInfo":   true,
		"ReadGCStats":     true,
		"SetGCPercent":    true,
		"SetMaxStack":     true,
		"SetMaxThreads":   true,
		"SetPanicOnFault": true,
		"SetTraceback":    true,
		"Stack":           true,
		"WriteHeapDump":   true,
	},
	"runtime/pprof": map[string]bool{
		"Do":                 true,
		"ForLabels":          true,
		"Label":              true,
		"LabelSet":           true,
		"Labels":             true,
		"Lookup":             true,
		"NewProfile":         true,
		"Profile":            true,
		"Profiles":           true,
		"SetGoroutineLabels": true,
		"StartCPUProfile":    true,
		"StopCPUProfile":     true,
		"WithLabels":         true,
		"WriteHeapProfile":   true,
	},
	"runtime/trace": map[string]bool{
		"IsEnabled":   true,
		"Log":         true,
		"Logf":        true,
		"NewTask":     true,
		"Region":      true,
		"Start":       true,
		"StartRegion": true,
		"Stop":        true,
		"Task":        true,
		"WithRegion":  true,
	},
	"sort": map[string]bool{
		"Float64Slice":      true,
		"Float64s":          true,
		"Float64sAreSorted": true,
		"IntSlice":          true,
		"Interface":         true,
		"Ints":              true,
		"IntsAreSorted":     true,
		"IsSorted":          true,
		"Reverse":           true,
		"Search":            true,
		"SearchFloat64s":    true,
		"SearchInts":        true,
		"SearchStrings":     true,
		"Slice":             true,
		"SliceIsSorted":     true,
		"SliceStable":       true,
		"Sort":              true,
		"Stable":            true,
		"StringSlice":       true,
		"Strings":           true,
		"StringsAreSorted":  true,
	},
	"strconv": map[string]bool{
		"AppendBool":               true,
		"AppendFloat":              true,
		"AppendInt":                true,
		"AppendQuote":              true,
		"AppendQuoteRune":          true,
		"AppendQuoteRuneToASCII":   true,
		"AppendQuoteRuneToGraphic": true,
		"AppendQuoteToASCII":       true,
		"AppendQuoteToGraphic":     true,
		"AppendUint":               true,
		"Atoi":                     true,
		"CanBackquote":             true,
		"ErrRange":                 true,
		"ErrSyntax":                true,
		"FormatBool":               true,
		"FormatFloat":              true,
		"FormatInt":                true,
		"FormatUint":               true,
		"IntSize":                  true,
		"IsGraphic":                true,
		"IsPrint":                  true,
		"Itoa":                     true,
		"NumError":                 true,
		"ParseBool":                true,
		"ParseFloat":               true,
		"ParseInt":                 true,
		"ParseUint":                true,
		"Quote":                    true,
		"QuoteRune":                true,
		"QuoteRuneToASCII":         true,
		"QuoteRuneToGraphic":       true,
		"QuoteToASCII":             true,
		"QuoteToGraphic":           true,
		"Unquote":                  true,
		"UnquoteChar":              true,
	},
	"strings": map[string]bool{
		"Builder":        true,
		"Compare":        true,
		"Contains":       true,
		"ContainsAny":    true,
		"ContainsRune":   true,
		"Count":          true,
		"EqualFold":      true,
		"Fields":         true,
		"FieldsFunc":     true,
		"HasPrefix":      true,
		"HasSuffix":      true,
		"Index":          true,
		"IndexAny":       true,
		"IndexByte":      true,
		"IndexFunc":      true,
		"IndexRune":      true,
		"Join":           true,
		"LastIndex":      true,
		"LastIndexAny":   true,
		"LastIndexByte":  true,
		"LastIndexFunc":  true,
		"Map":            true,
		"NewReader":      true,
		"NewReplacer":    true,
		"Reader":         true,
		"Repeat":         true,
		"Replace":        true,
		"ReplaceAll":     true,
		"Replacer":       true,
		"Split":          true,
		"SplitAfter":     true,
		"SplitAfterN":    true,
		"SplitN":         true,
		"Title":          true,
		"ToLower":        true,
		"ToLowerSpecial": true,
		"ToTitle":        true,
		"ToTitleSpecial": true,
		"ToUpper":        true,
		"ToUpperSpecial": true,
		"Trim":           true,
		"TrimFunc":       true,
		"TrimLeft":       true,
		"TrimLeftFunc":   true,
		"TrimPrefix":     true,
		"TrimRight":      true,
		"TrimRightFunc":  true,
		"TrimSpace":      true,
		"TrimSuffix":     true,
	},
	"sync": map[string]bool{
		"Cond":      true,
		"Locker":    true,
		"Map":       true,
		"Mutex":     true,
		"NewCond":   true,
		"Once":      true,
		"Pool":      true,
		"RWMutex":   true,
		"WaitGroup": true,
	},
	"sync/atomic": map[string]bool{
		"AddInt32":              true,
		"AddInt64":              true,
		"AddUint32":             true,
		"AddUint64":             true,
		"AddUintptr":            true,
		"CompareAndSwapInt32":   true,
		"CompareAndSwapInt64":   true,
		"CompareAndSwapPointer": true,
		"CompareAndSwapUint32":  true,
		"CompareAndSwapUint64":  true,
		"CompareAndSwapUintptr": true,
		"LoadInt32":             true,
		"LoadInt64":             true,
		"LoadPointer":           true,
		"LoadUint32":            true,
		"LoadUint64":            true,
		"LoadUintptr":           true,
		"StoreInt32":            true,
		"StoreInt64":            true,
		"StorePointer":          true,
		"StoreUint32":           true,
		"StoreUint64":           true,
		"StoreUintptr":          true,
		"SwapInt32":             true,
		"SwapInt64":             true,
		"SwapPointer":           true,
		"SwapUint32":            true,
		"SwapUint64":            true,
		"SwapUintptr":           true,
		"Value":                 true,
	},
	"syscall": map[string]bool{
		"AF_ALG":                              true,
		"AF_APPLETALK":                        true,
		"AF_ARP":                              true,
		"AF_ASH":                              true,
		"AF_ATM":                              true,
		"AF_ATMPVC":                           true,
		"AF_ATMSVC":                           true,
		"AF_AX25":                             true,
		"AF_BLUETOOTH":                        true,
		"AF_BRIDGE":                           true,
		"AF_CAIF":                             true,
		"AF_CAN":                              true,
		"AF_CCITT":                            true,
		"AF_CHAOS":                            true,
		"AF_CNT":                              true,
		"AF_COIP":                             true,
		"AF_DATAKIT":                          true,
		"AF_DECnet":                           true,
		"AF_DLI":                              true,
		"AF_E164":                             true,
		"AF_ECMA":                             true,
		"AF_ECONET":                           true,
		"AF_ENCAP":                            true,
		"AF_FILE":                             true,
		"AF_HYLINK":                           true,
		"AF_IEEE80211":                        true,
		"AF_IEEE802154":                       true,
		"AF_IMPLINK":                          true,
		"AF_INET":                             true,
		"AF_INET6":                            true,
		"AF_INET6_SDP":                        true,
		"AF_INET_SDP":                         true,
		"AF_IPX":                              true,
		"AF_IRDA":                             true,
		"AF_ISDN":                             true,
		"AF_ISO":                              true,
		"AF_IUCV":                             true,
		"AF_KEY":                              true,
		"AF_LAT":                              true,
		"AF_LINK":                             true,
		"AF_LLC":                              true,
		"AF_LOCAL":                            true,
		"AF_MAX":                              true,
		"AF_MPLS":                             true,
		"AF_NATM":                             true,
		"AF_NDRV":                             true,
		"AF_NETBEUI":                          true,
		"AF_NETBIOS":                          true,
		"AF_NETGRAPH":                         true,
		"AF_NETLINK":                          true,
		"AF_NETROM":                           true,
		"AF_NS":                               true,
		"AF_OROUTE":                           true,
		"AF_OSI":                              true,
		"AF_PACKET":                           true,
		"AF_PHONET":                           true,
		"AF_PPP":                              true,
		"AF_PPPOX":                            true,
		"AF_PUP":                              true,
		"AF_RDS":                              true,
		"AF_RESERVED_36":                      true,
		"AF_ROSE":                             true,
		"AF_ROUTE":                            true,
		"AF_RXRPC":                            true,
		"AF_SCLUSTER":                         true,
		"AF_SECURITY":                         true,
		"AF_SIP":                              true,
		"AF_SLOW":                             true,
		"AF_SNA":                              true,
		"AF_SYSTEM":                           true,
		"AF_TIPC":                             true,
		"AF_UNIX":                             true,
		"AF_UNSPEC":                           true,
		"AF_VENDOR00":                         true,
		"AF_VENDOR01":                         true,
		"AF_VENDOR02":                         true,
		"AF_VENDOR03":                         true,
		"AF_VENDOR04":                         true,
		"AF_VENDOR05":                         true,
		"AF_VENDOR06":                         true,
		"AF_VENDOR07":                         true,
		"AF_VENDOR08":                         true,
		"AF_VENDOR09":                         true,
		"AF_VENDOR10":                         true,
		"AF_VENDOR11":                         true,
		"AF_VENDOR12":                         true,
		"AF_VENDOR13":                         true,
		"AF_VENDOR14":                         true,
		"AF_VENDOR15":                         true,
		"AF_VENDOR16":                         true,
		"AF_VENDOR17":                         true,
		"AF_VENDOR18":                         true,
		"AF_VENDOR19":                         true,
		"AF_VENDOR20":                         true,
		"AF_VENDOR21":                         true,
		"AF_VENDOR22":                         true,
		"AF_VENDOR23":                         true,
		"AF_VENDOR24":                         true,
		"AF_VENDOR25":                         true,
		"AF_VENDOR26":                         true,
		"AF_VENDOR27":                         true,
		"AF_VENDOR28":                         true,
		"AF_VENDOR29":                         true,
		"AF_VENDOR30":                         true,
		"AF_VENDOR31":                         true,
		"AF_VENDOR32":                         true,
		"AF_VENDOR33":                         true,
		"AF_VENDOR34":                         true,
		"AF_VENDOR35":                         true,
		"AF_VENDOR36":                         true,
		"AF_VENDOR37":                         true,
		"AF_VENDOR38":                         true,
		"AF_VENDOR39":                         true,
		"AF_VENDOR40":                         true,
		"AF_VENDOR41":                         true,
		"AF_VENDOR42":                         true,
		"AF_VENDOR43":                         true,
		"AF_VENDOR44":                         true,
		"AF_VENDOR45":                         true,
		"AF_VENDOR46":                         true,
		"AF_VENDOR47":                         true,
		"AF_WANPIPE":                          true,
		"AF_X25":                              true,
		"AI_CANONNAME":                        true,
		"AI_NUMERICHOST":                      true,
		"AI_PASSIVE":                          true,
		"APPLICATION_ERROR":                   true,
		"ARPHRD_ADAPT":                        true,
		"ARPHRD_APPLETLK":                     true,
		"ARPHRD_ARCNET":                       true,
		"ARPHRD_ASH":                          true,
		"ARPHRD_ATM":                          true,
		"ARPHRD_AX25":                         true,
		"ARPHRD_BIF":                          true,
		"ARPHRD_CHAOS":                        true,
		"ARPHRD_CISCO":                        true,
		"ARPHRD_CSLIP":                        true,
		"ARPHRD_CSLIP6":                       true,
		"ARPHRD_DDCMP":                        true,
		"ARPHRD_DLCI":                         true,
		"ARPHRD_ECONET":                       true,
		"ARPHRD_EETHER":                       true,
		"ARPHRD_ETHER":                        true,
		"ARPHRD_EUI64":                        true,
		"ARPHRD_FCAL":                         true,
		"ARPHRD_FCFABRIC":                     true,
		"ARPHRD_FCPL":                         true,
		"ARPHRD_FCPP":                         true,
		"ARPHRD_FDDI":                         true,
		"ARPHRD_FRAD":                         true,
		"ARPHRD_FRELAY":                       true,
		"ARPHRD_HDLC":                         true,
		"ARPHRD_HIPPI":                        true,
		"ARPHRD_HWX25":                        true,
		"ARPHRD_IEEE1394":                     true,
		"ARPHRD_IEEE802":                      true,
		"ARPHRD_IEEE80211":                    true,
		"ARPHRD_IEEE80211_PRISM":              true,
		"ARPHRD_IEEE80211_RADIOTAP":           true,
		"ARPHRD_IEEE802154":                   true,
		"ARPHRD_IEEE802154_PHY":               true,
		"ARPHRD_IEEE802_TR":                   true,
		"ARPHRD_INFINIBAND":                   true,
		"ARPHRD_IPDDP":                        true,
		"ARPHRD_IPGRE":                        true,
		"ARPHRD_IRDA":                         true,
		"ARPHRD_LAPB":                         true,
		"ARPHRD_LOCALTLK":                     true,
		"ARPHRD_LOOPBACK":                     true,
		"ARPHRD_METRICOM":                     true,
		"ARPHRD_NETROM":                       true,
		"ARPHRD_NONE":                         true,
		"ARPHRD_PIMREG":                       true,
		"ARPHRD_PPP":                          true,
		"ARPHRD_PRONET":                       true,
		"ARPHRD_RAWHDLC":                      true,
		"ARPHRD_ROSE":                         true,
		"ARPHRD_RSRVD":                        true,
		"ARPHRD_SIT":                          true,
		"ARPHRD_SKIP":                         true,
		"ARPHRD_SLIP":                         true,
		"ARPHRD_SLIP6":                        true,
		"ARPHRD_STRIP":                        true,
		"ARPHRD_TUNNEL":                       true,
		"ARPHRD_TUNNEL6":                      true,
		"ARPHRD_VOID":                         true,
		"ARPHRD_X25":                          true,
		"AUTHTYPE_CLIENT":                     true,
		"AUTHTYPE_SERVER":                     true,
		"Accept":                              true,
		"Accept4":                             true,
		"AcceptEx":                            true,
		"Access":                              true,
		"Acct":                                true,
		"AddrinfoW":                           true,
		"Adjtime":                             true,
		"Adjtimex":                            true,
		"AttachLsf":                           true,
		"B0":                                  true,
		"B1000000":                            true,
		"B110":                                true,
		"B115200":                             true,
		"B1152000":                            true,
		"B1200":                               true,
		"B134":                                true,
		"B14400":                              true,
		"B150":                                true,
		"B1500000":                            true,
		"B1800":                               true,
		"B19200":                              true,
		"B200":                                true,
		"B2000000":                            true,
		"B230400":                             true,
		"B2400":                               true,
		"B2500000":                            true,
		"B28800":                              true,
		"B300":                                true,
		"B3000000":                            true,
		"B3500000":                            true,
		"B38400":                              true,
		"B4000000":                            true,
		"B460800":                             true,
		"B4800":                               true,
		"B50":                                 true,
		"B500000":                             true,
		"B57600":                              true,
		"B576000":                             true,
		"B600":                                true,
		"B7200":                               true,
		"B75":                                 true,
		"B76800":                              true,
		"B921600":                             true,
		"B9600":                               true,
		"BASE_PROTOCOL":                       true,
		"BIOCFEEDBACK":                        true,
		"BIOCFLUSH":                           true,
		"BIOCGBLEN":                           true,
		"BIOCGDIRECTION":                      true,
		"BIOCGDIRFILT":                        true,
		"BIOCGDLT":                            true,
		"BIOCGDLTLIST":                        true,
		"BIOCGETBUFMODE":                      true,
		"BIOCGETIF":                           true,
		"BIOCGETZMAX":                         true,
		"BIOCGFEEDBACK":                       true,
		"BIOCGFILDROP":                        true,
		"BIOCGHDRCMPLT":                       true,
		"BIOCGRSIG":                           true,
		"BIOCGRTIMEOUT":                       true,
		"BIOCGSEESENT":                        true,
		"BIOCGSTATS":                          true,
		"BIOCGSTATSOLD":                       true,
		"BIOCGTSTAMP":                         true,
		"BIOCIMMEDIATE":                       true,
		"BIOCLOCK":                            true,
		"BIOCPROMISC":                         true,
		"BIOCROTZBUF":                         true,
		"BIOCSBLEN":                           true,
		"BIOCSDIRECTION":                      true,
		"BIOCSDIRFILT":                        true,
		"BIOCSDLT":                            true,
		"BIOCSETBUFMODE":                      true,
		"BIOCSETF":                            true,
		"BIOCSETFNR":                          true,
		"BIOCSETIF":                           true,
		"BIOCSETWF":                           true,
		"BIOCSETZBUF":                         true,
		"BIOCSFEEDBACK":                       true,
		"BIOCSFILDROP":                        true,
		"BIOCSHDRCMPLT":                       true,
		"BIOCSRSIG":                           true,
		"BIOCSRTIMEOUT":                       true,
		"BIOCSSEESENT":                        true,
		"BIOCSTCPF":                           true,
		"BIOCSTSTAMP":                         true,
		"BIOCSUDPF":                           true,
		"BIOCVERSION":                         true,
		"BPF_A":                               true,
		"BPF_ABS":                             true,
		"BPF_ADD":                             true,
		"BPF_ALIGNMENT":                       true,
		"BPF_ALIGNMENT32":                     true,
		"BPF_ALU":                             true,
		"BPF_AND":                             true,
		"BPF_B":                               true,
		"BPF_BUFMODE_BUFFER":                  true,
		"BPF_BUFMODE_ZBUF":                    true,
		"BPF_DFLTBUFSIZE":                     true,
		"BPF_DIRECTION_IN":                    true,
		"BPF_DIRECTION_OUT":                   true,
		"BPF_DIV":                             true,
		"BPF_H":                               true,
		"BPF_IMM":                             true,
		"BPF_IND":                             true,
		"BPF_JA":                              true,
		"BPF_JEQ":                             true,
		"BPF_JGE":                             true,
		"BPF_JGT":                             true,
		"BPF_JMP":                             true,
		"BPF_JSET":                            true,
		"BPF_K":                               true,
		"BPF_LD":                              true,
		"BPF_LDX":                             true,
		"BPF_LEN":                             true,
		"BPF_LSH":                             true,
		"BPF_MAJOR_VERSION":                   true,
		"BPF_MAXBUFSIZE":                      true,
		"BPF_MAXINSNS":                        true,
		"BPF_MEM":                             true,
		"BPF_MEMWORDS":                        true,
		"BPF_MINBUFSIZE":                      true,
		"BPF_MINOR_VERSION":                   true,
		"BPF_MISC":                            true,
		"BPF_MSH":                             true,
		"BPF_MUL":                             true,
		"BPF_NEG":                             true,
		"BPF_OR":                              true,
		"BPF_RELEASE":                         true,
		"BPF_RET":                             true,
		"BPF_RSH":                             true,
		"BPF_ST":                              true,
		"BPF_STX":                             true,
		"BPF_SUB":                             true,
		"BPF_TAX":                             true,
		"BPF_TXA":                             true,
		"BPF_T_BINTIME":                       true,
		"BPF_T_BINTIME_FAST":                  true,
		"BPF_T_BINTIME_MONOTONIC":             true,
		"BPF_T_BINTIME_MONOTONIC_FAST":        true,
		"BPF_T_FAST":                          true,
		"BPF_T_FLAG_MASK":                     true,
		"BPF_T_FORMAT_MASK":                   true,
		"BPF_T_MICROTIME":                     true,
		"BPF_T_MICROTIME_FAST":                true,
		"BPF_T_MICROTIME_MONOTONIC":           true,
		"BPF_T_MICROTIME_MONOTONIC_FAST":      true,
		"BPF_T_MONOTONIC":                     true,
		"BPF_T_MONOTONIC_FAST":                true,
		"BPF_T_NANOTIME":                      true,
		"BPF_T_NANOTIME_FAST":                 true,
		"BPF_T_NANOTIME_MONOTONIC":            true,
		"BPF_T_NANOTIME_MONOTONIC_FAST":       true,
		"BPF_T_NONE":                          true,
		"BPF_T_NORMAL":                        true,
		"BPF_W":                               true,
		"BPF_X":                               true,
		"BRKINT":                              true,
		"Bind":                                true,
		"BindToDevice":                        true,
		"BpfBuflen":                           true,
		"BpfDatalink":                         true,
		"BpfHdr":                              true,
		"BpfHeadercmpl":                       true,
		"BpfInsn":                             true,
		"BpfInterface":                        true,
		"BpfJump":                             true,
		"BpfProgram":                          true,
		"BpfStat":                             true,
		"BpfStats":                            true,
		"BpfStmt":                             true,
		"BpfTimeout":                          true,
		"BpfTimeval":                          true,
		"BpfVersion":                          true,
		"BpfZbuf":                             true,
		"BpfZbufHeader":                       true,
		"ByHandleFileInformation":             true,
		"BytePtrFromString":                   true,
		"ByteSliceFromString":                 true,
		"CCR0_FLUSH":                          true,
		"CERT_CHAIN_POLICY_AUTHENTICODE":      true,
		"CERT_CHAIN_POLICY_AUTHENTICODE_TS":   true,
		"CERT_CHAIN_POLICY_BASE":              true,
		"CERT_CHAIN_POLICY_BASIC_CONSTRAINTS": true,
		"CERT_CHAIN_POLICY_EV":                true,
		"CERT_CHAIN_POLICY_MICROSOFT_ROOT":    true,
		"CERT_CHAIN_POLICY_NT_AUTH":           true,
		"CERT_CHAIN_POLICY_SSL":               true,
		"CERT_E_CN_NO_MATCH":                  true,
		"CERT_E_EXPIRED":                      true,
		"CERT_E_PURPOSE":                      true,
		"CERT_E_ROLE":                         true,
		"CERT_E_UNTRUSTEDROOT":                true,
		"CERT_STORE_ADD_ALWAYS":               true,
		"CERT_STORE_DEFER_CLOSE_UNTIL_LAST_FREE_FLAG":  true,
		"CERT_STORE_PROV_MEMORY":                       true,
		"CERT_TRUST_HAS_EXCLUDED_NAME_CONSTRAINT":      true,
		"CERT_TRUST_HAS_NOT_DEFINED_NAME_CONSTRAINT":   true,
		"CERT_TRUST_HAS_NOT_PERMITTED_NAME_CONSTRAINT": true,
		"CERT_TRUST_HAS_NOT_SUPPORTED_CRITICAL_EXT":    true,
		"CERT_TRUST_HAS_NOT_SUPPORTED_NAME_CONSTRAINT": true,
		"CERT_TRUST_INVALID_BASIC_CONSTRAINTS":         true,
		"CERT_TRUST_INVALID_EXTENSION":                 true,
		"CERT_TRUST_INVALID_NAME_CONSTRAINTS":          true,
		"CERT_TRUST_INVALID_POLICY_CONSTRAINTS":        true,
		"CERT_TRUST_IS_CYCLIC":                         true,
		"CERT_TRUST_IS_EXPLICIT_DISTRUST":              true,
		"CERT_TRUST_IS_NOT_SIGNATURE_VALID":            true,
		"CERT_TRUST_IS_NOT_TIME_VALID":                 true,
		"CERT_TRUST_IS_NOT_VALID_FOR_USAGE":            true,
		"CERT_TRUST_IS_OFFLINE_REVOCATION":             true,
		"CERT_TRUST_IS_REVOKED":                        true,
		"CERT_TRUST_IS_UNTRUSTED_ROOT":                 true,
		"CERT_TRUST_NO_ERROR":                          true,
		"CERT_TRUST_NO_ISSUANCE_CHAIN_POLICY":          true,
		"CERT_TRUST_REVOCATION_STATUS_UNKNOWN":         true,
		"CFLUSH":                                       true,
		"CLOCAL":                                       true,
		"CLONE_CHILD_CLEARTID":                         true,
		"CLONE_CHILD_SETTID":                           true,
		"CLONE_CSIGNAL":                                true,
		"CLONE_DETACHED":                               true,
		"CLONE_FILES":                                  true,
		"CLONE_FS":                                     true,
		"CLONE_IO":                                     true,
		"CLONE_NEWIPC":                                 true,
		"CLONE_NEWNET":                                 true,
		"CLONE_NEWNS":                                  true,
		"CLONE_NEWPID":                                 true,
		"CLONE_NEWUSER":                                true,
		"CLONE_NEWUTS":                                 true,
		"CLONE_PARENT":                                 true,
		"CLONE_PARENT_SETTID":                          true,
		"CLONE_PID":                                    true,
		"CLONE_PTRACE":                                 true,
		"CLONE_SETTLS":                                 true,
		"CLONE_SIGHAND":                                true,
		"CLONE_SYSVSEM":                                true,
		"CLONE_THREAD":                                 true,
		"CLONE_UNTRACED":                               true,
		"CLONE_VFORK":                                  true,
		"CLONE_VM":                                     true,
		"CPUID_CFLUSH":                                 true,
		"CREAD":                                        true,
		"CREATE_ALWAYS":                                true,
		"CREATE_NEW":                                   true,
		"CREATE_NEW_PROCESS_GROUP":                     true,
		"CREATE_UNICODE_ENVIRONMENT":                   true,
		"CRYPT_DEFAULT_CONTAINER_OPTIONAL":             true,
		"CRYPT_DELETEKEYSET":                           true,
		"CRYPT_MACHINE_KEYSET":                         true,
		"CRYPT_NEWKEYSET":                              true,
		"CRYPT_SILENT":                                 true,
		"CRYPT_VERIFYCONTEXT":                          true,
		"CS5":                                          true,
		"CS6":                                          true,
		"CS7":                                          true,
		"CS8":                                          true,
		"CSIZE":                                        true,
		"CSTART":                                       true,
		"CSTATUS":                                      true,
		"CSTOP":                                        true,
		"CSTOPB":                                       true,
		"CSUSP":                                        true,
		"CTL_MAXNAME":                                  true,
		"CTL_NET":                                      true,
		"CTL_QUERY":                                    true,
		"CTRL_BREAK_EVENT":                             true,
		"CTRL_C_EVENT":                                 true,
		"CancelIo":                                     true,
		"CancelIoEx":                                   true,
		"CertAddCertificateContextToStore":             true,
		"CertChainContext":                             true,
		"CertChainElement":                             true,
		"CertChainPara":                                true,
		"CertChainPolicyPara":                          true,
		"CertChainPolicyStatus":                        true,
		"CertCloseStore":                               true,
		"CertContext":                                  true,
		"CertCreateCertificateContext":                 true,
		"CertEnhKeyUsage":                              true,
		"CertEnumCertificatesInStore":                  true,
		"CertFreeCertificateChain":                     true,
		"CertFreeCertificateContext":                   true,
		"CertGetCertificateChain":                      true,
		"CertInfo":                                     true,
		"CertOpenStore":                                true,
		"CertOpenSystemStore":                          true,
		"CertRevocationCrlInfo":                        true,
		"CertRevocationInfo":                           true,
		"CertSimpleChain":                              true,
		"CertTrustListInfo":                            true,
		"CertTrustStatus":                              true,
		"CertUsageMatch":                               true,
		"CertVerifyCertificateChainPolicy":             true,
		"Chdir":                                        true,
		"CheckBpfVersion":                              true,
		"Chflags":                                      true,
		"Chmod":                                        true,
		"Chown":                                        true,
		"Chroot":                                       true,
		"Clearenv":                                     true,
		"Close":                                        true,
		"CloseHandle":                                  true,
		"CloseOnExec":                                  true,
		"Closesocket":                                  true,
		"CmsgLen":                                      true,
		"CmsgSpace":                                    true,
		"Cmsghdr":                                      true,
		"CommandLineToArgv":                            true,
		"ComputerName":                                 true,
		"Conn":                                         true,
		"Connect":                                      true,
		"ConnectEx":                                    true,
		"ConvertSidToStringSid":                        true,
		"ConvertStringSidToSid":                        true,
		"CopySid":                                      true,
		"Creat":                                        true,
		"CreateDirectory":                              true,
		"CreateFile":                                   true,
		"CreateFileMapping":                            true,
		"CreateHardLink":                               true,
		"CreateIoCompletionPort":                       true,
		"CreatePipe":                                   true,
		"CreateProcess":                                true,
		"CreateProcessAsUser":                          true,
		"CreateSymbolicLink":                           true,
		"CreateToolhelp32Snapshot":                     true,
		"Credential":                                   true,
		"CryptAcquireContext":                          true,
		"CryptGenRandom":                               true,
		"CryptReleaseContext":                          true,
		"DIOCBSFLUSH":                                  true,
		"DIOCOSFPFLUSH":                                true,
		"DLL":                                          true,
		"DLLError":                                     true,
		"DLT_A429":                                     true,
		"DLT_A653_ICM":                                 true,
		"DLT_AIRONET_HEADER":                           true,
		"DLT_AOS":                                      true,
		"DLT_APPLE_IP_OVER_IEEE1394":                   true,
		"DLT_ARCNET":                                   true,
		"DLT_ARCNET_LINUX":                             true,
		"DLT_ATM_CLIP":                                 true,
		"DLT_ATM_RFC1483":                              true,
		"DLT_AURORA":                                   true,
		"DLT_AX25":                                     true,
		"DLT_AX25_KISS":                                true,
		"DLT_BACNET_MS_TP":                             true,
		"DLT_BLUETOOTH_HCI_H4":                         true,
		"DLT_BLUETOOTH_HCI_H4_WITH_PHDR":               true,
		"DLT_CAN20B":                                   true,
		"DLT_CAN_SOCKETCAN":                            true,
		"DLT_CHAOS":                                    true,
		"DLT_CHDLC":                                    true,
		"DLT_CISCO_IOS":                                true,
		"DLT_C_HDLC":                                   true,
		"DLT_C_HDLC_WITH_DIR":                          true,
		"DLT_DBUS":                                     true,
		"DLT_DECT":                                     true,
		"DLT_DOCSIS":                                   true,
		"DLT_DVB_CI":                                   true,
		"DLT_ECONET":                                   true,
		"DLT_EN10MB":                                   true,
		"DLT_EN3MB":                                    true,
		"DLT_ENC":                                      true,
		"DLT_ERF":                                      true,
		"DLT_ERF_ETH":                                  true,
		"DLT_ERF_POS":                                  true,
		"DLT_FC_2":                                     true,
		"DLT_FC_2_WITH_FRAME_DELIMS":                   true,
		"DLT_FDDI":                                     true,
		"DLT_FLEXRAY":                                  true,
		"DLT_FRELAY":                                   true,
		"DLT_FRELAY_WITH_DIR":                          true,
		"DLT_GCOM_SERIAL":                              true,
		"DLT_GCOM_T1E1":                                true,
		"DLT_GPF_F":                                    true,
		"DLT_GPF_T":                                    true,
		"DLT_GPRS_LLC":                                 true,
		"DLT_GSMTAP_ABIS":                              true,
		"DLT_GSMTAP_UM":                                true,
		"DLT_HDLC":                                     true,
		"DLT_HHDLC":                                    true,
		"DLT_HIPPI":                                    true,
		"DLT_IBM_SN":                                   true,
		"DLT_IBM_SP":                                   true,
		"DLT_IEEE802":                                  true,
		"DLT_IEEE802_11":                               true,
		"DLT_IEEE802_11_RADIO":                         true,
		"DLT_IEEE802_11_RADIO_AVS":                     true,
		"DLT_IEEE802_15_4":                             true,
		"DLT_IEEE802_15_4_LINUX":                       true,
		"DLT_IEEE802_15_4_NOFCS":                       true,
		"DLT_IEEE802_15_4_NONASK_PHY":                  true,
		"DLT_IEEE802_16_MAC_CPS":                       true,
		"DLT_IEEE802_16_MAC_CPS_RADIO":                 true,
		"DLT_IPFILTER":                                 true,
		"DLT_IPMB":                                     true,
		"DLT_IPMB_LINUX":                               true,
		"DLT_IPNET":                                    true,
		"DLT_IPOIB":                                    true,
		"DLT_IPV4":                                     true,
		"DLT_IPV6":                                     true,
		"DLT_IP_OVER_FC":                               true,
		"DLT_JUNIPER_ATM1":                             true,
		"DLT_JUNIPER_ATM2":                             true,
		"DLT_JUNIPER_ATM_CEMIC":                        true,
		"DLT_JUNIPER_CHDLC":                            true,
		"DLT_JUNIPER_ES":                               true,
		"DLT_JUNIPER_ETHER":                            true,
		"DLT_JUNIPER_FIBRECHANNEL":                     true,
		"DLT_JUNIPER_FRELAY":                           true,
		"DLT_JUNIPER_GGSN":                             true,
		"DLT_JUNIPER_ISM":                              true,
		"DLT_JUNIPER_MFR":                              true,
		"DLT_JUNIPER_MLFR":                             true,
		"DLT_JUNIPER_MLPPP":                            true,
		"DLT_JUNIPER_MONITOR":                          true,
		"DLT_JUNIPER_PIC_PEER":                         true,
		"DLT_JUNIPER_PPP":                              true,
		"DLT_JUNIPER_PPPOE":                            true,
		"DLT_JUNIPER_PPPOE_ATM":                        true,
		"DLT_JUNIPER_SERVICES":                         true,
		"DLT_JUNIPER_SRX_E2E":                          true,
		"DLT_JUNIPER_ST":                               true,
		"DLT_JUNIPER_VP":                               true,
		"DLT_JUNIPER_VS":                               true,
		"DLT_LAPB_WITH_DIR":                            true,
		"DLT_LAPD":                                     true,
		"DLT_LIN":                                      true,
		"DLT_LINUX_EVDEV":                              true,
		"DLT_LINUX_IRDA":                               true,
		"DLT_LINUX_LAPD":                               true,
		"DLT_LINUX_PPP_WITHDIRECTION":                  true,
		"DLT_LINUX_SLL":                                true,
		"DLT_LOOP":                                     true,
		"DLT_LTALK":                                    true,
		"DLT_MATCHING_MAX":                             true,
		"DLT_MATCHING_MIN":                             true,
		"DLT_MFR":                                      true,
		"DLT_MOST":                                     true,
		"DLT_MPEG_2_TS":                                true,
		"DLT_MPLS":                                     true,
		"DLT_MTP2":                                     true,
		"DLT_MTP2_WITH_PHDR":                           true,
		"DLT_MTP3":                                     true,
		"DLT_MUX27010":                                 true,
		"DLT_NETANALYZER":                              true,
		"DLT_NETANALYZER_TRANSPARENT":                  true,
		"DLT_NFC_LLCP":                                 true,
		"DLT_NFLOG":                                    true,
		"DLT_NG40":                                     true,
		"DLT_NULL":                                     true,
		"DLT_PCI_EXP":                                  true,
		"DLT_PFLOG":                                    true,
		"DLT_PFSYNC":                                   true,
		"DLT_PPI":                                      true,
		"DLT_PPP":                                      true,
		"DLT_PPP_BSDOS":                                true,
		"DLT_PPP_ETHER":                                true,
		"DLT_PPP_PPPD":                                 true,
		"DLT_PPP_SERIAL":                               true,
		"DLT_PPP_WITH_DIR":                             true,
		"DLT_PPP_WITH_DIRECTION":                       true,
		"DLT_PRISM_HEADER":                             true,
		"DLT_PRONET":                                   true,
		"DLT_RAIF1":                                    true,
		"DLT_RAW":                                      true,
		"DLT_RAWAF_MASK":                               true,
		"DLT_RIO":                                      true,
		"DLT_SCCP":                                     true,
		"DLT_SITA":                                     true,
		"DLT_SLIP":                                     true,
		"DLT_SLIP_BSDOS":                               true,
		"DLT_STANAG_5066_D_PDU":                        true,
		"DLT_SUNATM":                                   true,
		"DLT_SYMANTEC_FIREWALL":                        true,
		"DLT_TZSP":                                     true,
		"DLT_USB":                                      true,
		"DLT_USB_LINUX":                                true,
		"DLT_USB_LINUX_MMAPPED":                        true,
		"DLT_USER0":                                    true,
		"DLT_USER1":                                    true,
		"DLT_USER10":                                   true,
		"DLT_USER11":                                   true,
		"DLT_USER12":                                   true,
		"DLT_USER13":                                   true,
		"DLT_USER14":                                   true,
		"DLT_USER15":                                   true,
		"DLT_USER2":                                    true,
		"DLT_USER3":                                    true,
		"DLT_USER4":                                    true,
		"DLT_USER5":                                    true,
		"DLT_USER6":                                    true,
		"DLT_USER7":                                    true,
		"DLT_USER8":                                    true,
		"DLT_USER9":                                    true,
		"DLT_WIHART":                                   true,
		"DLT_X2E_SERIAL":                               true,
		"DLT_X2E_XORAYA":                               true,
		"DNSMXData":                                    true,
		"DNSPTRData":                                   true,
		"DNSRecord":                                    true,
		"DNSSRVData":                                   true,
		"DNSTXTData":                                   true,
		"DNS_INFO_NO_RECORDS":                          true,
		"DNS_TYPE_A":                                   true,
		"DNS_TYPE_A6":                                  true,
		"DNS_TYPE_AAAA":                                true,
		"DNS_TYPE_ADDRS":                               true,
		"DNS_TYPE_AFSDB":                               true,
		"DNS_TYPE_ALL":                                 true,
		"DNS_TYPE_ANY":                                 true,
		"DNS_TYPE_ATMA":                                true,
		"DNS_TYPE_AXFR":                                true,
		"DNS_TYPE_CERT":                                true,
		"DNS_TYPE_CNAME":                               true,
		"DNS_TYPE_DHCID":                               true,
		"DNS_TYPE_DNAME":                               true,
		"DNS_TYPE_DNSKEY":                              true,
		"DNS_TYPE_DS":                                  true,
		"DNS_TYPE_EID":                                 true,
		"DNS_TYPE_GID":                                 true,
		"DNS_TYPE_GPOS":                                true,
		"DNS_TYPE_HINFO":                               true,
		"DNS_TYPE_ISDN":                                true,
		"DNS_TYPE_IXFR":                                true,
		"DNS_TYPE_KEY":                                 true,
		"DNS_TYPE_KX":                                  true,
		"DNS_TYPE_LOC":                                 true,
		"DNS_TYPE_MAILA":                               true,
		"DNS_TYPE_MAILB":                               true,
		"DNS_TYPE_MB":                                  true,
		"DNS_TYPE_MD":                                  true,
		"DNS_TYPE_MF":                                  true,
		"DNS_TYPE_MG":                                  true,
		"DNS_TYPE_MINFO":                               true,
		"DNS_TYPE_MR":                                  true,
		"DNS_TYPE_MX":                                  true,
		"DNS_TYPE_NAPTR":                               true,
		"DNS_TYPE_NBSTAT":                              true,
		"DNS_TYPE_NIMLOC":                              true,
		"DNS_TYPE_NS":                                  true,
		"DNS_TYPE_NSAP":                                true,
		"DNS_TYPE_NSAPPTR":                             true,
		"DNS_TYPE_NSEC":                                true,
		"DNS_TYPE_NULL":                                true,
		"DNS_TYPE_NXT":                                 true,
		"DNS_TYPE_OPT":                                 true,
		"DNS_TYPE_PTR":                                 true,
		"DNS_TYPE_PX":                                  true,
		"DNS_TYPE_RP":                                  true,
		"DNS_TYPE_RRSIG":                               true,
		"DNS_TYPE_RT":                                  true,
		"DNS_TYPE_SIG":                                 true,
		"DNS_TYPE_SINK":                                true,
		"DNS_TYPE_SOA":                                 true,
		"DNS_TYPE_SRV":                                 true,
		"DNS_TYPE_TEXT":                                true,
		"DNS_TYPE_TKEY":                                true,
		"DNS_TYPE_TSIG":                                true,
		"DNS_TYPE_UID":                                 true,
		"DNS_TYPE_UINFO":                               true,
		"DNS_TYPE_UNSPEC":                              true,
		"DNS_TYPE_WINS":                                true,
		"DNS_TYPE_WINSR":                               true,
		"DNS_TYPE_WKS":                                 true,
		"DNS_TYPE_X25":                                 true,
		"DT_BLK":                                       true,
		"DT_CHR":                                       true,
		"DT_DIR":                                       true,
		"DT_FIFO":                                      true,
		"DT_LNK":                                       true,
		"DT_REG":                                       true,
		"DT_SOCK":                                      true,
		"DT_UNKNOWN":                                   true,
		"DT_WHT":                                       true,
		"DUPLICATE_CLOSE_SOURCE":                       true,
		"DUPLICATE_SAME_ACCESS":                        true,
		"DeleteFile":                                   true,
		"DetachLsf":                                    true,
		"DeviceIoControl":                              true,
		"Dirent":                                       true,
		"DnsNameCompare":                               true,
		"DnsQuery":                                     true,
		"DnsRecordListFree":                            true,
		"DnsSectionAdditional":                         true,
		"DnsSectionAnswer":                             true,
		"DnsSectionAuthority":                          true,
		"DnsSectionQuestion":                           true,
		"Dup":                                          true,
		"Dup2":                                         true,
		"Dup3":                                         true,
		"DuplicateHandle":                              true,
		"E2BIG":                                        true,
		"EACCES":                                       true,
		"EADDRINUSE":                                   true,
		"EADDRNOTAVAIL":                                true,
		"EADV":                                         true,
		"EAFNOSUPPORT":                                 true,
		"EAGAIN":                                       true,
		"EALREADY":                                     true,
		"EAUTH":                                        true,
		"EBADARCH":                                     true,
		"EBADE":                                        true,
		"EBADEXEC":                                     true,
		"EBADF":                                        true,
		"EBADFD":                                       true,
		"EBADMACHO":                                    true,
		"EBADMSG":                                      true,
		"EBADR":                                        true,
		"EBADRPC":                                      true,
		"EBADRQC":                                      true,
		"EBADSLT":                                      true,
		"EBFONT":                                       true,
		"EBUSY":                                        true,
		"ECANCELED":                                    true,
		"ECAPMODE":                                     true,
		"ECHILD":                                       true,
		"ECHO":                                         true,
		"ECHOCTL":                                      true,
		"ECHOE":                                        true,
		"ECHOK":                                        true,
		"ECHOKE":                                       true,
		"ECHONL":                                       true,
		"ECHOPRT":                                      true,
		"ECHRNG":                                       true,
		"ECOMM":                                        true,
		"ECONNABORTED":                                 true,
		"ECONNREFUSED":                                 true,
		"ECONNRESET":                                   true,
		"EDEADLK":                                      true,
		"EDEADLOCK":                                    true,
		"EDESTADDRREQ":                                 true,
		"EDEVERR":                                      true,
		"EDOM":                                         true,
		"EDOOFUS":                                      true,
		"EDOTDOT":                                      true,
		"EDQUOT":                                       true,
		"EEXIST":                                       true,
		"EFAULT":                                       true,
		"EFBIG":                                        true,
		"EFER_LMA":                                     true,
		"EFER_LME":                                     true,
		"EFER_NXE":                                     true,
		"EFER_SCE":                                     true,
		"EFTYPE":                                       true,
		"EHOSTDOWN":                                    true,
		"EHOSTUNREACH":                                 true,
		"EHWPOISON":                                    true,
		"EIDRM":                                        true,
		"EILSEQ":                                       true,
		"EINPROGRESS":                                  true,
		"EINTR":                                        true,
		"EINVAL":                                       true,
		"EIO":                                          true,
		"EIPSEC":                                       true,
		"EISCONN":                                      true,
		"EISDIR":                                       true,
		"EISNAM":                                       true,
		"EKEYEXPIRED":                                  true,
		"EKEYREJECTED":                                 true,
		"EKEYREVOKED":                                  true,
		"EL2HLT":                                       true,
		"EL2NSYNC":                                     true,
		"EL3HLT":                                       true,
		"EL3RST":                                       true,
		"ELAST":                                        true,
		"ELF_NGREG":                                    true,
		"ELF_PRARGSZ":                                  true,
		"ELIBACC":                                      true,
		"ELIBBAD":                                      true,
		"ELIBEXEC":                                     true,
		"ELIBMAX":                                      true,
		"ELIBSCN":                                      true,
		"ELNRNG":                                       true,
		"ELOOP":                                        true,
		"EMEDIUMTYPE":                                  true,
		"EMFILE":                                       true,
		"EMLINK":                                       true,
		"EMSGSIZE":                                     true,
		"EMT_TAGOVF":                                   true,
		"EMULTIHOP":                                    true,
		"EMUL_ENABLED":                                 true,
		"EMUL_LINUX":                                   true,
		"EMUL_LINUX32":                                 true,
		"EMUL_MAXID":                                   true,
		"EMUL_NATIVE":                                  true,
		"ENAMETOOLONG":                                 true,
		"ENAVAIL":                                      true,
		"ENDRUNDISC":                                   true,
		"ENEEDAUTH":                                    true,
		"ENETDOWN":                                     true,
		"ENETRESET":                                    true,
		"ENETUNREACH":                                  true,
		"ENFILE":                                       true,
		"ENOANO":                                       true,
		"ENOATTR":                                      true,
		"ENOBUFS":                                      true,
		"ENOCSI":                                       true,
		"ENODATA":                                      true,
		"ENODEV":                                       true,
		"ENOENT":                                       true,
		"ENOEXEC":                                      true,
		"ENOKEY":                                       true,
		"ENOLCK":                                       true,
		"ENOLINK":                                      true,
		"ENOMEDIUM":                                    true,
		"ENOMEM":                                       true,
		"ENOMSG":                                       true,
		"ENONET":                                       true,
		"ENOPKG":                                       true,
		"ENOPOLICY":                                    true,
		"ENOPROTOOPT":                                  true,
		"ENOSPC":                                       true,
		"ENOSR":                                        true,
		"ENOSTR":                                       true,
		"ENOSYS":                                       true,
		"ENOTBLK":                                      true,
		"ENOTCAPABLE":                                  true,
		"ENOTCONN":                                     true,
		"ENOTDIR":                                      true,
		"ENOTEMPTY":                                    true,
		"ENOTNAM":                                      true,
		"ENOTRECOVERABLE":                              true,
		"ENOTSOCK":                                     true,
		"ENOTSUP":                                      true,
		"ENOTTY":                                       true,
		"ENOTUNIQ":                                     true,
		"ENXIO":                                        true,
		"EN_SW_CTL_INF":                                true,
		"EN_SW_CTL_PREC":                               true,
		"EN_SW_CTL_ROUND":                              true,
		"EN_SW_DATACHAIN":                              true,
		"EN_SW_DENORM":                                 true,
		"EN_SW_INVOP":                                  true,
		"EN_SW_OVERFLOW":                               true,
		"EN_SW_PRECLOSS":                               true,
		"EN_SW_UNDERFLOW":                              true,
		"EN_SW_ZERODIV":                                true,
		"EOPNOTSUPP":                                   true,
		"EOVERFLOW":                                    true,
		"EOWNERDEAD":                                   true,
		"EPERM":                                        true,
		"EPFNOSUPPORT":                                 true,
		"EPIPE":                                        true,
		"EPOLLERR":                                     true,
		"EPOLLET":                                      true,
		"EPOLLHUP":                                     true,
		"EPOLLIN":                                      true,
		"EPOLLMSG":                                     true,
		"EPOLLONESHOT":                                 true,
		"EPOLLOUT":                                     true,
		"EPOLLPRI":                                     true,
		"EPOLLRDBAND":                                  true,
		"EPOLLRDHUP":                                   true,
		"EPOLLRDNORM":                                  true,
		"EPOLLWRBAND":                                  true,
		"EPOLLWRNORM":                                  true,
		"EPOLL_CLOEXEC":                                true,
		"EPOLL_CTL_ADD":                                true,
		"EPOLL_CTL_DEL":                                true,
		"EPOLL_CTL_MOD":                                true,
		"EPOLL_NONBLOCK":                               true,
		"EPROCLIM":                                     true,
		"EPROCUNAVAIL":                                 true,
		"EPROGMISMATCH":                                true,
		"EPROGUNAVAIL":                                 true,
		"EPROTO":                                       true,
		"EPROTONOSUPPORT":                              true,
		"EPROTOTYPE":                                   true,
		"EPWROFF":                                      true,
		"ERANGE":                                       true,
		"EREMCHG":                                      true,
		"EREMOTE":                                      true,
		"EREMOTEIO":                                    true,
		"ERESTART":                                     true,
		"ERFKILL":                                      true,
		"EROFS":                                        true,
		"ERPCMISMATCH":                                 true,
		"ERROR_ACCESS_DENIED":                          true,
		"ERROR_ALREADY_EXISTS":                         true,
		"ERROR_BROKEN_PIPE":                            true,
		"ERROR_BUFFER_OVERFLOW":                        true,
		"ERROR_DIR_NOT_EMPTY":                          true,
		"ERROR_ENVVAR_NOT_FOUND":                       true,
		"ERROR_FILE_EXISTS":                            true,
		"ERROR_FILE_NOT_FOUND":                         true,
		"ERROR_HANDLE_EOF":                             true,
		"ERROR_INSUFFICIENT_BUFFER":                    true,
		"ERROR_IO_PENDING":                             true,
		"ERROR_MOD_NOT_FOUND":                          true,
		"ERROR_MORE_DATA":                              true,
		"ERROR_NETNAME_DELETED":                        true,
		"ERROR_NOT_FOUND":                              true,
		"ERROR_NO_MORE_FILES":                          true,
		"ERROR_OPERATION_ABORTED":                      true,
		"ERROR_PATH_NOT_FOUND":                         true,
		"ERROR_PRIVILEGE_NOT_HELD":                     true,
		"ERROR_PROC_NOT_FOUND":                         true,
		"ESHLIBVERS":                                   true,
		"ESHUTDOWN":                                    true,
		"ESOCKTNOSUPPORT":                              true,
		"ESPIPE":                                       true,
		"ESRCH":                                        true,
		"ESRMNT":                                       true,
		"ESTALE":                                       true,
		"ESTRPIPE":                                     true,
		"ETHERCAP_JUMBO_MTU":                           true,
		"ETHERCAP_VLAN_HWTAGGING":                      true,
		"ETHERCAP_VLAN_MTU":                            true,
		"ETHERMIN":                                     true,
		"ETHERMTU":                                     true,
		"ETHERMTU_JUMBO":                               true,
		"ETHERTYPE_8023":                               true,
		"ETHERTYPE_AARP":                               true,
		"ETHERTYPE_ACCTON":                             true,
		"ETHERTYPE_AEONIC":                             true,
		"ETHERTYPE_ALPHA":                              true,
		"ETHERTYPE_AMBER":                              true,
		"ETHERTYPE_AMOEBA":                             true,
		"ETHERTYPE_AOE":                                true,
		"ETHERTYPE_APOLLO":                             true,
		"ETHERTYPE_APOLLODOMAIN":                       true,
		"ETHERTYPE_APPLETALK":                          true,
		"ETHERTYPE_APPLITEK":                           true,
		"ETHERTYPE_ARGONAUT":                           true,
		"ETHERTYPE_ARP":                                true,
		"ETHERTYPE_AT":                                 true,
		"ETHERTYPE_ATALK":                              true,
		"ETHERTYPE_ATOMIC":                             true,
		"ETHERTYPE_ATT":                                true,
		"ETHERTYPE_ATTSTANFORD":                        true,
		"ETHERTYPE_AUTOPHON":                           true,
		"ETHERTYPE_AXIS":                               true,
		"ETHERTYPE_BCLOOP":                             true,
		"ETHERTYPE_BOFL":                               true,
		"ETHERTYPE_CABLETRON":                          true,
		"ETHERTYPE_CHAOS":                              true,
		"ETHERTYPE_COMDESIGN":                          true,
		"ETHERTYPE_COMPUGRAPHIC":                       true,
		"ETHERTYPE_COUNTERPOINT":                       true,
		"ETHERTYPE_CRONUS":                             true,
		"ETHERTYPE_CRONUSVLN":                          true,
		"ETHERTYPE_DCA":                                true,
		"ETHERTYPE_DDE":                                true,
		"ETHERTYPE_DEBNI":                              true,
		"ETHERTYPE_DECAM":                              true,
		"ETHERTYPE_DECCUST":                            true,
		"ETHERTYPE_DECDIAG":                            true,
		"ETHERTYPE_DECDNS":                             true,
		"ETHERTYPE_DECDTS":                             true,
		"ETHERTYPE_DECEXPER":                           true,
		"ETHERTYPE_DECLAST":                            true,
		"ETHERTYPE_DECLTM":                             true,
		"ETHERTYPE_DECMUMPS":                           true,
		"ETHERTYPE_DECNETBIOS":                         true,
		"ETHERTYPE_DELTACON":                           true,
		"ETHERTYPE_DIDDLE":                             true,
		"ETHERTYPE_DLOG1":                              true,
		"ETHERTYPE_DLOG2":                              true,
		"ETHERTYPE_DN":                                 true,
		"ETHERTYPE_DOGFIGHT":                           true,
		"ETHERTYPE_DSMD":                               true,
		"ETHERTYPE_ECMA":                               true,
		"ETHERTYPE_ENCRYPT":                            true,
		"ETHERTYPE_ES":                                 true,
		"ETHERTYPE_EXCELAN":                            true,
		"ETHERTYPE_EXPERDATA":                          true,
		"ETHERTYPE_FLIP":                               true,
		"ETHERTYPE_FLOWCONTROL":                        true,
		"ETHERTYPE_FRARP":                              true,
		"ETHERTYPE_GENDYN":                             true,
		"ETHERTYPE_HAYES":                              true,
		"ETHERTYPE_HIPPI_FP":                           true,
		"ETHERTYPE_HITACHI":                            true,
		"ETHERTYPE_HP":                                 true,
		"ETHERTYPE_IEEEPUP":                            true,
		"ETHERTYPE_IEEEPUPAT":                          true,
		"ETHERTYPE_IMLBL":                              true,
		"ETHERTYPE_IMLBLDIAG":                          true,
		"ETHERTYPE_IP":                                 true,
		"ETHERTYPE_IPAS":                               true,
		"ETHERTYPE_IPV6":                               true,
		"ETHERTYPE_IPX":                                true,
		"ETHERTYPE_IPXNEW":                             true,
		"ETHERTYPE_KALPANA":                            true,
		"ETHERTYPE_LANBRIDGE":                          true,
		"ETHERTYPE_LANPROBE":                           true,
		"ETHERTYPE_LAT":                                true,
		"ETHERTYPE_LBACK":                              true,
		"ETHERTYPE_LITTLE":                             true,
		"ETHERTYPE_LLDP":                               true,
		"ETHERTYPE_LOGICRAFT":                          true,
		"ETHERTYPE_LOOPBACK":                           true,
		"ETHERTYPE_MATRA":                              true,
		"ETHERTYPE_MAX":                                true,
		"ETHERTYPE_MERIT":                              true,
		"ETHERTYPE_MICP":                               true,
		"ETHERTYPE_MOPDL":                              true,
		"ETHERTYPE_MOPRC":                              true,
		"ETHERTYPE_MOTOROLA":                           true,
		"ETHERTYPE_MPLS":                               true,
		"ETHERTYPE_MPLS_MCAST":                         true,
		"ETHERTYPE_MUMPS":                              true,
		"ETHERTYPE_NBPCC":                              true,
		"ETHERTYPE_NBPCLAIM":                           true,
		"ETHERTYPE_NBPCLREQ":                           true,
		"ETHERTYPE_NBPCLRSP":                           true,
		"ETHERTYPE_NBPCREQ":                            true,
		"ETHERTYPE_NBPCRSP":                            true,
		"ETHERTYPE_NBPDG":                              true,
		"ETHERTYPE_NBPDGB":                             true,
		"ETHERTYPE_NBPDLTE":                            true,
		"ETHERTYPE_NBPRAR":                             true,
		"ETHERTYPE_NBPRAS":                             true,
		"ETHERTYPE_NBPRST":                             true,
		"ETHERTYPE_NBPSCD":                             true,
		"ETHERTYPE_NBPVCD":                             true,
		"ETHERTYPE_NBS":                                true,
		"ETHERTYPE_NCD":                                true,
		"ETHERTYPE_NESTAR":                             true,
		"ETHERTYPE_NETBEUI":                            true,
		"ETHERTYPE_NOVELL":                             true,
		"ETHERTYPE_NS":                                 true,
		"ETHERTYPE_NSAT":                               true,
		"ETHERTYPE_NSCOMPAT":                           true,
		"ETHERTYPE_NTRAILER":                           true,
		"ETHERTYPE_OS9":                                true,
		"ETHERTYPE_OS9NET":                             true,
		"ETHERTYPE_PACER":                              true,
		"ETHERTYPE_PAE":                                true,
		"ETHERTYPE_PCS":                                true,
		"ETHERTYPE_PLANNING":                           true,
		"ETHERTYPE_PPP":                                true,
		"ETHERTYPE_PPPOE":                              true,
		"ETHERTYPE_PPPOEDISC":                          true,
		"ETHERTYPE_PRIMENTS":                           true,
		"ETHERTYPE_PUP":                                true,
		"ETHERTYPE_PUPAT":                              true,
		"ETHERTYPE_QINQ":                               true,
		"ETHERTYPE_RACAL":                              true,
		"ETHERTYPE_RATIONAL":                           true,
		"ETHERTYPE_RAWFR":                              true,
		"ETHERTYPE_RCL":                                true,
		"ETHERTYPE_RDP":                                true,
		"ETHERTYPE_RETIX":                              true,
		"ETHERTYPE_REVARP":                             true,
		"ETHERTYPE_SCA":                                true,
		"ETHERTYPE_SECTRA":                             true,
		"ETHERTYPE_SECUREDATA":                         true,
		"ETHERTYPE_SGITW":                              true,
		"ETHERTYPE_SG_BOUNCE":                          true,
		"ETHERTYPE_SG_DIAG":                            true,
		"ETHERTYPE_SG_NETGAMES":                        true,
		"ETHERTYPE_SG_RESV":                            true,
		"ETHERTYPE_SIMNET":                             true,
		"ETHERTYPE_SLOW":                               true,
		"ETHERTYPE_SLOWPROTOCOLS":                      true,
		"ETHERTYPE_SNA":                                true,
		"ETHERTYPE_SNMP":                               true,
		"ETHERTYPE_SONIX":                              true,
		"ETHERTYPE_SPIDER":                             true,
		"ETHERTYPE_SPRITE":                             true,
		"ETHERTYPE_STP":                                true,
		"ETHERTYPE_TALARIS":                            true,
		"ETHERTYPE_TALARISMC":                          true,
		"ETHERTYPE_TCPCOMP":                            true,
		"ETHERTYPE_TCPSM":                              true,
		"ETHERTYPE_TEC":                                true,
		"ETHERTYPE_TIGAN":                              true,
		"ETHERTYPE_TRAIL":                              true,
		"ETHERTYPE_TRANSETHER":                         true,
		"ETHERTYPE_TYMSHARE":                           true,
		"ETHERTYPE_UBBST":                              true,
		"ETHERTYPE_UBDEBUG":                            true,
		"ETHERTYPE_UBDIAGLOOP":                         true,
		"ETHERTYPE_UBDL":                               true,
		"ETHERTYPE_UBNIU":                              true,
		"ETHERTYPE_UBNMC":                              true,
		"ETHERTYPE_VALID":                              true,
		"ETHERTYPE_VARIAN":                             true,
		"ETHERTYPE_VAXELN":                             true,
		"ETHERTYPE_VEECO":                              true,
		"ETHERTYPE_VEXP":                               true,
		"ETHERTYPE_VGLAB":                              true,
		"ETHERTYPE_VINES":                              true,
		"ETHERTYPE_VINESECHO":                          true,
		"ETHERTYPE_VINESLOOP":                          true,
		"ETHERTYPE_VITAL":                              true,
		"ETHERTYPE_VLAN":                               true,
		"ETHERTYPE_VLTLMAN":                            true,
		"ETHERTYPE_VPROD":                              true,
		"ETHERTYPE_VURESERVED":                         true,
		"ETHERTYPE_WATERLOO":                           true,
		"ETHERTYPE_WELLFLEET":                          true,
		"ETHERTYPE_X25":                                true,
		"ETHERTYPE_X75":                                true,
		"ETHERTYPE_XNSSM":                              true,
		"ETHERTYPE_XTP":                                true,
		"ETHER_ADDR_LEN":                               true,
		"ETHER_ALIGN":                                  true,
		"ETHER_CRC_LEN":                                true,
		"ETHER_CRC_POLY_BE":                            true,
		"ETHER_CRC_POLY_LE":                            true,
		"ETHER_HDR_LEN":                                true,
		"ETHER_MAX_DIX_LEN":                            true,
		"ETHER_MAX_LEN":                                true,
		"ETHER_MAX_LEN_JUMBO":                          true,
		"ETHER_MIN_LEN":                                true,
		"ETHER_PPPOE_ENCAP_LEN":                        true,
		"ETHER_TYPE_LEN":                               true,
		"ETHER_VLAN_ENCAP_LEN":                         true,
		"ETH_P_1588":                                   true,
		"ETH_P_8021Q":                                  true,
		"ETH_P_802_2":                                  true,
		"ETH_P_802_3":                                  true,
		"ETH_P_AARP":                                   true,
		"ETH_P_ALL":                                    true,
		"ETH_P_AOE":                                    true,
		"ETH_P_ARCNET":                                 true,
		"ETH_P_ARP":                                    true,
		"ETH_P_ATALK":                                  true,
		"ETH_P_ATMFATE":                                true,
		"ETH_P_ATMMPOA":                                true,
		"ETH_P_AX25":                                   true,
		"ETH_P_BPQ":                                    true,
		"ETH_P_CAIF":                                   true,
		"ETH_P_CAN":                                    true,
		"ETH_P_CONTROL":                                true,
		"ETH_P_CUST":                                   true,
		"ETH_P_DDCMP":                                  true,
		"ETH_P_DEC":                                    true,
		"ETH_P_DIAG":                                   true,
		"ETH_P_DNA_DL":                                 true,
		"ETH_P_DNA_RC":                                 true,
		"ETH_P_DNA_RT":                                 true,
		"ETH_P_DSA":                                    true,
		"ETH_P_ECONET":                                 true,
		"ETH_P_EDSA":                                   true,
		"ETH_P_FCOE":                                   true,
		"ETH_P_FIP":                                    true,
		"ETH_P_HDLC":                                   true,
		"ETH_P_IEEE802154":                             true,
		"ETH_P_IEEEPUP":                                true,
		"ETH_P_IEEEPUPAT":                              true,
		"ETH_P_IP":                                     true,
		"ETH_P_IPV6":                                   true,
		"ETH_P_IPX":                                    true,
		"ETH_P_IRDA":                                   true,
		"ETH_P_LAT":                                    true,
		"ETH_P_LINK_CTL":                               true,
		"ETH_P_LOCALTALK":                              true,
		"ETH_P_LOOP":                                   true,
		"ETH_P_MOBITEX":                                true,
		"ETH_P_MPLS_MC":                                true,
		"ETH_P_MPLS_UC":                                true,
		"ETH_P_PAE":                                    true,
		"ETH_P_PAUSE":                                  true,
		"ETH_P_PHONET":                                 true,
		"ETH_P_PPPTALK":                                true,
		"ETH_P_PPP_DISC":                               true,
		"ETH_P_PPP_MP":                                 true,
		"ETH_P_PPP_SES":                                true,
		"ETH_P_PUP":                                    true,
		"ETH_P_PUPAT":                                  true,
		"ETH_P_RARP":                                   true,
		"ETH_P_SCA":                                    true,
		"ETH_P_SLOW":                                   true,
		"ETH_P_SNAP":                                   true,
		"ETH_P_TEB":                                    true,
		"ETH_P_TIPC":                                   true,
		"ETH_P_TRAILER":                                true,
		"ETH_P_TR_802_2":                               true,
		"ETH_P_WAN_PPP":                                true,
		"ETH_P_WCCP":                                   true,
		"ETH_P_X25":                                    true,
		"ETIME":                                        true,
		"ETIMEDOUT":                                    true,
		"ETOOMANYREFS":                                 true,
		"ETXTBSY":                                      true,
		"EUCLEAN":                                      true,
		"EUNATCH":                                      true,
		"EUSERS":                                       true,
		"EVFILT_AIO":                                   true,
		"EVFILT_FS":                                    true,
		"EVFILT_LIO":                                   true,
		"EVFILT_MACHPORT":                              true,
		"EVFILT_PROC":                                  true,
		"EVFILT_READ":                                  true,
		"EVFILT_SIGNAL":                                true,
		"EVFILT_SYSCOUNT":                              true,
		"EVFILT_THREADMARKER":                          true,
		"EVFILT_TIMER":                                 true,
		"EVFILT_USER":                                  true,
		"EVFILT_VM":                                    true,
		"EVFILT_VNODE":                                 true,
		"EVFILT_WRITE":                                 true,
		"EV_ADD":                                       true,
		"EV_CLEAR":                                     true,
		"EV_DELETE":                                    true,
		"EV_DISABLE":                                   true,
		"EV_DISPATCH":                                  true,
		"EV_DROP":                                      true,
		"EV_ENABLE":                                    true,
		"EV_EOF":                                       true,
		"EV_ERROR":                                     true,
		"EV_FLAG0":                                     true,
		"EV_FLAG1":                                     true,
		"EV_ONESHOT":                                   true,
		"EV_OOBAND":                                    true,
		"EV_POLL":                                      true,
		"EV_RECEIPT":                                   true,
		"EV_SYSFLAGS":                                  true,
		"EWINDOWS":                                     true,
		"EWOULDBLOCK":                                  true,
		"EXDEV":                                        true,
		"EXFULL":                                       true,
		"EXTA":                                         true,
		"EXTB":                                         true,
		"EXTPROC":                                      true,
		"Environ":                                      true,
		"EpollCreate":                                  true,
		"EpollCreate1":                                 true,
		"EpollCtl":                                     true,
		"EpollEvent":                                   true,
		"EpollWait":                                    true,
		"Errno":                                        true,
		"EscapeArg":                                    true,
		"Exchangedata":                                 true,
		"Exec":                                         true,
		"Exit":                                         true,
		"ExitProcess":                                  true,
		"FD_CLOEXEC":                                   true,
		"FD_SETSIZE":                                   true,
		"FILE_ACTION_ADDED":                            true,
		"FILE_ACTION_MODIFIED":                         true,
		"FILE_ACTION_REMOVED":                          true,
		"FILE_ACTION_RENAMED_NEW_NAME":                 true,
		"FILE_ACTION_RENAMED_OLD_NAME":                 true,
		"FILE_APPEND_DATA":                             true,
		"FILE_ATTRIBUTE_ARCHIVE":                       true,
		"FILE_ATTRIBUTE_DIRECTORY":                     true,
		"FILE_ATTRIBUTE_HIDDEN":                        true,
		"FILE_ATTRIBUTE_NORMAL":                        true,
		"FILE_ATTRIBUTE_READONLY":                      true,
		"FILE_ATTRIBUTE_REPARSE_POINT":                 true,
		"FILE_ATTRIBUTE_SYSTEM":                        true,
		"FILE_BEGIN":                                   true,
		"FILE_CURRENT":                                 true,
		"FILE_END":                                     true,
		"FILE_FLAG_BACKUP_SEMANTICS":                   true,
		"FILE_FLAG_OPEN_REPARSE_POINT":                 true,
		"FILE_FLAG_OVERLAPPED":                         true,
		"FILE_LIST_DIRECTORY":                          true,
		"FILE_MAP_COPY":                                true,
		"FILE_MAP_EXECUTE":                             true,
		"FILE_MAP_READ":                                true,
		"FILE_MAP_WRITE":                               true,
		"FILE_NOTIFY_CHANGE_ATTRIBUTES":                true,
		"FILE_NOTIFY_CHANGE_CREATION":                  true,
		"FILE_NOTIFY_CHANGE_DIR_NAME":                  true,
		"FILE_NOTIFY_CHANGE_FILE_NAME":                 true,
		"FILE_NOTIFY_CHANGE_LAST_ACCESS":               true,
		"FILE_NOTIFY_CHANGE_LAST_WRITE":                true,
		"FILE_NOTIFY_CHANGE_SIZE":                      true,
		"FILE_SHARE_DELETE":                            true,
		"FILE_SHARE_READ":                              true,
		"FILE_SHARE_WRITE":                             true,
		"FILE_SKIP_COMPLETION_PORT_ON_SUCCESS":         true,
		"FILE_SKIP_SET_EVENT_ON_HANDLE":                true,
		"FILE_TYPE_CHAR":                               true,
		"FILE_TYPE_DISK":                               true,
		"FILE_TYPE_PIPE":                               true,
		"FILE_TYPE_REMOTE":                             true,
		"FILE_TYPE_UNKNOWN":                            true,
		"FILE_WRITE_ATTRIBUTES":                        true,
		"FLUSHO":                                       true,
		"FORMAT_MESSAGE_ALLOCATE_BUFFER":               true,
		"FORMAT_MESSAGE_ARGUMENT_ARRAY":                true,
		"FORMAT_MESSAGE_FROM_HMODULE":                  true,
		"FORMAT_MESSAGE_FROM_STRING":                   true,
		"FORMAT_MESSAGE_FROM_SYSTEM":                   true,
		"FORMAT_MESSAGE_IGNORE_INSERTS":                true,
		"FORMAT_MESSAGE_MAX_WIDTH_MASK":                true,
		"FSCTL_GET_REPARSE_POINT":                      true,
		"F_ADDFILESIGS":                                true,
		"F_ADDSIGS":                                    true,
		"F_ALLOCATEALL":                                true,
		"F_ALLOCATECONTIG":                             true,
		"F_CANCEL":                                     true,
		"F_CHKCLEAN":                                   true,
		"F_CLOSEM":                                     true,
		"F_DUP2FD":                                     true,
		"F_DUP2FD_CLOEXEC":                             true,
		"F_DUPFD":                                      true,
		"F_DUPFD_CLOEXEC":                              true,
		"F_EXLCK":                                      true,
		"F_FLUSH_DATA":                                 true,
		"F_FREEZE_FS":                                  true,
		"F_FSCTL":                                      true,
		"F_FSDIRMASK":                                  true,
		"F_FSIN":                                       true,
		"F_FSINOUT":                                    true,
		"F_FSOUT":                                      true,
		"F_FSPRIV":                                     true,
		"F_FSVOID":                                     true,
		"F_FULLFSYNC":                                  true,
		"F_GETFD":                                      true,
		"F_GETFL":                                      true,
		"F_GETLEASE":                                   true,
		"F_GETLK":                                      true,
		"F_GETLK64":                                    true,
		"F_GETLKPID":                                   true,
		"F_GETNOSIGPIPE":                               true,
		"F_GETOWN":                                     true,
		"F_GETOWN_EX":                                  true,
		"F_GETPATH":                                    true,
		"F_GETPATH_MTMINFO":                            true,
		"F_GETPIPE_SZ":                                 true,
		"F_GETPROTECTIONCLASS":                         true,
		"F_GETSIG":                                     true,
		"F_GLOBAL_NOCACHE":                             true,
		"F_LOCK":                                       true,
		"F_LOG2PHYS":                                   true,
		"F_LOG2PHYS_EXT":                               true,
		"F_MARKDEPENDENCY":                             true,
		"F_MAXFD":                                      true,
		"F_NOCACHE":                                    true,
		"F_NODIRECT":                                   true,
		"F_NOTIFY":                                     true,
		"F_OGETLK":                                     true,
		"F_OK":                                         true,
		"F_OSETLK":                                     true,
		"F_OSETLKW":                                    true,
		"F_PARAM_MASK":                                 true,
		"F_PARAM_MAX":                                  true,
		"F_PATHPKG_CHECK":                              true,
		"F_PEOFPOSMODE":                                true,
		"F_PREALLOCATE":                                true,
		"F_RDADVISE":                                   true,
		"F_RDAHEAD":                                    true,
		"F_RDLCK":                                      true,
		"F_READAHEAD":                                  true,
		"F_READBOOTSTRAP":                              true,
		"F_SETBACKINGSTORE":                            true,
		"F_SETFD":                                      true,
		"F_SETFL":                                      true,
		"F_SETLEASE":                                   true,
		"F_SETLK":                                      true,
		"F_SETLK64":                                    true,
		"F_SETLKW":                                     true,
		"F_SETLKW64":                                   true,
		"F_SETLK_REMOTE":                               true,
		"F_SETNOSIGPIPE":                               true,
		"F_SETOWN":                                     true,
		"F_SETOWN_EX":                                  true,
		"F_SETPIPE_SZ":                                 true,
		"F_SETPROTECTIONCLASS":                         true,
		"F_SETSIG":                                     true,
		"F_SETSIZE":                                    true,
		"F_SHLCK":                                      true,
		"F_TEST":                                       true,
		"F_THAW_FS":                                    true,
		"F_TLOCK":                                      true,
		"F_ULOCK":                                      true,
		"F_UNLCK":                                      true,
		"F_UNLCKSYS":                                   true,
		"F_VOLPOSMODE":                                 true,
		"F_WRITEBOOTSTRAP":                             true,
		"F_WRLCK":                                      true,
		"Faccessat":                                    true,
		"Fallocate":                                    true,
		"Fbootstraptransfer_t":                         true,
		"Fchdir":                                       true,
		"Fchflags":                                     true,
		"Fchmod":                                       true,
		"Fchmodat":                                     true,
		"Fchown":                                       true,
		"Fchownat":                                     true,
		"FcntlFlock":                                   true,
		"FdSet":                                        true,
		"Fdatasync":                                    true,
		"FileNotifyInformation":                        true,
		"Filetime":                                     true,
		"FindClose":                                    true,
		"FindFirstFile":                                true,
		"FindNextFile":                                 true,
		"Flock":                                        true,
		"Flock_t":                                      true,
		"FlushBpf":                                     true,
		"FlushFileBuffers":                             true,
		"FlushViewOfFile":                              true,
		"ForkExec":                                     true,
		"ForkLock":                                     true,
		"FormatMessage":                                true,
		"Fpathconf":                                    true,
		"FreeAddrInfoW":                                true,
		"FreeEnvironmentStrings":                       true,
		"FreeLibrary":                                  true,
		"Fsid":                                         true,
		"Fstat":                                        true,
		"Fstatat":                                      true,
		"Fstatfs":                                      true,
		"Fstore_t":                                     true,
		"Fsync":                                        true,
		"Ftruncate":                                    true,
		"FullPath":                                     true,
		"Futimes":                                      true,
		"Futimesat":                                    true,
		"GENERIC_ALL":                                  true,
		"GENERIC_EXECUTE":                              true,
		"GENERIC_READ":                                 true,
		"GENERIC_WRITE":                                true,
		"GUID":                                         true,
		"GetAcceptExSockaddrs":                         true,
		"GetAdaptersInfo":                              true,
		"GetAddrInfoW":                                 true,
		"GetCommandLine":                               true,
		"GetComputerName":                              true,
		"GetConsoleMode":                               true,
		"GetCurrentDirectory":                          true,
		"GetCurrentProcess":                            true,
		"GetEnvironmentStrings":                        true,
		"GetEnvironmentVariable":                       true,
		"GetExitCodeProcess":                           true,
		"GetFileAttributes":                            true,
		"GetFileAttributesEx":                          true,
		"GetFileExInfoStandard":                        true,
		"GetFileExMaxInfoLevel":                        true,
		"GetFileInformationByHandle":                   true,
		"GetFileType":                                  true,
		"GetFullPathName":                              true,
		"GetHostByName":                                true,
		"GetIfEntry":                                   true,
		"GetLastError":                                 true,
		"GetLengthSid":                                 true,
		"GetLongPathName":                              true,
		"GetProcAddress":                               true,
		"GetProcessTimes":                              true,
		"GetProtoByName":                               true,
		"GetQueuedCompletionStatus":                    true,
		"GetServByName":                                true,
		"GetShortPathName":                             true,
		"GetStartupInfo":                               true,
		"GetStdHandle":                                 true,
		"GetSystemTimeAsFileTime":                      true,
		"GetTempPath":                                  true,
		"GetTimeZoneInformation":                       true,
		"GetTokenInformation":                          true,
		"GetUserNameEx":                                true,
		"GetUserProfileDirectory":                      true,
		"GetVersion":                                   true,
		"Getcwd":                                       true,
		"Getdents":                                     true,
		"Getdirentries":                                true,
		"Getdtablesize":                                true,
		"Getegid":                                      true,
		"Getenv":                                       true,
		"Geteuid":                                      true,
		"Getfsstat":                                    true,
		"Getgid":                                       true,
		"Getgroups":                                    true,
		"Getpagesize":                                  true,
		"Getpeername":                                  true,
		"Getpgid":                                      true,
		"Getpgrp":                                      true,
		"Getpid":                                       true,
		"Getppid":                                      true,
		"Getpriority":                                  true,
		"Getrlimit":                                    true,
		"Getrusage":                                    true,
		"Getsid":                                       true,
		"Getsockname":                                  true,
		"Getsockopt":                                   true,
		"GetsockoptByte":                               true,
		"GetsockoptICMPv6Filter":                       true,
		"GetsockoptIPMreq":                             true,
		"GetsockoptIPMreqn":                            true,
		"GetsockoptIPv6MTUInfo":                        true,
		"GetsockoptIPv6Mreq":                           true,
		"GetsockoptInet4Addr":                          true,
		"GetsockoptInt":                                true,
		"GetsockoptUcred":                              true,
		"Gettid":                                       true,
		"Gettimeofday":                                 true,
		"Getuid":                                       true,
		"Getwd":                                        true,
		"Getxattr":                                     true,
		"HANDLE_FLAG_INHERIT":                          true,
		"HKEY_CLASSES_ROOT":                            true,
		"HKEY_CURRENT_CONFIG":                          true,
		"HKEY_CURRENT_USER":                            true,
		"HKEY_DYN_DATA":                                true,
		"HKEY_LOCAL_MACHINE":                           true,
		"HKEY_PERFORMANCE_DATA":                        true,
		"HKEY_USERS":                                   true,
		"HUPCL":                                        true,
		"Handle":                                       true,
		"Hostent":                                      true,
		"ICANON":                                       true,
		"ICMP6_FILTER":                                 true,
		"ICMPV6_FILTER":                                true,
		"ICMPv6Filter":                                 true,
		"ICRNL":                                        true,
		"IEXTEN":                                       true,
		"IFAN_ARRIVAL":                                 true,
		"IFAN_DEPARTURE":                               true,
		"IFA_ADDRESS":                                  true,
		"IFA_ANYCAST":                                  true,
		"IFA_BROADCAST":                                true,
		"IFA_CACHEINFO":                                true,
		"IFA_F_DADFAILED":                              true,
		"IFA_F_DEPRECATED":                             true,
		"IFA_F_HOMEADDRESS":                            true,
		"IFA_F_NODAD":                                  true,
		"IFA_F_OPTIMISTIC":                             true,
		"IFA_F_PERMANENT":                              true,
		"IFA_F_SECONDARY":                              true,
		"IFA_F_TEMPORARY":                              true,
		"IFA_F_TENTATIVE":                              true,
		"IFA_LABEL":                                    true,
		"IFA_LOCAL":                                    true,
		"IFA_MAX":                                      true,
		"IFA_MULTICAST":                                true,
		"IFA_ROUTE":                                    true,
		"IFA_UNSPEC":                                   true,
		"IFF_ALLMULTI":                                 true,
		"IFF_ALTPHYS":                                  true,
		"IFF_AUTOMEDIA":                                true,
		"IFF_BROADCAST":                                true,
		"IFF_CANTCHANGE":                               true,
		"IFF_CANTCONFIG":                               true,
		"IFF_DEBUG":                                    true,
		"IFF_DRV_OACTIVE":                              true,
		"IFF_DRV_RUNNING":                              true,
		"IFF_DYING":                                    true,
		"IFF_DYNAMIC":                                  true,
		"IFF_LINK0":                                    true,
		"IFF_LINK1":                                    true,
		"IFF_LINK2":                                    true,
		"IFF_LOOPBACK":                                 true,
		"IFF_MASTER":                                   true,
		"IFF_MONITOR":                                  true,
		"IFF_MULTICAST":                                true,
		"IFF_NOARP":                                    true,
		"IFF_NOTRAILERS":                               true,
		"IFF_NO_PI":                                    true,
		"IFF_OACTIVE":                                  true,
		"IFF_ONE_QUEUE":                                true,
		"IFF_POINTOPOINT":                              true,
		"IFF_POINTTOPOINT":                             true,
		"IFF_PORTSEL":                                  true,
		"IFF_PPROMISC":                                 true,
		"IFF_PROMISC":                                  true,
		"IFF_RENAMING":                                 true,
		"IFF_RUNNING":                                  true,
		"IFF_SIMPLEX":                                  true,
		"IFF_SLAVE":                                    true,
		"IFF_SMART":                                    true,
		"IFF_STATICARP":                                true,
		"IFF_TAP":                                      true,
		"IFF_TUN":                                      true,
		"IFF_TUN_EXCL":                                 true,
		"IFF_UP":                                       true,
		"IFF_VNET_HDR":                                 true,
		"IFLA_ADDRESS":                                 true,
		"IFLA_BROADCAST":                               true,
		"IFLA_COST":                                    true,
		"IFLA_IFALIAS":                                 true,
		"IFLA_IFNAME":                                  true,
		"IFLA_LINK":                                    true,
		"IFLA_LINKINFO":                                true,
		"IFLA_LINKMODE":                                true,
		"IFLA_MAP":                                     true,
		"IFLA_MASTER":                                  true,
		"IFLA_MAX":                                     true,
		"IFLA_MTU":                                     true,
		"IFLA_NET_NS_PID":                              true,
		"IFLA_OPERSTATE":                               true,
		"IFLA_PRIORITY":                                true,
		"IFLA_PROTINFO":                                true,
		"IFLA_QDISC":                                   true,
		"IFLA_STATS":                                   true,
		"IFLA_TXQLEN":                                  true,
		"IFLA_UNSPEC":                                  true,
		"IFLA_WEIGHT":                                  true,
		"IFLA_WIRELESS":                                true,
		"IFNAMSIZ":                                     true,
		"IFT_1822":                                     true,
		"IFT_A12MPPSWITCH":                             true,
		"IFT_AAL2":                                     true,
		"IFT_AAL5":                                     true,
		"IFT_ADSL":                                     true,
		"IFT_AFLANE8023":                               true,
		"IFT_AFLANE8025":                               true,
		"IFT_ARAP":                                     true,
		"IFT_ARCNET":                                   true,
		"IFT_ARCNETPLUS":                               true,
		"IFT_ASYNC":                                    true,
		"IFT_ATM":                                      true,
		"IFT_ATMDXI":                                   true,
		"IFT_ATMFUNI":                                  true,
		"IFT_ATMIMA":                                   true,
		"IFT_ATMLOGICAL":                               true,
		"IFT_ATMRADIO":                                 true,
		"IFT_ATMSUBINTERFACE":                          true,
		"IFT_ATMVCIENDPT":                              true,
		"IFT_ATMVIRTUAL":                               true,
		"IFT_BGPPOLICYACCOUNTING":                      true,
		"IFT_BLUETOOTH":                                true,
		"IFT_BRIDGE":                                   true,
		"IFT_BSC":                                      true,
		"IFT_CARP":                                     true,
		"IFT_CCTEMUL":                                  true,
		"IFT_CELLULAR":                                 true,
		"IFT_CEPT":                                     true,
		"IFT_CES":                                      true,
		"IFT_CHANNEL":                                  true,
		"IFT_CNR":                                      true,
		"IFT_COFFEE":                                   true,
		"IFT_COMPOSITELINK":                            true,
		"IFT_DCN":                                      true,
		"IFT_DIGITALPOWERLINE":                         true,
		"IFT_DIGITALWRAPPEROVERHEADCHANNEL":            true,
		"IFT_DLSW":                                     true,
		"IFT_DOCSCABLEDOWNSTREAM":                      true,
		"IFT_DOCSCABLEMACLAYER":                        true,
		"IFT_DOCSCABLEUPSTREAM":                        true,
		"IFT_DOCSCABLEUPSTREAMCHANNEL":                 true,
		"IFT_DS0":                                      true,
		"IFT_DS0BUNDLE":                                true,
		"IFT_DS1FDL":                                   true,
		"IFT_DS3":                                      true,
		"IFT_DTM":                                      true,
		"IFT_DUMMY":                                    true,
		"IFT_DVBASILN":                                 true,
		"IFT_DVBASIOUT":                                true,
		"IFT_DVBRCCDOWNSTREAM":                         true,
		"IFT_DVBRCCMACLAYER":                           true,
		"IFT_DVBRCCUPSTREAM":                           true,
		"IFT_ECONET":                                   true,
		"IFT_ENC":                                      true,
		"IFT_EON":                                      true,
		"IFT_EPLRS":                                    true,
		"IFT_ESCON":                                    true,
		"IFT_ETHER":                                    true,
		"IFT_FAITH":                                    true,
		"IFT_FAST":                                     true,
		"IFT_FASTETHER":                                true,
		"IFT_FASTETHERFX":                              true,
		"IFT_FDDI":                                     true,
		"IFT_FIBRECHANNEL":                             true,
		"IFT_FRAMERELAYINTERCONNECT":                   true,
		"IFT_FRAMERELAYMPI":                            true,
		"IFT_FRDLCIENDPT":                              true,
		"IFT_FRELAY":                                   true,
		"IFT_FRELAYDCE":                                true,
		"IFT_FRF16MFRBUNDLE":                           true,
		"IFT_FRFORWARD":                                true,
		"IFT_G703AT2MB":                                true,
		"IFT_G703AT64K":                                true,
		"IFT_GIF":                                      true,
		"IFT_GIGABITETHERNET":                          true,
		"IFT_GR303IDT":                                 true,
		"IFT_GR303RDT":                                 true,
		"IFT_H323GATEKEEPER":                           true,
		"IFT_H323PROXY":                                true,
		"IFT_HDH1822":                                  true,
		"IFT_HDLC":                                     true,
		"IFT_HDSL2":                                    true,
		"IFT_HIPERLAN2":                                true,
		"IFT_HIPPI":                                    true,
		"IFT_HIPPIINTERFACE":                           true,
		"IFT_HOSTPAD":                                  true,
		"IFT_HSSI":                                     true,
		"IFT_HY":                                       true,
		"IFT_IBM370PARCHAN":                            true,
		"IFT_IDSL":                                     true,
		"IFT_IEEE1394":                                 true,
		"IFT_IEEE80211":                                true,
		"IFT_IEEE80212":                                true,
		"IFT_IEEE8023ADLAG":                            true,
		"IFT_IFGSN":                                    true,
		"IFT_IMT":                                      true,
		"IFT_INFINIBAND":                               true,
		"IFT_INTERLEAVE":                               true,
		"IFT_IP":                                       true,
		"IFT_IPFORWARD":                                true,
		"IFT_IPOVERATM":                                true,
		"IFT_IPOVERCDLC":                               true,
		"IFT_IPOVERCLAW":                               true,
		"IFT_IPSWITCH":                                 true,
		"IFT_IPXIP":                                    true,
		"IFT_ISDN":                                     true,
		"IFT_ISDNBASIC":                                true,
		"IFT_ISDNPRIMARY":                              true,
		"IFT_ISDNS":                                    true,
		"IFT_ISDNU":                                    true,
		"IFT_ISO88022LLC":                              true,
		"IFT_ISO88023":                                 true,
		"IFT_ISO88024":                                 true,
		"IFT_ISO88025":                                 true,
		"IFT_ISO88025CRFPINT":                          true,
		"IFT_ISO88025DTR":                              true,
		"IFT_ISO88025FIBER":                            true,
		"IFT_ISO88026":                                 true,
		"IFT_ISUP":                                     true,
		"IFT_L2VLAN":                                   true,
		"IFT_L3IPVLAN":                                 true,
		"IFT_L3IPXVLAN":                                true,
		"IFT_LAPB":                                     true,
		"IFT_LAPD":                                     true,
		"IFT_LAPF":                                     true,
		"IFT_LINEGROUP":                                true,
		"IFT_LOCALTALK":                                true,
		"IFT_LOOP":                                     true,
		"IFT_MEDIAMAILOVERIP":                          true,
		"IFT_MFSIGLINK":                                true,
		"IFT_MIOX25":                                   true,
		"IFT_MODEM":                                    true,
		"IFT_MPC":                                      true,
		"IFT_MPLS":                                     true,
		"IFT_MPLSTUNNEL":                               true,
		"IFT_MSDSL":                                    true,
		"IFT_MVL":                                      true,
		"IFT_MYRINET":                                  true,
		"IFT_NFAS":                                     true,
		"IFT_NSIP":                                     true,
		"IFT_OPTICALCHANNEL":                           true,
		"IFT_OPTICALTRANSPORT":                         true,
		"IFT_OTHER":                                    true,
		"IFT_P10":                                      true,
		"IFT_P80":                                      true,
		"IFT_PARA":                                     true,
		"IFT_PDP":                                      true,
		"IFT_PFLOG":                                    true,
		"IFT_PFLOW":                                    true,
		"IFT_PFSYNC":                                   true,
		"IFT_PLC":                                      true,
		"IFT_PON155":                                   true,
		"IFT_PON622":                                   true,
		"IFT_POS":                                      true,
		"IFT_PPP":                                      true,
		"IFT_PPPMULTILINKBUNDLE":                       true,
		"IFT_PROPATM":                                  true,
		"IFT_PROPBWAP2MP":                              true,
		"IFT_PROPCNLS":                                 true,
		"IFT_PROPDOCSWIRELESSDOWNSTREAM":               true,
		"IFT_PROPDOCSWIRELESSMACLAYER":                 true,
		"IFT_PROPDOCSWIRELESSUPSTREAM":                 true,
		"IFT_PROPMUX":                                  true,
		"IFT_PROPVIRTUAL":                              true,
		"IFT_PROPWIRELESSP2P":                          true,
		"IFT_PTPSERIAL":                                true,
		"IFT_PVC":                                      true,
		"IFT_Q2931":                                    true,
		"IFT_QLLC":                                     true,
		"IFT_RADIOMAC":                                 true,
		"IFT_RADSL":                                    true,
		"IFT_REACHDSL":                                 true,
		"IFT_RFC1483":                                  true,
		"IFT_RS232":                                    true,
		"IFT_RSRB":                                     true,
		"IFT_SDLC":                                     true,
		"IFT_SDSL":                                     true,
		"IFT_SHDSL":                                    true,
		"IFT_SIP":                                      true,
		"IFT_SIPSIG":                                   true,
		"IFT_SIPTG":                                    true,
		"IFT_SLIP":                                     true,
		"IFT_SMDSDXI":                                  true,
		"IFT_SMDSICIP":                                 true,
		"IFT_SONET":                                    true,
		"IFT_SONETOVERHEADCHANNEL":                     true,
		"IFT_SONETPATH":                                true,
		"IFT_SONETVT":                                  true,
		"IFT_SRP":                                      true,
		"IFT_SS7SIGLINK":                               true,
		"IFT_STACKTOSTACK":                             true,
		"IFT_STARLAN":                                  true,
		"IFT_STF":                                      true,
		"IFT_T1":                                       true,
		"IFT_TDLC":                                     true,
		"IFT_TELINK":                                   true,
		"IFT_TERMPAD":                                  true,
		"IFT_TR008":                                    true,
		"IFT_TRANSPHDLC":                               true,
		"IFT_TUNNEL":                                   true,
		"IFT_ULTRA":                                    true,
		"IFT_USB":                                      true,
		"IFT_V11":                                      true,
		"IFT_V35":                                      true,
		"IFT_V36":                                      true,
		"IFT_V37":                                      true,
		"IFT_VDSL":                                     true,
		"IFT_VIRTUALIPADDRESS":                         true,
		"IFT_VIRTUALTG":                                true,
		"IFT_VOICEDID":                                 true,
		"IFT_VOICEEM":                                  true,
		"IFT_VOICEEMFGD":                               true,
		"IFT_VOICEENCAP":                               true,
		"IFT_VOICEFGDEANA":                             true,
		"IFT_VOICEFXO":                                 true,
		"IFT_VOICEFXS":                                 true,
		"IFT_VOICEOVERATM":                             true,
		"IFT_VOICEOVERCABLE":                           true,
		"IFT_VOICEOVERFRAMERELAY":                      true,
		"IFT_VOICEOVERIP":                              true,
		"IFT_X213":                                     true,
		"IFT_X25":                                      true,
		"IFT_X25DDN":                                   true,
		"IFT_X25HUNTGROUP":                             true,
		"IFT_X25MLP":                                   true,
		"IFT_X25PLE":                                   true,
		"IFT_XETHER":                                   true,
		"IGNBRK":                                       true,
		"IGNCR":                                        true,
		"IGNORE":                                       true,
		"IGNPAR":                                       true,
		"IMAXBEL":                                      true,
		"INFINITE":                                     true,
		"INLCR":                                        true,
		"INPCK":                                        true,
		"INVALID_FILE_ATTRIBUTES":                      true,
		"IN_ACCESS":                                    true,
		"IN_ALL_EVENTS":                                true,
		"IN_ATTRIB":                                    true,
		"IN_CLASSA_HOST":                               true,
		"IN_CLASSA_MAX":                                true,
		"IN_CLASSA_NET":                                true,
		"IN_CLASSA_NSHIFT":                             true,
		"IN_CLASSB_HOST":                               true,
		"IN_CLASSB_MAX":                                true,
		"IN_CLASSB_NET":                                true,
		"IN_CLASSB_NSHIFT":                             true,
		"IN_CLASSC_HOST":                               true,
		"IN_CLASSC_NET":                                true,
		"IN_CLASSC_NSHIFT":                             true,
		"IN_CLASSD_HOST":                               true,
		"IN_CLASSD_NET":                                true,
		"IN_CLASSD_NSHIFT":                             true,
		"IN_CLOEXEC":                                   true,
		"IN_CLOSE":                                     true,
		"IN_CLOSE_NOWRITE":                             true,
		"IN_CLOSE_WRITE":                               true,
		"IN_CREATE":                                    true,
		"IN_DELETE":                                    true,
		"IN_DELETE_SELF":                               true,
		"IN_DONT_FOLLOW":                               true,
		"IN_EXCL_UNLINK":                               true,
		"IN_IGNORED":                                   true,
		"IN_ISDIR":                                     true,
		"IN_LINKLOCALNETNUM":                           true,
		"IN_LOOPBACKNET":                               true,
		"IN_MASK_ADD":                                  true,
		"IN_MODIFY":                                    true,
		"IN_MOVE":                                      true,
		"IN_MOVED_FROM":                                true,
		"IN_MOVED_TO":                                  true,
		"IN_MOVE_SELF":                                 true,
		"IN_NONBLOCK":                                  true,
		"IN_ONESHOT":                                   true,
		"IN_ONLYDIR":                                   true,
		"IN_OPEN":                                      true,
		"IN_Q_OVERFLOW":                                true,
		"IN_RFC3021_HOST":                              true,
		"IN_RFC3021_MASK":                              true,
		"IN_RFC3021_NET":                               true,
		"IN_RFC3021_NSHIFT":                            true,
		"IN_UNMOUNT":                                   true,
		"IOC_IN":                                       true,
		"IOC_INOUT":                                    true,
		"IOC_OUT":                                      true,
		"IOC_VENDOR":                                   true,
		"IOC_WS2":                                      true,
		"IO_REPARSE_TAG_SYMLINK":                       true,
		"IPMreq":                                       true,
		"IPMreqn":                                      true,
		"IPPROTO_3PC":                                  true,
		"IPPROTO_ADFS":                                 true,
		"IPPROTO_AH":                                   true,
		"IPPROTO_AHIP":                                 true,
		"IPPROTO_APES":                                 true,
		"IPPROTO_ARGUS":                                true,
		"IPPROTO_AX25":                                 true,
		"IPPROTO_BHA":                                  true,
		"IPPROTO_BLT":                                  true,
		"IPPROTO_BRSATMON":                             true,
		"IPPROTO_CARP":                                 true,
		"IPPROTO_CFTP":                                 true,
		"IPPROTO_CHAOS":                                true,
		"IPPROTO_CMTP":                                 true,
		"IPPROTO_COMP":                                 true,
		"IPPROTO_CPHB":                                 true,
		"IPPROTO_CPNX":                                 true,
		"IPPROTO_DCCP":                                 true,
		"IPPROTO_DDP":                                  true,
		"IPPROTO_DGP":                                  true,
		"IPPROTO_DIVERT":                               true,
		"IPPROTO_DIVERT_INIT":                          true,
		"IPPROTO_DIVERT_RESP":                          true,
		"IPPROTO_DONE":                                 true,
		"IPPROTO_DSTOPTS":                              true,
		"IPPROTO_EGP":                                  true,
		"IPPROTO_EMCON":                                true,
		"IPPROTO_ENCAP":                                true,
		"IPPROTO_EON":                                  true,
		"IPPROTO_ESP":                                  true,
		"IPPROTO_ETHERIP":                              true,
		"IPPROTO_FRAGMENT":                             true,
		"IPPROTO_GGP":                                  true,
		"IPPROTO_GMTP":                                 true,
		"IPPROTO_GRE":                                  true,
		"IPPROTO_HELLO":                                true,
		"IPPROTO_HMP":                                  true,
		"IPPROTO_HOPOPTS":                              true,
		"IPPROTO_ICMP":                                 true,
		"IPPROTO_ICMPV6":                               true,
		"IPPROTO_IDP":                                  true,
		"IPPROTO_IDPR":                                 true,
		"IPPROTO_IDRP":                                 true,
		"IPPROTO_IGMP":                                 true,
		"IPPROTO_IGP":                                  true,
		"IPPROTO_IGRP":                                 true,
		"IPPROTO_IL":                                   true,
		"IPPROTO_INLSP":                                true,
		"IPPROTO_INP":                                  true,
		"IPPROTO_IP":                                   true,
		"IPPROTO_IPCOMP":                               true,
		"IPPROTO_IPCV":                                 true,
		"IPPROTO_IPEIP":                                true,
		"IPPROTO_IPIP":                                 true,
		"IPPROTO_IPPC":                                 true,
		"IPPROTO_IPV4":                                 true,
		"IPPROTO_IPV6":                                 true,
		"IPPROTO_IPV6_ICMP":                            true,
		"IPPROTO_IRTP":                                 true,
		"IPPROTO_KRYPTOLAN":                            true,
		"IPPROTO_LARP":                                 true,
		"IPPROTO_LEAF1":                                true,
		"IPPROTO_LEAF2":                                true,
		"IPPROTO_MAX":                                  true,
		"IPPROTO_MAXID":                                true,
		"IPPROTO_MEAS":                                 true,
		"IPPROTO_MH":                                   true,
		"IPPROTO_MHRP":                                 true,
		"IPPROTO_MICP":                                 true,
		"IPPROTO_MOBILE":                               true,
		"IPPROTO_MPLS":                                 true,
		"IPPROTO_MTP":                                  true,
		"IPPROTO_MUX":                                  true,
		"IPPROTO_ND":                                   true,
		"IPPROTO_NHRP":                                 true,
		"IPPROTO_NONE":                                 true,
		"IPPROTO_NSP":                                  true,
		"IPPROTO_NVPII":                                true,
		"IPPROTO_OLD_DIVERT":                           true,
		"IPPROTO_OSPFIGP":                              true,
		"IPPROTO_PFSYNC":                               true,
		"IPPROTO_PGM":                                  true,
		"IPPROTO_PIGP":                                 true,
		"IPPROTO_PIM":                                  true,
		"IPPROTO_PRM":                                  true,
		"IPPROTO_PUP":                                  true,
		"IPPROTO_PVP":                                  true,
		"IPPROTO_RAW":                                  true,
		"IPPROTO_RCCMON":                               true,
		"IPPROTO_RDP":                                  true,
		"IPPROTO_ROUTING":                              true,
		"IPPROTO_RSVP":                                 true,
		"IPPROTO_RVD":                                  true,
		"IPPROTO_SATEXPAK":                             true,
		"IPPROTO_SATMON":                               true,
		"IPPROTO_SCCSP":                                true,
		"IPPROTO_SCTP":                                 true,
		"IPPROTO_SDRP":                                 true,
		"IPPROTO_SEND":                                 true,
		"IPPROTO_SEP":                                  true,
		"IPPROTO_SKIP":                                 true,
		"IPPROTO_SPACER":                               true,
		"IPPROTO_SRPC":                                 true,
		"IPPROTO_ST":                                   true,
		"IPPROTO_SVMTP":                                true,
		"IPPROTO_SWIPE":                                true,
		"IPPROTO_TCF":                                  true,
		"IPPROTO_TCP":                                  true,
		"IPPROTO_TLSP":                                 true,
		"IPPROTO_TP":                                   true,
		"IPPROTO_TPXX":                                 true,
		"IPPROTO_TRUNK1":                               true,
		"IPPROTO_TRUNK2":                               true,
		"IPPROTO_TTP":                                  true,
		"IPPROTO_UDP":                                  true,
		"IPPROTO_UDPLITE":                              true,
		"IPPROTO_VINES":                                true,
		"IPPROTO_VISA":                                 true,
		"IPPROTO_VMTP":                                 true,
		"IPPROTO_VRRP":                                 true,
		"IPPROTO_WBEXPAK":                              true,
		"IPPROTO_WBMON":                                true,
		"IPPROTO_WSN":                                  true,
		"IPPROTO_XNET":                                 true,
		"IPPROTO_XTP":                                  true,
		"IPV6_2292DSTOPTS":                             true,
		"IPV6_2292HOPLIMIT":                            true,
		"IPV6_2292HOPOPTS":                             true,
		"IPV6_2292NEXTHOP":                             true,
		"IPV6_2292PKTINFO":                             true,
		"IPV6_2292PKTOPTIONS":                          true,
		"IPV6_2292RTHDR":                               true,
		"IPV6_ADDRFORM":                                true,
		"IPV6_ADD_MEMBERSHIP":                          true,
		"IPV6_AUTHHDR":                                 true,
		"IPV6_AUTH_LEVEL":                              true,
		"IPV6_AUTOFLOWLABEL":                           true,
		"IPV6_BINDANY":                                 true,
		"IPV6_BINDV6ONLY":                              true,
		"IPV6_BOUND_IF":                                true,
		"IPV6_CHECKSUM":                                true,
		"IPV6_DEFAULT_MULTICAST_HOPS":                  true,
		"IPV6_DEFAULT_MULTICAST_LOOP":                  true,
		"IPV6_DEFHLIM":                                 true,
		"IPV6_DONTFRAG":                                true,
		"IPV6_DROP_MEMBERSHIP":                         true,
		"IPV6_DSTOPTS":                                 true,
		"IPV6_ESP_NETWORK_LEVEL":                       true,
		"IPV6_ESP_TRANS_LEVEL":                         true,
		"IPV6_FAITH":                                   true,
		"IPV6_FLOWINFO_MASK":                           true,
		"IPV6_FLOWLABEL_MASK":                          true,
		"IPV6_FRAGTTL":                                 true,
		"IPV6_FW_ADD":                                  true,
		"IPV6_FW_DEL":                                  true,
		"IPV6_FW_FLUSH":                                true,
		"IPV6_FW_GET":                                  true,
		"IPV6_FW_ZERO":                                 true,
		"IPV6_HLIMDEC":                                 true,
		"IPV6_HOPLIMIT":                                true,
		"IPV6_HOPOPTS":                                 true,
		"IPV6_IPCOMP_LEVEL":                            true,
		"IPV6_IPSEC_POLICY":                            true,
		"IPV6_JOIN_ANYCAST":                            true,
		"IPV6_JOIN_GROUP":                              true,
		"IPV6_LEAVE_ANYCAST":                           true,
		"IPV6_LEAVE_GROUP":                             true,
		"IPV6_MAXHLIM":                                 true,
		"IPV6_MAXOPTHDR":                               true,
		"IPV6_MAXPACKET":                               true,
		"IPV6_MAX_GROUP_SRC_FILTER":                    true,
		"IPV6_MAX_MEMBERSHIPS":                         true,
		"IPV6_MAX_SOCK_SRC_FILTER":                     true,
		"IPV6_MIN_MEMBERSHIPS":                         true,
		"IPV6_MMTU":                                    true,
		"IPV6_MSFILTER":                                true,
		"IPV6_MTU":                                     true,
		"IPV6_MTU_DISCOVER":                            true,
		"IPV6_MULTICAST_HOPS":                          true,
		"IPV6_MULTICAST_IF":                            true,
		"IPV6_MULTICAST_LOOP":                          true,
		"IPV6_NEXTHOP":                                 true,
		"IPV6_OPTIONS":                                 true,
		"IPV6_PATHMTU":                                 true,
		"IPV6_PIPEX":                                   true,
		"IPV6_PKTINFO":                                 true,
		"IPV6_PMTUDISC_DO":                             true,
		"IPV6_PMTUDISC_DONT":                           true,
		"IPV6_PMTUDISC_PROBE":                          true,
		"IPV6_PMTUDISC_WANT":                           true,
		"IPV6_PORTRANGE":                               true,
		"IPV6_PORTRANGE_DEFAULT":                       true,
		"IPV6_PORTRANGE_HIGH":                          true,
		"IPV6_PORTRANGE_LOW":                           true,
		"IPV6_PREFER_TEMPADDR":                         true,
		"IPV6_RECVDSTOPTS":                             true,
		"IPV6_RECVDSTPORT":                             true,
		"IPV6_RECVERR":                                 true,
		"IPV6_RECVHOPLIMIT":                            true,
		"IPV6_RECVHOPOPTS":                             true,
		"IPV6_RECVPATHMTU":                             true,
		"IPV6_RECVPKTINFO":                             true,
		"IPV6_RECVRTHDR":                               true,
		"IPV6_RECVTCLASS":                              true,
		"IPV6_ROUTER_ALERT":                            true,
		"IPV6_RTABLE":                                  true,
		"IPV6_RTHDR":                                   true,
		"IPV6_RTHDRDSTOPTS":                            true,
		"IPV6_RTHDR_LOOSE":                             true,
		"IPV6_RTHDR_STRICT":                            true,
		"IPV6_RTHDR_TYPE_0":                            true,
		"IPV6_RXDSTOPTS":                               true,
		"IPV6_RXHOPOPTS":                               true,
		"IPV6_SOCKOPT_RESERVED1":                       true,
		"IPV6_TCLASS":                                  true,
		"IPV6_UNICAST_HOPS":                            true,
		"IPV6_USE_MIN_MTU":                             true,
		"IPV6_V6ONLY":                                  true,
		"IPV6_VERSION":                                 true,
		"IPV6_VERSION_MASK":                            true,
		"IPV6_XFRM_POLICY":                             true,
		"IP_ADD_MEMBERSHIP":                            true,
		"IP_ADD_SOURCE_MEMBERSHIP":                     true,
		"IP_AUTH_LEVEL":                                true,
		"IP_BINDANY":                                   true,
		"IP_BLOCK_SOURCE":                              true,
		"IP_BOUND_IF":                                  true,
		"IP_DEFAULT_MULTICAST_LOOP":                    true,
		"IP_DEFAULT_MULTICAST_TTL":                     true,
		"IP_DF":                                        true,
		"IP_DIVERTFL":                                  true,
		"IP_DONTFRAG":                                  true,
		"IP_DROP_MEMBERSHIP":                           true,
		"IP_DROP_SOURCE_MEMBERSHIP":                    true,
		"IP_DUMMYNET3":                                 true,
		"IP_DUMMYNET_CONFIGURE":                        true,
		"IP_DUMMYNET_DEL":                              true,
		"IP_DUMMYNET_FLUSH":                            true,
		"IP_DUMMYNET_GET":                              true,
		"IP_EF":                                        true,
		"IP_ERRORMTU":                                  true,
		"IP_ESP_NETWORK_LEVEL":                         true,
		"IP_ESP_TRANS_LEVEL":                           true,
		"IP_FAITH":                                     true,
		"IP_FREEBIND":                                  true,
		"IP_FW3":                                       true,
		"IP_FW_ADD":                                    true,
		"IP_FW_DEL":                                    true,
		"IP_FW_FLUSH":                                  true,
		"IP_FW_GET":                                    true,
		"IP_FW_NAT_CFG":                                true,
		"IP_FW_NAT_DEL":                                true,
		"IP_FW_NAT_GET_CONFIG":                         true,
		"IP_FW_NAT_GET_LOG":                            true,
		"IP_FW_RESETLOG":                               true,
		"IP_FW_TABLE_ADD":                              true,
		"IP_FW_TABLE_DEL":                              true,
		"IP_FW_TABLE_FLUSH":                            true,
		"IP_FW_TABLE_GETSIZE":                          true,
		"IP_FW_TABLE_LIST":                             true,
		"IP_FW_ZERO":                                   true,
		"IP_HDRINCL":                                   true,
		"IP_IPCOMP_LEVEL":                              true,
		"IP_IPSECFLOWINFO":                             true,
		"IP_IPSEC_LOCAL_AUTH":                          true,
		"IP_IPSEC_LOCAL_CRED":                          true,
		"IP_IPSEC_LOCAL_ID":                            true,
		"IP_IPSEC_POLICY":                              true,
		"IP_IPSEC_REMOTE_AUTH":                         true,
		"IP_IPSEC_REMOTE_CRED":                         true,
		"IP_IPSEC_REMOTE_ID":                           true,
		"IP_MAXPACKET":                                 true,
		"IP_MAX_GROUP_SRC_FILTER":                      true,
		"IP_MAX_MEMBERSHIPS":                           true,
		"IP_MAX_SOCK_MUTE_FILTER":                      true,
		"IP_MAX_SOCK_SRC_FILTER":                       true,
		"IP_MAX_SOURCE_FILTER":                         true,
		"IP_MF":                                        true,
		"IP_MINFRAGSIZE":                               true,
		"IP_MINTTL":                                    true,
		"IP_MIN_MEMBERSHIPS":                           true,
		"IP_MSFILTER":                                  true,
		"IP_MSS":                                       true,
		"IP_MTU":                                       true,
		"IP_MTU_DISCOVER":                              true,
		"IP_MULTICAST_IF":                              true,
		"IP_MULTICAST_IFINDEX":                         true,
		"IP_MULTICAST_LOOP":                            true,
		"IP_MULTICAST_TTL":                             true,
		"IP_MULTICAST_VIF":                             true,
		"IP_NAT__XXX":                                  true,
		"IP_OFFMASK":                                   true,
		"IP_OLD_FW_ADD":                                true,
		"IP_OLD_FW_DEL":                                true,
		"IP_OLD_FW_FLUSH":                              true,
		"IP_OLD_FW_GET":                                true,
		"IP_OLD_FW_RESETLOG":                           true,
		"IP_OLD_FW_ZERO":                               true,
		"IP_ONESBCAST":                                 true,
		"IP_OPTIONS":                                   true,
		"IP_ORIGDSTADDR":                               true,
		"IP_PASSSEC":                                   true,
		"IP_PIPEX":                                     true,
		"IP_PKTINFO":                                   true,
		"IP_PKTOPTIONS":                                true,
		"IP_PMTUDISC":                                  true,
		"IP_PMTUDISC_DO":                               true,
		"IP_PMTUDISC_DONT":                             true,
		"IP_PMTUDISC_PROBE":                            true,
		"IP_PMTUDISC_WANT":                             true,
		"IP_PORTRANGE":                                 true,
		"IP_PORTRANGE_DEFAULT":                         true,
		"IP_PORTRANGE_HIGH":                            true,
		"IP_PORTRANGE_LOW":                             true,
		"IP_RECVDSTADDR":                               true,
		"IP_RECVDSTPORT":                               true,
		"IP_RECVERR":                                   true,
		"IP_RECVIF":                                    true,
		"IP_RECVOPTS":                                  true,
		"IP_RECVORIGDSTADDR":                           true,
		"IP_RECVPKTINFO":                               true,
		"IP_RECVRETOPTS":                               true,
		"IP_RECVRTABLE":                                true,
		"IP_RECVTOS":                                   true,
		"IP_RECVTTL":                                   true,
		"IP_RETOPTS":                                   true,
		"IP_RF":                                        true,
		"IP_ROUTER_ALERT":                              true,
		"IP_RSVP_OFF":                                  true,
		"IP_RSVP_ON":                                   true,
		"IP_RSVP_VIF_OFF":                              true,
		"IP_RSVP_VIF_ON":                               true,
		"IP_RTABLE":                                    true,
		"IP_SENDSRCADDR":                               true,
		"IP_STRIPHDR":                                  true,
		"IP_TOS":                                       true,
		"IP_TRAFFIC_MGT_BACKGROUND":                    true,
		"IP_TRANSPARENT":                               true,
		"IP_TTL":                                       true,
		"IP_UNBLOCK_SOURCE":                            true,
		"IP_XFRM_POLICY":                               true,
		"IPv6MTUInfo":                                  true,
		"IPv6Mreq":                                     true,
		"ISIG":                                         true,
		"ISTRIP":                                       true,
		"IUCLC":                                        true,
		"IUTF8":                                        true,
		"IXANY":                                        true,
		"IXOFF":                                        true,
		"IXON":                                         true,
		"IfAddrmsg":                                    true,
		"IfAnnounceMsghdr":                             true,
		"IfData":                                       true,
		"IfInfomsg":                                    true,
		"IfMsghdr":                                     true,
		"IfaMsghdr":                                    true,
		"IfmaMsghdr":                                   true,
		"IfmaMsghdr2":                                  true,
		"ImplementsGetwd":                              true,
		"Inet4Pktinfo":                                 true,
		"Inet6Pktinfo":                                 true,
		"InotifyAddWatch":                              true,
		"InotifyEvent":                                 true,
		"InotifyInit":                                  true,
		"InotifyInit1":                                 true,
		"InotifyRmWatch":                               true,
		"InterfaceAddrMessage":                         true,
		"InterfaceAnnounceMessage":                     true,
		"InterfaceInfo":                                true,
		"InterfaceMessage":                             true,
		"InterfaceMulticastAddrMessage":                true,
		"InvalidHandle":                                true,
		"Ioperm":                                       true,
		"Iopl":                                         true,
		"Iovec":                                        true,
		"IpAdapterInfo":                                true,
		"IpAddrString":                                 true,
		"IpAddressString":                              true,
		"IpMaskString":                                 true,
		"Issetugid":                                    true,
		"KEY_ALL_ACCESS":                               true,
		"KEY_CREATE_LINK":                              true,
		"KEY_CREATE_SUB_KEY":                           true,
		"KEY_ENUMERATE_SUB_KEYS":                       true,
		"KEY_EXECUTE":                                  true,
		"KEY_NOTIFY":                                   true,
		"KEY_QUERY_VALUE":                              true,
		"KEY_READ":                                     true,
		"KEY_SET_VALUE":                                true,
		"KEY_WOW64_32KEY":                              true,
		"KEY_WOW64_64KEY":                              true,
		"KEY_WRITE":                                    true,
		"Kevent":                                       true,
		"Kevent_t":                                     true,
		"Kill":                                         true,
		"Klogctl":                                      true,
		"Kqueue":                                       true,
		"LANG_ENGLISH":                                 true,
		"LAYERED_PROTOCOL":                             true,
		"LCNT_OVERLOAD_FLUSH":                          true,
		"LINUX_REBOOT_CMD_CAD_OFF":                     true,
		"LINUX_REBOOT_CMD_CAD_ON":                      true,
		"LINUX_REBOOT_CMD_HALT":                        true,
		"LINUX_REBOOT_CMD_KEXEC":                       true,
		"LINUX_REBOOT_CMD_POWER_OFF":                   true,
		"LINUX_REBOOT_CMD_RESTART":                     true,
		"LINUX_REBOOT_CMD_RESTART2":                    true,
		"LINUX_REBOOT_CMD_SW_SUSPEND":                  true,
		"LINUX_REBOOT_MAGIC1":                          true,
		"LINUX_REBOOT_MAGIC2":                          true,
		"LOCK_EX":                                      true,
		"LOCK_NB":                                      true,
		"LOCK_SH":                                      true,
		"LOCK_UN":                                      true,
		"LazyDLL":                                      true,
		"LazyProc":                                     true,
		"Lchown":                                       true,
		"Linger":                                       true,
		"Link":                                         true,
		"Listen":                                       true,
		"Listxattr":                                    true,
		"LoadCancelIoEx":                               true,
		"LoadConnectEx":                                true,
		"LoadCreateSymbolicLink":                       true,
		"LoadDLL":                                      true,
		"LoadGetAddrInfo":                              true,
		"LoadLibrary":                                  true,
		"LoadSetFileCompletionNotificationModes":       true,
		"LocalFree":                                    true,
		"Log2phys_t":                                   true,
		"LookupAccountName":                            true,
		"LookupAccountSid":                             true,
		"LookupSID":                                    true,
		"LsfJump":                                      true,
		"LsfSocket":                                    true,
		"LsfStmt":                                      true,
		"Lstat":                                        true,
		"MADV_AUTOSYNC":                                true,
		"MADV_CAN_REUSE":                               true,
		"MADV_CORE":                                    true,
		"MADV_DOFORK":                                  true,
		"MADV_DONTFORK":                                true,
		"MADV_DONTNEED":                                true,
		"MADV_FREE":                                    true,
		"MADV_FREE_REUSABLE":                           true,
		"MADV_FREE_REUSE":                              true,
		"MADV_HUGEPAGE":                                true,
		"MADV_HWPOISON":                                true,
		"MADV_MERGEABLE":                               true,
		"MADV_NOCORE":                                  true,
		"MADV_NOHUGEPAGE":                              true,
		"MADV_NORMAL":                                  true,
		"MADV_NOSYNC":                                  true,
		"MADV_PROTECT":                                 true,
		"MADV_RANDOM":                                  true,
		"MADV_REMOVE":                                  true,
		"MADV_SEQUENTIAL":                              true,
		"MADV_SPACEAVAIL":                              true,
		"MADV_UNMERGEABLE":                             true,
		"MADV_WILLNEED":                                true,
		"MADV_ZERO_WIRED_PAGES":                        true,
		"MAP_32BIT":                                    true,
		"MAP_ALIGNED_SUPER":                            true,
		"MAP_ALIGNMENT_16MB":                           true,
		"MAP_ALIGNMENT_1TB":                            true,
		"MAP_ALIGNMENT_256TB":                          true,
		"MAP_ALIGNMENT_4GB":                            true,
		"MAP_ALIGNMENT_64KB":                           true,
		"MAP_ALIGNMENT_64PB":                           true,
		"MAP_ALIGNMENT_MASK":                           true,
		"MAP_ALIGNMENT_SHIFT":                          true,
		"MAP_ANON":                                     true,
		"MAP_ANONYMOUS":                                true,
		"MAP_COPY":                                     true,
		"MAP_DENYWRITE":                                true,
		"MAP_EXECUTABLE":                               true,
		"MAP_FILE":                                     true,
		"MAP_FIXED":                                    true,
		"MAP_FLAGMASK":                                 true,
		"MAP_GROWSDOWN":                                true,
		"MAP_HASSEMAPHORE":                             true,
		"MAP_HUGETLB":                                  true,
		"MAP_INHERIT":                                  true,
		"MAP_INHERIT_COPY":                             true,
		"MAP_INHERIT_DEFAULT":                          true,
		"MAP_INHERIT_DONATE_COPY":                      true,
		"MAP_INHERIT_NONE":                             true,
		"MAP_INHERIT_SHARE":                            true,
		"MAP_JIT":                                      true,
		"MAP_LOCKED":                                   true,
		"MAP_NOCACHE":                                  true,
		"MAP_NOCORE":                                   true,
		"MAP_NOEXTEND":                                 true,
		"MAP_NONBLOCK":                                 true,
		"MAP_NORESERVE":                                true,
		"MAP_NOSYNC":                                   true,
		"MAP_POPULATE":                                 true,
		"MAP_PREFAULT_READ":                            true,
		"MAP_PRIVATE":                                  true,
		"MAP_RENAME":                                   true,
		"MAP_RESERVED0080":                             true,
		"MAP_RESERVED0100":                             true,
		"MAP_SHARED":                                   true,
		"MAP_STACK":                                    true,
		"MAP_TRYFIXED":                                 true,
		"MAP_TYPE":                                     true,
		"MAP_WIRED":                                    true,
		"MAXIMUM_REPARSE_DATA_BUFFER_SIZE":             true,
		"MAXLEN_IFDESCR":                               true,
		"MAXLEN_PHYSADDR":                              true,
		"MAX_ADAPTER_ADDRESS_LENGTH":                   true,
		"MAX_ADAPTER_DESCRIPTION_LENGTH":               true,
		"MAX_ADAPTER_NAME_LENGTH":                      true,
		"MAX_COMPUTERNAME_LENGTH":                      true,
		"MAX_INTERFACE_NAME_LEN":                       true,
		"MAX_LONG_PATH":                                true,
		"MAX_PATH":                                     true,
		"MAX_PROTOCOL_CHAIN":                           true,
		"MCL_CURRENT":                                  true,
		"MCL_FUTURE":                                   true,
		"MNT_DETACH":                                   true,
		"MNT_EXPIRE":                                   true,
		"MNT_FORCE":                                    true,
		"MSG_BCAST":                                    true,
		"MSG_CMSG_CLOEXEC":                             true,
		"MSG_COMPAT":                                   true,
		"MSG_CONFIRM":                                  true,
		"MSG_CONTROLMBUF":                              true,
		"MSG_CTRUNC":                                   true,
		"MSG_DONTROUTE":                                true,
		"MSG_DONTWAIT":                                 true,
		"MSG_EOF":                                      true,
		"MSG_EOR":                                      true,
		"MSG_ERRQUEUE":                                 true,
		"MSG_FASTOPEN":                                 true,
		"MSG_FIN":                                      true,
		"MSG_FLUSH":                                    true,
		"MSG_HAVEMORE":                                 true,
		"MSG_HOLD":                                     true,
		"MSG_IOVUSRSPACE":                              true,
		"MSG_LENUSRSPACE":                              true,
		"MSG_MCAST":                                    true,
		"MSG_MORE":                                     true,
		"MSG_NAMEMBUF":                                 true,
		"MSG_NBIO":                                     true,
		"MSG_NEEDSA":                                   true,
		"MSG_NOSIGNAL":                                 true,
		"MSG_NOTIFICATION":                             true,
		"MSG_OOB":                                      true,
		"MSG_PEEK":                                     true,
		"MSG_PROXY":                                    true,
		"MSG_RCVMORE":                                  true,
		"MSG_RST":                                      true,
		"MSG_SEND":                                     true,
		"MSG_SYN":                                      true,
		"MSG_TRUNC":                                    true,
		"MSG_TRYHARD":                                  true,
		"MSG_USERFLAGS":                                true,
		"MSG_WAITALL":                                  true,
		"MSG_WAITFORONE":                               true,
		"MSG_WAITSTREAM":                               true,
		"MS_ACTIVE":                                    true,
		"MS_ASYNC":                                     true,
		"MS_BIND":                                      true,
		"MS_DEACTIVATE":                                true,
		"MS_DIRSYNC":                                   true,
		"MS_INVALIDATE":                                true,
		"MS_I_VERSION":                                 true,
		"MS_KERNMOUNT":                                 true,
		"MS_KILLPAGES":                                 true,
		"MS_MANDLOCK":                                  true,
		"MS_MGC_MSK":                                   true,
		"MS_MGC_VAL":                                   true,
		"MS_MOVE":                                      true,
		"MS_NOATIME":                                   true,
		"MS_NODEV":                                     true,
		"MS_NODIRATIME":                                true,
		"MS_NOEXEC":                                    true,
		"MS_NOSUID":                                    true,
		"MS_NOUSER":                                    true,
		"MS_POSIXACL":                                  true,
		"MS_PRIVATE":                                   true,
		"MS_RDONLY":                                    true,
		"MS_REC":                                       true,
		"MS_RELATIME":                                  true,
		"MS_REMOUNT":                                   true,
		"MS_RMT_MASK":                                  true,
		"MS_SHARED":                                    true,
		"MS_SILENT":                                    true,
		"MS_SLAVE":                                     true,
		"MS_STRICTATIME":                               true,
		"MS_SYNC":                                      true,
		"MS_SYNCHRONOUS":                               true,
		"MS_UNBINDABLE":                                true,
		"Madvise":                                      true,
		"MapViewOfFile":                                true,
		"MaxTokenInfoClass":                            true,
		"Mclpool":                                      true,
		"MibIfRow":                                     true,
		"Mkdir":                                        true,
		"Mkdirat":                                      true,
		"Mkfifo":                                       true,
		"Mknod":                                        true,
		"Mknodat":                                      true,
		"Mlock":                                        true,
		"Mlockall":                                     true,
		"Mmap":                                         true,
		"Mount":                                        true,
		"MoveFile":                                     true,
		"Mprotect":                                     true,
		"Msghdr":                                       true,
		"Munlock":                                      true,
		"Munlockall":                                   true,
		"Munmap":                                       true,
		"MustLoadDLL":                                  true,
		"NAME_MAX":                                     true,
		"NETLINK_ADD_MEMBERSHIP":                       true,
		"NETLINK_AUDIT":                                true,
		"NETLINK_BROADCAST_ERROR":                      true,
		"NETLINK_CONNECTOR":                            true,
		"NETLINK_DNRTMSG":                              true,
		"NETLINK_DROP_MEMBERSHIP":                      true,
		"NETLINK_ECRYPTFS":                             true,
		"NETLINK_FIB_LOOKUP":                           true,
		"NETLINK_FIREWALL":                             true,
		"NETLINK_GENERIC":                              true,
		"NETLINK_INET_DIAG":                            true,
		"NETLINK_IP6_FW":                               true,
		"NETLINK_ISCSI":                                true,
		"NETLINK_KOBJECT_UEVENT":                       true,
		"NETLINK_NETFILTER":                            true,
		"NETLINK_NFLOG":                                true,
		"NETLINK_NO_ENOBUFS":                           true,
		"NETLINK_PKTINFO":                              true,
		"NETLINK_RDMA":                                 true,
		"NETLINK_ROUTE":                                true,
		"NETLINK_SCSITRANSPORT":                        true,
		"NETLINK_SELINUX":                              true,
		"NETLINK_UNUSED":                               true,
		"NETLINK_USERSOCK":                             true,
		"NETLINK_XFRM":                                 true,
		"NET_RT_DUMP":                                  true,
		"NET_RT_DUMP2":                                 true,
		"NET_RT_FLAGS":                                 true,
		"NET_RT_IFLIST":                                true,
		"NET_RT_IFLIST2":                               true,
		"NET_RT_IFLISTL":                               true,
		"NET_RT_IFMALIST":                              true,
		"NET_RT_MAXID":                                 true,
		"NET_RT_OIFLIST":                               true,
		"NET_RT_OOIFLIST":                              true,
		"NET_RT_STAT":                                  true,
		"NET_RT_STATS":                                 true,
		"NET_RT_TABLE":                                 true,
		"NET_RT_TRASH":                                 true,
		"NLA_ALIGNTO":                                  true,
		"NLA_F_NESTED":                                 true,
		"NLA_F_NET_BYTEORDER":                          true,
		"NLA_HDRLEN":                                   true,
		"NLMSG_ALIGNTO":                                true,
		"NLMSG_DONE":                                   true,
		"NLMSG_ERROR":                                  true,
		"NLMSG_HDRLEN":                                 true,
		"NLMSG_MIN_TYPE":                               true,
		"NLMSG_NOOP":                                   true,
		"NLMSG_OVERRUN":                                true,
		"NLM_F_ACK":                                    true,
		"NLM_F_APPEND":                                 true,
		"NLM_F_ATOMIC":                                 true,
		"NLM_F_CREATE":                                 true,
		"NLM_F_DUMP":                                   true,
		"NLM_F_ECHO":                                   true,
		"NLM_F_EXCL":                                   true,
		"NLM_F_MATCH":                                  true,
		"NLM_F_MULTI":                                  true,
		"NLM_F_REPLACE":                                true,
		"NLM_F_REQUEST":                                true,
		"NLM_F_ROOT":                                   true,
		"NOFLSH":                                       true,
		"NOTE_ABSOLUTE":                                true,
		"NOTE_ATTRIB":                                  true,
		"NOTE_CHILD":                                   true,
		"NOTE_DELETE":                                  true,
		"NOTE_EOF":                                     true,
		"NOTE_EXEC":                                    true,
		"NOTE_EXIT":                                    true,
		"NOTE_EXITSTATUS":                              true,
		"NOTE_EXTEND":                                  true,
		"NOTE_FFAND":                                   true,
		"NOTE_FFCOPY":                                  true,
		"NOTE_FFCTRLMASK":                              true,
		"NOTE_FFLAGSMASK":                              true,
		"NOTE_FFNOP":                                   true,
		"NOTE_FFOR":                                    true,
		"NOTE_FORK":                                    true,
		"NOTE_LINK":                                    true,
		"NOTE_LOWAT":                                   true,
		"NOTE_NONE":                                    true,
		"NOTE_NSECONDS":                                true,
		"NOTE_PCTRLMASK":                               true,
		"NOTE_PDATAMASK":                               true,
		"NOTE_REAP":                                    true,
		"NOTE_RENAME":                                  true,
		"NOTE_RESOURCEEND":                             true,
		"NOTE_REVOKE":                                  true,
		"NOTE_SECONDS":                                 true,
		"NOTE_SIGNAL":                                  true,
		"NOTE_TRACK":                                   true,
		"NOTE_TRACKERR":                                true,
		"NOTE_TRIGGER":                                 true,
		"NOTE_TRUNCATE":                                true,
		"NOTE_USECONDS":                                true,
		"NOTE_VM_ERROR":                                true,
		"NOTE_VM_PRESSURE":                             true,
		"NOTE_VM_PRESSURE_SUDDEN_TERMINATE":            true,
		"NOTE_VM_PRESSURE_TERMINATE":                   true,
		"NOTE_WRITE":                                   true,
		"NameCanonical":                                true,
		"NameCanonicalEx":                              true,
		"NameDisplay":                                  true,
		"NameDnsDomain":                                true,
		"NameFullyQualifiedDN":                         true,
		"NameSamCompatible":                            true,
		"NameServicePrincipal":                         true,
		"NameUniqueId":                                 true,
		"NameUnknown":                                  true,
		"NameUserPrincipal":                            true,
		"Nanosleep":                                    true,
		"NetApiBufferFree":                             true,
		"NetGetJoinInformation":                        true,
		"NetSetupDomainName":                           true,
		"NetSetupUnjoined":                             true,
		"NetSetupUnknownStatus":                        true,
		"NetSetupWorkgroupName":                        true,
		"NetUserGetInfo":                               true,
		"NetlinkMessage":                               true,
		"NetlinkRIB":                                   true,
		"NetlinkRouteAttr":                             true,
		"NetlinkRouteRequest":                          true,
		"NewCallback":                                  true,
		"NewCallbackCDecl":                             true,
		"NewLazyDLL":                                   true,
		"NlAttr":                                       true,
		"NlMsgerr":                                     true,
		"NlMsghdr":                                     true,
		"NsecToFiletime":                               true,
		"NsecToTimespec":                               true,
		"NsecToTimeval":                                true,
		"Ntohs":                                        true,
		"OCRNL":                                        true,
		"OFDEL":                                        true,
		"OFILL":                                        true,
		"OFIOGETBMAP":                                  true,
		"OID_PKIX_KP_SERVER_AUTH":                      true,
		"OID_SERVER_GATED_CRYPTO":                      true,
		"OID_SGC_NETSCAPE":                             true,
		"OLCUC":                                        true,
		"ONLCR":                                        true,
		"ONLRET":                                       true,
		"ONOCR":                                        true,
		"ONOEOT":                                       true,
		"OPEN_ALWAYS":                                  true,
		"OPEN_EXISTING":                                true,
		"OPOST":                                        true,
		"O_ACCMODE":                                    true,
		"O_ALERT":                                      true,
		"O_ALT_IO":                                     true,
		"O_APPEND":                                     true,
		"O_ASYNC":                                      true,
		"O_CLOEXEC":                                    true,
		"O_CREAT":                                      true,
		"O_DIRECT":                                     true,
		"O_DIRECTORY":                                  true,
		"O_DSYNC":                                      true,
		"O_EVTONLY":                                    true,
		"O_EXCL":                                       true,
		"O_EXEC":                                       true,
		"O_EXLOCK":                                     true,
		"O_FSYNC":                                      true,
		"O_LARGEFILE":                                  true,
		"O_NDELAY":                                     true,
		"O_NOATIME":                                    true,
		"O_NOCTTY":                                     true,
		"O_NOFOLLOW":                                   true,
		"O_NONBLOCK":                                   true,
		"O_NOSIGPIPE":                                  true,
		"O_POPUP":                                      true,
		"O_RDONLY":                                     true,
		"O_RDWR":                                       true,
		"O_RSYNC":                                      true,
		"O_SHLOCK":                                     true,
		"O_SYMLINK":                                    true,
		"O_SYNC":                                       true,
		"O_TRUNC":                                      true,
		"O_TTY_INIT":                                   true,
		"O_WRONLY":                                     true,
		"Open":                                         true,
		"OpenCurrentProcessToken":                      true,
		"OpenProcess":                                  true,
		"OpenProcessToken":                             true,
		"Openat":                                       true,
		"Overlapped":                                   true,
		"PACKET_ADD_MEMBERSHIP":                        true,
		"PACKET_BROADCAST":                             true,
		"PACKET_DROP_MEMBERSHIP":                       true,
		"PACKET_FASTROUTE":                             true,
		"PACKET_HOST":                                  true,
		"PACKET_LOOPBACK":                              true,
		"PACKET_MR_ALLMULTI":                           true,
		"PACKET_MR_MULTICAST":                          true,
		"PACKET_MR_PROMISC":                            true,
		"PACKET_MULTICAST":                             true,
		"PACKET_OTHERHOST":                             true,
		"PACKET_OUTGOING":                              true,
		"PACKET_RECV_OUTPUT":                           true,
		"PACKET_RX_RING":                               true,
		"PACKET_STATISTICS":                            true,
		"PAGE_EXECUTE_READ":                            true,
		"PAGE_EXECUTE_READWRITE":                       true,
		"PAGE_EXECUTE_WRITECOPY":                       true,
		"PAGE_READONLY":                                true,
		"PAGE_READWRITE":                               true,
		"PAGE_WRITECOPY":                               true,
		"PARENB":                                       true,
		"PARMRK":                                       true,
		"PARODD":                                       true,
		"PENDIN":                                       true,
		"PFL_HIDDEN":                                   true,
		"PFL_MATCHES_PROTOCOL_ZERO":                    true,
		"PFL_MULTIPLE_PROTO_ENTRIES":                   true,
		"PFL_NETWORKDIRECT_PROVIDER":                   true,
		"PFL_RECOMMENDED_PROTO_ENTRY":                  true,
		"PF_FLUSH":                                     true,
		"PKCS_7_ASN_ENCODING":                          true,
		"PMC5_PIPELINE_FLUSH":                          true,
		"PRIO_PGRP":                                    true,
		"PRIO_PROCESS":                                 true,
		"PRIO_USER":                                    true,
		"PRI_IOFLUSH":                                  true,
		"PROCESS_QUERY_INFORMATION":                    true,
		"PROCESS_TERMINATE":                            true,
		"PROT_EXEC":                                    true,
		"PROT_GROWSDOWN":                               true,
		"PROT_GROWSUP":                                 true,
		"PROT_NONE":                                    true,
		"PROT_READ":                                    true,
		"PROT_WRITE":                                   true,
		"PROV_DH_SCHANNEL":                             true,
		"PROV_DSS":                                     true,
		"PROV_DSS_DH":                                  true,
		"PROV_EC_ECDSA_FULL":                           true,
		"PROV_EC_ECDSA_SIG":                            true,
		"PROV_EC_ECNRA_FULL":                           true,
		"PROV_EC_ECNRA_SIG":                            true,
		"PROV_FORTEZZA":                                true,
		"PROV_INTEL_SEC":                               true,
		"PROV_MS_EXCHANGE":                             true,
		"PROV_REPLACE_OWF":                             true,
		"PROV_RNG":                                     true,
		"PROV_RSA_AES":                                 true,
		"PROV_RSA_FULL":                                true,
		"PROV_RSA_SCHANNEL":                            true,
		"PROV_RSA_SIG":                                 true,
		"PROV_SPYRUS_LYNKS":                            true,
		"PROV_SSL":                                     true,
		"PR_CAPBSET_DROP":                              true,
		"PR_CAPBSET_READ":                              true,
		"PR_CLEAR_SECCOMP_FILTER":                      true,
		"PR_ENDIAN_BIG":                                true,
		"PR_ENDIAN_LITTLE":                             true,
		"PR_ENDIAN_PPC_LITTLE":                         true,
		"PR_FPEMU_NOPRINT":                             true,
		"PR_FPEMU_SIGFPE":                              true,
		"PR_FP_EXC_ASYNC":                              true,
		"PR_FP_EXC_DISABLED":                           true,
		"PR_FP_EXC_DIV":                                true,
		"PR_FP_EXC_INV":                                true,
		"PR_FP_EXC_NONRECOV":                           true,
		"PR_FP_EXC_OVF":                                true,
		"PR_FP_EXC_PRECISE":                            true,
		"PR_FP_EXC_RES":                                true,
		"PR_FP_EXC_SW_ENABLE":                          true,
		"PR_FP_EXC_UND":                                true,
		"PR_GET_DUMPABLE":                              true,
		"PR_GET_ENDIAN":                                true,
		"PR_GET_FPEMU":                                 true,
		"PR_GET_FPEXC":                                 true,
		"PR_GET_KEEPCAPS":                              true,
		"PR_GET_NAME":                                  true,
		"PR_GET_PDEATHSIG":                             true,
		"PR_GET_SECCOMP":                               true,
		"PR_GET_SECCOMP_FILTER":                        true,
		"PR_GET_SECUREBITS":                            true,
		"PR_GET_TIMERSLACK":                            true,
		"PR_GET_TIMING":                                true,
		"PR_GET_TSC":                                   true,
		"PR_GET_UNALIGN":                               true,
		"PR_MCE_KILL":                                  true,
		"PR_MCE_KILL_CLEAR":                            true,
		"PR_MCE_KILL_DEFAULT":                          true,
		"PR_MCE_KILL_EARLY":                            true,
		"PR_MCE_KILL_GET":                              true,
		"PR_MCE_KILL_LATE":                             true,
		"PR_MCE_KILL_SET":                              true,
		"PR_SECCOMP_FILTER_EVENT":                      true,
		"PR_SECCOMP_FILTER_SYSCALL":                    true,
		"PR_SET_DUMPABLE":                              true,
		"PR_SET_ENDIAN":                                true,
		"PR_SET_FPEMU":                                 true,
		"PR_SET_FPEXC":                                 true,
		"PR_SET_KEEPCAPS":                              true,
		"PR_SET_NAME":                                  true,
		"PR_SET_PDEATHSIG":                             true,
		"PR_SET_PTRACER":                               true,
		"PR_SET_SECCOMP":                               true,
		"PR_SET_SECCOMP_FILTER":                        true,
		"PR_SET_SECUREBITS":                            true,
		"PR_SET_TIMERSLACK":                            true,
		"PR_SET_TIMING":                                true,
		"PR_SET_TSC":                                   true,
		"PR_SET_UNALIGN":                               true,
		"PR_TASK_PERF_EVENTS_DISABLE":                  true,
		"PR_TASK_PERF_EVENTS_ENABLE":                   true,
		"PR_TIMING_STATISTICAL":                        true,
		"PR_TIMING_TIMESTAMP":                          true,
		"PR_TSC_ENABLE":                                true,
		"PR_TSC_SIGSEGV":                               true,
		"PR_UNALIGN_NOPRINT":                           true,
		"PR_UNALIGN_SIGBUS":                            true,
		"PTRACE_ARCH_PRCTL":                            true,
		"PTRACE_ATTACH":                                true,
		"PTRACE_CONT":                                  true,
		"PTRACE_DETACH":                                true,
		"PTRACE_EVENT_CLONE":                           true,
		"PTRACE_EVENT_EXEC":                            true,
		"PTRACE_EVENT_EXIT":                            true,
		"PTRACE_EVENT_FORK":                            true,
		"PTRACE_EVENT_VFORK":                           true,
		"PTRACE_EVENT_VFORK_DONE":                      true,
		"PTRACE_GETCRUNCHREGS":                         true,
		"PTRACE_GETEVENTMSG":                           true,
		"PTRACE_GETFPREGS":                             true,
		"PTRACE_GETFPXREGS":                            true,
		"PTRACE_GETHBPREGS":                            true,
		"PTRACE_GETREGS":                               true,
		"PTRACE_GETREGSET":                             true,
		"PTRACE_GETSIGINFO":                            true,
		"PTRACE_GETVFPREGS":                            true,
		"PTRACE_GETWMMXREGS":                           true,
		"PTRACE_GET_THREAD_AREA":                       true,
		"PTRACE_KILL":                                  true,
		"PTRACE_OLDSETOPTIONS":                         true,
		"PTRACE_O_MASK":                                true,
		"PTRACE_O_TRACECLONE":                          true,
		"PTRACE_O_TRACEEXEC":                           true,
		"PTRACE_O_TRACEEXIT":                           true,
		"PTRACE_O_TRACEFORK":                           true,
		"PTRACE_O_TRACESYSGOOD":                        true,
		"PTRACE_O_TRACEVFORK":                          true,
		"PTRACE_O_TRACEVFORKDONE":                      true,
		"PTRACE_PEEKDATA":                              true,
		"PTRACE_PEEKTEXT":                              true,
		"PTRACE_PEEKUSR":                               true,
		"PTRACE_POKEDATA":                              true,
		"PTRACE_POKETEXT":                              true,
		"PTRACE_POKEUSR":                               true,
		"PTRACE_SETCRUNCHREGS":                         true,
		"PTRACE_SETFPREGS":                             true,
		"PTRACE_SETFPXREGS":                            true,
		"PTRACE_SETHBPREGS":                            true,
		"PTRACE_SETOPTIONS":                            true,
		"PTRACE_SETREGS":                               true,
		"PTRACE_SETREGSET":                             true,
		"PTRACE_SETSIGINFO":                            true,
		"PTRACE_SETVFPREGS":                            true,
		"PTRACE_SETWMMXREGS":                           true,
		"PTRACE_SET_SYSCALL":                           true,
		"PTRACE_SET_THREAD_AREA":                       true,
		"PTRACE_SINGLEBLOCK":                           true,
		"PTRACE_SINGLESTEP":                            true,
		"PTRACE_SYSCALL":                               true,
		"PTRACE_SYSEMU":                                true,
		"PTRACE_SYSEMU_SINGLESTEP":                     true,
		"PTRACE_TRACEME":                               true,
		"PT_ATTACH":                                    true,
		"PT_ATTACHEXC":                                 true,
		"PT_CONTINUE":                                  true,
		"PT_DATA_ADDR":                                 true,
		"PT_DENY_ATTACH":                               true,
		"PT_DETACH":                                    true,
		"PT_FIRSTMACH":                                 true,
		"PT_FORCEQUOTA":                                true,
		"PT_KILL":                                      true,
		"PT_MASK":                                      true,
		"PT_READ_D":                                    true,
		"PT_READ_I":                                    true,
		"PT_READ_U":                                    true,
		"PT_SIGEXC":                                    true,
		"PT_STEP":                                      true,
		"PT_TEXT_ADDR":                                 true,
		"PT_TEXT_END_ADDR":                             true,
		"PT_THUPDATE":                                  true,
		"PT_TRACE_ME":                                  true,
		"PT_WRITE_D":                                   true,
		"PT_WRITE_I":                                   true,
		"PT_WRITE_U":                                   true,
		"ParseDirent":                                  true,
		"ParseNetlinkMessage":                          true,
		"ParseNetlinkRouteAttr":                        true,
		"ParseRoutingMessage":                          true,
		"ParseRoutingSockaddr":                         true,
		"ParseSocketControlMessage":                    true,
		"ParseUnixCredentials":                         true,
		"ParseUnixRights":                              true,
		"PathMax":                                      true,
		"Pathconf":                                     true,
		"Pause":                                        true,
		"Pipe":                                         true,
		"Pipe2":                                        true,
		"PivotRoot":                                    true,
		"Pointer":                                      true,
		"PostQueuedCompletionStatus":                   true,
		"Pread":                                        true,
		"Proc":                                         true,
		"ProcAttr":                                     true,
		"Process32First":                               true,
		"Process32Next":                                true,
		"ProcessEntry32":                               true,
		"ProcessInformation":                           true,
		"Protoent":                                     true,
		"PtraceAttach":                                 true,
		"PtraceCont":                                   true,
		"PtraceDetach":                                 true,
		"PtraceGetEventMsg":                            true,
		"PtraceGetRegs":                                true,
		"PtracePeekData":                               true,
		"PtracePeekText":                               true,
		"PtracePokeData":                               true,
		"PtracePokeText":                               true,
		"PtraceRegs":                                   true,
		"PtraceSetOptions":                             true,
		"PtraceSetRegs":                                true,
		"PtraceSingleStep":                             true,
		"PtraceSyscall":                                true,
		"Pwrite":                                       true,
		"REG_BINARY":                                   true,
		"REG_DWORD":                                    true,
		"REG_DWORD_BIG_ENDIAN":                         true,
		"REG_DWORD_LITTLE_ENDIAN":                      true,
		"REG_EXPAND_SZ":                                true,
		"REG_FULL_RESOURCE_DESCRIPTOR":                 true,
		"REG_LINK":                                     true,
		"REG_MULTI_SZ":                                 true,
		"REG_NONE":                                     true,
		"REG_QWORD":                                    true,
		"REG_QWORD_LITTLE_ENDIAN":                      true,
		"REG_RESOURCE_LIST":                            true,
		"REG_RESOURCE_REQUIREMENTS_LIST":               true,
		"REG_SZ":                                       true,
		"RLIMIT_AS":                                    true,
		"RLIMIT_CORE":                                  true,
		"RLIMIT_CPU":                                   true,
		"RLIMIT_DATA":                                  true,
		"RLIMIT_FSIZE":                                 true,
		"RLIMIT_NOFILE":                                true,
		"RLIMIT_STACK":                                 true,
		"RLIM_INFINITY":                                true,
		"RTAX_ADVMSS":                                  true,
		"RTAX_AUTHOR":                                  true,
		"RTAX_BRD":                                     true,
		"RTAX_CWND":                                    true,
		"RTAX_DST":                                     true,
		"RTAX_FEATURES":                                true,
		"RTAX_FEATURE_ALLFRAG":                         true,
		"RTAX_FEATURE_ECN":                             true,
		"RTAX_FEATURE_SACK":                            true,
		"RTAX_FEATURE_TIMESTAMP":                       true,
		"RTAX_GATEWAY":                                 true,
		"RTAX_GENMASK":                                 true,
		"RTAX_HOPLIMIT":                                true,
		"RTAX_IFA":                                     true,
		"RTAX_IFP":                                     true,
		"RTAX_INITCWND":                                true,
		"RTAX_INITRWND":                                true,
		"RTAX_LABEL":                                   true,
		"RTAX_LOCK":                                    true,
		"RTAX_MAX":                                     true,
		"RTAX_MTU":                                     true,
		"RTAX_NETMASK":                                 true,
		"RTAX_REORDERING":                              true,
		"RTAX_RTO_MIN":                                 true,
		"RTAX_RTT":                                     true,
		"RTAX_RTTVAR":                                  true,
		"RTAX_SRC":                                     true,
		"RTAX_SRCMASK":                                 true,
		"RTAX_SSTHRESH":                                true,
		"RTAX_TAG":                                     true,
		"RTAX_UNSPEC":                                  true,
		"RTAX_WINDOW":                                  true,
		"RTA_ALIGNTO":                                  true,
		"RTA_AUTHOR":                                   true,
		"RTA_BRD":                                      true,
		"RTA_CACHEINFO":                                true,
		"RTA_DST":                                      true,
		"RTA_FLOW":                                     true,
		"RTA_GATEWAY":                                  true,
		"RTA_GENMASK":                                  true,
		"RTA_IFA":                                      true,
		"RTA_IFP":                                      true,
		"RTA_IIF":                                      true,
		"RTA_LABEL":                                    true,
		"RTA_MAX":                                      true,
		"RTA_METRICS":                                  true,
		"RTA_MULTIPATH":                                true,
		"RTA_NETMASK":                                  true,
		"RTA_OIF":                                      true,
		"RTA_PREFSRC":                                  true,
		"RTA_PRIORITY":                                 true,
		"RTA_SRC":                                      true,
		"RTA_SRCMASK":                                  true,
		"RTA_TABLE":                                    true,
		"RTA_TAG":                                      true,
		"RTA_UNSPEC":                                   true,
		"RTCF_DIRECTSRC":                               true,
		"RTCF_DOREDIRECT":                              true,
		"RTCF_LOG":                                     true,
		"RTCF_MASQ":                                    true,
		"RTCF_NAT":                                     true,
		"RTCF_VALVE":                                   true,
		"RTF_ADDRCLASSMASK":                            true,
		"RTF_ADDRCONF":                                 true,
		"RTF_ALLONLINK":                                true,
		"RTF_ANNOUNCE":                                 true,
		"RTF_BLACKHOLE":                                true,
		"RTF_BROADCAST":                                true,
		"RTF_CACHE":                                    true,
		"RTF_CLONED":                                   true,
		"RTF_CLONING":                                  true,
		"RTF_CONDEMNED":                                true,
		"RTF_DEFAULT":                                  true,
		"RTF_DELCLONE":                                 true,
		"RTF_DONE":                                     true,
		"RTF_DYNAMIC":                                  true,
		"RTF_FLOW":                                     true,
		"RTF_FMASK":                                    true,
		"RTF_GATEWAY":                                  true,
		"RTF_GWFLAG_COMPAT":                            true,
		"RTF_HOST":                                     true,
		"RTF_IFREF":                                    true,
		"RTF_IFSCOPE":                                  true,
		"RTF_INTERFACE":                                true,
		"RTF_IRTT":                                     true,
		"RTF_LINKRT":                                   true,
		"RTF_LLDATA":                                   true,
		"RTF_LLINFO":                                   true,
		"RTF_LOCAL":                                    true,
		"RTF_MASK":                                     true,
		"RTF_MODIFIED":                                 true,
		"RTF_MPATH":                                    true,
		"RTF_MPLS":                                     true,
		"RTF_MSS":                                      true,
		"RTF_MTU":                                      true,
		"RTF_MULTICAST":                                true,
		"RTF_NAT":                                      true,
		"RTF_NOFORWARD":                                true,
		"RTF_NONEXTHOP":                                true,
		"RTF_NOPMTUDISC":                               true,
		"RTF_PERMANENT_ARP":                            true,
		"RTF_PINNED":                                   true,
		"RTF_POLICY":                                   true,
		"RTF_PRCLONING":                                true,
		"RTF_PROTO1":                                   true,
		"RTF_PROTO2":                                   true,
		"RTF_PROTO3":                                   true,
		"RTF_REINSTATE":                                true,
		"RTF_REJECT":                                   true,
		"RTF_RNH_LOCKED":                               true,
		"RTF_SOURCE":                                   true,
		"RTF_SRC":                                      true,
		"RTF_STATIC":                                   true,
		"RTF_STICKY":                                   true,
		"RTF_THROW":                                    true,
		"RTF_TUNNEL":                                   true,
		"RTF_UP":                                       true,
		"RTF_USETRAILERS":                              true,
		"RTF_WASCLONED":                                true,
		"RTF_WINDOW":                                   true,
		"RTF_XRESOLVE":                                 true,
		"RTM_ADD":                                      true,
		"RTM_BASE":                                     true,
		"RTM_CHANGE":                                   true,
		"RTM_CHGADDR":                                  true,
		"RTM_DELACTION":                                true,
		"RTM_DELADDR":                                  true,
		"RTM_DELADDRLABEL":                             true,
		"RTM_DELETE":                                   true,
		"RTM_DELLINK":                                  true,
		"RTM_DELMADDR":                                 true,
		"RTM_DELNEIGH":                                 true,
		"RTM_DELQDISC":                                 true,
		"RTM_DELROUTE":                                 true,
		"RTM_DELRULE":                                  true,
		"RTM_DELTCLASS":                                true,
		"RTM_DELTFILTER":                               true,
		"RTM_DESYNC":                                   true,
		"RTM_F_CLONED":                                 true,
		"RTM_F_EQUALIZE":                               true,
		"RTM_F_NOTIFY":                                 true,
		"RTM_F_PREFIX":                                 true,
		"RTM_GET":                                      true,
		"RTM_GET2":                                     true,
		"RTM_GETACTION":                                true,
		"RTM_GETADDR":                                  true,
		"RTM_GETADDRLABEL":                             true,
		"RTM_GETANYCAST":                               true,
		"RTM_GETDCB":                                   true,
		"RTM_GETLINK":                                  true,
		"RTM_GETMULTICAST":                             true,
		"RTM_GETNEIGH":                                 true,
		"RTM_GETNEIGHTBL":                              true,
		"RTM_GETQDISC":                                 true,
		"RTM_GETROUTE":                                 true,
		"RTM_GETRULE":                                  true,
		"RTM_GETTCLASS":                                true,
		"RTM_GETTFILTER":                               true,
		"RTM_IEEE80211":                                true,
		"RTM_IFANNOUNCE":                               true,
		"RTM_IFINFO":                                   true,
		"RTM_IFINFO2":                                  true,
		"RTM_LLINFO_UPD":                               true,
		"RTM_LOCK":                                     true,
		"RTM_LOSING":                                   true,
		"RTM_MAX":                                      true,
		"RTM_MAXSIZE":                                  true,
		"RTM_MISS":                                     true,
		"RTM_NEWACTION":                                true,
		"RTM_NEWADDR":                                  true,
		"RTM_NEWADDRLABEL":                             true,
		"RTM_NEWLINK":                                  true,
		"RTM_NEWMADDR":                                 true,
		"RTM_NEWMADDR2":                                true,
		"RTM_NEWNDUSEROPT":                             true,
		"RTM_NEWNEIGH":                                 true,
		"RTM_NEWNEIGHTBL":                              true,
		"RTM_NEWPREFIX":                                true,
		"RTM_NEWQDISC":                                 true,
		"RTM_NEWROUTE":                                 true,
		"RTM_NEWRULE":                                  true,
		"RTM_NEWTCLASS":                                true,
		"RTM_NEWTFILTER":                               true,
		"RTM_NR_FAMILIES":                              true,
		"RTM_NR_MSGTYPES":                              true,
		"RTM_OIFINFO":                                  true,
		"RTM_OLDADD":                                   true,
		"RTM_OLDDEL":                                   true,
		"RTM_OOIFINFO":                                 true,
		"RTM_REDIRECT":                                 true,
		"RTM_RESOLVE":                                  true,
		"RTM_RTTUNIT":                                  true,
		"RTM_SETDCB":                                   true,
		"RTM_SETGATE":                                  true,
		"RTM_SETLINK":                                  true,
		"RTM_SETNEIGHTBL":                              true,
		"RTM_VERSION":                                  true,
		"RTNH_ALIGNTO":                                 true,
		"RTNH_F_DEAD":                                  true,
		"RTNH_F_ONLINK":                                true,
		"RTNH_F_PERVASIVE":                             true,
		"RTNLGRP_IPV4_IFADDR":                          true,
		"RTNLGRP_IPV4_MROUTE":                          true,
		"RTNLGRP_IPV4_ROUTE":                           true,
		"RTNLGRP_IPV4_RULE":                            true,
		"RTNLGRP_IPV6_IFADDR":                          true,
		"RTNLGRP_IPV6_IFINFO":                          true,
		"RTNLGRP_IPV6_MROUTE":                          true,
		"RTNLGRP_IPV6_PREFIX":                          true,
		"RTNLGRP_IPV6_ROUTE":                           true,
		"RTNLGRP_IPV6_RULE":                            true,
		"RTNLGRP_LINK":                                 true,
		"RTNLGRP_ND_USEROPT":                           true,
		"RTNLGRP_NEIGH":                                true,
		"RTNLGRP_NONE":                                 true,
		"RTNLGRP_NOTIFY":                               true,
		"RTNLGRP_TC":                                   true,
		"RTN_ANYCAST":                                  true,
		"RTN_BLACKHOLE":                                true,
		"RTN_BROADCAST":                                true,
		"RTN_LOCAL":                                    true,
		"RTN_MAX":                                      true,
		"RTN_MULTICAST":                                true,
		"RTN_NAT":                                      true,
		"RTN_PROHIBIT":                                 true,
		"RTN_THROW":                                    true,
		"RTN_UNICAST":                                  true,
		"RTN_UNREACHABLE":                              true,
		"RTN_UNSPEC":                                   true,
		"RTN_XRESOLVE":                                 true,
		"RTPROT_BIRD":                                  true,
		"RTPROT_BOOT":                                  true,
		"RTPROT_DHCP":                                  true,
		"RTPROT_DNROUTED":                              true,
		"RTPROT_GATED":                                 true,
		"RTPROT_KERNEL":                                true,
		"RTPROT_MRT":                                   true,
		"RTPROT_NTK":                                   true,
		"RTPROT_RA":                                    true,
		"RTPROT_REDIRECT":                              true,
		"RTPROT_STATIC":                                true,
		"RTPROT_UNSPEC":                                true,
		"RTPROT_XORP":                                  true,
		"RTPROT_ZEBRA":                                 true,
		"RTV_EXPIRE":                                   true,
		"RTV_HOPCOUNT":                                 true,
		"RTV_MTU":                                      true,
		"RTV_RPIPE":                                    true,
		"RTV_RTT":                                      true,
		"RTV_RTTVAR":                                   true,
		"RTV_SPIPE":                                    true,
		"RTV_SSTHRESH":                                 true,
		"RTV_WEIGHT":                                   true,
		"RT_CACHING_CONTEXT":                           true,
		"RT_CLASS_DEFAULT":                             true,
		"RT_CLASS_LOCAL":                               true,
		"RT_CLASS_MAIN":                                true,
		"RT_CLASS_MAX":                                 true,
		"RT_CLASS_UNSPEC":                              true,
		"RT_DEFAULT_FIB":                               true,
		"RT_NORTREF":                                   true,
		"RT_SCOPE_HOST":                                true,
		"RT_SCOPE_LINK":                                true,
		"RT_SCOPE_NOWHERE":                             true,
		"RT_SCOPE_SITE":                                true,
		"RT_SCOPE_UNIVERSE":                            true,
		"RT_TABLEID_MAX":                               true,
		"RT_TABLE_COMPAT":                              true,
		"RT_TABLE_DEFAULT":                             true,
		"RT_TABLE_LOCAL":                               true,
		"RT_TABLE_MAIN":                                true,
		"RT_TABLE_MAX":                                 true,
		"RT_TABLE_UNSPEC":                              true,
		"RUSAGE_CHILDREN":                              true,
		"RUSAGE_SELF":                                  true,
		"RUSAGE_THREAD":                                true,
		"Radvisory_t":                                  true,
		"RawConn":                                      true,
		"RawSockaddr":                                  true,
		"RawSockaddrAny":                               true,
		"RawSockaddrDatalink":                          true,
		"RawSockaddrInet4":                             true,
		"RawSockaddrInet6":                             true,
		"RawSockaddrLinklayer":                         true,
		"RawSockaddrNetlink":                           true,
		"RawSockaddrUnix":                              true,
		"RawSyscall":                                   true,
		"RawSyscall6":                                  true,
		"Read":                                         true,
		"ReadConsole":                                  true,
		"ReadDirectoryChanges":                         true,
		"ReadDirent":                                   true,
		"ReadFile":                                     true,
		"Readlink":                                     true,
		"Reboot":                                       true,
		"Recvfrom":                                     true,
		"Recvmsg":                                      true,
		"RegCloseKey":                                  true,
		"RegEnumKeyEx":                                 true,
		"RegOpenKeyEx":                                 true,
		"RegQueryInfoKey":                              true,
		"RegQueryValueEx":                              true,
		"RemoveDirectory":                              true,
		"Removexattr":                                  true,
		"Rename":                                       true,
		"Renameat":                                     true,
		"Revoke":                                       true,
		"Rlimit":                                       true,
		"Rmdir":                                        true,
		"RouteMessage":                                 true,
		"RouteRIB":                                     true,
		"RtAttr":                                       true,
		"RtGenmsg":                                     true,
		"RtMetrics":                                    true,
		"RtMsg":                                        true,
		"RtMsghdr":                                     true,
		"RtNexthop":                                    true,
		"Rusage":                                       true,
		"SCM_BINTIME":                                  true,
		"SCM_CREDENTIALS":                              true,
		"SCM_CREDS":                                    true,
		"SCM_RIGHTS":                                   true,
		"SCM_TIMESTAMP":                                true,
		"SCM_TIMESTAMPING":                             true,
		"SCM_TIMESTAMPNS":                              true,
		"SCM_TIMESTAMP_MONOTONIC":                      true,
		"SHUT_RD":                                      true,
		"SHUT_RDWR":                                    true,
		"SHUT_WR":                                      true,
		"SID":                                          true,
		"SIDAndAttributes":                             true,
		"SIGABRT":                                      true,
		"SIGALRM":                                      true,
		"SIGBUS":                                       true,
		"SIGCHLD":                                      true,
		"SIGCLD":                                       true,
		"SIGCONT":                                      true,
		"SIGEMT":                                       true,
		"SIGFPE":                                       true,
		"SIGHUP":                                       true,
		"SIGILL":                                       true,
		"SIGINFO":                                      true,
		"SIGINT":                                       true,
		"SIGIO":                                        true,
		"SIGIOT":                                       true,
		"SIGKILL":                                      true,
		"SIGLIBRT":                                     true,
		"SIGLWP":                                       true,
		"SIGPIPE":                                      true,
		"SIGPOLL":                                      true,
		"SIGPROF":                                      true,
		"SIGPWR":                                       true,
		"SIGQUIT":                                      true,
		"SIGSEGV":                                      true,
		"SIGSTKFLT":                                    true,
		"SIGSTOP":                                      true,
		"SIGSYS":                                       true,
		"SIGTERM":                                      true,
		"SIGTHR":                                       true,
		"SIGTRAP":                                      true,
		"SIGTSTP":                                      true,
		"SIGTTIN":                                      true,
		"SIGTTOU":                                      true,
		"SIGUNUSED":                                    true,
		"SIGURG":                                       true,
		"SIGUSR1":                                      true,
		"SIGUSR2":                                      true,
		"SIGVTALRM":                                    true,
		"SIGWINCH":                                     true,
		"SIGXCPU":                                      true,
		"SIGXFSZ":                                      true,
		"SIOCADDDLCI":                                  true,
		"SIOCADDMULTI":                                 true,
		"SIOCADDRT":                                    true,
		"SIOCAIFADDR":                                  true,
		"SIOCAIFGROUP":                                 true,
		"SIOCALIFADDR":                                 true,
		"SIOCARPIPLL":                                  true,
		"SIOCATMARK":                                   true,
		"SIOCAUTOADDR":                                 true,
		"SIOCAUTONETMASK":                              true,
		"SIOCBRDGADD":                                  true,
		"SIOCBRDGADDS":                                 true,
		"SIOCBRDGARL":                                  true,
		"SIOCBRDGDADDR":                                true,
		"SIOCBRDGDEL":                                  true,
		"SIOCBRDGDELS":                                 true,
		"SIOCBRDGFLUSH":                                true,
		"SIOCBRDGFRL":                                  true,
		"SIOCBRDGGCACHE":                               true,
		"SIOCBRDGGFD":                                  true,
		"SIOCBRDGGHT":                                  true,
		"SIOCBRDGGIFFLGS":                              true,
		"SIOCBRDGGMA":                                  true,
		"SIOCBRDGGPARAM":                               true,
		"SIOCBRDGGPRI":                                 true,
		"SIOCBRDGGRL":                                  true,
		"SIOCBRDGGSIFS":                                true,
		"SIOCBRDGGTO":                                  true,
		"SIOCBRDGIFS":                                  true,
		"SIOCBRDGRTS":                                  true,
		"SIOCBRDGSADDR":                                true,
		"SIOCBRDGSCACHE":                               true,
		"SIOCBRDGSFD":                                  true,
		"SIOCBRDGSHT":                                  true,
		"SIOCBRDGSIFCOST":                              true,
		"SIOCBRDGSIFFLGS":                              true,
		"SIOCBRDGSIFPRIO":                              true,
		"SIOCBRDGSMA":                                  true,
		"SIOCBRDGSPRI":                                 true,
		"SIOCBRDGSPROTO":                               true,
		"SIOCBRDGSTO":                                  true,
		"SIOCBRDGSTXHC":                                true,
		"SIOCDARP":                                     true,
		"SIOCDELDLCI":                                  true,
		"SIOCDELMULTI":                                 true,
		"SIOCDELRT":                                    true,
		"SIOCDEVPRIVATE":                               true,
		"SIOCDIFADDR":                                  true,
		"SIOCDIFGROUP":                                 true,
		"SIOCDIFPHYADDR":                               true,
		"SIOCDLIFADDR":                                 true,
		"SIOCDRARP":                                    true,
		"SIOCGARP":                                     true,
		"SIOCGDRVSPEC":                                 true,
		"SIOCGETKALIVE":                                true,
		"SIOCGETLABEL":                                 true,
		"SIOCGETPFLOW":                                 true,
		"SIOCGETPFSYNC":                                true,
		"SIOCGETSGCNT":                                 true,
		"SIOCGETVIFCNT":                                true,
		"SIOCGETVLAN":                                  true,
		"SIOCGHIWAT":                                   true,
		"SIOCGIFADDR":                                  true,
		"SIOCGIFADDRPREF":                              true,
		"SIOCGIFALIAS":                                 true,
		"SIOCGIFALTMTU":                                true,
		"SIOCGIFASYNCMAP":                              true,
		"SIOCGIFBOND":                                  true,
		"SIOCGIFBR":                                    true,
		"SIOCGIFBRDADDR":                               true,
		"SIOCGIFCAP":                                   true,
		"SIOCGIFCONF":                                  true,
		"SIOCGIFCOUNT":                                 true,
		"SIOCGIFDATA":                                  true,
		"SIOCGIFDESCR":                                 true,
		"SIOCGIFDEVMTU":                                true,
		"SIOCGIFDLT":                                   true,
		"SIOCGIFDSTADDR":                               true,
		"SIOCGIFENCAP":                                 true,
		"SIOCGIFFIB":                                   true,
		"SIOCGIFFLAGS":                                 true,
		"SIOCGIFGATTR":                                 true,
		"SIOCGIFGENERIC":                               true,
		"SIOCGIFGMEMB":                                 true,
		"SIOCGIFGROUP":                                 true,
		"SIOCGIFHARDMTU":                               true,
		"SIOCGIFHWADDR":                                true,
		"SIOCGIFINDEX":                                 true,
		"SIOCGIFKPI":                                   true,
		"SIOCGIFMAC":                                   true,
		"SIOCGIFMAP":                                   true,
		"SIOCGIFMEDIA":                                 true,
		"SIOCGIFMEM":                                   true,
		"SIOCGIFMETRIC":                                true,
		"SIOCGIFMTU":                                   true,
		"SIOCGIFNAME":                                  true,
		"SIOCGIFNETMASK":                               true,
		"SIOCGIFPDSTADDR":                              true,
		"SIOCGIFPFLAGS":                                true,
		"SIOCGIFPHYS":                                  true,
		"SIOCGIFPRIORITY":                              true,
		"SIOCGIFPSRCADDR":                              true,
		"SIOCGIFRDOMAIN":                               true,
		"SIOCGIFRTLABEL":                               true,
		"SIOCGIFSLAVE":                                 true,
		"SIOCGIFSTATUS":                                true,
		"SIOCGIFTIMESLOT":                              true,
		"SIOCGIFTXQLEN":                                true,
		"SIOCGIFVLAN":                                  true,
		"SIOCGIFWAKEFLAGS":                             true,
		"SIOCGIFXFLAGS":                                true,
		"SIOCGLIFADDR":                                 true,
		"SIOCGLIFPHYADDR":                              true,
		"SIOCGLIFPHYRTABLE":                            true,
		"SIOCGLIFPHYTTL":                               true,
		"SIOCGLINKSTR":                                 true,
		"SIOCGLOWAT":                                   true,
		"SIOCGPGRP":                                    true,
		"SIOCGPRIVATE_0":                               true,
		"SIOCGPRIVATE_1":                               true,
		"SIOCGRARP":                                    true,
		"SIOCGSPPPPARAMS":                              true,
		"SIOCGSTAMP":                                   true,
		"SIOCGSTAMPNS":                                 true,
		"SIOCGVH":                                      true,
		"SIOCGVNETID":                                  true,
		"SIOCIFCREATE":                                 true,
		"SIOCIFCREATE2":                                true,
		"SIOCIFDESTROY":                                true,
		"SIOCIFGCLONERS":                               true,
		"SIOCINITIFADDR":                               true,
		"SIOCPROTOPRIVATE":                             true,
		"SIOCRSLVMULTI":                                true,
		"SIOCRTMSG":                                    true,
		"SIOCSARP":                                     true,
		"SIOCSDRVSPEC":                                 true,
		"SIOCSETKALIVE":                                true,
		"SIOCSETLABEL":                                 true,
		"SIOCSETPFLOW":                                 true,
		"SIOCSETPFSYNC":                                true,
		"SIOCSETVLAN":                                  true,
		"SIOCSHIWAT":                                   true,
		"SIOCSIFADDR":                                  true,
		"SIOCSIFADDRPREF":                              true,
		"SIOCSIFALTMTU":                                true,
		"SIOCSIFASYNCMAP":                              true,
		"SIOCSIFBOND":                                  true,
		"SIOCSIFBR":                                    true,
		"SIOCSIFBRDADDR":                               true,
		"SIOCSIFCAP":                                   true,
		"SIOCSIFDESCR":                                 true,
		"SIOCSIFDSTADDR":                               true,
		"SIOCSIFENCAP":                                 true,
		"SIOCSIFFIB":                                   true,
		"SIOCSIFFLAGS":                                 true,
		"SIOCSIFGATTR":                                 true,
		"SIOCSIFGENERIC":                               true,
		"SIOCSIFHWADDR":                                true,
		"SIOCSIFHWBROADCAST":                           true,
		"SIOCSIFKPI":                                   true,
		"SIOCSIFLINK":                                  true,
		"SIOCSIFLLADDR":                                true,
		"SIOCSIFMAC":                                   true,
		"SIOCSIFMAP":                                   true,
		"SIOCSIFMEDIA":                                 true,
		"SIOCSIFMEM":                                   true,
		"SIOCSIFMETRIC":                                true,
		"SIOCSIFMTU":                                   true,
		"SIOCSIFNAME":                                  true,
		"SIOCSIFNETMASK":                               true,
		"SIOCSIFPFLAGS":                                true,
		"SIOCSIFPHYADDR":                               true,
		"SIOCSIFPHYS":                                  true,
		"SIOCSIFPRIORITY":                              true,
		"SIOCSIFRDOMAIN":                               true,
		"SIOCSIFRTLABEL":                               true,
		"SIOCSIFRVNET":                                 true,
		"SIOCSIFSLAVE":                                 true,
		"SIOCSIFTIMESLOT":                              true,
		"SIOCSIFTXQLEN":                                true,
		"SIOCSIFVLAN":                                  true,
		"SIOCSIFVNET":                                  true,
		"SIOCSIFXFLAGS":                                true,
		"SIOCSLIFPHYADDR":                              true,
		"SIOCSLIFPHYRTABLE":                            true,
		"SIOCSLIFPHYTTL":                               true,
		"SIOCSLINKSTR":                                 true,
		"SIOCSLOWAT":                                   true,
		"SIOCSPGRP":                                    true,
		"SIOCSRARP":                                    true,
		"SIOCSSPPPPARAMS":                              true,
		"SIOCSVH":                                      true,
		"SIOCSVNETID":                                  true,
		"SIOCZIFDATA":                                  true,
		"SIO_GET_EXTENSION_FUNCTION_POINTER":           true,
		"SIO_GET_INTERFACE_LIST":                       true,
		"SIO_KEEPALIVE_VALS":                           true,
		"SIO_UDP_CONNRESET":                            true,
		"SOCK_CLOEXEC":                                 true,
		"SOCK_DCCP":                                    true,
		"SOCK_DGRAM":                                   true,
		"SOCK_FLAGS_MASK":                              true,
		"SOCK_MAXADDRLEN":                              true,
		"SOCK_NONBLOCK":                                true,
		"SOCK_NOSIGPIPE":                               true,
		"SOCK_PACKET":                                  true,
		"SOCK_RAW":                                     true,
		"SOCK_RDM":                                     true,
		"SOCK_SEQPACKET":                               true,
		"SOCK_STREAM":                                  true,
		"SOL_AAL":                                      true,
		"SOL_ATM":                                      true,
		"SOL_DECNET":                                   true,
		"SOL_ICMPV6":                                   true,
		"SOL_IP":                                       true,
		"SOL_IPV6":                                     true,
		"SOL_IRDA":                                     true,
		"SOL_PACKET":                                   true,
		"SOL_RAW":                                      true,
		"SOL_SOCKET":                                   true,
		"SOL_TCP":                                      true,
		"SOL_X25":                                      true,
		"SOMAXCONN":                                    true,
		"SO_ACCEPTCONN":                                true,
		"SO_ACCEPTFILTER":                              true,
		"SO_ATTACH_FILTER":                             true,
		"SO_BINDANY":                                   true,
		"SO_BINDTODEVICE":                              true,
		"SO_BINTIME":                                   true,
		"SO_BROADCAST":                                 true,
		"SO_BSDCOMPAT":                                 true,
		"SO_DEBUG":                                     true,
		"SO_DETACH_FILTER":                             true,
		"SO_DOMAIN":                                    true,
		"SO_DONTROUTE":                                 true,
		"SO_DONTTRUNC":                                 true,
		"SO_ERROR":                                     true,
		"SO_KEEPALIVE":                                 true,
		"SO_LABEL":                                     true,
		"SO_LINGER":                                    true,
		"SO_LINGER_SEC":                                true,
		"SO_LISTENINCQLEN":                             true,
		"SO_LISTENQLEN":                                true,
		"SO_LISTENQLIMIT":                              true,
		"SO_MARK":                                      true,
		"SO_NETPROC":                                   true,
		"SO_NKE":                                       true,
		"SO_NOADDRERR":                                 true,
		"SO_NOHEADER":                                  true,
		"SO_NOSIGPIPE":                                 true,
		"SO_NOTIFYCONFLICT":                            true,
		"SO_NO_CHECK":                                  true,
		"SO_NO_DDP":                                    true,
		"SO_NO_OFFLOAD":                                true,
		"SO_NP_EXTENSIONS":                             true,
		"SO_NREAD":                                     true,
		"SO_NWRITE":                                    true,
		"SO_OOBINLINE":                                 true,
		"SO_OVERFLOWED":                                true,
		"SO_PASSCRED":                                  true,
		"SO_PASSSEC":                                   true,
		"SO_PEERCRED":                                  true,
		"SO_PEERLABEL":                                 true,
		"SO_PEERNAME":                                  true,
		"SO_PEERSEC":                                   true,
		"SO_PRIORITY":                                  true,
		"SO_PROTOCOL":                                  true,
		"SO_PROTOTYPE":                                 true,
		"SO_RANDOMPORT":                                true,
		"SO_RCVBUF":                                    true,
		"SO_RCVBUFFORCE":                               true,
		"SO_RCVLOWAT":                                  true,
		"SO_RCVTIMEO":                                  true,
		"SO_RESTRICTIONS":                              true,
		"SO_RESTRICT_DENYIN":                           true,
		"SO_RESTRICT_DENYOUT":                          true,
		"SO_RESTRICT_DENYSET":                          true,
		"SO_REUSEADDR":                                 true,
		"SO_REUSEPORT":                                 true,
		"SO_REUSESHAREUID":                             true,
		"SO_RTABLE":                                    true,
		"SO_RXQ_OVFL":                                  true,
		"SO_SECURITY_AUTHENTICATION":                   true,
		"SO_SECURITY_ENCRYPTION_NETWORK":               true,
		"SO_SECURITY_ENCRYPTION_TRANSPORT":             true,
		"SO_SETFIB":                                    true,
		"SO_SNDBUF":                                    true,
		"SO_SNDBUFFORCE":                               true,
		"SO_SNDLOWAT":                                  true,
		"SO_SNDTIMEO":                                  true,
		"SO_SPLICE":                                    true,
		"SO_TIMESTAMP":                                 true,
		"SO_TIMESTAMPING":                              true,
		"SO_TIMESTAMPNS":                               true,
		"SO_TIMESTAMP_MONOTONIC":                       true,
		"SO_TYPE":                                      true,
		"SO_UPCALLCLOSEWAIT":                           true,
		"SO_UPDATE_ACCEPT_CONTEXT":                     true,
		"SO_UPDATE_CONNECT_CONTEXT":                    true,
		"SO_USELOOPBACK":                               true,
		"SO_USER_COOKIE":                               true,
		"SO_VENDOR":                                    true,
		"SO_WANTMORE":                                  true,
		"SO_WANTOOBFLAG":                               true,
		"SSLExtraCertChainPolicyPara":                  true,
		"STANDARD_RIGHTS_ALL":                          true,
		"STANDARD_RIGHTS_EXECUTE":                      true,
		"STANDARD_RIGHTS_READ":                         true,
		"STANDARD_RIGHTS_REQUIRED":                     true,
		"STANDARD_RIGHTS_WRITE":                        true,
		"STARTF_USESHOWWINDOW":                         true,
		"STARTF_USESTDHANDLES":                         true,
		"STD_ERROR_HANDLE":                             true,
		"STD_INPUT_HANDLE":                             true,
		"STD_OUTPUT_HANDLE":                            true,
		"SUBLANG_ENGLISH_US":                           true,
		"SW_FORCEMINIMIZE":                             true,
		"SW_HIDE":                                      true,
		"SW_MAXIMIZE":                                  true,
		"SW_MINIMIZE":                                  true,
		"SW_NORMAL":                                    true,
		"SW_RESTORE":                                   true,
		"SW_SHOW":                                      true,
		"SW_SHOWDEFAULT":                               true,
		"SW_SHOWMAXIMIZED":                             true,
		"SW_SHOWMINIMIZED":                             true,
		"SW_SHOWMINNOACTIVE":                           true,
		"SW_SHOWNA":                                    true,
		"SW_SHOWNOACTIVATE":                            true,
		"SW_SHOWNORMAL":                                true,
		"SYMBOLIC_LINK_FLAG_DIRECTORY":                 true,
		"SYNCHRONIZE":                                  true,
		"SYSCTL_VERSION":                               true,
		"SYSCTL_VERS_0":                                true,
		"SYSCTL_VERS_1":                                true,
		"SYSCTL_VERS_MASK":                             true,
		"SYS_ABORT2":                                   true,
		"SYS_ACCEPT":                                   true,
		"SYS_ACCEPT4":                                  true,
		"SYS_ACCEPT_NOCANCEL":                          true,
		"SYS_ACCESS":                                   true,
		"SYS_ACCESS_EXTENDED":                          true,
		"SYS_ACCT":                                     true,
		"SYS_ADD_KEY":                                  true,
		"SYS_ADD_PROFIL":                               true,
		"SYS_ADJFREQ":                                  true,
		"SYS_ADJTIME":                                  true,
		"SYS_ADJTIMEX":                                 true,
		"SYS_AFS_SYSCALL":                              true,
		"SYS_AIO_CANCEL":                               true,
		"SYS_AIO_ERROR":                                true,
		"SYS_AIO_FSYNC":                                true,
		"SYS_AIO_READ":                                 true,
		"SYS_AIO_RETURN":                               true,
		"SYS_AIO_SUSPEND":                              true,
		"SYS_AIO_SUSPEND_NOCANCEL":                     true,
		"SYS_AIO_WRITE":                                true,
		"SYS_ALARM":                                    true,
		"SYS_ARCH_PRCTL":                               true,
		"SYS_ARM_FADVISE64_64":                         true,
		"SYS_ARM_SYNC_FILE_RANGE":                      true,
		"SYS_ATGETMSG":                                 true,
		"SYS_ATPGETREQ":                                true,
		"SYS_ATPGETRSP":                                true,
		"SYS_ATPSNDREQ":                                true,
		"SYS_ATPSNDRSP":                                true,
		"SYS_ATPUTMSG":                                 true,
		"SYS_ATSOCKET":                                 true,
		"SYS_AUDIT":                                    true,
		"SYS_AUDITCTL":                                 true,
		"SYS_AUDITON":                                  true,
		"SYS_AUDIT_SESSION_JOIN":                       true,
		"SYS_AUDIT_SESSION_PORT":                       true,
		"SYS_AUDIT_SESSION_SELF":                       true,
		"SYS_BDFLUSH":                                  true,
		"SYS_BIND":                                     true,
		"SYS_BINDAT":                                   true,
		"SYS_BREAK":                                    true,
		"SYS_BRK":                                      true,
		"SYS_BSDTHREAD_CREATE":                         true,
		"SYS_BSDTHREAD_REGISTER":                       true,
		"SYS_BSDTHREAD_TERMINATE":                      true,
		"SYS_CAPGET":                                   true,
		"SYS_CAPSET":                                   true,
		"SYS_CAP_ENTER":                                true,
		"SYS_CAP_FCNTLS_GET":                           true,
		"SYS_CAP_FCNTLS_LIMIT":                         true,
		"SYS_CAP_GETMODE":                              true,
		"SYS_CAP_GETRIGHTS":                            true,
		"SYS_CAP_IOCTLS_GET":                           true,
		"SYS_CAP_IOCTLS_LIMIT":                         true,
		"SYS_CAP_NEW":                                  true,
		"SYS_CAP_RIGHTS_GET":                           true,
		"SYS_CAP_RIGHTS_LIMIT":                         true,
		"SYS_CHDIR":                                    true,
		"SYS_CHFLAGS":                                  true,
		"SYS_CHFLAGSAT":                                true,
		"SYS_CHMOD":                                    true,
		"SYS_CHMOD_EXTENDED":                           true,
		"SYS_CHOWN":                                    true,
		"SYS_CHOWN32":                                  true,
		"SYS_CHROOT":                                   true,
		"SYS_CHUD":                                     true,
		"SYS_CLOCK_ADJTIME":                            true,
		"SYS_CLOCK_GETCPUCLOCKID2":                     true,
		"SYS_CLOCK_GETRES":                             true,
		"SYS_CLOCK_GETTIME":                            true,
		"SYS_CLOCK_NANOSLEEP":                          true,
		"SYS_CLOCK_SETTIME":                            true,
		"SYS_CLONE":                                    true,
		"SYS_CLOSE":                                    true,
		"SYS_CLOSEFROM":                                true,
		"SYS_CLOSE_NOCANCEL":                           true,
		"SYS_CONNECT":                                  true,
		"SYS_CONNECTAT":                                true,
		"SYS_CONNECT_NOCANCEL":                         true,
		"SYS_COPYFILE":                                 true,
		"SYS_CPUSET":                                   true,
		"SYS_CPUSET_GETAFFINITY":                       true,
		"SYS_CPUSET_GETID":                             true,
		"SYS_CPUSET_SETAFFINITY":                       true,
		"SYS_CPUSET_SETID":                             true,
		"SYS_CREAT":                                    true,
		"SYS_CREATE_MODULE":                            true,
		"SYS_CSOPS":                                    true,
		"SYS_DELETE":                                   true,
		"SYS_DELETE_MODULE":                            true,
		"SYS_DUP":                                      true,
		"SYS_DUP2":                                     true,
		"SYS_DUP3":                                     true,
		"SYS_EACCESS":                                  true,
		"SYS_EPOLL_CREATE":                             true,
		"SYS_EPOLL_CREATE1":                            true,
		"SYS_EPOLL_CTL":                                true,
		"SYS_EPOLL_CTL_OLD":                            true,
		"SYS_EPOLL_PWAIT":                              true,
		"SYS_EPOLL_WAIT":                               true,
		"SYS_EPOLL_WAIT_OLD":                           true,
		"SYS_EVENTFD":                                  true,
		"SYS_EVENTFD2":                                 true,
		"SYS_EXCHANGEDATA":                             true,
		"SYS_EXECVE":                                   true,
		"SYS_EXIT":                                     true,
		"SYS_EXIT_GROUP":                               true,
		"SYS_EXTATTRCTL":                               true,
		"SYS_EXTATTR_DELETE_FD":                        true,
		"SYS_EXTATTR_DELETE_FILE":                      true,
		"SYS_EXTATTR_DELETE_LINK":                      true,
		"SYS_EXTATTR_GET_FD":                           true,
		"SYS_EXTATTR_GET_FILE":                         true,
		"SYS_EXTATTR_GET_LINK":                         true,
		"SYS_EXTATTR_LIST_FD":                          true,
		"SYS_EXTATTR_LIST_FILE":                        true,
		"SYS_EXTATTR_LIST_LINK":                        true,
		"SYS_EXTATTR_SET_FD":                           true,
		"SYS_EXTATTR_SET_FILE":                         true,
		"SYS_EXTATTR_SET_LINK":                         true,
		"SYS_FACCESSAT":                                true,
		"SYS_FADVISE64":                                true,
		"SYS_FADVISE64_64":                             true,
		"SYS_FALLOCATE":                                true,
		"SYS_FANOTIFY_INIT":                            true,
		"SYS_FANOTIFY_MARK":                            true,
		"SYS_FCHDIR":                                   true,
		"SYS_FCHFLAGS":                                 true,
		"SYS_FCHMOD":                                   true,
		"SYS_FCHMODAT":                                 true,
		"SYS_FCHMOD_EXTENDED":                          true,
		"SYS_FCHOWN":                                   true,
		"SYS_FCHOWN32":                                 true,
		"SYS_FCHOWNAT":                                 true,
		"SYS_FCHROOT":                                  true,
		"SYS_FCNTL":                                    true,
		"SYS_FCNTL64":                                  true,
		"SYS_FCNTL_NOCANCEL":                           true,
		"SYS_FDATASYNC":                                true,
		"SYS_FEXECVE":                                  true,
		"SYS_FFCLOCK_GETCOUNTER":                       true,
		"SYS_FFCLOCK_GETESTIMATE":                      true,
		"SYS_FFCLOCK_SETESTIMATE":                      true,
		"SYS_FFSCTL":                                   true,
		"SYS_FGETATTRLIST":                             true,
		"SYS_FGETXATTR":                                true,
		"SYS_FHOPEN":                                   true,
		"SYS_FHSTAT":                                   true,
		"SYS_FHSTATFS":                                 true,
		"SYS_FILEPORT_MAKEFD":                          true,
		"SYS_FILEPORT_MAKEPORT":                        true,
		"SYS_FKTRACE":                                  true,
		"SYS_FLISTXATTR":                               true,
		"SYS_FLOCK":                                    true,
		"SYS_FORK":                                     true,
		"SYS_FPATHCONF":                                true,
		"SYS_FREEBSD6_FTRUNCATE":                       true,
		"SYS_FREEBSD6_LSEEK":                           true,
		"SYS_FREEBSD6_MMAP":                            true,
		"SYS_FREEBSD6_PREAD":                           true,
		"SYS_FREEBSD6_PWRITE":                          true,
		"SYS_FREEBSD6_TRUNCATE":                        true,
		"SYS_FREMOVEXATTR":                             true,
		"SYS_FSCTL":                                    true,
		"SYS_FSETATTRLIST":                             true,
		"SYS_FSETXATTR":                                true,
		"SYS_FSGETPATH":                                true,
		"SYS_FSTAT":                                    true,
		"SYS_FSTAT64":                                  true,
		"SYS_FSTAT64_EXTENDED":                         true,
		"SYS_FSTATAT":                                  true,
		"SYS_FSTATAT64":                                true,
		"SYS_FSTATFS":                                  true,
		"SYS_FSTATFS64":                                true,
		"SYS_FSTATV":                                   true,
		"SYS_FSTATVFS1":                                true,
		"SYS_FSTAT_EXTENDED":                           true,
		"SYS_FSYNC":                                    true,
		"SYS_FSYNC_NOCANCEL":                           true,
		"SYS_FSYNC_RANGE":                              true,
		"SYS_FTIME":                                    true,
		"SYS_FTRUNCATE":                                true,
		"SYS_FTRUNCATE64":                              true,
		"SYS_FUTEX":                                    true,
		"SYS_FUTIMENS":                                 true,
		"SYS_FUTIMES":                                  true,
		"SYS_FUTIMESAT":                                true,
		"SYS_GETATTRLIST":                              true,
		"SYS_GETAUDIT":                                 true,
		"SYS_GETAUDIT_ADDR":                            true,
		"SYS_GETAUID":                                  true,
		"SYS_GETCONTEXT":                               true,
		"SYS_GETCPU":                                   true,
		"SYS_GETCWD":                                   true,
		"SYS_GETDENTS":                                 true,
		"SYS_GETDENTS64":                               true,
		"SYS_GETDIRENTRIES":                            true,
		"SYS_GETDIRENTRIES64":                          true,
		"SYS_GETDIRENTRIESATTR":                        true,
		"SYS_GETDTABLECOUNT":                           true,
		"SYS_GETDTABLESIZE":                            true,
		"SYS_GETEGID":                                  true,
		"SYS_GETEGID32":                                true,
		"SYS_GETEUID":                                  true,
		"SYS_GETEUID32":                                true,
		"SYS_GETFH":                                    true,
		"SYS_GETFSSTAT":                                true,
		"SYS_GETFSSTAT64":                              true,
		"SYS_GETGID":                                   true,
		"SYS_GETGID32":                                 true,
		"SYS_GETGROUPS":                                true,
		"SYS_GETGROUPS32":                              true,
		"SYS_GETHOSTUUID":                              true,
		"SYS_GETITIMER":                                true,
		"SYS_GETLCID":                                  true,
		"SYS_GETLOGIN":                                 true,
		"SYS_GETLOGINCLASS":                            true,
		"SYS_GETPEERNAME":                              true,
		"SYS_GETPGID":                                  true,
		"SYS_GETPGRP":                                  true,
		"SYS_GETPID":                                   true,
		"SYS_GETPMSG":                                  true,
		"SYS_GETPPID":                                  true,
		"SYS_GETPRIORITY":                              true,
		"SYS_GETRESGID":                                true,
		"SYS_GETRESGID32":                              true,
		"SYS_GETRESUID":                                true,
		"SYS_GETRESUID32":                              true,
		"SYS_GETRLIMIT":                                true,
		"SYS_GETRTABLE":                                true,
		"SYS_GETRUSAGE":                                true,
		"SYS_GETSGROUPS":                               true,
		"SYS_GETSID":                                   true,
		"SYS_GETSOCKNAME":                              true,
		"SYS_GETSOCKOPT":                               true,
		"SYS_GETTHRID":                                 true,
		"SYS_GETTID":                                   true,
		"SYS_GETTIMEOFDAY":                             true,
		"SYS_GETUID":                                   true,
		"SYS_GETUID32":                                 true,
		"SYS_GETVFSSTAT":                               true,
		"SYS_GETWGROUPS":                               true,
		"SYS_GETXATTR":                                 true,
		"SYS_GET_KERNEL_SYMS":                          true,
		"SYS_GET_MEMPOLICY":                            true,
		"SYS_GET_ROBUST_LIST":                          true,
		"SYS_GET_THREAD_AREA":                          true,
		"SYS_GTTY":                                     true,
		"SYS_IDENTITYSVC":                              true,
		"SYS_IDLE":                                     true,
		"SYS_INITGROUPS":                               true,
		"SYS_INIT_MODULE":                              true,
		"SYS_INOTIFY_ADD_WATCH":                        true,
		"SYS_INOTIFY_INIT":                             true,
		"SYS_INOTIFY_INIT1":                            true,
		"SYS_INOTIFY_RM_WATCH":                         true,
		"SYS_IOCTL":                                    true,
		"SYS_IOPERM":                                   true,
		"SYS_IOPL":                                     true,
		"SYS_IOPOLICYSYS":                              true,
		"SYS_IOPRIO_GET":                               true,
		"SYS_IOPRIO_SET":                               true,
		"SYS_IO_CANCEL":                                true,
		"SYS_IO_DESTROY":                               true,
		"SYS_IO_GETEVENTS":                             true,
		"SYS_IO_SETUP":                                 true,
		"SYS_IO_SUBMIT":                                true,
		"SYS_IPC":                                      true,
		"SYS_ISSETUGID":                                true,
		"SYS_JAIL":                                     true,
		"SYS_JAIL_ATTACH":                              true,
		"SYS_JAIL_GET":                                 true,
		"SYS_JAIL_REMOVE":                              true,
		"SYS_JAIL_SET":                                 true,
		"SYS_KDEBUG_TRACE":                             true,
		"SYS_KENV":                                     true,
		"SYS_KEVENT":                                   true,
		"SYS_KEVENT64":                                 true,
		"SYS_KEXEC_LOAD":                               true,
		"SYS_KEYCTL":                                   true,
		"SYS_KILL":                                     true,
		"SYS_KLDFIND":                                  true,
		"SYS_KLDFIRSTMOD":                              true,
		"SYS_KLDLOAD":                                  true,
		"SYS_KLDNEXT":                                  true,
		"SYS_KLDSTAT":                                  true,
		"SYS_KLDSYM":                                   true,
		"SYS_KLDUNLOAD":                                true,
		"SYS_KLDUNLOADF":                               true,
		"SYS_KQUEUE":                                   true,
		"SYS_KQUEUE1":                                  true,
		"SYS_KTIMER_CREATE":                            true,
		"SYS_KTIMER_DELETE":                            true,
		"SYS_KTIMER_GETOVERRUN":                        true,
		"SYS_KTIMER_GETTIME":                           true,
		"SYS_KTIMER_SETTIME":                           true,
		"SYS_KTRACE":                                   true,
		"SYS_LCHFLAGS":                                 true,
		"SYS_LCHMOD":                                   true,
		"SYS_LCHOWN":                                   true,
		"SYS_LCHOWN32":                                 true,
		"SYS_LGETFH":                                   true,
		"SYS_LGETXATTR":                                true,
		"SYS_LINK":                                     true,
		"SYS_LINKAT":                                   true,
		"SYS_LIO_LISTIO":                               true,
		"SYS_LISTEN":                                   true,
		"SYS_LISTXATTR":                                true,
		"SYS_LLISTXATTR":                               true,
		"SYS_LOCK":                                     true,
		"SYS_LOOKUP_DCOOKIE":                           true,
		"SYS_LPATHCONF":                                true,
		"SYS_LREMOVEXATTR":                             true,
		"SYS_LSEEK":                                    true,
		"SYS_LSETXATTR":                                true,
		"SYS_LSTAT":                                    true,
		"SYS_LSTAT64":                                  true,
		"SYS_LSTAT64_EXTENDED":                         true,
		"SYS_LSTATV":                                   true,
		"SYS_LSTAT_EXTENDED":                           true,
		"SYS_LUTIMES":                                  true,
		"SYS_MAC_SYSCALL":                              true,
		"SYS_MADVISE":                                  true,
		"SYS_MADVISE1":                                 true,
		"SYS_MAXSYSCALL":                               true,
		"SYS_MBIND":                                    true,
		"SYS_MIGRATE_PAGES":                            true,
		"SYS_MINCORE":                                  true,
		"SYS_MINHERIT":                                 true,
		"SYS_MKCOMPLEX":                                true,
		"SYS_MKDIR":                                    true,
		"SYS_MKDIRAT":                                  true,
		"SYS_MKDIR_EXTENDED":                           true,
		"SYS_MKFIFO":                                   true,
		"SYS_MKFIFOAT":                                 true,
		"SYS_MKFIFO_EXTENDED":                          true,
		"SYS_MKNOD":                                    true,
		"SYS_MKNODAT":                                  true,
		"SYS_MLOCK":                                    true,
		"SYS_MLOCKALL":                                 true,
		"SYS_MMAP":                                     true,
		"SYS_MMAP2":                                    true,
		"SYS_MODCTL":                                   true,
		"SYS_MODFIND":                                  true,
		"SYS_MODFNEXT":                                 true,
		"SYS_MODIFY_LDT":                               true,
		"SYS_MODNEXT":                                  true,
		"SYS_MODSTAT":                                  true,
		"SYS_MODWATCH":                                 true,
		"SYS_MOUNT":                                    true,
		"SYS_MOVE_PAGES":                               true,
		"SYS_MPROTECT":                                 true,
		"SYS_MPX":                                      true,
		"SYS_MQUERY":                                   true,
		"SYS_MQ_GETSETATTR":                            true,
		"SYS_MQ_NOTIFY":                                true,
		"SYS_MQ_OPEN":                                  true,
		"SYS_MQ_TIMEDRECEIVE":                          true,
		"SYS_MQ_TIMEDSEND":                             true,
		"SYS_MQ_UNLINK":                                true,
		"SYS_MREMAP":                                   true,
		"SYS_MSGCTL":                                   true,
		"SYS_MSGGET":                                   true,
		"SYS_MSGRCV":                                   true,
		"SYS_MSGRCV_NOCANCEL":                          true,
		"SYS_MSGSND":                                   true,
		"SYS_MSGSND_NOCANCEL":                          true,
		"SYS_MSGSYS":                                   true,
		"SYS_MSYNC":                                    true,
		"SYS_MSYNC_NOCANCEL":                           true,
		"SYS_MUNLOCK":                                  true,
		"SYS_MUNLOCKALL":                               true,
		"SYS_MUNMAP":                                   true,
		"SYS_NAME_TO_HANDLE_AT":                        true,
		"SYS_NANOSLEEP":                                true,
		"SYS_NEWFSTATAT":                               true,
		"SYS_NFSCLNT":                                  true,
		"SYS_NFSSERVCTL":                               true,
		"SYS_NFSSVC":                                   true,
		"SYS_NFSTAT":                                   true,
		"SYS_NICE":                                     true,
		"SYS_NLSTAT":                                   true,
		"SYS_NMOUNT":                                   true,
		"SYS_NSTAT":                                    true,
		"SYS_NTP_ADJTIME":                              true,
		"SYS_NTP_GETTIME":                              true,
		"SYS_OABI_SYSCALL_BASE":                        true,
		"SYS_OBREAK":                                   true,
		"SYS_OLDFSTAT":                                 true,
		"SYS_OLDLSTAT":                                 true,
		"SYS_OLDOLDUNAME":                              true,
		"SYS_OLDSTAT":                                  true,
		"SYS_OLDUNAME":                                 true,
		"SYS_OPEN":                                     true,
		"SYS_OPENAT":                                   true,
		"SYS_OPENBSD_POLL":                             true,
		"SYS_OPEN_BY_HANDLE_AT":                        true,
		"SYS_OPEN_EXTENDED":                            true,
		"SYS_OPEN_NOCANCEL":                            true,
		"SYS_OVADVISE":                                 true,
		"SYS_PACCEPT":                                  true,
		"SYS_PATHCONF":                                 true,
		"SYS_PAUSE":                                    true,
		"SYS_PCICONFIG_IOBASE":                         true,
		"SYS_PCICONFIG_READ":                           true,
		"SYS_PCICONFIG_WRITE":                          true,
		"SYS_PDFORK":                                   true,
		"SYS_PDGETPID":                                 true,
		"SYS_PDKILL":                                   true,
		"SYS_PERF_EVENT_OPEN":                          true,
		"SYS_PERSONALITY":                              true,
		"SYS_PID_HIBERNATE":                            true,
		"SYS_PID_RESUME":                               true,
		"SYS_PID_SHUTDOWN_SOCKETS":                     true,
		"SYS_PID_SUSPEND":                              true,
		"SYS_PIPE":                                     true,
		"SYS_PIPE2":                                    true,
		"SYS_PIVOT_ROOT":                               true,
		"SYS_PMC_CONTROL":                              true,
		"SYS_PMC_GET_INFO":                             true,
		"SYS_POLL":                                     true,
		"SYS_POLLTS":                                   true,
		"SYS_POLL_NOCANCEL":                            true,
		"SYS_POSIX_FADVISE":                            true,
		"SYS_POSIX_FALLOCATE":                          true,
		"SYS_POSIX_OPENPT":                             true,
		"SYS_POSIX_SPAWN":                              true,
		"SYS_PPOLL":                                    true,
		"SYS_PRCTL":                                    true,
		"SYS_PREAD":                                    true,
		"SYS_PREAD64":                                  true,
		"SYS_PREADV":                                   true,
		"SYS_PREAD_NOCANCEL":                           true,
		"SYS_PRLIMIT64":                                true,
		"SYS_PROCCTL":                                  true,
		"SYS_PROCESS_POLICY":                           true,
		"SYS_PROCESS_VM_READV":                         true,
		"SYS_PROCESS_VM_WRITEV":                        true,
		"SYS_PROC_INFO":                                true,
		"SYS_PROF":                                     true,
		"SYS_PROFIL":                                   true,
		"SYS_PSELECT":                                  true,
		"SYS_PSELECT6":                                 true,
		"SYS_PSET_ASSIGN":                              true,
		"SYS_PSET_CREATE":                              true,
		"SYS_PSET_DESTROY":                             true,
		"SYS_PSYNCH_CVBROAD":                           true,
		"SYS_PSYNCH_CVCLRPREPOST":                      true,
		"SYS_PSYNCH_CVSIGNAL":                          true,
		"SYS_PSYNCH_CVWAIT":                            true,
		"SYS_PSYNCH_MUTEXDROP":                         true,
		"SYS_PSYNCH_MUTEXWAIT":                         true,
		"SYS_PSYNCH_RW_DOWNGRADE":                      true,
		"SYS_PSYNCH_RW_LONGRDLOCK":                     true,
		"SYS_PSYNCH_RW_RDLOCK":                         true,
		"SYS_PSYNCH_RW_UNLOCK":                         true,
		"SYS_PSYNCH_RW_UNLOCK2":                        true,
		"SYS_PSYNCH_RW_UPGRADE":                        true,
		"SYS_PSYNCH_RW_WRLOCK":                         true,
		"SYS_PSYNCH_RW_YIELDWRLOCK":                    true,
		"SYS_PTRACE":                                   true,
		"SYS_PUTPMSG":                                  true,
		"SYS_PWRITE":                                   true,
		"SYS_PWRITE64":                                 true,
		"SYS_PWRITEV":                                  true,
		"SYS_PWRITE_NOCANCEL":                          true,
		"SYS_QUERY_MODULE":                             true,
		"SYS_QUOTACTL":                                 true,
		"SYS_RASCTL":                                   true,
		"SYS_RCTL_ADD_RULE":                            true,
		"SYS_RCTL_GET_LIMITS":                          true,
		"SYS_RCTL_GET_RACCT":                           true,
		"SYS_RCTL_GET_RULES":                           true,
		"SYS_RCTL_REMOVE_RULE":                         true,
		"SYS_READ":                                     true,
		"SYS_READAHEAD":                                true,
		"SYS_READDIR":                                  true,
		"SYS_READLINK":                                 true,
		"SYS_READLINKAT":                               true,
		"SYS_READV":                                    true,
		"SYS_READV_NOCANCEL":                           true,
		"SYS_READ_NOCANCEL":                            true,
		"SYS_REBOOT":                                   true,
		"SYS_RECV":                                     true,
		"SYS_RECVFROM":                                 true,
		"SYS_RECVFROM_NOCANCEL":                        true,
		"SYS_RECVMMSG":                                 true,
		"SYS_RECVMSG":                                  true,
		"SYS_RECVMSG_NOCANCEL":                         true,
		"SYS_REMAP_FILE_PAGES":                         true,
		"SYS_REMOVEXATTR":                              true,
		"SYS_RENAME":                                   true,
		"SYS_RENAMEAT":                                 true,
		"SYS_REQUEST_KEY":                              true,
		"SYS_RESTART_SYSCALL":                          true,
		"SYS_REVOKE":                                   true,
		"SYS_RFORK":                                    true,
		"SYS_RMDIR":                                    true,
		"SYS_RTPRIO":                                   true,
		"SYS_RTPRIO_THREAD":                            true,
		"SYS_RT_SIGACTION":                             true,
		"SYS_RT_SIGPENDING":                            true,
		"SYS_RT_SIGPROCMASK":                           true,
		"SYS_RT_SIGQUEUEINFO":                          true,
		"SYS_RT_SIGRETURN":                             true,
		"SYS_RT_SIGSUSPEND":                            true,
		"SYS_RT_SIGTIMEDWAIT":                          true,
		"SYS_RT_TGSIGQUEUEINFO":                        true,
		"SYS_SBRK":                                     true,
		"SYS_SCHED_GETAFFINITY":                        true,
		"SYS_SCHED_GETPARAM":                           true,
		"SYS_SCHED_GETSCHEDULER":                       true,
		"SYS_SCHED_GET_PRIORITY_MAX":                   true,
		"SYS_SCHED_GET_PRIORITY_MIN":                   true,
		"SYS_SCHED_RR_GET_INTERVAL":                    true,
		"SYS_SCHED_SETAFFINITY":                        true,
		"SYS_SCHED_SETPARAM":                           true,
		"SYS_SCHED_SETSCHEDULER":                       true,
		"SYS_SCHED_YIELD":                              true,
		"SYS_SCTP_GENERIC_RECVMSG":                     true,
		"SYS_SCTP_GENERIC_SENDMSG":                     true,
		"SYS_SCTP_GENERIC_SENDMSG_IOV":                 true,
		"SYS_SCTP_PEELOFF":                             true,
		"SYS_SEARCHFS":                                 true,
		"SYS_SECURITY":                                 true,
		"SYS_SELECT":                                   true,
		"SYS_SELECT_NOCANCEL":                          true,
		"SYS_SEMCONFIG":                                true,
		"SYS_SEMCTL":                                   true,
		"SYS_SEMGET":                                   true,
		"SYS_SEMOP":                                    true,
		"SYS_SEMSYS":                                   true,
		"SYS_SEMTIMEDOP":                               true,
		"SYS_SEM_CLOSE":                                true,
		"SYS_SEM_DESTROY":                              true,
		"SYS_SEM_GETVALUE":                             true,
		"SYS_SEM_INIT":                                 true,
		"SYS_SEM_OPEN":                                 true,
		"SYS_SEM_POST":                                 true,
		"SYS_SEM_TRYWAIT":                              true,
		"SYS_SEM_UNLINK":                               true,
		"SYS_SEM_WAIT":                                 true,
		"SYS_SEM_WAIT_NOCANCEL":                        true,
		"SYS_SEND":                                     true,
		"SYS_SENDFILE":                                 true,
		"SYS_SENDFILE64":                               true,
		"SYS_SENDMMSG":                                 true,
		"SYS_SENDMSG":                                  true,
		"SYS_SENDMSG_NOCANCEL":                         true,
		"SYS_SENDTO":                                   true,
		"SYS_SENDTO_NOCANCEL":                          true,
		"SYS_SETATTRLIST":                              true,
		"SYS_SETAUDIT":                                 true,
		"SYS_SETAUDIT_ADDR":                            true,
		"SYS_SETAUID":                                  true,
		"SYS_SETCONTEXT":                               true,
		"SYS_SETDOMAINNAME":                            true,
		"SYS_SETEGID":                                  true,
		"SYS_SETEUID":                                  true,
		"SYS_SETFIB":                                   true,
		"SYS_SETFSGID":                                 true,
		"SYS_SETFSGID32":                               true,
		"SYS_SETFSUID":                                 true,
		"SYS_SETFSUID32":                               true,
		"SYS_SETGID":                                   true,
		"SYS_SETGID32":                                 true,
		"SYS_SETGROUPS":                                true,
		"SYS_SETGROUPS32":                              true,
		"SYS_SETHOSTNAME":                              true,
		"SYS_SETITIMER":                                true,
		"SYS_SETLCID":                                  true,
		"SYS_SETLOGIN":                                 true,
		"SYS_SETLOGINCLASS":                            true,
		"SYS_SETNS":                                    true,
		"SYS_SETPGID":                                  true,
		"SYS_SETPRIORITY":                              true,
		"SYS_SETPRIVEXEC":                              true,
		"SYS_SETREGID":                                 true,
		"SYS_SETREGID32":                               true,
		"SYS_SETRESGID":                                true,
		"SYS_SETRESGID32":                              true,
		"SYS_SETRESUID":                                true,
		"SYS_SETRESUID32":                              true,
		"SYS_SETREUID":                                 true,
		"SYS_SETREUID32":                               true,
		"SYS_SETRLIMIT":                                true,
		"SYS_SETRTABLE":                                true,
		"SYS_SETSGROUPS":                               true,
		"SYS_SETSID":                                   true,
		"SYS_SETSOCKOPT":                               true,
		"SYS_SETTID":                                   true,
		"SYS_SETTID_WITH_PID":                          true,
		"SYS_SETTIMEOFDAY":                             true,
		"SYS_SETUID":                                   true,
		"SYS_SETUID32":                                 true,
		"SYS_SETWGROUPS":                               true,
		"SYS_SETXATTR":                                 true,
		"SYS_SET_MEMPOLICY":                            true,
		"SYS_SET_ROBUST_LIST":                          true,
		"SYS_SET_THREAD_AREA":                          true,
		"SYS_SET_TID_ADDRESS":                          true,
		"SYS_SGETMASK":                                 true,
		"SYS_SHARED_REGION_CHECK_NP":                   true,
		"SYS_SHARED_REGION_MAP_AND_SLIDE_NP":           true,
		"SYS_SHMAT":                                    true,
		"SYS_SHMCTL":                                   true,
		"SYS_SHMDT":                                    true,
		"SYS_SHMGET":                                   true,
		"SYS_SHMSYS":                                   true,
		"SYS_SHM_OPEN":                                 true,
		"SYS_SHM_UNLINK":                               true,
		"SYS_SHUTDOWN":                                 true,
		"SYS_SIGACTION":                                true,
		"SYS_SIGALTSTACK":                              true,
		"SYS_SIGNAL":                                   true,
		"SYS_SIGNALFD":                                 true,
		"SYS_SIGNALFD4":                                true,
		"SYS_SIGPENDING":                               true,
		"SYS_SIGPROCMASK":                              true,
		"SYS_SIGQUEUE":                                 true,
		"SYS_SIGQUEUEINFO":                             true,
		"SYS_SIGRETURN":                                true,
		"SYS_SIGSUSPEND":                               true,
		"SYS_SIGSUSPEND_NOCANCEL":                      true,
		"SYS_SIGTIMEDWAIT":                             true,
		"SYS_SIGWAIT":                                  true,
		"SYS_SIGWAITINFO":                              true,
		"SYS_SOCKET":                                   true,
		"SYS_SOCKETCALL":                               true,
		"SYS_SOCKETPAIR":                               true,
		"SYS_SPLICE":                                   true,
		"SYS_SSETMASK":                                 true,
		"SYS_SSTK":                                     true,
		"SYS_STACK_SNAPSHOT":                           true,
		"SYS_STAT":                                     true,
		"SYS_STAT64":                                   true,
		"SYS_STAT64_EXTENDED":                          true,
		"SYS_STATFS":                                   true,
		"SYS_STATFS64":                                 true,
		"SYS_STATV":                                    true,
		"SYS_STATVFS1":                                 true,
		"SYS_STAT_EXTENDED":                            true,
		"SYS_STIME":                                    true,
		"SYS_STTY":                                     true,
		"SYS_SWAPCONTEXT":                              true,
		"SYS_SWAPCTL":                                  true,
		"SYS_SWAPOFF":                                  true,
		"SYS_SWAPON":                                   true,
		"SYS_SYMLINK":                                  true,
		"SYS_SYMLINKAT":                                true,
		"SYS_SYNC":                                     true,
		"SYS_SYNCFS":                                   true,
		"SYS_SYNC_FILE_RANGE":                          true,
		"SYS_SYSARCH":                                  true,
		"SYS_SYSCALL":                                  true,
		"SYS_SYSCALL_BASE":                             true,
		"SYS_SYSFS":                                    true,
		"SYS_SYSINFO":                                  true,
		"SYS_SYSLOG":                                   true,
		"SYS_TEE":                                      true,
		"SYS_TGKILL":                                   true,
		"SYS_THREAD_SELFID":                            true,
		"SYS_THR_CREATE":                               true,
		"SYS_THR_EXIT":                                 true,
		"SYS_THR_KILL":                                 true,
		"SYS_THR_KILL2":                                true,
		"SYS_THR_NEW":                                  true,
		"SYS_THR_SELF":                                 true,
		"SYS_THR_SET_NAME":                             true,
		"SYS_THR_SUSPEND":                              true,
		"SYS_THR_WAKE":                                 true,
		"SYS_TIME":                                     true,
		"SYS_TIMERFD_CREATE":                           true,
		"SYS_TIMERFD_GETTIME":                          true,
		"SYS_TIMERFD_SETTIME":                          true,
		"SYS_TIMER_CREATE":                             true,
		"SYS_TIMER_DELETE":                             true,
		"SYS_TIMER_GETOVERRUN":                         true,
		"SYS_TIMER_GETTIME":                            true,
		"SYS_TIMER_SETTIME":                            true,
		"SYS_TIMES":                                    true,
		"SYS_TKILL":                                    true,
		"SYS_TRUNCATE":                                 true,
		"SYS_TRUNCATE64":                               true,
		"SYS_TUXCALL":                                  true,
		"SYS_UGETRLIMIT":                               true,
		"SYS_ULIMIT":                                   true,
		"SYS_UMASK":                                    true,
		"SYS_UMASK_EXTENDED":                           true,
		"SYS_UMOUNT":                                   true,
		"SYS_UMOUNT2":                                  true,
		"SYS_UNAME":                                    true,
		"SYS_UNDELETE":                                 true,
		"SYS_UNLINK":                                   true,
		"SYS_UNLINKAT":                                 true,
		"SYS_UNMOUNT":                                  true,
		"SYS_UNSHARE":                                  true,
		"SYS_USELIB":                                   true,
		"SYS_USTAT":                                    true,
		"SYS_UTIME":                                    true,
		"SYS_UTIMENSAT":                                true,
		"SYS_UTIMES":                                   true,
		"SYS_UTRACE":                                   true,
		"SYS_UUIDGEN":                                  true,
		"SYS_VADVISE":                                  true,
		"SYS_VFORK":                                    true,
		"SYS_VHANGUP":                                  true,
		"SYS_VM86":                                     true,
		"SYS_VM86OLD":                                  true,
		"SYS_VMSPLICE":                                 true,
		"SYS_VM_PRESSURE_MONITOR":                      true,
		"SYS_VSERVER":                                  true,
		"SYS_WAIT4":                                    true,
		"SYS_WAIT4_NOCANCEL":                           true,
		"SYS_WAIT6":                                    true,
		"SYS_WAITEVENT":                                true,
		"SYS_WAITID":                                   true,
		"SYS_WAITID_NOCANCEL":                          true,
		"SYS_WAITPID":                                  true,
		"SYS_WATCHEVENT":                               true,
		"SYS_WORKQ_KERNRETURN":                         true,
		"SYS_WORKQ_OPEN":                               true,
		"SYS_WRITE":                                    true,
		"SYS_WRITEV":                                   true,
		"SYS_WRITEV_NOCANCEL":                          true,
		"SYS_WRITE_NOCANCEL":                           true,
		"SYS_YIELD":                                    true,
		"SYS__LLSEEK":                                  true,
		"SYS__LWP_CONTINUE":                            true,
		"SYS__LWP_CREATE":                              true,
		"SYS__LWP_CTL":                                 true,
		"SYS__LWP_DETACH":                              true,
		"SYS__LWP_EXIT":                                true,
		"SYS__LWP_GETNAME":                             true,
		"SYS__LWP_GETPRIVATE":                          true,
		"SYS__LWP_KILL":                                true,
		"SYS__LWP_PARK":                                true,
		"SYS__LWP_SELF":                                true,
		"SYS__LWP_SETNAME":                             true,
		"SYS__LWP_SETPRIVATE":                          true,
		"SYS__LWP_SUSPEND":                             true,
		"SYS__LWP_UNPARK":                              true,
		"SYS__LWP_UNPARK_ALL":                          true,
		"SYS__LWP_WAIT":                                true,
		"SYS__LWP_WAKEUP":                              true,
		"SYS__NEWSELECT":                               true,
		"SYS__PSET_BIND":                               true,
		"SYS__SCHED_GETAFFINITY":                       true,
		"SYS__SCHED_GETPARAM":                          true,
		"SYS__SCHED_SETAFFINITY":                       true,
		"SYS__SCHED_SETPARAM":                          true,
		"SYS__SYSCTL":                                  true,
		"SYS__UMTX_LOCK":                               true,
		"SYS__UMTX_OP":                                 true,
		"SYS__UMTX_UNLOCK":                             true,
		"SYS___ACL_ACLCHECK_FD":                        true,
		"SYS___ACL_ACLCHECK_FILE":                      true,
		"SYS___ACL_ACLCHECK_LINK":                      true,
		"SYS___ACL_DELETE_FD":                          true,
		"SYS___ACL_DELETE_FILE":                        true,
		"SYS___ACL_DELETE_LINK":                        true,
		"SYS___ACL_GET_FD":                             true,
		"SYS___ACL_GET_FILE":                           true,
		"SYS___ACL_GET_LINK":                           true,
		"SYS___ACL_SET_FD":                             true,
		"SYS___ACL_SET_FILE":                           true,
		"SYS___ACL_SET_LINK":                           true,
		"SYS___CLONE":                                  true,
		"SYS___DISABLE_THREADSIGNAL":                   true,
		"SYS___GETCWD":                                 true,
		"SYS___GETLOGIN":                               true,
		"SYS___GET_TCB":                                true,
		"SYS___MAC_EXECVE":                             true,
		"SYS___MAC_GETFSSTAT":                          true,
		"SYS___MAC_GET_FD":                             true,
		"SYS___MAC_GET_FILE":                           true,
		"SYS___MAC_GET_LCID":                           true,
		"SYS___MAC_GET_LCTX":                           true,
		"SYS___MAC_GET_LINK":                           true,
		"SYS___MAC_GET_MOUNT":                          true,
		"SYS___MAC_GET_PID":                            true,
		"SYS___MAC_GET_PROC":                           true,
		"SYS___MAC_MOUNT":                              true,
		"SYS___MAC_SET_FD":                             true,
		"SYS___MAC_SET_FILE":                           true,
		"SYS___MAC_SET_LCTX":                           true,
		"SYS___MAC_SET_LINK":                           true,
		"SYS___MAC_SET_PROC":                           true,
		"SYS___MAC_SYSCALL":                            true,
		"SYS___OLD_SEMWAIT_SIGNAL":                     true,
		"SYS___OLD_SEMWAIT_SIGNAL_NOCANCEL":            true,
		"SYS___POSIX_CHOWN":                            true,
		"SYS___POSIX_FCHOWN":                           true,
		"SYS___POSIX_LCHOWN":                           true,
		"SYS___POSIX_RENAME":                           true,
		"SYS___PTHREAD_CANCELED":                       true,
		"SYS___PTHREAD_CHDIR":                          true,
		"SYS___PTHREAD_FCHDIR":                         true,
		"SYS___PTHREAD_KILL":                           true,
		"SYS___PTHREAD_MARKCANCEL":                     true,
		"SYS___PTHREAD_SIGMASK":                        true,
		"SYS___QUOTACTL":                               true,
		"SYS___SEMCTL":                                 true,
		"SYS___SEMWAIT_SIGNAL":                         true,
		"SYS___SEMWAIT_SIGNAL_NOCANCEL":                true,
		"SYS___SETLOGIN":                               true,
		"SYS___SETUGID":                                true,
		"SYS___SET_TCB":                                true,
		"SYS___SIGACTION_SIGTRAMP":                     true,
		"SYS___SIGTIMEDWAIT":                           true,
		"SYS___SIGWAIT":                                true,
		"SYS___SIGWAIT_NOCANCEL":                       true,
		"SYS___SYSCTL":                                 true,
		"SYS___TFORK":                                  true,
		"SYS___THREXIT":                                true,
		"SYS___THRSIGDIVERT":                           true,
		"SYS___THRSLEEP":                               true,
		"SYS___THRWAKEUP":                              true,
		"S_ARCH1":                                      true,
		"S_ARCH2":                                      true,
		"S_BLKSIZE":                                    true,
		"S_IEXEC":                                      true,
		"S_IFBLK":                                      true,
		"S_IFCHR":                                      true,
		"S_IFDIR":                                      true,
		"S_IFIFO":                                      true,
		"S_IFLNK":                                      true,
		"S_IFMT":                                       true,
		"S_IFREG":                                      true,
		"S_IFSOCK":                                     true,
		"S_IFWHT":                                      true,
		"S_IREAD":                                      true,
		"S_IRGRP":                                      true,
		"S_IROTH":                                      true,
		"S_IRUSR":                                      true,
		"S_IRWXG":                                      true,
		"S_IRWXO":                                      true,
		"S_IRWXU":                                      true,
		"S_ISGID":                                      true,
		"S_ISTXT":                                      true,
		"S_ISUID":                                      true,
		"S_ISVTX":                                      true,
		"S_IWGRP":                                      true,
		"S_IWOTH":                                      true,
		"S_IWRITE":                                     true,
		"S_IWUSR":                                      true,
		"S_IXGRP":                                      true,
		"S_IXOTH":                                      true,
		"S_IXUSR":                                      true,
		"S_LOGIN_SET":                                  true,
		"SecurityAttributes":                           true,
		"Seek":                                         true,
		"Select":                                       true,
		"Sendfile":                                     true,
		"Sendmsg":                                      true,
		"SendmsgN":                                     true,
		"Sendto":                                       true,
		"Servent":                                      true,
		"SetBpf":                                       true,
		"SetBpfBuflen":                                 true,
		"SetBpfDatalink":                               true,
		"SetBpfHeadercmpl":                             true,
		"SetBpfImmediate":                              true,
		"SetBpfInterface":                              true,
		"SetBpfPromisc":                                true,
		"SetBpfTimeout":                                true,
		"SetCurrentDirectory":                          true,
		"SetEndOfFile":                                 true,
		"SetEnvironmentVariable":                       true,
		"SetFileAttributes":                            true,
		"SetFileCompletionNotificationModes":           true,
		"SetFilePointer":                               true,
		"SetFileTime":                                  true,
		"SetHandleInformation":                         true,
		"SetKevent":                                    true,
		"SetLsfPromisc":                                true,
		"SetNonblock":                                  true,
		"Setdomainname":                                true,
		"Setegid":                                      true,
		"Setenv":                                       true,
		"Seteuid":                                      true,
		"Setfsgid":                                     true,
		"Setfsuid":                                     true,
		"Setgid":                                       true,
		"Setgroups":                                    true,
		"Sethostname":                                  true,
		"Setlogin":                                     true,
		"Setpgid":                                      true,
		"Setpriority":                                  true,
		"Setprivexec":                                  true,
		"Setregid":                                     true,
		"Setresgid":                                    true,
		"Setresuid":                                    true,
		"Setreuid":                                     true,
		"Setrlimit":                                    true,
		"Setsid":                                       true,
		"Setsockopt":                                   true,
		"SetsockoptByte":                               true,
		"SetsockoptICMPv6Filter":                       true,
		"SetsockoptIPMreq":                             true,
		"SetsockoptIPMreqn":                            true,
		"SetsockoptIPv6Mreq":                           true,
		"SetsockoptInet4Addr":                          true,
		"SetsockoptInt":                                true,
		"SetsockoptLinger":                             true,
		"SetsockoptString":                             true,
		"SetsockoptTimeval":                            true,
		"Settimeofday":                                 true,
		"Setuid":                                       true,
		"Setxattr":                                     true,
		"Shutdown":                                     true,
		"SidTypeAlias":                                 true,
		"SidTypeComputer":                              true,
		"SidTypeDeletedAccount":                        true,
		"SidTypeDomain":                                true,
		"SidTypeGroup":                                 true,
		"SidTypeInvalid":                               true,
		"SidTypeLabel":                                 true,
		"SidTypeUnknown":                               true,
		"SidTypeUser":                                  true,
		"SidTypeWellKnownGroup":                        true,
		"Signal":                                       true,
		"SizeofBpfHdr":                                 true,
		"SizeofBpfInsn":                                true,
		"SizeofBpfProgram":                             true,
		"SizeofBpfStat":                                true,
		"SizeofBpfVersion":                             true,
		"SizeofBpfZbuf":                                true,
		"SizeofBpfZbufHeader":                          true,
		"SizeofCmsghdr":                                true,
		"SizeofICMPv6Filter":                           true,
		"SizeofIPMreq":                                 true,
		"SizeofIPMreqn":                                true,
		"SizeofIPv6MTUInfo":                            true,
		"SizeofIPv6Mreq":                               true,
		"SizeofIfAddrmsg":                              true,
		"SizeofIfAnnounceMsghdr":                       true,
		"SizeofIfData":                                 true,
		"SizeofIfInfomsg":                              true,
		"SizeofIfMsghdr":                               true,
		"SizeofIfaMsghdr":                              true,
		"SizeofIfmaMsghdr":                             true,
		"SizeofIfmaMsghdr2":                            true,
		"SizeofInet4Pktinfo":                           true,
		"SizeofInet6Pktinfo":                           true,
		"SizeofInotifyEvent":                           true,
		"SizeofLinger":                                 true,
		"SizeofMsghdr":                                 true,
		"SizeofNlAttr":                                 true,
		"SizeofNlMsgerr":                               true,
		"SizeofNlMsghdr":                               true,
		"SizeofRtAttr":                                 true,
		"SizeofRtGenmsg":                               true,
		"SizeofRtMetrics":                              true,
		"SizeofRtMsg":                                  true,
		"SizeofRtMsghdr":                               true,
		"SizeofRtNexthop":                              true,
		"SizeofSockFilter":                             true,
		"SizeofSockFprog":                              true,
		"SizeofSockaddrAny":                            true,
		"SizeofSockaddrDatalink":                       true,
		"SizeofSockaddrInet4":                          true,
		"SizeofSockaddrInet6":                          true,
		"SizeofSockaddrLinklayer":                      true,
		"SizeofSockaddrNetlink":                        true,
		"SizeofSockaddrUnix":                           true,
		"SizeofTCPInfo":                                true,
		"SizeofUcred":                                  true,
		"SlicePtrFromStrings":                          true,
		"SockFilter":                                   true,
		"SockFprog":                                    true,
		"SockaddrDatalink":                             true,
		"SockaddrGen":                                  true,
		"SockaddrInet4":                                true,
		"SockaddrInet6":                                true,
		"SockaddrLinklayer":                            true,
		"SockaddrNetlink":                              true,
		"SockaddrUnix":                                 true,
		"Socket":                                       true,
		"SocketControlMessage":                         true,
		"SocketDisableIPv6":                            true,
		"Socketpair":                                   true,
		"Splice":                                       true,
		"StartProcess":                                 true,
		"StartupInfo":                                  true,
		"Stat":                                         true,
		"Stat_t":                                       true,
		"Statfs":                                       true,
		"Statfs_t":                                     true,
		"Stderr":                                       true,
		"Stdin":                                        true,
		"Stdout":                                       true,
		"StringBytePtr":                                true,
		"StringByteSlice":                              true,
		"StringSlicePtr":                               true,
		"StringToSid":                                  true,
		"StringToUTF16":                                true,
		"StringToUTF16Ptr":                             true,
		"Symlink":                                      true,
		"Sync":                                         true,
		"SyncFileRange":                                true,
		"SysProcAttr":                                  true,
		"SysProcIDMap":                                 true,
		"Syscall":                                      true,
		"Syscall12":                                    true,
		"Syscall15":                                    true,
		"Syscall18":                                    true,
		"Syscall6":                                     true,
		"Syscall9":                                     true,
		"Sysctl":                                       true,
		"SysctlUint32":                                 true,
		"Sysctlnode":                                   true,
		"Sysinfo":                                      true,
		"Sysinfo_t":                                    true,
		"Systemtime":                                   true,
		"TCGETS":                                       true,
		"TCIFLUSH":                                     true,
		"TCIOFLUSH":                                    true,
		"TCOFLUSH":                                     true,
		"TCPInfo":                                      true,
		"TCPKeepalive":                                 true,
		"TCP_CA_NAME_MAX":                              true,
		"TCP_CONGCTL":                                  true,
		"TCP_CONGESTION":                               true,
		"TCP_CONNECTIONTIMEOUT":                        true,
		"TCP_CORK":                                     true,
		"TCP_DEFER_ACCEPT":                             true,
		"TCP_INFO":                                     true,
		"TCP_KEEPALIVE":                                true,
		"TCP_KEEPCNT":                                  true,
		"TCP_KEEPIDLE":                                 true,
		"TCP_KEEPINIT":                                 true,
		"TCP_KEEPINTVL":                                true,
		"TCP_LINGER2":                                  true,
		"TCP_MAXBURST":                                 true,
		"TCP_MAXHLEN":                                  true,
		"TCP_MAXOLEN":                                  true,
		"TCP_MAXSEG":                                   true,
		"TCP_MAXWIN":                                   true,
		"TCP_MAX_SACK":                                 true,
		"TCP_MAX_WINSHIFT":                             true,
		"TCP_MD5SIG":                                   true,
		"TCP_MD5SIG_MAXKEYLEN":                         true,
		"TCP_MINMSS":                                   true,
		"TCP_MINMSSOVERLOAD":                           true,
		"TCP_MSS":                                      true,
		"TCP_NODELAY":                                  true,
		"TCP_NOOPT":                                    true,
		"TCP_NOPUSH":                                   true,
		"TCP_NSTATES":                                  true,
		"TCP_QUICKACK":                                 true,
		"TCP_RXT_CONNDROPTIME":                         true,
		"TCP_RXT_FINDROP":                              true,
		"TCP_SACK_ENABLE":                              true,
		"TCP_SYNCNT":                                   true,
		"TCP_VENDOR":                                   true,
		"TCP_WINDOW_CLAMP":                             true,
		"TCSAFLUSH":                                    true,
		"TCSETS":                                       true,
		"TF_DISCONNECT":                                true,
		"TF_REUSE_SOCKET":                              true,
		"TF_USE_DEFAULT_WORKER":                        true,
		"TF_USE_KERNEL_APC":                            true,
		"TF_USE_SYSTEM_THREAD":                         true,
		"TF_WRITE_BEHIND":                              true,
		"TH32CS_INHERIT":                               true,
		"TH32CS_SNAPALL":                               true,
		"TH32CS_SNAPHEAPLIST":                          true,
		"TH32CS_SNAPMODULE":                            true,
		"TH32CS_SNAPMODULE32":                          true,
		"TH32CS_SNAPPROCESS":                           true,
		"TH32CS_SNAPTHREAD":                            true,
		"TIME_ZONE_ID_DAYLIGHT":                        true,
		"TIME_ZONE_ID_STANDARD":                        true,
		"TIME_ZONE_ID_UNKNOWN":                         true,
		"TIOCCBRK":                                     true,
		"TIOCCDTR":                                     true,
		"TIOCCONS":                                     true,
		"TIOCDCDTIMESTAMP":                             true,
		"TIOCDRAIN":                                    true,
		"TIOCDSIMICROCODE":                             true,
		"TIOCEXCL":                                     true,
		"TIOCEXT":                                      true,
		"TIOCFLAG_CDTRCTS":                             true,
		"TIOCFLAG_CLOCAL":                              true,
		"TIOCFLAG_CRTSCTS":                             true,
		"TIOCFLAG_MDMBUF":                              true,
		"TIOCFLAG_PPS":                                 true,
		"TIOCFLAG_SOFTCAR":                             true,
		"TIOCFLUSH":                                    true,
		"TIOCGDEV":                                     true,
		"TIOCGDRAINWAIT":                               true,
		"TIOCGETA":                                     true,
		"TIOCGETD":                                     true,
		"TIOCGFLAGS":                                   true,
		"TIOCGICOUNT":                                  true,
		"TIOCGLCKTRMIOS":                               true,
		"TIOCGLINED":                                   true,
		"TIOCGPGRP":                                    true,
		"TIOCGPTN":                                     true,
		"TIOCGQSIZE":                                   true,
		"TIOCGRANTPT":                                  true,
		"TIOCGRS485":                                   true,
		"TIOCGSERIAL":                                  true,
		"TIOCGSID":                                     true,
		"TIOCGSIZE":                                    true,
		"TIOCGSOFTCAR":                                 true,
		"TIOCGTSTAMP":                                  true,
		"TIOCGWINSZ":                                   true,
		"TIOCINQ":                                      true,
		"TIOCIXOFF":                                    true,
		"TIOCIXON":                                     true,
		"TIOCLINUX":                                    true,
		"TIOCMBIC":                                     true,
		"TIOCMBIS":                                     true,
		"TIOCMGDTRWAIT":                                true,
		"TIOCMGET":                                     true,
		"TIOCMIWAIT":                                   true,
		"TIOCMODG":                                     true,
		"TIOCMODS":                                     true,
		"TIOCMSDTRWAIT":                                true,
		"TIOCMSET":                                     true,
		"TIOCM_CAR":                                    true,
		"TIOCM_CD":                                     true,
		"TIOCM_CTS":                                    true,
		"TIOCM_DCD":                                    true,
		"TIOCM_DSR":                                    true,
		"TIOCM_DTR":                                    true,
		"TIOCM_LE":                                     true,
		"TIOCM_RI":                                     true,
		"TIOCM_RNG":                                    true,
		"TIOCM_RTS":                                    true,
		"TIOCM_SR":                                     true,
		"TIOCM_ST":                                     true,
		"TIOCNOTTY":                                    true,
		"TIOCNXCL":                                     true,
		"TIOCOUTQ":                                     true,
		"TIOCPKT":                                      true,
		"TIOCPKT_DATA":                                 true,
		"TIOCPKT_DOSTOP":                               true,
		"TIOCPKT_FLUSHREAD":                            true,
		"TIOCPKT_FLUSHWRITE":                           true,
		"TIOCPKT_IOCTL":                                true,
		"TIOCPKT_NOSTOP":                               true,
		"TIOCPKT_START":                                true,
		"TIOCPKT_STOP":                                 true,
		"TIOCPTMASTER":                                 true,
		"TIOCPTMGET":                                   true,
		"TIOCPTSNAME":                                  true,
		"TIOCPTYGNAME":                                 true,
		"TIOCPTYGRANT":                                 true,
		"TIOCPTYUNLK":                                  true,
		"TIOCRCVFRAME":                                 true,
		"TIOCREMOTE":                                   true,
		"TIOCSBRK":                                     true,
		"TIOCSCONS":                                    true,
		"TIOCSCTTY":                                    true,
		"TIOCSDRAINWAIT":                               true,
		"TIOCSDTR":                                     true,
		"TIOCSERCONFIG":                                true,
		"TIOCSERGETLSR":                                true,
		"TIOCSERGETMULTI":                              true,
		"TIOCSERGSTRUCT":                               true,
		"TIOCSERGWILD":                                 true,
		"TIOCSERSETMULTI":                              true,
		"TIOCSERSWILD":                                 true,
		"TIOCSER_TEMT":                                 true,
		"TIOCSETA":                                     true,
		"TIOCSETAF":                                    true,
		"TIOCSETAW":                                    true,
		"TIOCSETD":                                     true,
		"TIOCSFLAGS":                                   true,
		"TIOCSIG":                                      true,
		"TIOCSLCKTRMIOS":                               true,
		"TIOCSLINED":                                   true,
		"TIOCSPGRP":                                    true,
		"TIOCSPTLCK":                                   true,
		"TIOCSQSIZE":                                   true,
		"TIOCSRS485":                                   true,
		"TIOCSSERIAL":                                  true,
		"TIOCSSIZE":                                    true,
		"TIOCSSOFTCAR":                                 true,
		"TIOCSTART":                                    true,
		"TIOCSTAT":                                     true,
		"TIOCSTI":                                      true,
		"TIOCSTOP":                                     true,
		"TIOCSTSTAMP":                                  true,
		"TIOCSWINSZ":                                   true,
		"TIOCTIMESTAMP":                                true,
		"TIOCUCNTL":                                    true,
		"TIOCVHANGUP":                                  true,
		"TIOCXMTFRAME":                                 true,
		"TOKEN_ADJUST_DEFAULT":                         true,
		"TOKEN_ADJUST_GROUPS":                          true,
		"TOKEN_ADJUST_PRIVILEGES":                      true,
		"TOKEN_ADJUST_SESSIONID":                       true,
		"TOKEN_ALL_ACCESS":                             true,
		"TOKEN_ASSIGN_PRIMARY":                         true,
		"TOKEN_DUPLICATE":                              true,
		"TOKEN_EXECUTE":                                true,
		"TOKEN_IMPERSONATE":                            true,
		"TOKEN_QUERY":                                  true,
		"TOKEN_QUERY_SOURCE":                           true,
		"TOKEN_READ":                                   true,
		"TOKEN_WRITE":                                  true,
		"TOSTOP":                                       true,
		"TRUNCATE_EXISTING":                            true,
		"TUNATTACHFILTER":                              true,
		"TUNDETACHFILTER":                              true,
		"TUNGETFEATURES":                               true,
		"TUNGETIFF":                                    true,
		"TUNGETSNDBUF":                                 true,
		"TUNGETVNETHDRSZ":                              true,
		"TUNSETDEBUG":                                  true,
		"TUNSETGROUP":                                  true,
		"TUNSETIFF":                                    true,
		"TUNSETLINK":                                   true,
		"TUNSETNOCSUM":                                 true,
		"TUNSETOFFLOAD":                                true,
		"TUNSETOWNER":                                  true,
		"TUNSETPERSIST":                                true,
		"TUNSETSNDBUF":                                 true,
		"TUNSETTXFILTER":                               true,
		"TUNSETVNETHDRSZ":                              true,
		"Tee":                                          true,
		"TerminateProcess":                             true,
		"Termios":                                      true,
		"Tgkill":                                       true,
		"Time":                                         true,
		"Time_t":                                       true,
		"Times":                                        true,
		"Timespec":                                     true,
		"TimespecToNsec":                               true,
		"Timeval":                                      true,
		"Timeval32":                                    true,
		"TimevalToNsec":                                true,
		"Timex":                                        true,
		"Timezoneinformation":                          true,
		"Tms":                                          true,
		"Token":                                        true,
		"TokenAccessInformation":                       true,
		"TokenAuditPolicy":                             true,
		"TokenDefaultDacl":                             true,
		"TokenElevation":                               true,
		"TokenElevationType":                           true,
		"TokenGroups":                                  true,
		"TokenGroupsAndPrivileges":                     true,
		"TokenHasRestrictions":                         true,
		"TokenImpersonationLevel":                      true,
		"TokenIntegrityLevel":                          true,
		"TokenLinkedToken":                             true,
		"TokenLogonSid":                                true,
		"TokenMandatoryPolicy":                         true,
		"TokenOrigin":                                  true,
		"TokenOwner":                                   true,
		"TokenPrimaryGroup":                            true,
		"TokenPrivileges":                              true,
		"TokenRestrictedSids":                          true,
		"TokenSandBoxInert":                            true,
		"TokenSessionId":                               true,
		"TokenSessionReference":                        true,
		"TokenSource":                                  true,
		"TokenStatistics":                              true,
		"TokenType":                                    true,
		"TokenUIAccess":                                true,
		"TokenUser":                                    true,
		"TokenVirtualizationAllowed":                   true,
		"TokenVirtualizationEnabled":                   true,
		"Tokenprimarygroup":                            true,
		"Tokenuser":                                    true,
		"TranslateAccountName":                         true,
		"TranslateName":                                true,
		"TransmitFile":                                 true,
		"TransmitFileBuffers":                          true,
		"Truncate":                                     true,
		"UNIX_PATH_MAX":                                true,
		"USAGE_MATCH_TYPE_AND":                         true,
		"USAGE_MATCH_TYPE_OR":                          true,
		"UTF16FromString":                              true,
		"UTF16PtrFromString":                           true,
		"UTF16ToString":                                true,
		"Ucred":                                        true,
		"Umask":                                        true,
		"Uname":                                        true,
		"Undelete":                                     true,
		"UnixCredentials":                              true,
		"UnixRights":                                   true,
		"Unlink":                                       true,
		"Unlinkat":                                     true,
		"UnmapViewOfFile":                              true,
		"Unmount":                                      true,
		"Unsetenv":                                     true,
		"Unshare":                                      true,
		"UserInfo10":                                   true,
		"Ustat":                                        true,
		"Ustat_t":                                      true,
		"Utimbuf":                                      true,
		"Utime":                                        true,
		"Utimes":                                       true,
		"UtimesNano":                                   true,
		"Utsname":                                      true,
		"VDISCARD":                                     true,
		"VDSUSP":                                       true,
		"VEOF":                                         true,
		"VEOL":                                         true,
		"VEOL2":                                        true,
		"VERASE":                                       true,
		"VERASE2":                                      true,
		"VINTR":                                        true,
		"VKILL":                                        true,
		"VLNEXT":                                       true,
		"VMIN":                                         true,
		"VQUIT":                                        true,
		"VREPRINT":                                     true,
		"VSTART":                                       true,
		"VSTATUS":                                      true,
		"VSTOP":                                        true,
		"VSUSP":                                        true,
		"VSWTC":                                        true,
		"VT0":                                          true,
		"VT1":                                          true,
		"VTDLY":                                        true,
		"VTIME":                                        true,
		"VWERASE":                                      true,
		"VirtualLock":                                  true,
		"VirtualUnlock":                                true,
		"WAIT_ABANDONED":                               true,
		"WAIT_FAILED":                                  true,
		"WAIT_OBJECT_0":                                true,
		"WAIT_TIMEOUT":                                 true,
		"WALL":                                         true,
		"WALLSIG":                                      true,
		"WALTSIG":                                      true,
		"WCLONE":                                       true,
		"WCONTINUED":                                   true,
		"WCOREFLAG":                                    true,
		"WEXITED":                                      true,
		"WLINUXCLONE":                                  true,
		"WNOHANG":                                      true,
		"WNOTHREAD":                                    true,
		"WNOWAIT":                                      true,
		"WNOZOMBIE":                                    true,
		"WOPTSCHECKED":                                 true,
		"WORDSIZE":                                     true,
		"WSABuf":                                       true,
		"WSACleanup":                                   true,
		"WSADESCRIPTION_LEN":                           true,
		"WSAData":                                      true,
		"WSAEACCES":                                    true,
		"WSAECONNABORTED":                              true,
		"WSAECONNRESET":                                true,
		"WSAEnumProtocols":                             true,
		"WSAID_CONNECTEX":                              true,
		"WSAIoctl":                                     true,
		"WSAPROTOCOL_LEN":                              true,
		"WSAProtocolChain":                             true,
		"WSAProtocolInfo":                              true,
		"WSARecv":                                      true,
		"WSARecvFrom":                                  true,
		"WSASYS_STATUS_LEN":                            true,
		"WSASend":                                      true,
		"WSASendTo":                                    true,
		"WSASendto":                                    true,
		"WSAStartup":                                   true,
		"WSTOPPED":                                     true,
		"WTRAPPED":                                     true,
		"WUNTRACED":                                    true,
		"Wait4":                                        true,
		"WaitForSingleObject":                          true,
		"WaitStatus":                                   true,
		"Win32FileAttributeData":                       true,
		"Win32finddata":                                true,
		"Write":                                        true,
		"WriteConsole":                                 true,
		"WriteFile":                                    true,
		"X509_ASN_ENCODING":                            true,
		"XCASE":                                        true,
		"XP1_CONNECTIONLESS":                           true,
		"XP1_CONNECT_DATA":                             true,
		"XP1_DISCONNECT_DATA":                          true,
		"XP1_EXPEDITED_DATA":                           true,
		"XP1_GRACEFUL_CLOSE":                           true,
		"XP1_GUARANTEED_DELIVERY":                      true,
		"XP1_GUARANTEED_ORDER":                         true,
		"XP1_IFS_HANDLES":                              true,
		"XP1_MESSAGE_ORIENTED":                         true,
		"XP1_MULTIPOINT_CONTROL_PLANE":                 true,
		"XP1_MULTIPOINT_DATA_PLANE":                    true,
		"XP1_PARTIAL_MESSAGE":                          true,
		"XP1_PSEUDO_STREAM":                            true,
		"XP1_QOS_SUPPORTED":                            true,
		"XP1_SAN_SUPPORT_SDP":                          true,
		"XP1_SUPPORT_BROADCAST":                        true,
		"XP1_SUPPORT_MULTIPOINT":                       true,
		"XP1_UNI_RECV":                                 true,
		"XP1_UNI_SEND":                                 true,
	},
	"syscall/js": map[string]bool{
		"Error":         true,
		"Func":          true,
		"FuncOf":        true,
		"Global":        true,
		"Null":          true,
		"Type":          true,
		"TypeBoolean":   true,
		"TypeFunction":  true,
		"TypeNull":      true,
		"TypeNumber":    true,
		"TypeObject":    true,
		"TypeString":    true,
		"TypeSymbol":    true,
		"TypeUndefined": true,
		"TypedArray":    true,
		"TypedArrayOf":  true,
		"Undefined":     true,
		"Value":         true,
		"ValueError":    true,
		"ValueOf":       true,
		"Wrapper":       true,
	},
	"testing": map[string]bool{
		"AllocsPerRun":      true,
		"B":                 true,
		"Benchmark":         true,
		"BenchmarkResult":   true,
		"Cover":             true,
		"CoverBlock":        true,
		"CoverMode":         true,
		"Coverage":          true,
		"InternalBenchmark": true,
		"InternalExample":   true,
		"InternalTest":      true,
		"M":                 true,
		"Main":              true,
		"MainStart":         true,
		"PB":                true,
		"RegisterCover":     true,
		"RunBenchmarks":     true,
		"RunExamples":       true,
		"RunTests":          true,
		"Short":             true,
		"T":                 true,
		"Verbose":           true,
	},
	"testing/iotest": map[string]bool{
		"DataErrReader":  true,
		"ErrTimeout":     true,
		"HalfReader":     true,
		"NewReadLogger":  true,
		"NewWriteLogger": true,
		"OneByteReader":  true,
		"TimeoutReader":  true,
		"TruncateWriter": true,
	},
	"testing/quick": map[string]bool{
		"Check":           true,
		"CheckEqual":      true,
		"CheckEqualError": true,
		"CheckError":      true,
		"Config":          true,
		"Generator":       true,
		"SetupError":      true,
		"Value":           true,
	},
	"text/scanner": map[string]bool{
		"Char":           true,
		"Comment":        true,
		"EOF":            true,
		"Float":          true,
		"GoTokens":       true,
		"GoWhitespace":   true,
		"Ident":          true,
		"Int":            true,
		"Position":       true,
		"RawString":      true,
		"ScanChars":      true,
		"ScanComments":   true,
		"ScanFloats":     true,
		"ScanIdents":     true,
		"ScanInts":       true,
		"ScanRawStrings": true,
		"ScanStrings":    true,
		"Scanner":        true,
		"SkipComments":   true,
		"String":         true,
		"TokenString":    true,
	},
	"text/tabwriter": map[string]bool{
		"AlignRight":          true,
		"Debug":               true,
		"DiscardEmptyColumns": true,
		"Escape":              true,
		"FilterHTML":          true,
		"NewWriter":           true,
		"StripEscape":         true,
		"TabIndent":           true,
		"Writer":              true,
	},
	"text/template": map[string]bool{
		"ExecError":        true,
		"FuncMap":          true,
		"HTMLEscape":       true,
		"HTMLEscapeString": true,
		"HTMLEscaper":      true,
		"IsTrue":           true,
		"JSEscape":         true,
		"JSEscapeString":   true,
		"JSEscaper":        true,
		"Must":             true,
		"New":              true,
		"ParseFiles":       true,
		"ParseGlob":        true,
		"Template":         true,
		"URLQueryEscaper":  true,
	},
	"text/template/parse": map[string]bool{
		"ActionNode":     true,
		"BoolNode":       true,
		"BranchNode":     true,
		"ChainNode":      true,
		"CommandNode":    true,
		"DotNode":        true,
		"FieldNode":      true,
		"IdentifierNode": true,
		"IfNode":         true,
		"IsEmptyTree":    true,
		"ListNode":       true,
		"New":            true,
		"NewIdentifier":  true,
		"NilNode":        true,
		"Node":           true,
		"NodeAction":     true,
		"NodeBool":       true,
		"NodeChain":      true,
		"NodeCommand":    true,
		"NodeDot":        true,
		"NodeField":      true,
		"NodeIdentifier": true,
		"NodeIf":         true,
		"NodeList":       true,
		"NodeNil":        true,
		"NodeNumber":     true,
		"NodePipe":       true,
		"NodeRange":      true,
		"NodeString":     true,
		"NodeTemplate":   true,
		"NodeText":       true,
		"NodeType":       true,
		"NodeVariable":   true,
		"NodeWith":       true,
		"NumberNode":     true,
		"Parse":          true,
		"PipeNode":       true,
		"Pos":            true,
		"RangeNode":      true,
		"StringNode":     true,
		"TemplateNode":   true,
		"TextNode":       true,
		"Tree":           true,
		"VariableNode":   true,
		"WithNode":       true,
	},
	"time": map[string]bool{
		"ANSIC":                  true,
		"After":                  true,
		"AfterFunc":              true,
		"April":                  true,
		"August":                 true,
		"Date":                   true,
		"December":               true,
		"Duration":               true,
		"February":               true,
		"FixedZone":              true,
		"Friday":                 true,
		"Hour":                   true,
		"January":                true,
		"July":                   true,
		"June":                   true,
		"Kitchen":                true,
		"LoadLocation":           true,
		"LoadLocationFromTZData": true,
		"Local":                  true,
		"Location":               true,
		"March":                  true,
		"May":                    true,
		"Microsecond":            true,
		"Millisecond":            true,
		"Minute":                 true,
		"Monday":                 true,
		"Month":                  true,
		"Nanosecond":             true,
		"NewTicker":              true,
		"NewTimer":               true,
		"November":               true,
		"Now":                    true,
		"October":                true,
		"Parse":                  true,
		"ParseDuration":          true,
		"ParseError":             true,
		"ParseInLocation":        true,
		"RFC1123":                true,
		"RFC1123Z":               true,
		"RFC3339":                true,
		"RFC3339Nano":            true,
		"RFC822":                 true,
		"RFC822Z":                true,
		"RFC850":                 true,
		"RubyDate":               true,
		"Saturday":               true,
		"Second":                 true,
		"September":              true,
		"Since":                  true,
		"Sleep":                  true,
		"Stamp":                  true,
		"StampMicro":             true,
		"StampMilli":             true,
		"StampNano":              true,
		"Sunday":                 true,
		"Thursday":               true,
		"Tick":                   true,
		"Ticker":                 true,
		"Time":                   true,
		"Timer":                  true,
		"Tuesday":                true,
		"UTC":                    true,
		"Unix":                   true,
		"UnixDate":               true,
		"Until":                  true,
		"Wednesday":              true,
		"Weekday":                true,
	},
	"unicode": map[string]bool{
		"ASCII_Hex_Digit":                    true,
		"Adlam":                              true,
		"Ahom":                               true,
		"Anatolian_Hieroglyphs":              true,
		"Arabic":                             true,
		"Armenian":                           true,
		"Avestan":                            true,
		"AzeriCase":                          true,
		"Balinese":                           true,
		"Bamum":                              true,
		"Bassa_Vah":                          true,
		"Batak":                              true,
		"Bengali":                            true,
		"Bhaiksuki":                          true,
		"Bidi_Control":                       true,
		"Bopomofo":                           true,
		"Brahmi":                             true,
		"Braille":                            true,
		"Buginese":                           true,
		"Buhid":                              true,
		"C":                                  true,
		"Canadian_Aboriginal":                true,
		"Carian":                             true,
		"CaseRange":                          true,
		"CaseRanges":                         true,
		"Categories":                         true,
		"Caucasian_Albanian":                 true,
		"Cc":                                 true,
		"Cf":                                 true,
		"Chakma":                             true,
		"Cham":                               true,
		"Cherokee":                           true,
		"Co":                                 true,
		"Common":                             true,
		"Coptic":                             true,
		"Cs":                                 true,
		"Cuneiform":                          true,
		"Cypriot":                            true,
		"Cyrillic":                           true,
		"Dash":                               true,
		"Deprecated":                         true,
		"Deseret":                            true,
		"Devanagari":                         true,
		"Diacritic":                          true,
		"Digit":                              true,
		"Duployan":                           true,
		"Egyptian_Hieroglyphs":               true,
		"Elbasan":                            true,
		"Ethiopic":                           true,
		"Extender":                           true,
		"FoldCategory":                       true,
		"FoldScript":                         true,
		"Georgian":                           true,
		"Glagolitic":                         true,
		"Gothic":                             true,
		"Grantha":                            true,
		"GraphicRanges":                      true,
		"Greek":                              true,
		"Gujarati":                           true,
		"Gurmukhi":                           true,
		"Han":                                true,
		"Hangul":                             true,
		"Hanunoo":                            true,
		"Hatran":                             true,
		"Hebrew":                             true,
		"Hex_Digit":                          true,
		"Hiragana":                           true,
		"Hyphen":                             true,
		"IDS_Binary_Operator":                true,
		"IDS_Trinary_Operator":               true,
		"Ideographic":                        true,
		"Imperial_Aramaic":                   true,
		"In":                                 true,
		"Inherited":                          true,
		"Inscriptional_Pahlavi":              true,
		"Inscriptional_Parthian":             true,
		"Is":                                 true,
		"IsControl":                          true,
		"IsDigit":                            true,
		"IsGraphic":                          true,
		"IsLetter":                           true,
		"IsLower":                            true,
		"IsMark":                             true,
		"IsNumber":                           true,
		"IsOneOf":                            true,
		"IsPrint":                            true,
		"IsPunct":                            true,
		"IsSpace":                            true,
		"IsSymbol":                           true,
		"IsTitle":                            true,
		"IsUpper":                            true,
		"Javanese":                           true,
		"Join_Control":                       true,
		"Kaithi":                             true,
		"Kannada":                            true,
		"Katakana":                           true,
		"Kayah_Li":                           true,
		"Kharoshthi":                         true,
		"Khmer":                              true,
		"Khojki":                             true,
		"Khudawadi":                          true,
		"L":                                  true,
		"Lao":                                true,
		"Latin":                              true,
		"Lepcha":                             true,
		"Letter":                             true,
		"Limbu":                              true,
		"Linear_A":                           true,
		"Linear_B":                           true,
		"Lisu":                               true,
		"Ll":                                 true,
		"Lm":                                 true,
		"Lo":                                 true,
		"Logical_Order_Exception":            true,
		"Lower":                              true,
		"LowerCase":                          true,
		"Lt":                                 true,
		"Lu":                                 true,
		"Lycian":                             true,
		"Lydian":                             true,
		"M":                                  true,
		"Mahajani":                           true,
		"Malayalam":                          true,
		"Mandaic":                            true,
		"Manichaean":                         true,
		"Marchen":                            true,
		"Mark":                               true,
		"Masaram_Gondi":                      true,
		"MaxASCII":                           true,
		"MaxCase":                            true,
		"MaxLatin1":                          true,
		"MaxRune":                            true,
		"Mc":                                 true,
		"Me":                                 true,
		"Meetei_Mayek":                       true,
		"Mende_Kikakui":                      true,
		"Meroitic_Cursive":                   true,
		"Meroitic_Hieroglyphs":               true,
		"Miao":                               true,
		"Mn":                                 true,
		"Modi":                               true,
		"Mongolian":                          true,
		"Mro":                                true,
		"Multani":                            true,
		"Myanmar":                            true,
		"N":                                  true,
		"Nabataean":                          true,
		"Nd":                                 true,
		"New_Tai_Lue":                        true,
		"Newa":                               true,
		"Nko":                                true,
		"Nl":                                 true,
		"No":                                 true,
		"Noncharacter_Code_Point":            true,
		"Number":                             true,
		"Nushu":                              true,
		"Ogham":                              true,
		"Ol_Chiki":                           true,
		"Old_Hungarian":                      true,
		"Old_Italic":                         true,
		"Old_North_Arabian":                  true,
		"Old_Permic":                         true,
		"Old_Persian":                        true,
		"Old_South_Arabian":                  true,
		"Old_Turkic":                         true,
		"Oriya":                              true,
		"Osage":                              true,
		"Osmanya":                            true,
		"Other":                              true,
		"Other_Alphabetic":                   true,
		"Other_Default_Ignorable_Code_Point": true,
		"Other_Grapheme_Extend":              true,
		"Other_ID_Continue":                  true,
		"Other_ID_Start":                     true,
		"Other_Lowercase":                    true,
		"Other_Math":                         true,
		"Other_Uppercase":                    true,
		"P":                                  true,
		"Pahawh_Hmong":                       true,
		"Palmyrene":                          true,
		"Pattern_Syntax":                     true,
		"Pattern_White_Space":                true,
		"Pau_Cin_Hau":                        true,
		"Pc":                                 true,
		"Pd":                                 true,
		"Pe":                                 true,
		"Pf":                                 true,
		"Phags_Pa":                           true,
		"Phoenician":                         true,
		"Pi":                                 true,
		"Po":                                 true,
		"Prepended_Concatenation_Mark":       true,
		"PrintRanges":                        true,
		"Properties":                         true,
		"Ps":                                 true,
		"Psalter_Pahlavi":                    true,
		"Punct":                              true,
		"Quotation_Mark":                     true,
		"Radical":                            true,
		"Range16":                            true,
		"Range32":                            true,
		"RangeTable":                         true,
		"Regional_Indicator":                 true,
		"Rejang":                             true,
		"ReplacementChar":                    true,
		"Runic":                              true,
		"S":                                  true,
		"STerm":                              true,
		"Samaritan":                          true,
		"Saurashtra":                         true,
		"Sc":                                 true,
		"Scripts":                            true,
		"Sentence_Terminal":                  true,
		"Sharada":                            true,
		"Shavian":                            true,
		"Siddham":                            true,
		"SignWriting":                        true,
		"SimpleFold":                         true,
		"Sinhala":                            true,
		"Sk":                                 true,
		"Sm":                                 true,
		"So":                                 true,
		"Soft_Dotted":                        true,
		"Sora_Sompeng":                       true,
		"Soyombo":                            true,
		"Space":                              true,
		"SpecialCase":                        true,
		"Sundanese":                          true,
		"Syloti_Nagri":                       true,
		"Symbol":                             true,
		"Syriac":                             true,
		"Tagalog":                            true,
		"Tagbanwa":                           true,
		"Tai_Le":                             true,
		"Tai_Tham":                           true,
		"Tai_Viet":                           true,
		"Takri":                              true,
		"Tamil":                              true,
		"Tangut":                             true,
		"Telugu":                             true,
		"Terminal_Punctuation":               true,
		"Thaana":                             true,
		"Thai":                               true,
		"Tibetan":                            true,
		"Tifinagh":                           true,
		"Tirhuta":                            true,
		"Title":                              true,
		"TitleCase":                          true,
		"To":                                 true,
		"ToLower":                            true,
		"ToTitle":                            true,
		"ToUpper":                            true,
		"TurkishCase":                        true,
		"Ugaritic":                           true,
		"Unified_Ideograph":                  true,
		"Upper":                              true,
		"UpperCase":                          true,
		"UpperLower":                         true,
		"Vai":                                true,
		"Variation_Selector":                 true,
		"Version":                            true,
		"Warang_Citi":                        true,
		"White_Space":                        true,
		"Yi":                                 true,
		"Z":                                  true,
		"Zanabazar_Square":                   true,
		"Zl":                                 true,
		"Zp":                                 true,
		"Zs":                                 true,
	},
	"unicode/utf16": map[string]bool{
		"Decode":      true,
		"DecodeRune":  true,
		"Encode":      true,
		"EncodeRune":  true,
		"IsSurrogate": true,
	},
	"unicode/utf8": map[string]bool{
		"DecodeLastRune":         true,
		"DecodeLastRuneInString": true,
		"DecodeRune":             true,
		"DecodeRuneInString":     true,
		"EncodeRune":             true,
		"FullRune":               true,
		"FullRuneInString":       true,
		"MaxRune":                true,
		"RuneCount":              true,
		"RuneCountInString":      true,
		"RuneError":              true,
		"RuneLen":                true,
		"RuneSelf":               true,
		"RuneStart":              true,
		"UTFMax":                 true,
		"Valid":                  true,
		"ValidRune":              true,
		"ValidString":            true,
	},
	"unsafe": map[string]bool{
		"Alignof":       true,
		"ArbitraryType": true,
		"Offsetof":      true,
		"Pointer":       true,
		"Sizeof":        true,
	},
}
